﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Threading;

namespace MetaSharp.Build
{
	public sealed class ActionQueue : IActionQueue
	{
		private Dispatcher dispatcher = Dispatcher.CurrentDispatcher;
		private Dictionary<string, QueueGroup> groups = new Dictionary<string, QueueGroup>();

		private class QueueGroup
		{
			public Request next;
			public Request processing;
		}

		public void Enqueue(string queueGroup, Func<object> action, Action<Result> complete)
		{
			var req = new Request(action, complete);

			var group = default(QueueGroup);
			if (!this.groups.TryGetValue(queueGroup, out group))
				this.groups.Add(queueGroup, group = new QueueGroup());

			groups[queueGroup].next = req;
			this.ProcessNext(queueGroup);
		}

		private void ProcessNext(string queueGroup)
		{
			var group = groups[queueGroup];
			if (group.processing == null && group.next != null)
			{
				group.processing = group.next;
				group.next = null;
				group.processing.Task.BeginInvoke(RequestComplete, queueGroup);
			}
		}

		private void RequestComplete(IAsyncResult result)
		{
			var queueGroup = (string)result.AsyncState;
			var group = this.groups[queueGroup];

			Exception error = null;
			object value = null;
			try
			{
				value = group.processing.Task.EndInvoke(result);
			}
			catch (Exception ex) { error = ex; }

			this.dispatcher.BeginInvoke(new Action(() =>
			{
				group.processing.Complete(new Result(error, value));
				group.processing = null;
				this.ProcessNext(queueGroup);
			}));
		}

		private class Request
		{
			public Func<object> Task { get; private set; }
			public Action<Result> Complete { get; private set; }

			public Request(Func<object> task, Action<Result> complete)
			{
				this.Task = task;
				this.Complete = complete;
			}
		}
	}

	public class Result
	{
		public Exception Error { get; private set; }
		public object Value { get; private set; }

		public Result(Exception error, object value)
		{
			this.Error = error;
			this.Value = value;
		}
	}

	public interface IActionQueue
	{
		void Enqueue(string queueGroup, Func<object> action, Action<Result> complete);
	}
}

