﻿namespace MetaSharp.Build
{
	using System;
	using System.IO;
	using System.Linq;
	using System.Runtime.Serialization;
	using System.Text;
	using System.Xml;
	using Microsoft.Build.Framework;
	using Microsoft.Build.Utilities;
	using Transformation.Grammar;

	internal class GeneratedFileManager : IDisposable
	{
		internal const string DataContractNamespace = "urn:metasharp/2011/06";

		private const string GeneratedFilesFilename = "generated-files.txt";

		private IFileService fileService;
		private string projectFileDirectory;
		private string generationDirectory;
		private string generatedFilesFilename;
		private string generatedFileExtension;
		private TaskLoggingHelper log;
		private bool isCleanTask;
		private GeneratedFiles generatedFiles;

		private Uri projectFileDirectoryUri;

		public GeneratedFileManager(IFileService fileService, string componentName, string projectFilePath, string intermediateOutputPath, string generatedFileExtension, TaskLoggingHelper log, bool isCleanTask)
		{
			this.fileService = fileService;
			this.projectFileDirectory = Path.GetDirectoryName(projectFilePath);
			this.projectFileDirectoryUri = new Uri(this.projectFileDirectory);
			this.generationDirectory = Path.Combine(intermediateOutputPath, componentName);
			this.generatedFilesFilename = Path.Combine(generationDirectory, GeneratedFilesFilename);
			this.generatedFileExtension = generatedFileExtension.StartsWith(".") ? generatedFileExtension.Substring(1) : generatedFileExtension;
			this.log = log;
			this.isCleanTask = isCleanTask;
			this.Initialize();
		}

		public virtual bool IsFileRegenerationNeeded(string sourceFilename)
		{
			var result = false;
			var generatedFileRelativePath = GetRelativePath(sourceFilename);
			var generatedFileItem = generatedFiles.Files.FirstOrDefault(i => String.Compare(i.SourceFilename, generatedFileRelativePath, StringComparison.OrdinalIgnoreCase) == 0);

			if (generatedFileItem == null)
			{
				log.LogMessage(MessageImportance.Normal, "CodeGen needed for: {0}", sourceFilename);

				result = true;
			}
			else
			{
				var lastSourceWriteTime = this.fileService.GetLastWriteTimeUtc(sourceFilename);
				var lastWriteTimeOfTransformAssembly = this.fileService.GetLastWriteTimeUtc(typeof(GrammarTransformer).Assembly.Location);

				if (lastSourceWriteTime > generatedFileItem.SourceLastModificationDate)
				{
					log.LogMessage(MessageImportance.Normal,
								   "CodeGen needed for: {0}, last update time: {1}, last generation time: {2}", sourceFilename,
								   lastSourceWriteTime.ToString("u"), generatedFileItem.SourceLastModificationDate.ToString("u"));

					result = true;
				}
				else if ((!generatedFiles.LastUpdateOfGrammarTransformer.HasValue && (lastWriteTimeOfTransformAssembly > generatedFileItem.SourceLastModificationDate)) ||
					(generatedFiles.LastUpdateOfGrammarTransformer.HasValue && (lastWriteTimeOfTransformAssembly > generatedFiles.LastUpdateOfGrammarTransformer.Value)) ||
					!generatedFiles.LastUpdateOfGrammarTransformer.HasValue)
				{
					log.LogMessage(MessageImportance.Normal,
								   "CodeGen needed for: {0}, last update time of GrammarTransformer: {1}, last generation time: {2}", sourceFilename,
								   lastWriteTimeOfTransformAssembly.ToString("u"), generatedFileItem.SourceLastModificationDate.ToString("u"));

					result = true;
				}
			}

			if (!result)
			{
				log.LogMessage(MessageImportance.Normal, "No CodeGen needed for: {0}", sourceFilename);
			}

			return result;
		}

		public virtual string AddGeneratedFile(string sourceFilename, string fileContent)
		{
			return AddGeneratedFile(sourceFilename, fileContent, Encoding.UTF8);
		}

		public virtual string AddGeneratedFile(string sourceFilename, string fileContent, Encoding encoding)
		{
			var generatedFilename = GetGeneratedFilename(sourceFilename);
			var sourceFileRelativePath = GetRelativePath(sourceFilename);
			var generatedFileRelativePath = GetRelativePath(generatedFilename);

			this.fileService.WriteAllText(generatedFilename, fileContent, encoding);

			generatedFiles.AddOrUpdate(new GeneratedFile
			{
				SourceFilename = sourceFileRelativePath,
				SourceLastModificationDate = this.fileService.GetLastWriteTimeUtc(sourceFilename),
				TargetFilename = generatedFileRelativePath,
				TargetCreationDate = this.fileService.GetLastWriteTimeUtc(generatedFilename),
				Type = OutputType.Compile
			});

			return generatedFilename;
		}

		public virtual string AddGeneratedFile(string sourceFilename, byte[] fileContent)
		{
			var generatedFilename = GetGeneratedFilename(sourceFilename);
			this.fileService.WriteAllBytes(generatedFilename, fileContent);
			return generatedFilename;
		}

		public string GetGeneratedFilename(string sourceFilename)
		{
			var generatedFilename = String.Format("{0}.{1}", Path.GetFileName(sourceFilename), this.generatedFileExtension);
			var generatedFilePath = Path.Combine(projectFileDirectory, generationDirectory, generatedFilename);

			return generatedFilePath;
		}

		public void Dispose()
		{
			if (this.isCleanTask)
			{
				CleanFiles();
			}
			else
			{
				WriteFile(this.generatedFiles);
			}
		}

		private void Initialize()
		{
			if (!this.fileService.Exists(generationDirectory))
				this.fileService.CreateDirectory(generationDirectory);
			this.generatedFiles = ReadFile();
		}

		private GeneratedFiles ReadFile()
		{
			var result = new GeneratedFiles();
			if (this.fileService.Exists(generatedFilesFilename))
			{
				var serializer = new DataContractSerializer(typeof(GeneratedFiles));
				using (var stream = this.fileService.Open(generatedFilesFilename, FileMode.Open, FileAccess.Read, FileShare.None))
				{
					result = serializer.ReadObject(stream) as GeneratedFiles;
				}
			}

			return result;
		}

		private void WriteFile(GeneratedFiles generatedFiles)
		{
			var serializer = new DataContractSerializer(typeof(GeneratedFiles));

			using (var stream = this.fileService.Open(generatedFilesFilename, FileMode.Create, FileAccess.Write, FileShare.None))
			{
				using (var xmlTextWriter = new XmlTextWriter(stream, Encoding.UTF8))
				{
					xmlTextWriter.Formatting = Formatting.Indented;
					xmlTextWriter.Indentation = 1;
					xmlTextWriter.IndentChar = '\t';
					serializer.WriteObject(xmlTextWriter, generatedFiles);
				}
			}
		}

		private void CleanFiles()
		{
			if (this.fileService.Exists(generatedFilesFilename))
			{
				var generatedFiles = ReadFile();

				foreach (var generatedFile in generatedFiles.Files)
				{
					var filename = Path.Combine(generationDirectory, generatedFile.TargetFilename);

					if (this.fileService.Exists(filename))
					{
						log.LogMessage("Deleting '{0}'.", filename);

						try
						{
							this.fileService.Delete(filename);
							log.LogMessage("'{0}' deleted.", filename);
						}
						catch
						{
							log.LogError("Unable to delete '{0}'.", filename);
						}
					}
					else
					{
						log.LogMessage("Deleting '{0}': Skipping, does not exists.", filename);
					}
				}

				try
				{
					this.fileService.Delete(this.generatedFilesFilename);
				}
				catch (Exception ex)
				{
					log.LogError("Unable to delete '{0}'.", this.generatedFilesFilename);
					log.LogErrorFromException(ex);
				}
			}

			if (this.fileService.Exists(generationDirectory))
			{
				log.LogMessage("Deleting directory '{0}'.", generationDirectory);

				try
				{
					this.fileService.Delete(generationDirectory);
					log.LogMessage("Directory '{0}' deleted.", generationDirectory);
				}
				catch (Exception ex)
				{
					log.LogError("Unable to delete directory '{0}'.", this.generationDirectory);
					log.LogErrorFromException(ex);
				}
			}
		}

		private string GetRelativePath(string sourceFilename)
		{
			var sourceFilenameUri = new Uri(sourceFilename, UriKind.RelativeOrAbsolute);
			return projectFileDirectoryUri.MakeRelativeUri(sourceFilenameUri).OriginalString;
		}

		internal void UpdateGrammarTransformerLastWriteTime()
		{
			var lastWriteTimeOfTransformAssembly = this.fileService.GetLastWriteTimeUtc(typeof(GrammarTransformer).Assembly.Location);
			generatedFiles.LastUpdateOfGrammarTransformer = lastWriteTimeOfTransformAssembly;
		}
	}
}
