﻿using System;
using System.CodeDom;
using System.CodeDom.Compiler;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using MetaSharp.Transformation;
using MetaSharp.Transformation.Grammar;
using MetaSharp.Transformation.Grammar.Ast;
using MetaSharp.Transformation.Lang.Ast;
using Microsoft.CSharp;
using System.Text;
using System.Security.Cryptography;
using System.IO;

namespace MetaSharp.Build.Grammar
{
	public class DslBuildProvider :
		GrammarBuildProvider,
		IDslBuildProvider
	{
		IGrammarBuildProvider grammarBuilder;
		IActionQueue actionQueue;
		IErrorProvider errorProvider;
		IFileService fileService;
		IDynamicCompiler dynamicCompiler;

		static readonly MD5 md5 = MD5.Create();
		Dictionary<string, TransformerBuildResult> projectBuildResults = new Dictionary<string, TransformerBuildResult>();
		Dictionary<string, string> documents = new Dictionary<string, string>();

		public DslBuildProvider(IActionQueue actionQueue, IGrammarBuildProvider grammarBuildProvider, IDynamicCompiler dynamicCompiler, IErrorProvider errorProvider, IFileService fileService)
			: base(actionQueue)
		{
			this.grammarBuilder = grammarBuildProvider;
			this.grammarBuilder.BuildComplete += OnGrammarBuilderBuildComplete;
			this.grammarBuilder.Renamed += OnGrammarBuilderRenamed;
			this.dynamicCompiler = dynamicCompiler;
			this.errorProvider = errorProvider;
			this.actionQueue = actionQueue;
			this.fileService = fileService;
		}

		class TransformerBuildResult
		{
			private HashSet<string> needsUpdates = new HashSet<string>();
			private HashSet<string> updating = new HashSet<string>();

			public Dictionary<string, Unit> DocumentNodes = new Dictionary<string, Unit>();
			public IImmutableArray<string> TransformTypes = new ArrayList<string>();
			public string SelectedType;
			public Assembly Assembly;

			public void SetNeedsUpdate(string documentPath)
			{
				if (!needsUpdates.Contains(documentPath))
					needsUpdates.Add(documentPath);
			}

			public bool NeedsUpdate(string documentPath)
			{
				return needsUpdates.Contains(documentPath);
			}

			public bool IsUpdating(string documentPath)
			{
				return updating.Contains(documentPath);
			}

			public void StartUpdate(string documentPath)
			{
				needsUpdates.Remove(documentPath);
				updating.Add(documentPath);
			}

			public void DoneUpdating(string documentPath)
			{
				updating.Remove(documentPath);
			}
		}

		protected override Transformer GetTransformer(string documentPath)
		{
			if (documents.TryGetValue(documentPath, out documentPath))
			{
				var projectPath = this.grammarBuilder.GetDocumentProject(documentPath);
				if (projectPath != null)
				{
					var buildResult = this.projectBuildResults[projectPath];

					if (buildResult.SelectedType == null || buildResult.Assembly == null)
						return null;

					var selected = buildResult.Assembly.GetType(buildResult.SelectedType);
					if (selected != null)
						return (Transformer)Activator.CreateInstance(selected);
				}
			}

			return null;
		}

		private void OnGrammarBuilderRenamed(string oldDocumentPath, string newDocumentPath)
		{
			var oldDslPath = DslBuildProvider.GetDslPath(oldDocumentPath);
			var newDslPath = DslBuildProvider.GetDslPath(newDocumentPath);
			var projectPath = this.grammarBuilder.GetDocumentProject(newDocumentPath);

			documents.Remove(oldDslPath);
			documents.Add(newDslPath, newDocumentPath);

			var buildResult = default(TransformerBuildResult);
			if (projectBuildResults.TryGetValue(projectPath, out buildResult))
			{
				var unit = (Unit)grammarBuilder.GetResult(newDocumentPath);
				if (buildResult.DocumentNodes.ContainsKey(oldDocumentPath))
					buildResult.DocumentNodes.Remove(oldDocumentPath);
				buildResult.DocumentNodes.Add(newDocumentPath, unit);
			}

			if (this.fileService.Exists(newDslPath))
				this.fileService.Delete(newDslPath);

			this.fileService.Move(oldDslPath, newDslPath);

			base.Rename(oldDslPath, newDslPath);
		}

		private void OnGrammarBuilderBuildComplete(string documentPath, CompileResult result)
		{
			var dslPath = DslBuildProvider.GetDslPath(documentPath);
			var projectPath = this.grammarBuilder.GetDocumentProject(documentPath);

			if (!documents.ContainsKey(dslPath))
				documents.Add(dslPath, documentPath);

			var buildResult = default(TransformerBuildResult);
			if (!projectBuildResults.TryGetValue(projectPath, out buildResult))
				projectBuildResults.Add(projectPath, buildResult = new TransformerBuildResult());

			var unit = (Unit)result.Result;
			if (!buildResult.DocumentNodes.ContainsKey(documentPath))
				buildResult.DocumentNodes.Add(documentPath, null);
			buildResult.DocumentNodes[documentPath] = unit;

			if (buildResult.IsUpdating(documentPath))
			{
				buildResult.SetNeedsUpdate(documentPath);
			}
			else
			{
				buildResult.StartUpdate(documentPath);

				//unit = new Unit();
				//foreach (var result in grammarBuilder.GetResults(projectPath).Cast<Unit>())
				//    unit = Combine(unit, result);

				this.errorProvider.Clear(documentPath);
				this.actionQueue.Enqueue("Dsl" + documentPath,
					() => CompileGrammar(this.fileService, this.dynamicCompiler, unit, dslPath),
					r =>
					{
						buildResult.DoneUpdating(documentPath);
						if (buildResult.NeedsUpdate(documentPath))
						{
							OnGrammarBuilderBuildComplete(documentPath, result);
							return;
						}

						var compileGrammarResult = (DslCompileResult)r.Value;
						if (compileGrammarResult.errors.Count > 0)
						{
							foreach (CompilerError error in compileGrammarResult.errors)
								this.errorProvider.Add(documentPath, new Build.SourceError(error));
						}
						else
						{
							buildResult.Assembly = compileGrammarResult.assembly;
							buildResult.TransformTypes = compileGrammarResult.types;
							Update(dslPath, null);
						}
					});
			}
		}

		public static string GetDslPath(string grammarPath)
		{
			var metaSharpTempPath = Path.Combine(
				Environment.GetFolderPath(Environment.SpecialFolder.LocalApplicationData),
				"MetaSharp",
				"Temp",
				"Console");

			var hash = GetMd5Hash(grammarPath);
			var dslPath = Path.Combine(
				metaSharpTempPath,
				hash + Path.GetExtension(grammarPath) + ".mdsl");

			return dslPath;
		}

		public static string GetMd5Hash(string input)
		{
			byte[] data = md5.ComputeHash(Encoding.UTF8.GetBytes(input));
			StringBuilder sBuilder = new StringBuilder();
			for (int i = 0; i < data.Length; i++)
			{
				sBuilder.Append(data[i].ToString("x2"));
			}

			return sBuilder.ToString();
		}

		public IImmutableArray<string> GetTransformTypes(string documentPath)
		{
			var grammarTypes = Enumerable.Empty<string>();
			if (documents.TryGetValue(documentPath, out documentPath))
			{
				var projectPath = this.grammarBuilder.GetDocumentProject(documentPath);
				if (projectPath != null)
				{
					var buildResult = this.projectBuildResults[projectPath];
					var documentUnit = buildResult.DocumentNodes[documentPath];

					grammarTypes = documentUnit.Namespaces
						.OfType<CodeNamespace>()
						.SelectMany(n => n.Types.OfType<GrammarDeclaration>().Select(g => n.Name + "." + g.Name))
						.ToArray();
				}
			}

			return new ArrayList<string>(grammarTypes);
		}

		public void SetTransformType(string documentPath, string type)
		{
			var mappedDocumentPath = default(string);
			if (documents.TryGetValue(documentPath, out mappedDocumentPath))
			{
				var projectPath = this.grammarBuilder.GetDocumentProject(mappedDocumentPath);
				var buildResult = this.projectBuildResults[projectPath];
				if (buildResult.SelectedType != type)
				{
					buildResult.SelectedType = type;
					this.Update(documentPath, null);
				}
			}
		}

		private static Unit Combine(Unit u1, Unit u2)
		{
			if (u2 == null)
				return u1;

			var u3 = new Unit();
			u3.AssemblyCustomAttributes.AddRange(u1.AssemblyCustomAttributes
				.Cast<CodeAttributeDeclaration>().Concat(u2.AssemblyCustomAttributes.Cast<CodeAttributeDeclaration>())
				.ToArray());
			u3.StartDirectives.AddRange(u1.StartDirectives
				.Cast<CodeDirective>().Concat(u2.StartDirectives.Cast<CodeDirective>())
				.ToArray());
			u3.EndDirectives.AddRange(u1.EndDirectives
				.Cast<CodeDirective>().Concat(u2.EndDirectives.Cast<CodeDirective>())
				.ToArray());
			u3.ReferencedAssemblies.AddRange(u1.ReferencedAssemblies
				.Cast<string>().Concat(u2.ReferencedAssemblies.Cast<string>())
				.Distinct()
				.ToArray());
			u3.Namespaces.AddRange(u1.Namespaces.Cast<CodeNamespace>()
				.Concat(u2.Namespaces.Cast<CodeNamespace>())
				.ToArray());

			return u3;
		}

		private static readonly GrammarTransformer transformer = new GrammarTransformer();

		private class DslCompileResult
		{
			public Assembly assembly;
			public IImmutableArray<string> types = new ArrayList<string>();
			public CompilerErrorCollection errors = new CompilerErrorCollection();
		}

		private static DslCompileResult CompileGrammar(IFileService fileService, IDynamicCompiler dynamicCompiler, Unit ast, string codePath)
		{
			var compileResult = new DslCompileResult();
			var generatedCodeFile = codePath + ".cs";
			var generatedAssembly = codePath + ".dll";
			var generatedPdb = Path.ChangeExtension(generatedAssembly, "pdb");
			try
			{
				using (var provider = new CSharpCodeProvider())
				{
					var unit = transformer.Transform(ast);
					var generatedCode = unit.Print(provider);
					fileService.WriteAllText(generatedCodeFile, generatedCode, Encoding.UTF8);

					var parameters = new CompilerParameters(
						new string[] { 
							typeof(string).Assembly.Location, 
							typeof(Pattern).Assembly.Location, 
							typeof(Enumerable).Assembly.Location,
							typeof(CodeDomProvider).Assembly.Location,
						},
						generatedAssembly,
						true);

					var results = provider.CompileAssemblyFromDom(parameters, unit);
					if (results.Errors.HasErrors || results.Errors.HasWarnings)
					{
						compileResult.errors = results.Errors;
						return compileResult;
					}

					var dynamicAssembly = dynamicCompiler.DynamicLoad(generatedAssembly, generatedPdb);
					var transformTypes = dynamicAssembly.GetTypes().Where(t => typeof(Transformer).IsAssignableFrom(t)).Select(t => t.FullName);
					compileResult.types = new ArrayList<string>(transformTypes);
					compileResult.assembly = dynamicAssembly;
				}
			}
			finally
			{
				if (fileService.Exists(generatedCodeFile))
					fileService.Delete(generatedCodeFile);
				if (fileService.Exists(generatedAssembly))
					fileService.Delete(generatedAssembly);
				if (fileService.Exists(generatedPdb))
					fileService.Delete(generatedPdb);
			}

			return compileResult;
		}
	}
}
