﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using MetaSharp.Transformation.Grammar;
using MetaSharp.Transformation;
using System.IO;

namespace MetaSharp.Build.Grammar
{
	public class GrammarBuildProvider : IGrammarBuildProvider
	{
		private IActionQueue actionQueue;
		private GrammarParser parser = new GrammarParser();
		private Dictionary<string, string> documentProjects = new Dictionary<string, string>();
		private Dictionary<string, List<string>> projects = new Dictionary<string, List<string>>();
		private Dictionary<string, CompileResult> documents = new Dictionary<string, CompileResult>();

		public GrammarBuildProvider(IActionQueue actionQueue)
		{
			this.actionQueue = actionQueue;
		}

		public string GetDocumentProject(string documentPath)
		{
			var projectPath = default(string);
			if (documentProjects.TryGetValue(documentPath, out projectPath))
				return projectPath;

			return null;
		}

		public void Update(string documentPath, Func<string> text)
		{
			var lastResult = default(CompileResult);
			if (documents.TryGetValue(documentPath, out lastResult))
			{
				if (lastResult.Updating)
				{
					lastResult.NeedsUpdate = true;
				}
				else
				{
					if (text == null)
						text = lastResult.Text;

					var transformer = this.GetTransformer(documentPath);
					if (transformer != null)
					{
						lastResult.Updating = true;
						lastResult.NeedsUpdate = false;
						var code = text();
						this.actionQueue.Enqueue("Grammar",
							() =>
							{
								var context = new PatternContext();
								var result = transformer.Transform(context, new StringStream(code));
								var compileResult = new CompileResult { Context = context, Result = result, Text = text };
								return compileResult;
							},
							r =>
							{
								var result = (CompileResult)r.Value;
								if (r.Error != null)
								{
									throw r.Error;
								}
								else
								{
									documents[documentPath] = result;
									if (lastResult.NeedsUpdate)
										Update(documentPath, text);
								}

								this.BuildComplete(documentPath, result);
							});
					}
					else
					{
						this.BuildComplete(documentPath, new CompileResult());
					}
				}
			}
		}

		protected virtual Transformer GetTransformer(string projectItemPath)
		{
			return this.parser;
		}

		public IEnumerable<RuleMetadata> GetMetadata(string documentPath)
		{
			var result = documents[documentPath];
			if (result.Context != null && result.Result != null)
				return result.Context.Memos.GetMetadata(result.Result);

			return Enumerable.Empty<RuleMetadata>();
		}

		public IEnumerable<Error> GetErrors(string documentPath)
		{
			var result = documents[documentPath];
			if (result.Context != null)
				return result.Context.Errors;

			return Enumerable.Empty<Error>();
		}

		public INode GetResult(string documentPath)
		{
			if (documents.ContainsKey(documentPath))
				return documents[documentPath].Result;

			return null;
		}

		public IEnumerable<INode> GetResults(string projectPath)
		{
			List<string> items = null;
			if (this.projects.TryGetValue(projectPath, out items))
			{
				foreach (var key in items)
					yield return documents[key].Result;
			}
		}

		public string GetText(string documentPath)
		{
			return documents[documentPath].Text();
		}

		public event Action<string, CompileResult> BuildComplete = (b, r) => { };
		public event Action<string> Closing = b => { };
		public event Action<string, string> Renamed = (o, n) => { };

		public void Add(string projectPath, string documentPath, Func<string> text)
		{
			List<string> items = null;
			if (!this.projects.TryGetValue(projectPath, out items))
				this.projects.Add(projectPath, items = new List<string>());

			if (!items.Contains(documentPath))
			{
				items.Add(documentPath);
				documents.Add(documentPath, new CompileResult { Text = text });
				documentProjects.Add(documentPath, projectPath);
				Update(documentPath, text);
			}
		}

		public void Remove(string documentPath)
		{
			var projectPath = default(string);
			if (documentProjects.TryGetValue(documentPath, out projectPath))
			{
				List<string> items = projects[projectPath];

				documents.Remove(documentPath);
				documentProjects.Remove(documentPath);
				items.Remove(documentPath);
				if (items.Count == 0)
					projects.Remove(projectPath);
			}
		}

		public void Rename(string oldDocumentPath, string newDocumentPath)
		{
			var projectPath = default(string);
			if (documentProjects.TryGetValue(oldDocumentPath, out projectPath))
			{
				List<string> items = projects[projectPath];

				var buildResult = documents[oldDocumentPath];
				documents.Remove(oldDocumentPath);
				documentProjects.Remove(oldDocumentPath);
				items.Remove(oldDocumentPath);

				documents.Add(newDocumentPath, buildResult);
				documentProjects.Add(newDocumentPath, projectPath);
				items.Add(newDocumentPath);
			}

			Renamed(oldDocumentPath, newDocumentPath);
		}
	}

	public class CompileResult
	{
		public PatternContext Context;
		public INode Result;
		public bool Updating;
		public bool NeedsUpdate;
		public Func<string> Text;
	}
}
