﻿namespace MetaSharp.Build
{
	using System;
	using System.CodeDom.Compiler;
	using System.Collections.Generic;
	using System.Diagnostics;
	using System.IO;
	using System.Linq;
	using System.Reflection;
	using Microsoft.Build.Framework;
	using Microsoft.Build.Utilities;
	using Resources;
	using Transformation;
	using Transformation.Lang.Ast;
	using System.Text.RegularExpressions;

	public class MetaSharpBuildTask : Task
	{
		private IFileService fileService;

		public MetaSharpBuildTask()
			: this(new DefaultFileService())
		{ }

		public MetaSharpBuildTask(IFileService fileService)
		{
			this.fileService = fileService;
		}

		/// <summary>
		/// Gets or sets the current project path
		/// </summary>
		[Required]
		public ITaskItem ProjectPath { get; set; }

		/// <summary>
		/// Gets or sets the language of the current project (i.e. 'C#' or 'VB').
		/// </summary>
		[Required]
		public ITaskItem Language { get; set; }

		/// <summary>
		/// Gets or sets the configuration of the current build (i.e. 'Debug' or 'Release').
		/// </summary>
		[Required]
		public ITaskItem Configuration { get; set; }

		/// <summary>
		/// Gets or sets the default namespace of the current project.
		/// </summary>
		[Required]
		public ITaskItem DefaultNamespace { get; set; }

		/// <summary>
		/// Gets or sets the path at which to put intermediate files.
		/// </summary>
		[Required]
		public ITaskItem IntermediateOutputPath { get; set; }

		/// <summary>
		/// Gets or sets the name of the assembly being created.
		/// </summary>
		[Required]
		public ITaskItem AssemblyName { get; set; }

		/// <summary>
		/// Gets or sets the type name of the transformer that will be used.
		/// </summary>
		public ITaskItem TransformerTypeName { get; set; }

		/// <summary>
		/// Gets or sets the assemblies to load for the specified transformer.
		/// </summary>
		public ITaskItem[] TransformerAssemblies { get; set; }

		/// <summary>
		/// Gets or sets the input files used as source by this task.
		/// </summary>
		[Required]
		public ITaskItem[] InputFiles { get; set; }

		/// <summary>
		/// Gets the Generated compileable files generated by this task.
		/// </summary>
		[Output]
		public ITaskItem[] GeneratedCompiles
		{
			get { return this.generatedCompiles.ToArray(); }
		}

		private IList<ITaskItem> generatedCompiles = new List<ITaskItem>();

		public override bool Execute()
		{
			if (this.InputFiles == null || this.InputFiles.Count() == 0)
			{
				this.Log.LogError(
					Strings.MetaSharp,
					"MS1001",
					String.Empty,
					String.Empty,
					0,
					0,
					0,
					0,
					BuildErrors.MS1001);

				return false;
			}

			if (this.Language == null || String.IsNullOrWhiteSpace(this.Language.ItemSpec))
			{
				this.Log.LogError(
					Strings.MetaSharp,
					"MS1002",
					String.Empty,
					String.Empty,
					0,
					0,
					0,
					0,
					BuildErrors.MS1002);

				return false;
			}

			if (this.Configuration == null || String.IsNullOrWhiteSpace(this.Configuration.ItemSpec))
			{
				this.Log.LogError(
					Strings.MetaSharp,
					"MS1003",
					String.Empty,
					String.Empty,
					0,
					0,
					0,
					0,
					BuildErrors.MS1003);

				return false;
			}

			// Find the provider based on the current projects language.
			if (!CodeDomProvider.IsDefinedLanguage(this.Language.ItemSpec))
			{
				this.Log.LogError(
					Strings.MetaSharp,
					"MS1004",
					String.Empty,
					String.Empty,
					0,
					0,
					0,
					0,
					String.Format(BuildErrors.MS1004, this.Language.ItemSpec));

				return false;
			}

			var errors = false;
			var provider = CodeDomProvider.CreateProvider(this.Language.ItemSpec);

			using (var generatedFileManager = new GeneratedFileManager(
				this.fileService,
				Strings.MetaSharp,
				ProjectPath.GetMetadata("FullPath"),
				this.IntermediateOutputPath.ItemSpec,
				provider.FileExtension,
				this.Log,
				false))
			{
				var currentFilePath = default(string);

				try
				{
					foreach (var inputFile in this.InputFiles)
					{
						currentFilePath = inputFile.ItemSpec;

						if (generatedFileManager.IsFileRegenerationNeeded(inputFile.GetMetadata("FullPath")))
						{
							this.Log.LogMessage(String.Format(Strings.TransformingFile, currentFilePath));

							var code = this.fileService.ReadAllText(currentFilePath);
							var context = new PatternContext();
							context.Register<IBuildContext>(new MSBuildContext(
								this.DefaultNamespace.ItemSpec,
								this.Configuration.ItemSpec,
								this.Language.ItemSpec,
								this.AssemblyName.ItemSpec,
								inputFile.ItemSpec));

							Transformer transformer = new GrammarPipeline();

							if (this.TransformerTypeName != null && !string.IsNullOrEmpty(this.TransformerTypeName.ItemSpec))
							{
								if (this.TransformerAssemblies != null)
								{
									foreach (var reference in this.TransformerAssemblies)
										Assembly.LoadFrom(reference.ItemSpec);
								}

								Type transformerType = Type.GetType(this.TransformerTypeName.ItemSpec);
								transformer = (Transformer)Activator.CreateInstance(transformerType);
							}

							var ast = transformer.Transform(context, new StringStream(code));

							foreach (var error in context.Errors)
							{
								this.LogError(new SourceError(error, code, currentFilePath));
							}

							if (context.Errors.Any())
								errors = true;

							Unit unit = ast as Unit;
							if (unit == null)
							{
								unit = new Unit();
								Namespace ns = ast as Namespace;
								if (ns == null)
								{
									ns = new Namespace(this.DefaultNamespace.ItemSpec);
									TypeDeclaration td = ast as TypeDeclaration;

									if (td != null)
										ns.Types.Add(td);
								}

								unit.Namespaces.Add(ns);
							}

							if (unit != null)
							{
								var generatedCode = unit.Print(provider);
								var generatedFilePath = generatedFileManager.AddGeneratedFile(inputFile.GetMetadata("FullPath"), generatedCode);
								generatedCompiles.Add(new TaskItem(generatedFilePath));
							}
							else
							{
								// TODO: Handle other generated file types, such as Content.
							}
						}
						else
						{
							this.Log.LogMessage(String.Format(Strings.SkipTransformingFile, currentFilePath));
							var generatedFilePath = generatedFileManager.GetGeneratedFilename(inputFile.GetMetadata("FullPath"));
							generatedCompiles.Add(new TaskItem(generatedFilePath));
						}
					}

					generatedFileManager.UpdateGrammarTransformerLastWriteTime();
				}
				catch (Exception ex)
				{
					Log.LogErrorFromException(ex, true, true, currentFilePath);
					return false;
				}
			}

			return !errors;
		}

		private void LogError(SourceError error)
		{
			switch (error.Severity)
			{
				case ErrorSeverity.Error:
					this.Log.LogError(
						error.Subcategory,
						error.ErrorCode,
						error.HelpKeyword,
						error.File,
						error.LineNumber,
						error.ColumnNumber,
						error.EndLineNumber,
						error.EndColumnNumber,
						error.Message);
					break;
				case ErrorSeverity.Warning:
					this.Log.LogWarning(
						error.Subcategory,
						error.ErrorCode,
						error.HelpKeyword,
						error.File,
						error.LineNumber,
						error.ColumnNumber,
						error.EndLineNumber,
						error.EndColumnNumber,
						error.Message);
					break;
				case ErrorSeverity.Info:
					this.Log.LogMessage(MessageImportance.Low, error.Message);
					break;
				default:
					throw new NotSupportedException("invalid severity.");
			}
		}

		private static string Escape(char c)
		{
			switch (c)
			{
				case '\'':
					return "'";
				case '\"':
					return "\\\"";
				case '\\':
					return "\\\\";
				case '\0':
					return "\\0";
				case '\a':
					return "\\a";
				case '\b':
					return "\\b";
				case '\f':
					return "\\f";
				case '\n':
					return "\\n";
				case '\r':
					return "\\r";
				case '\t':
					return "\\t";
				case '\v':
					return "\\v";
				default:
					return c.ToString();
			}
		}
	}
}
