﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.VisualStudio.Text.Tagging;
using MetaSharp.Transformation;
using Microsoft.VisualStudio.Text;
using MetaSharp.Build.Grammar;
using MetaSharp.Build;

namespace MetaSharp.Integration.VisualStudio.Grammar
{
	public abstract class MetaSharpTagger<TMetadata, TTag> : ITagger<TTag>
		where TMetadata : IHaveSourceSpan
		where TTag : ITag
	{
		string documentPath;
		ITextBuffer buffer;
		IActionQueue queue;
		IGrammarBuildProvider buildProvider;
		List<Tag> lastTags = new List<Tag>();

		public MetaSharpTagger(ITextBuffer buffer, IActionQueue queue, IGrammarBuildProvider buildProvider)
		{
			this.buffer = buffer;
			this.queue = queue;
			this.buildProvider = buildProvider;
			this.buildProvider.BuildComplete += OnBuildProviderBuildComplete;
			this.buildProvider.Closing += OnBuildProvderClosing;

			var textDocument = (ITextDocument)buffer.Properties[typeof(ITextDocument)];
			this.documentPath = textDocument.FilePath;
		}

		private void OnBuildProvderClosing(string documentPath)
		{
			if (this.documentPath == documentPath)
			{
				this.buildProvider.BuildComplete -= OnBuildProviderBuildComplete;
				this.buildProvider.Closing -= OnBuildProvderClosing;
			}
		}

		public event EventHandler<SnapshotSpanEventArgs> TagsChanged = (o, e) => { };

		protected abstract IEnumerable<TMetadata> GetMetadata(CompileResult result);
		protected abstract IEnumerable<Tag> CreateTags(int start, int end, TMetadata metadata);
		protected abstract TTag CreateMetadataTag(Tag tag);

		private void OnBuildProviderBuildComplete(string documentPath, CompileResult result)
		{
			if (this.documentPath == documentPath && result.Context != null)
			{
				var lastTags = this.lastTags.ToArray();
				var snapshot = buffer.CurrentSnapshot;
				var metadata = GetMetadata(result);
				var tags = new List<Tag>();
				foreach (var m in metadata)
				{
					tags.AddRange(CreateTags(
						(int)((IIndexed)m.Span.Start).Index,
						(int)((IIndexed)m.Span.End).Index,
						m));
				}

				tags.Sort((l, r) => l.Start.CompareTo(r.Start));

				var changes = new List<Tag>(tags.Concat(lastTags));
				int x = 0;
				int y = 0;
				while (x < tags.Count && y < lastTags.Length)
				{
					var newH = tags[x];
					var oldH = lastTags[y];

					if (newH.Start > oldH.Start)
					{
						y++;
						continue;
					}

					if (oldH.Start > newH.Start)
					{
						x++;
						continue;
					}

					x++;
					y++;
					if (newH.Equals(oldH))
					{
						changes.Remove(newH);
						changes.Remove(oldH);
					}
				}

				var lines = new HashSet<int>();
				foreach (var h in changes)
				{
					var start = snapshot.GetLineNumberFromPosition(Math.Min(h.Start, snapshot.Length));
					var end = snapshot.GetLineNumberFromPosition(Math.Min(h.End, snapshot.Length));
					lines.Add(start);
					lines.Add(end);

					foreach (var line in Enumerable.Range(start, end))
						lines.Add(line);
				}

				foreach (var line in lines)
				{
					if (line < snapshot.LineCount)
					{
						var snapshotLine = snapshot.GetLineFromLineNumber(line);
						if (snapshotLine != null)
							this.TagsChanged(this, new SnapshotSpanEventArgs(snapshotLine.ExtentIncludingLineBreak));
					}
				}

				this.lastTags = tags;
			}
		}

		IEnumerable<ITagSpan<TTag>> ITagger<TTag>.GetTags(NormalizedSnapshotSpanCollection spans)
		{
			var lastIndex = 0;
			foreach (SnapshotSpan span in spans)
			{
				var start = span.Start.Position;
				var end = span.End.Position;
				for (var x = lastIndex; x < this.lastTags.Count; x++)
				{
					var t = this.lastTags[x];
					if ((t.Start >= start && t.Start < end) ||
						(t.End < end && t.End > start))
					{
						var i = Math.Max(t.Start, start);
						var len = Math.Min(Math.Max(t.End - t.Start, 1), Math.Max(end - i, 0));
						var tokenSpan = new SnapshotSpan(span.Snapshot, new Span(i, len));
						yield return new TagSpan<TTag>(
							tokenSpan,
							CreateMetadataTag(t));
					}

					lastIndex = x;
				}
			}
		}
	}
}
