﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.VisualStudio.ComponentModelHost;
using EnvDTE;
using System.IO;
using Microsoft.VisualStudio.Shell;
using Microsoft.VisualStudio.Shell.Interop;
using Microsoft.VisualStudio.Editor;
using MetaSharp.Build.Grammar;
using System.Diagnostics;
using EnvDTE80;

namespace MetaSharp.Integration.VisualStudio.Grammar
{
	class VisualStudioGrammarProvider : IDisposable
	{
		Solution currentSolution;
		EnvDTE.DTE dte;
		Events2 events;
		SolutionEvents solutionEvents; // store else GC will collect it and unhook the events.
		ProjectItemsEvents miscItemsEvents;
		ProjectItemsEvents solutionItemsEvents;
		ProjectItemsEvents projectItemsEvents;

		IVsEditorAdaptersFactoryService editorService;
		ServiceProvider serviceProvider;
		IGrammarBuildProvider buildProvider;

		Dictionary<uint, ProjectItem> items = new Dictionary<uint, ProjectItem>();

		public VisualStudioGrammarProvider(EnvDTE.DTE dte, IComponentModel componentModel)
		{
			this.dte = dte;
			this.events = dte.Events as Events2;
			this.solutionEvents = events.SolutionEvents;
			this.miscItemsEvents = events.MiscFilesEvents;
			this.solutionItemsEvents = events.SolutionItemsEvents;
			this.projectItemsEvents = events.ProjectItemsEvents;

			events.SolutionEvents.AfterClosing += SolutionEvents_AfterClosing;
			events.SolutionEvents.Opened += SolutionEvents_Opened;
			events.SolutionEvents.ProjectAdded += SolutionEvents_ProjectAdded;
			events.SolutionEvents.ProjectRemoved += SolutionEvents_ProjectRemoved;
			events.SolutionEvents.ProjectRenamed += SolutionEvents_ProjectRenamed;
			events.MiscFilesEvents.ItemAdded += MiscFilesEvents_ItemAdded;
			events.MiscFilesEvents.ItemRemoved += MiscFilesEvents_ItemRemoved;
			events.MiscFilesEvents.ItemRenamed += MiscFilesEvents_ItemRenamed;
			events.SolutionItemsEvents.ItemAdded += SolutionItemsEvents_ItemAdded;
			events.SolutionItemsEvents.ItemRemoved += SolutionItemsEvents_ItemRemoved;
			events.SolutionItemsEvents.ItemRenamed += SolutionItemsEvents_ItemRenamed;
			events.ProjectItemsEvents.ItemAdded += ProjectItemsEvents_ItemAdded;
			events.ProjectItemsEvents.ItemRemoved += ProjectItemsEvents_ItemRemoved;
			events.ProjectItemsEvents.ItemRenamed += ProjectItemsEvents_ItemRenamed;

			this.editorService = componentModel.GetService<IVsEditorAdaptersFactoryService>();
			this.buildProvider = componentModel.GetService<IGrammarBuildProvider>();
			this.serviceProvider = new ServiceProvider(MetaSharpPackage.OleServiceProvider);
			if (this.dte.Solution != null)
			{
				this.RegisterSolution(this.dte.Solution);
			}
		}

		void ProjectItemsEvents_ItemRenamed(ProjectItem projectItem, string oldName)
		{
			RenameGrammarFile(projectItem, oldName);
		}

		void ProjectItemsEvents_ItemRemoved(ProjectItem ProjectItem)
		{
			UnregisterProjectItem(ProjectItem);
		}

		void ProjectItemsEvents_ItemAdded(ProjectItem projectItem)
		{
			RegisterProjectItem(projectItem.ContainingProject, projectItem);
		}

		void SolutionItemsEvents_ItemRenamed(ProjectItem projectItem, string oldName)
		{
			RenameGrammarFile(projectItem, oldName);
		}

		void SolutionItemsEvents_ItemRemoved(ProjectItem projectItem)
		{
			UnregisterProjectItem(projectItem);
		}

		void SolutionItemsEvents_ItemAdded(ProjectItem projectItem)
		{
			RegisterProjectItem(projectItem.ContainingProject, projectItem);
		}

		void MiscFilesEvents_ItemRenamed(ProjectItem projectItem, string oldName)
		{
			RenameGrammarFile(projectItem, oldName);
		}

		void MiscFilesEvents_ItemRemoved(ProjectItem projectItem)
		{
			UnregisterProjectItem(projectItem);
		}

		void MiscFilesEvents_ItemAdded(ProjectItem projectItem)
		{
			RegisterProjectItem(projectItem.ContainingProject, projectItem);
		}

		void SolutionEvents_ProjectRenamed(Project Project, string OldName)
		{
			// TODO: implement project rename support...
			//throw new NotImplementedException();
		}

		void SolutionEvents_ProjectRemoved(Project project)
		{
			UnregisterProject(project);
		}

		void SolutionEvents_ProjectAdded(Project project)
		{
			RegisterProject(project);
		}

		void SolutionEvents_Opened()
		{
			if (this.currentSolution != null)
				UnregisterSolution(this.currentSolution);

			RegisterSolution(this.dte.Solution);
		}

		void SolutionEvents_AfterClosing()
		{
			UnregisterSolution(currentSolution);
			currentSolution = null;
		}

		private void UnregisterSolution(Solution solution)
		{
			foreach (Project project in solution.Projects)
			{
				UnregisterProject(project);
			}
		}

		private void RenameGrammarFile(ProjectItem item, string oldName)
		{
			var fullPathProperty = item.Properties.Item("FullPath");
			if (fullPathProperty != null)
			{
				var documentPath = (string)fullPathProperty.Value;
				var fullOldPath = Path.Combine(Path.GetDirectoryName(documentPath), oldName);
				var ext = Path.GetExtension(documentPath);
				if (ext == ".g")
					this.buildProvider.Rename(fullOldPath, documentPath);
			}
		}

		private void UnregisterProject(Project project)
		{
			if (project.ProjectItems != null)
			{
				foreach (ProjectItem item in project.ProjectItems)
				{
					UnregisterProjectItem(item);
				}
			}
		}

		private void UnregisterProjectItem(ProjectItem item)
		{
			if (item.Properties != null)
			{
				var fullPathProperty = item.Properties.Item("FullPath");
				if (fullPathProperty != null)
				{
					var documentPath = (string)fullPathProperty.Value;
					var ext = Path.GetExtension(documentPath);
					if (ext == ".g")
						this.buildProvider.Remove(documentPath);
				}
			}
		}

		private void RegisterSolution(Solution solution)
		{
			this.currentSolution = solution;
			foreach (Project project in solution.Projects)
			{
				RegisterProject(project);
			}
		}

		private void RegisterProject(Project project)
		{
			if (project.ProjectItems != null)
			{
				foreach (ProjectItem item in project.ProjectItems)
				{
					RegisterProjectItem(project, item);
				}
			}
		}

		private void RegisterProjectItem(Project project, ProjectItem item)
		{
			if (item.Properties != null)
			{
				var fullPathProperty = item.Properties.Item("FullPath");
				if (fullPathProperty != null)
				{
					var documentPath = (string)fullPathProperty.Value;
					var ext = Path.GetExtension(documentPath);
					if (ext == ".g")
					{
						IVsUIHierarchy uiHierarchy;
						uint itemID;
						IVsWindowFrame windowFrame;

						// open documents will be tracked using the other build providers.
						if (!VsShellUtilities.IsDocumentOpen(
							serviceProvider,
							documentPath,
							Guid.Empty,
							out uiHierarchy,
							out itemID,
							out windowFrame) &&
							File.Exists(documentPath))
						{
							var code = new Func<string>(() => File.ReadAllText(documentPath));
							this.buildProvider.Add(project.FullName, documentPath, code);
						}
					}
				}
			}
		}

		public void Dispose()
		{
			events.SolutionEvents.AfterClosing -= SolutionEvents_AfterClosing;
			events.SolutionEvents.Opened -= SolutionEvents_Opened;
			events.SolutionEvents.ProjectAdded -= SolutionEvents_ProjectAdded;
			events.SolutionEvents.ProjectRemoved -= SolutionEvents_ProjectRemoved;
			events.SolutionEvents.ProjectRenamed -= SolutionEvents_ProjectRenamed;
			events.MiscFilesEvents.ItemAdded -= MiscFilesEvents_ItemAdded;
			events.MiscFilesEvents.ItemRemoved -= MiscFilesEvents_ItemRemoved;
			events.MiscFilesEvents.ItemRenamed -= MiscFilesEvents_ItemRenamed;
			events.SolutionItemsEvents.ItemAdded -= SolutionItemsEvents_ItemAdded;
			events.SolutionItemsEvents.ItemRemoved -= SolutionItemsEvents_ItemRemoved;
			events.SolutionItemsEvents.ItemRenamed -= SolutionItemsEvents_ItemRenamed;
			events.ProjectItemsEvents.ItemAdded -= ProjectItemsEvents_ItemAdded;
			events.ProjectItemsEvents.ItemRemoved -= ProjectItemsEvents_ItemRemoved;
			events.ProjectItemsEvents.ItemRenamed -= ProjectItemsEvents_ItemRenamed;

			this.dte = null;
			this.events = null;
			this.solutionEvents = null;
			this.miscItemsEvents = null;
			this.solutionItemsEvents = null;
			this.projectItemsEvents = null;
		}
	}
}
