﻿namespace MetaSharp.Integration.VisualStudio
{
	using System;
	using System.Linq;
	using System.ComponentModel.Design;
	using System.Runtime.InteropServices;
	using MetaSharp.Integration.VisualStudio.Commands;
	using Microsoft.VisualStudio;
	using Microsoft.VisualStudio.Shell;
	using EnvDTE;
	using Microsoft.VisualStudio.Shell.Interop;
	using System.Diagnostics;
	using System.IO;
	using Microsoft.VisualStudio.Utilities;
	using System.ComponentModel.Composition;
	using MetaSharp.Integration.VisualStudio.Grammar;
	using MetaSharp.Integration.VisualStudio.Views;
	using Microsoft.VisualStudio.OLE.Interop;
	using Microsoft.VisualStudio.ComponentModelHost;
	using MetaSharp.Build.Grammar;
	using Microsoft.VisualStudio.Editor;
	using MetaSharp.Build;
	using System.Runtime.Serialization.Formatters.Binary;

	[PackageRegistration(UseManagedResourcesOnly = true)]
	[InstalledProductRegistration("#110", "#112", "0.1", IconResourceID = 400)]
	[ProvideMenuResource("Menus.ctmenu", 1)]
	[ProvideToolWindow(typeof(GrammarToolWindow))]
	[Guid(GuidList.guidMetaSharpPackagePkgString)]
	[ProvideAutoLoad(VSConstants.UICONTEXT.SolutionExistsAndFullyLoaded_string)]
	public sealed class MetaSharpPackage : Package
	{
		private VisualStudioGrammarProvider grammarProvider;
		private static Microsoft.VisualStudio.OLE.Interop.IServiceProvider cachedOleServiceProvider;

		private static IGrammarBuildProvider grammarBuildProvider;
		private static IDslBuildProvider dslBuildProvider;
		private static IActionQueue actionQueue;
		private static IErrorProvider errorProvider;
		private static IMetaSharpOptions options;
		private static IFileService fileService;
		private static IDynamicCompiler dynamicCompiler;

		[Export]
		public IGrammarBuildProvider GrammarBuildProvider
		{
			get
			{
				if (grammarBuildProvider == null)
					grammarBuildProvider = new GrammarBuildProvider(this.ActionQueue);
				return grammarBuildProvider;
			}
		}

		[Export]
		public IDslBuildProvider DslBuildProvider
		{
			get
			{
				if (dslBuildProvider == null)
					dslBuildProvider = new DslBuildProvider(this.ActionQueue, this.GrammarBuildProvider, this.DynamicCompiler, this.ErrorProvider, this.FileService);
				return dslBuildProvider;
			}
		}

		[Export]
		public IActionQueue ActionQueue
		{
			get
			{
				if (actionQueue == null)
					actionQueue = new ActionQueue();
				return actionQueue;
			}
		}

		[Export]
		public IErrorProvider ErrorProvider
		{
			get
			{
				if (errorProvider == null)
					errorProvider = new VisualStudioErrorProvider();
				return errorProvider;
			}
		}

		[Export]
		public IMetaSharpOptions Options
		{
			get
			{
				if (options == null)
					options = new MetaSharpOptions();
				return options;
			}
		}

		[Export]
		public IFileService FileService
		{
			get
			{
				if (fileService == null)
					fileService = new DefaultFileService();
				return fileService;
			}
		}

		public IDynamicCompiler DynamicCompiler
		{
			get
			{
				if (dynamicCompiler == null)
					dynamicCompiler = new CciDynamicCompiler(this.FileService);
				return dynamicCompiler;
			}
		}

		public DTE DTE { get; private set; }
		public IVsSolution Solution { get; private set; }

		static MetaSharpPackage()
		{
			AppDomain.CurrentDomain.AssemblyResolve += OnCurrentDomainAssemblyResolve;
		}

		static System.Reflection.Assembly OnCurrentDomainAssemblyResolve(object sender, ResolveEventArgs args)
		{
			var found = AppDomain.CurrentDomain
				.GetAssemblies()
				.Where(a => a.FullName == args.Name)
				.FirstOrDefault();

			return found;
		}

		public MetaSharpPackage()
		{
			this.AddOptionKey("MetaSharp");
		}

		#region Package Members

		protected override void Initialize()
		{
			base.Initialize();

			DTE = GetService<SDTE, DTE>();
			Solution = GetService<SVsSolution, IVsSolution>();

			RegisterCommands();
			InitializeGrammarBuilder();
		}

		protected override void Dispose(bool disposing)
		{
			if (disposing)
				this.grammarProvider.Dispose();

			base.Dispose(disposing);
		}

		private void RegisterCommands()
		{
			var menuCommandService = GetService<IMenuCommandService, OleMenuCommandService>();
			menuCommandService.AddCommand(new MetaSharpConsoleCommand(this).MenuCommand);
		}

		private void InitializeGrammarBuilder()
		{
			var componentModel = GetService<SComponentModel, IComponentModel>();
			this.grammarProvider = new VisualStudioGrammarProvider(DTE, componentModel);
		}

		#endregion

		public TResult GetService<T, TResult>() where TResult : class
		{
			var result = GetService(typeof(T)) as TResult;

			Debug.Assert(result != null);

			return result;
		}

		public TResult GetGlobalService<T, TResult>() where TResult : class
		{
			var result = GetGlobalService(typeof(T)) as TResult;

			Debug.Assert(result != null);

			return result;
		}

		/// <summary>
		/// The shell's service provider as an OLE service provider (needed to create the editor bits).
		/// </summary>
		internal static Microsoft.VisualStudio.OLE.Interop.IServiceProvider OleServiceProvider
		{
			get
			{
				if (cachedOleServiceProvider == null)
				{
					//ServiceProvider.GlobalProvider is a System.IServiceProvider, but the editor pieces want an OLE.IServiceProvider, luckily the
					//global provider is also IObjectWithSite and we can use that to extract its underlying (OLE) IServiceProvider object.
					var objWithSite = (IObjectWithSite)ServiceProvider.GlobalProvider;

					Guid interfaceIID = typeof(Microsoft.VisualStudio.OLE.Interop.IServiceProvider).GUID;
					IntPtr rawSP;
					objWithSite.GetSite(ref interfaceIID, out rawSP);
					try
					{
						if (rawSP != IntPtr.Zero)
						{
							//Get an RCW over the raw OLE service provider pointer.
							cachedOleServiceProvider = (Microsoft.VisualStudio.OLE.Interop.IServiceProvider)Marshal.GetObjectForIUnknown(rawSP);
						}
					}
					finally
					{
						if (rawSP != IntPtr.Zero)
						{
							//Release the raw pointer we got from IObjectWithSite so we don't cause leaks.
							Marshal.Release(rawSP);
						}
					}
				}

				return cachedOleServiceProvider;
			}
		}

		protected override void OnLoadOptions(string key, Stream stream)
		{
			if (key == "MetaSharp")
				this.Options.Load(stream);

			base.OnLoadOptions(key, stream);
		}

		protected override void OnSaveOptions(string key, Stream stream)
		{
			if (key == "MetaSharp")
				this.Options.Save(stream);

			base.OnSaveOptions(key, stream);
		}
	}
}
