﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using MetaSharp.Transformation;
using MetaSharp.Integration.VisualStudio.Grammar;
using System.Collections.ObjectModel;
using Microsoft.VisualStudio.Text;
using System.CodeDom;
using MetaSharp.Build;
using MetaSharp.Build.Grammar;
using System.CodeDom.Compiler;

namespace MetaSharp.Integration.VisualStudio.Models
{
	public class GrammarHost : ObservableObject
	{
		private INode node;
		private string documentPath;
		private string selectedType;
		private ObservableCollection<HostAction> actions = new ObservableCollection<HostAction>();
		private ObservableCollection<string> transformTypes = new ObservableCollection<string>();
		private IErrorProvider errorProvider;
		private IDslBuildProvider buildProvider;
		private IMetaSharpOptions options;

		public GrammarHost(IErrorProvider errorProvider, IDslBuildProvider buildProvider, IMetaSharpOptions options)
		{
			this.errorProvider = errorProvider;
			this.errorProvider.Changed += ErrorsChanged;

			this.buildProvider = buildProvider;
			this.buildProvider.BuildComplete += UpdateDocument;

			this.options = options;
		}

		public INode Node
		{
			get { return this.node; }
			set { this.SetAndNotify(ref this.node, value, () => Node); }
		}

		public ObservableCollection<HostAction> Actions
		{
			get { return this.actions; }
		}

		public ObservableCollection<string> TransformTypes
		{
			get { return this.transformTypes; }
		}

		public bool TransformTypeChoice
		{
			get { return this.transformTypes.Count > 1; }
		}

		public string SelectedType
		{
			get { return this.selectedType; }
			set
			{
				this.SetAndNotify(ref this.selectedType, value, () => SelectedType);
				this.buildProvider.SetTransformType(this.documentPath, value);

				if (value != null)
				{
					var key = DslBuildProvider.GetMd5Hash(documentPath);
					if (!this.options.SelectedGrammars.ContainsKey(key))
						this.options.SelectedGrammars.Add(key, value);
					else
						this.options.SelectedGrammars[key] = value;
				}
			}
		}

		public void ErrorsChanged()
		{
			this.BeginInvoke(() =>
			{
				if (errorProvider.Any())
				{
					if (!actions.Any())
						actions.Add(new ErrorAction(this.errorProvider));
				}
				else if (actions.Any())
				{
					actions.Clear();
				}
			});
		}

		public void LoadDocument(string filePath)
		{
			this.buildProvider.Update(filePath, null);
		}

		private void UpdateDocument(string documentPath, CompileResult result)
		{
			this.documentPath = documentPath;
			this.Node = result.Result;
			this.UpdateErrors(documentPath, result);
			this.UpdateTransformTypes(documentPath);

			var key = DslBuildProvider.GetMd5Hash(documentPath);
			var val = default(string);
			if (this.options.SelectedGrammars.TryGetValue(key, out val))
			{
				this.SelectedType = val;
			}
			else if (this.transformTypes.Any() &&
				(string.IsNullOrEmpty(this.SelectedType) || !this.transformTypes.Contains(this.SelectedType)))
			{
				this.SelectedType = this.transformTypes[0];
			}
			else if (string.IsNullOrEmpty(documentPath))
			{
				this.SelectedType = null;
			}
		}

		public void ClearDocument()
		{
			this.Node = null;
			this.errorProvider.Clear(this.documentPath);
			this.transformTypes.Clear();

			this.OnPropertyChanged(() => this.TransformTypeChoice);
		}

		private void UpdateErrors(string documentPath, CompileResult result)
		{
			if (documentPath == null)
				throw new ArgumentNullException("documentPath");

			if (result.Context != null)
			{
				var errors = result.Context.Errors;
				var code = this.buildProvider.GetText(documentPath);
				errorProvider.Clear(documentPath);
				foreach (var e in errors)
				{
					var sourceError = new SourceError(e, code, documentPath);
					errorProvider.Add(documentPath, sourceError);
				}
			}
			else
			{
				errorProvider.Clear(documentPath);
				errorProvider.Add(
					documentPath,
					new SourceError(new CompilerError(
						documentPath,
						0, 0,
						"0",
						"Unable to find a valid Transformer.")));
			}
		}

		public void UpdateTransformTypes(string documentPath)
		{
			var newTypes = this.buildProvider.GetTransformTypes(documentPath);
			foreach (var type in this.transformTypes.Where(t => !newTypes.Contains(t)).ToArray())
				this.transformTypes.Remove(type);

			foreach (var type in newTypes.Where(t => !this.transformTypes.Contains(t)))
				this.transformTypes.Add(type);

			this.OnPropertyChanged(() => this.TransformTypeChoice);
		}
	}
}
