﻿using System;
using System.Linq;
using System.Text;
using System.IO;
using System.Collections;
using MetaSharp.Transformation;

namespace MetaSharp.Integration.VisualStudio.Models
{
	static class GraphPrinter
	{
		public static string Print(object node)
		{
			var builder = new StringBuilder();
			using (var writer = new StringWriter(builder))
				Print(writer, Node.Unwrap(node));
			return builder.ToString();
		}

		private static void Print(StringWriter writer, object node)
		{
			if (node == null)
			{
				writer.Write("(null)");
				return;
			}

			if (node is Fail)
			{
				writer.Write("(fail)");
				return;
			}

			var str = node as string;
			if (str != null)
			{
				writer.Write("\"");
				writer.Write(str);
				writer.Write("\"");
				return;
			}

			if (node is char)
			{
				writer.Write("'");
				char c = (char)node;
				switch (c)
				{
					case '\t':
						writer.Write("\\t");
						break;
					case '\r':
						writer.Write("\\r");
						break;
					case '\n':
						writer.Write("\\n");
						break;
					case '\b':
						writer.Write("\\b");
						break;
					default:
						writer.Write((char)node);
						break;
				}
				writer.Write("'");
				return;
			}


			if (node.GetType().IsPrimitive)
			{
				writer.Write(node.ToString());
				return;
			}

			var type = node.GetType();
			var typename = (type.Namespace + "." + type.Name)
				.Replace("__N__.", "")
				.Replace("MetaSharp.Transformation.", "");

			var enumerable = node as IEnumerable;
			var dictionary = node as IDictionary;
			if (type.IsEnum)
			{
				writer.Write(typename);
				if (Enum.IsDefined(type, node))
					writer.Write(Enum.GetName(type, node));
			}
			else if (dictionary != null)
			{
				writer.Write(" {");
				writer.Indent();
				int x = 0;
				foreach (var key in dictionary.Keys)
				{
					if (x++ > 0)
						writer.WriteLine(",");
					else
						writer.WriteLine();

					writer.Write(" {");
					Print(writer, key);
					writer.Write(" = ");
					Print(writer, Node.Unwrap(dictionary[key]));
					writer.Write(" }");
				}

				writer.Dedent();
				writer.WriteLine();
				writer.Write("}");
			}
			else if (enumerable != null)
			{
				bool allPrimitives = enumerable.Cast<object>().All(o => o == null || o.GetType().IsPrimitive);
				writer.Indent();
				writer.Write("[");
				if (allPrimitives)
					writer.WriteLine();

				int x = 0;
				foreach (var item in enumerable)
				{
					if (x++ > 0)
						writer.Write(", ");

					if (allPrimitives)
						writer.WriteLine();

					Print(writer, Node.Unwrap(item));
				}
				writer.Dedent();
				writer.Write("]");
				return;
			}
			else
			{
				writer.Write(typename);

				writer.Write(" {");
				writer.Indent();
				int x = 0;
				foreach (var property in node.GetType().GetProperties())
				{
					if (x++ > 0)
						writer.WriteLine(",");
					else
						writer.WriteLine();

					writer.Write(property.Name);
					writer.Write(" = ");
					Print(writer, Node.Unwrap(property.GetValue(node, null)));
				}

				foreach (var field in node.GetType().GetFields(System.Reflection.BindingFlags.Public | System.Reflection.BindingFlags.Instance))
				{
					if (x++ > 0)
						writer.WriteLine(",");
					else
						writer.WriteLine();

					writer.Write(field.Name);
					writer.Write(" = ");
					Print(writer, Node.Unwrap(field.GetValue(node)));
				}
				writer.Dedent();
				writer.WriteLine();
				writer.Write("}");
			}
		}

		private static void Indent(this StringWriter writer)
		{
			writer.NewLine += " ";
		}

		private static void Dedent(this StringWriter writer)
		{
			if (writer.NewLine.EndsWith(" "))
				writer.NewLine = writer.NewLine.Remove(writer.NewLine.Length - 1, 1);
		}
	}
}
