﻿using System;
using System.Diagnostics;
using System.Runtime.InteropServices;
using System.Windows;
using System.Windows.Controls;
using Microsoft.VisualStudio;
using Microsoft.VisualStudio.ComponentModelHost;
using Microsoft.VisualStudio.Editor;
using Microsoft.VisualStudio.OLE.Interop;
using Microsoft.VisualStudio.Shell;
using Microsoft.VisualStudio.Shell.Interop;
using Microsoft.VisualStudio.Text.Editor;
using Microsoft.VisualStudio.TextManager.Interop;
using Microsoft.VisualStudio.Text;
using System.IO;
using System.Security.Cryptography;
using System.Text;
using System.ComponentModel.Composition;
using MetaSharp.Integration.VisualStudio.Models;
using MetaSharp.Integration.VisualStudio.Views;
using MetaSharp.Integration.VisualStudio.Grammar;
using System.Linq;
using MetaSharp.Build;
using System.Collections.Generic;
using EnvDTE;
using MetaSharp.Build.Grammar;
using System.Windows.Threading;

namespace MetaSharp.Integration.VisualStudio.Views
{
	[Guid("feaa0ebb-adb8-475e-b907-fba301db8e44")]
	public class GrammarToolWindow :
		ToolWindowPane,
		IOleCommandTarget,
		IVsFindTarget,
		IVsFindTarget2,
		IVsRunningDocTableEvents,
		IVsRunningDocTableEvents4
	{
		#region Constants

		private const int WM_KEYFIRST = 0x0100;
		private const int WM_KEYLAST = 0x0109;

		#endregion

		#region Private Fields

		GrammarHostView control;
		GrammarHost model;
		IOleCommandTarget cachedEditorCommandTarget;
		IVsTextView textView;
		IVsCodeWindow codeWindow;
		IVsInvisibleEditor invisibleEditor;
		IVsFindTarget cachedEditorFindTarget;
		EnvDTE.Window activeWindow;
		WindowEvents windowEvents;
		string displayedFilePath;
		uint docEventsCookie;

		#endregion

		#region Ctor

		public GrammarToolWindow()
			: base(null)
		{
			this.Caption = Properties.Resources.GrammarConsoleTitle; // Resources.ToolWindowTitle;
			this.BitmapResourceID = 301;
			this.BitmapIndex = 0;
		}

		#endregion

		#region Public Methods

		/// <summary>
		/// Method to call to cause us to place the file at the given path into our hosted editor.
		/// </summary>
		public void SetDisplayedFile(string filePath)
		{
			if (filePath != displayedFilePath)
			{
				displayedFilePath = filePath;
				ClearOldEditor();

				//Get an invisible editor over the file, this makes it much easier than having to manually figure out the right content type, 
				//language service, and it will automatically associate the document with its owning project, meaning we will get intellisense
				//in our editor with no extra work.
				IVsInvisibleEditorManager invisibleEditorManager = (IVsInvisibleEditorManager)GetService(typeof(SVsInvisibleEditorManager));
				ErrorHandler.ThrowOnFailure(invisibleEditorManager.RegisterInvisibleEditor(
					filePath,
					pProject: null,
					dwFlags: (uint)_EDITORREGFLAGS.RIEF_ENABLECACHING,
					pFactory: null,
					ppEditor: out this.invisibleEditor));

				//The doc data is the IVsTextLines that represents the in-memory version of the file we opened in our invisibe editor, we need
				//to extract that so that we can create our real (visible) editor.
				IntPtr docDataPointer = IntPtr.Zero;
				Guid guidIVSTextLines = typeof(IVsTextLines).GUID;
				ErrorHandler.ThrowOnFailure(this.invisibleEditor.GetDocData(fEnsureWritable: 1, riid: ref guidIVSTextLines, ppDocData: out docDataPointer));
				try
				{
					IVsTextLines docData = (IVsTextLines)Marshal.GetObjectForIUnknown(docDataPointer);

					//Get the component model so we can request the editor adapter factory which we can use to spin up an editor instance.
					IComponentModel componentModel = (IComponentModel)GetService(typeof(SComponentModel));
					IVsEditorAdaptersFactoryService editorAdapterFactoryService = componentModel.GetService<IVsEditorAdaptersFactoryService>();

					//Create a code window adapter.
					this.codeWindow = editorAdapterFactoryService.CreateVsCodeWindowAdapter(MetaSharpPackage.OleServiceProvider);

					//Disable the splitter control on the editor as leaving it enabled causes a crash if the user
					//tries to use it here :(
					IVsCodeWindowEx codeWindowEx = (IVsCodeWindowEx)this.codeWindow;
					INITVIEW[] initView = new INITVIEW[1];
					codeWindowEx.Initialize(
						(uint)_codewindowbehaviorflags.CWB_DISABLESPLITTER,
						VSUSERCONTEXTATTRIBUTEUSAGE.VSUC_Usage_Filter,
						szNameAuxUserContext: "",
						szValueAuxUserContext: "",
						InitViewFlags: 0,
						pInitView: initView);

					//Associate our IVsTextLines with our new code window.
					ErrorHandler.ThrowOnFailure(this.codeWindow.SetBuffer((IVsTextLines)docData));

					//Get our text view for our editor which we will use to get the WPF control that hosts said editor.
					ErrorHandler.ThrowOnFailure(this.codeWindow.GetPrimaryView(out this.textView));

					//Get our WPF host from our text view (from our code window).
					Dispatcher.CurrentDispatcher.BeginInvoke(new Action(() =>
					{
						IWpfTextViewHost textViewHost = editorAdapterFactoryService.GetWpfTextViewHost(this.textView);
						this.control.InsertNewEditor(textViewHost.HostControl);
						this.model.LoadDocument(filePath);
					}));
				}
				finally
				{
					if (docDataPointer != IntPtr.Zero)
					{
						//Release the doc data from the invisible editor since it gave us a ref-counted copy.
						Marshal.Release(docDataPointer);
					}
				}
			}
		}

		public void Save(string filePath)
		{
			if (this.invisibleEditor != null)
			{
				IntPtr docDataPointer = IntPtr.Zero;
				Guid guidIVSTextLines = typeof(IVsTextLines).GUID;
				ErrorHandler.ThrowOnFailure(this.invisibleEditor.GetDocData(fEnsureWritable: 1, riid: ref guidIVSTextLines, ppDocData: out docDataPointer));
				try
				{
					IVsPersistDocData docData = (IVsPersistDocData)Marshal.GetObjectForIUnknown(docDataPointer);

					string pbstrMkDocumentNew = null;
					int pfSaveCanceled = 0;
					docData.SaveDocData(
						VSSAVEFLAGS.VSSAVE_SilentSave,
						out pbstrMkDocumentNew,
						out pfSaveCanceled);
				}
				finally
				{
					if (docDataPointer != IntPtr.Zero)
					{
						//Release the doc data from the invisible editor since it gave us a ref-counted copy.
						Marshal.Release(docDataPointer);
					}
				}
			}
		}
		#endregion

		#region Protected Overrides

		/// <summary>
		/// Set our Content property to be an initially empty version of our editor hosting control.
		/// </summary>
		protected override void Initialize()
		{
			base.Initialize();

			var dte = (EnvDTE.DTE)this.GetService(typeof(EnvDTE.DTE));
			windowEvents = dte.Events.WindowEvents;
			dte.Events.WindowEvents.WindowActivated += WindowEvents_WindowActivated;
			dte.Events.WindowEvents.WindowClosing += WindowEvents_WindowClosing;

			var componentModel = (IComponentModel)GetService(typeof(SComponentModel));
			var errorProvider = componentModel.GetService<IErrorProvider>();
			var buildProvider = componentModel.GetService<IDslBuildProvider>();
			var options = componentModel.GetService<IMetaSharpOptions>();

			var documentTable = this.GetService(typeof(SVsRunningDocumentTable)) as IVsRunningDocumentTable;
			documentTable.AdviseRunningDocTableEvents(this, out docEventsCookie);

			base.Content = this.control = new GrammarHostView { Host = model = new GrammarHost(errorProvider, buildProvider, options) };
			this.control.InsertNewEditor(new Grid());

			if (dte.ActiveDocument != null)
			{
				this.ActivateWindow(dte.ActiveWindow);
			}
		}

		protected override void OnClose()
		{
			base.OnClose();

			var dte = (EnvDTE.DTE)this.GetService(typeof(EnvDTE.DTE));
			if (dte != null)
				dte.Events.WindowEvents.WindowActivated -= WindowEvents_WindowActivated;

			var documentTable = this.GetService(typeof(SVsRunningDocumentTable)) as IVsRunningDocumentTable;
			if (documentTable != null)
				documentTable.UnadviseRunningDocTableEvents(docEventsCookie);
		}

		private void WindowEvents_WindowActivated(EnvDTE.Window GotFocus, EnvDTE.Window LostFocus)
		{
			ActivateWindow(GotFocus);
		}

		private void WindowEvents_WindowClosing(EnvDTE.Window closing)
		{
			if (closing == activeWindow)
			{
				displayedFilePath = null;
				activeWindow = null;
				this.model.Node = null;
				ClearOldEditor();
			}
		}

		internal void ActivateWindow(EnvDTE.Window window)
		{
			if (window != null && window.Document != null)
			{
				var openWindowPath = Path.Combine(window.Document.Path, window.Document.Name);
				var buffer = GetBufferAt(openWindowPath);
				if (buffer != null && buffer.ContentType.TypeName == "Grammar")
				{
					var grammarPath = Path.Combine(window.Document.Path, window.Document.Name);
					var dslPath = DslBuildProvider.GetDslPath(grammarPath);

					var dir = Path.GetDirectoryName(dslPath);
					if (!Directory.Exists(dir))
						Directory.CreateDirectory(dir);

					if (!File.Exists(dslPath))
						using (File.Create(dslPath)) { }

					activeWindow = window;
					this.SetDisplayedFile(dslPath);
				}
			}
		}

		internal ITextBuffer GetBufferAt(string filePath)
		{
			var componentModel = (IComponentModel)GetService(typeof(SComponentModel));
			var editorAdapterFactoryService = componentModel.GetService<IVsEditorAdaptersFactoryService>();
			var serviceProvider = new Microsoft.VisualStudio.Shell.ServiceProvider(MetaSharpPackage.OleServiceProvider);

			IVsUIHierarchy uiHierarchy;
			uint itemID;
			IVsWindowFrame windowFrame;
			if (VsShellUtilities.IsDocumentOpen(
				serviceProvider,
				filePath,
				Guid.Empty,
				out uiHierarchy,
				out itemID,
				out windowFrame))
			{
				IVsTextView view = VsShellUtilities.GetTextView(windowFrame);
				IVsTextLines lines;
				if (view.GetBuffer(out lines) == 0)
				{
					var buffer = lines as IVsTextBuffer;
					if (buffer != null)
						return editorAdapterFactoryService.GetDataBuffer(buffer);
				}
			}

			return null;
		}

		protected override bool PreProcessMessage(ref System.Windows.Forms.Message m)
		{
			//	if (TextViewHost != null)
			if (this.textView != null && m.Msg >= GrammarToolWindow.WM_KEYFIRST && m.Msg <= GrammarToolWindow.WM_KEYLAST)
			{
				// copy the Message into a MSG[] array, so we can pass
				// it along to the active core editor's IVsWindowPane.TranslateAccelerator
				var pMsg = new MSG[1];
				pMsg[0].hwnd = m.HWnd;
				pMsg[0].message = (uint)m.Msg;
				pMsg[0].wParam = m.WParam;
				pMsg[0].lParam = m.LParam;

				var vsWindowPane = (IVsWindowPane)this.textView;
				var retVal = vsWindowPane.TranslateAccelerator(pMsg);
				return retVal == 0;
			}

			return base.PreProcessMessage(ref m);
		}

		public override void OnToolWindowCreated()
		{
			var windowFrame = (IVsWindowFrame)Frame;
			var cmdUi = Microsoft.VisualStudio.VSConstants.GUID_TextEditorFactory;
			windowFrame.SetGuidProperty(
				(int)__VSFPROPID.VSFPROPID_InheritKeyBindings,
				ref cmdUi);

			base.OnToolWindowCreated();
		}

		#endregion

		#region IOleCommandTarget Members

		/// <summary>
		/// When our tool window is active it will be the 'focus command target' of the shell's command route, as such we need to handle any
		/// commands we want here and forward the rest to the editor (since most all typing is translated into a command for the editor to
		/// deal with).
		/// </summary>
		int IOleCommandTarget.Exec(ref Guid pguidCmdGroup, uint nCmdID, uint nCmdexecopt, IntPtr pvaIn, IntPtr pvaOut)
		{
			if (this.control.IsKeyboardFocusWithin && (EditorCommandTarget != null))
			{
				int res = EditorCommandTarget.Exec(ref pguidCmdGroup, nCmdID, nCmdexecopt, pvaIn, pvaOut);

				var cmdId = (VSConstants.VSStd97CmdID)nCmdID;
				if (res == (int)Microsoft.VisualStudio.OLE.Interop.Constants.OLECMDERR_E_NOTSUPPORTED &&
					pguidCmdGroup == Microsoft.VisualStudio.VSConstants.CMDSETID.StandardCommandSet97_guid &&
					(cmdId == VSConstants.VSStd97CmdID.SaveProjectItem ||
					cmdId == VSConstants.VSStd97CmdID.Save))
				{
					Save(this.displayedFilePath);
					res = (int)VSConstants.S_OK;
				}

				return res;
			}

			return (int)Microsoft.VisualStudio.OLE.Interop.Constants.OLECMDERR_E_NOTSUPPORTED;
		}

		/// <summary>
		/// When our tool window is active it will be the 'focus command target' of the shell's command route, as such we need to set the state
		/// of any commands we want here and forward the rest to the editor (since most all typing is translated into a command for the editor to
		/// deal with).
		/// </summary>
		int IOleCommandTarget.QueryStatus(ref Guid pguidCmdGroup, uint cCmds, OLECMD[] prgCmds, IntPtr pCmdText)
		{
			if (this.control.IsKeyboardFocusWithin && (EditorCommandTarget != null))
			{
				return EditorCommandTarget.QueryStatus(ref pguidCmdGroup, cCmds, prgCmds, pCmdText);
			}

			return (int)Microsoft.VisualStudio.OLE.Interop.Constants.OLECMDERR_E_NOTSUPPORTED;
		}

		#endregion

		#region IVsFindTarget Members

		public int Find(string pszSearch, uint grfOptions, int fResetStartPoint, IVsFindHelper pHelper, out uint pResult)
		{
			if (EditorFindTarget != null)
				return EditorFindTarget.Find(pszSearch, grfOptions, fResetStartPoint, pHelper, out pResult);

			pResult = 0;
			return 0;
		}

		public int GetCapabilities(bool[] pfImage, uint[] pgrfOptions)
		{
			if (EditorFindTarget != null)
				return EditorFindTarget.GetCapabilities(pfImage, pgrfOptions);

			return 0;
		}

		public int GetCurrentSpan(TextSpan[] pts)
		{
			if (EditorFindTarget != null)
				return EditorFindTarget.GetCurrentSpan(pts);

			return 0;
		}

		public int GetFindState(out object ppunk)
		{
			if (EditorFindTarget != null)
				return EditorFindTarget.GetFindState(out ppunk);

			ppunk = 0;
			return 0;
		}

		public int GetMatchRect(RECT[] prc)
		{
			if (EditorFindTarget != null)
				return EditorFindTarget.GetMatchRect(prc);

			return 0;
		}

		public int GetProperty(uint propid, out object pvar)
		{
			if (EditorFindTarget != null)
				return EditorFindTarget.GetProperty(propid, out pvar);

			pvar = 0;
			return 0;
		}

		public int GetSearchImage(uint grfOptions, IVsTextSpanSet[] ppSpans, out IVsTextImage ppTextImage)
		{
			if (EditorFindTarget != null)
				return EditorFindTarget.GetSearchImage(grfOptions, ppSpans, out ppTextImage);

			ppTextImage = null;
			return 0;
		}

		public int MarkSpan(TextSpan[] pts)
		{
			if (EditorFindTarget != null)
				return EditorFindTarget.MarkSpan(pts);

			return 0;
		}

		public int NavigateTo(TextSpan[] pts)
		{
			if (EditorFindTarget != null)
				return EditorFindTarget.NavigateTo(pts);

			return 0;
		}

		public int NotifyFindTarget(uint notification)
		{
			if (EditorFindTarget != null)
				return EditorFindTarget.NotifyFindTarget(notification);

			return 0;
		}

		public int Replace(string pszSearch, string pszReplace, uint grfOptions, int fResetStartPoint, IVsFindHelper pHelper, out int pfReplaced)
		{
			if (EditorFindTarget != null)
				return EditorFindTarget.Replace(pszSearch, pszReplace, grfOptions, fResetStartPoint, pHelper, out pfReplaced);

			pfReplaced = 0;
			return 0;
		}

		public int SetFindState(object pUnk)
		{
			if (EditorFindTarget != null)
				return EditorFindTarget.SetFindState(pUnk);

			return 0;
		}
		#endregion

		#region IVsFindTarget2 Members

		public int NavigateTo2(IVsTextSpanSet pSpans, TextSelMode iSelMode)
		{
			return ((IVsFindTarget2)EditorFindTarget).NavigateTo2(pSpans, iSelMode);
		}

		#endregion

		#region Private Properties

		/// <summary>
		/// The IOleCommandTarget for the editor that our tool window will forward all command requests to when it is the active tool window
		/// and the editor we are hosting has keyboard focus.
		/// </summary>
		private IOleCommandTarget EditorCommandTarget
		{
			get
			{
				return (this.cachedEditorCommandTarget ?? (this.cachedEditorCommandTarget = this.textView as IOleCommandTarget));
			}
		}

		/// <summary>
		/// The IVsFindTarget for the editor that our tool window will forward all find releated requests to when it is the active tool window
		/// and the editor we are hosting has keyboard focus.
		/// </summary>
		private IVsFindTarget EditorFindTarget
		{
			get
			{
				return (this.cachedEditorFindTarget ?? (this.cachedEditorFindTarget = this.textView as IVsFindTarget));
			}
		}

		#endregion

		#region Private Methods

		/// <summary>
		/// Cleans up an existing editor if we are about to put a new one in place, used to close down the old editor bits as well as
		/// nulling out any cached objects that we have that came from the now dead editor.
		/// </summary>
		private void ClearOldEditor()
		{
			if (this.codeWindow != null)
			{
				this.codeWindow.Close();
				this.codeWindow = null;
			}

			if (this.textView != null)
			{
				this.textView.CloseView();
				this.textView = null;
			}

			this.cachedEditorCommandTarget = null;
			this.cachedEditorFindTarget = null;
			this.invisibleEditor = null;
		}

		#endregion

		#region IVsRunningDocTableEvents Members
		int IVsRunningDocTableEvents.OnAfterAttributeChange(uint docCookie, uint grfAttribs)
		{
			return (int)Microsoft.VisualStudio.OLE.Interop.Constants.OLECMDERR_E_NOTSUPPORTED;
		}

		int IVsRunningDocTableEvents.OnAfterDocumentWindowHide(uint docCookie, IVsWindowFrame pFrame)
		{
			return (int)Microsoft.VisualStudio.OLE.Interop.Constants.OLECMDERR_E_NOTSUPPORTED;
		}

		int IVsRunningDocTableEvents.OnAfterFirstDocumentLock(uint docCookie, uint dwRDTLockType, uint dwReadLocksRemaining, uint dwEditLocksRemaining)
		{
			return (int)Microsoft.VisualStudio.OLE.Interop.Constants.OLECMDERR_E_NOTSUPPORTED;
		}

		int IVsRunningDocTableEvents.OnAfterSave(uint docCookie)
		{
			Save(this.displayedFilePath);
			return (int)VSConstants.S_OK;
		}

		int IVsRunningDocTableEvents.OnBeforeDocumentWindowShow(uint docCookie, int fFirstShow, IVsWindowFrame pFrame)
		{
			return (int)Microsoft.VisualStudio.OLE.Interop.Constants.OLECMDERR_E_NOTSUPPORTED;
		}

		int IVsRunningDocTableEvents.OnBeforeLastDocumentUnlock(uint docCookie, uint dwRDTLockType, uint dwReadLocksRemaining, uint dwEditLocksRemaining)
		{
			return (int)Microsoft.VisualStudio.OLE.Interop.Constants.OLECMDERR_E_NOTSUPPORTED;
		}

		int IVsRunningDocTableEvents4.OnAfterLastDocumentUnlock(IVsHierarchy pHier, uint itemid, string pszMkDocument, int fClosedWithoutSaving)
		{
			return (int)Microsoft.VisualStudio.OLE.Interop.Constants.OLECMDERR_E_NOTSUPPORTED;
		}

		int IVsRunningDocTableEvents4.OnAfterSaveAll()
		{
			Save(this.displayedFilePath);
			return (int)VSConstants.S_OK;
		}

		int IVsRunningDocTableEvents4.OnBeforeFirstDocumentLock(IVsHierarchy pHier, uint itemid, string pszMkDocument)
		{
			return (int)Microsoft.VisualStudio.OLE.Interop.Constants.OLECMDERR_E_NOTSUPPORTED;
		}
		#endregion
	}
}