﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Xunit.Extensions;
using MetaSharp.Build;
using Xunit;
using NSubstitute;
using Microsoft.Build.Framework;
using Microsoft.Build.Utilities;
using System.IO;
using System.Threading;
using System.Windows.Threading;

namespace MetaSharp.Transformation.Tests.Build
{
	public class MetaSharpActionQueueTests : AsyncTests
	{
		private IActionQueue actionQueue = new ActionQueue();

		[Fact]
		public void CanQueueASingleItem()
		{
			int count = 0;
			using (var token = Wait())
			{
				actionQueue.Enqueue("",
					() => ++count,
					r =>
					{
						Assert.Null(r.Error);
						Assert.Equal(1, r.Value);
						count++;
						token.Done();
					});
			}

			Assert.Equal(2, count);
		}

		[Fact]
		public void CanQueueMultipleItems()
		{
			int count = 0;
			var task = new Func<object>(() => ++count);
			using (var token = Wait())
			{
				actionQueue.Enqueue("", task,
					r =>
					{
						Assert.Null(r.Error);
						Assert.Equal(1, r.Value);
						count++;
					});

				actionQueue.Enqueue("", task,
					r =>
					{
						Assert.Null(r.Error);
						Assert.Equal(3, r.Value);
						count++;
						token.Done();
					});
			}

			Assert.Equal(4, count);
		}

		[Fact]
		public void ActionsAreGuaranteedToBeProcessedSequentially()
		{
			int last = 0;
			int count = 0;
			using (var token = Wait())
			{
				for (int x = 0; x < 100; x++)
				{
					int y = x;
					actionQueue.Enqueue("",
						() => count++,
						r =>
						{
							Assert.Equal(last + 1, count);
							last = count;
							if (count == 2)
								token.Done();
						});
				}
			}

			Assert.Equal(2, count);
		}

		[Fact]
		public void ActionGroupsAreProcessedInParallel()
		{
			using (var w1 = Wait())
			{
				WaitToken w2 = Wait();
				WaitToken w3 = Wait();
				actionQueue.Enqueue("g1",
					() => { w2.Dispose(); return null; },
					r => w3.Done());

				actionQueue.Enqueue("g2",
					() => { w2.Done(); w3.Dispose(); return null; },
					r => w1.Done());
			}
		}
	}

	public class AsyncTests
	{
		public WaitToken Wait()
		{
			return new WaitToken(TimeSpan.FromSeconds(2));
		}

		public WaitToken Wait(TimeSpan timeout)
		{
			return new WaitToken(timeout);
		}

		public class WaitToken : IDisposable
		{
			bool _done;
			TimeSpan _timeout;
			public WaitToken(TimeSpan timeout)
			{
				_timeout = timeout;
			}

			public void Done()
			{
				_done = true;
			}

			public void Dispose()
			{
				var start = DateTime.Now;
				while (!_done)
				{
					if (!System.Diagnostics.Debugger.IsAttached &&
						DateTime.Now - start > _timeout)
						throw new TimeoutException("Async test timed out...");

					Dispatcher.CurrentDispatcher.Invoke(
						new Action(() => { }),
						DispatcherPriority.ContextIdle);
				}
			}
		}
	}
}