﻿using System.CodeDom;
using System.Linq;
using MetaSharp.Transformation.Grammar;
using MetaSharp.Transformation.Grammar.Ast;
using Xunit;
using MetaSharp.Transformation.Lang.Ast;

namespace MetaSharp.Transformation.Tests.Grammar
{
	public class GrammarParserTests : IUseFixture<GrammarParser>
	{
		private PatternContext context = new PatternContext();
		private GrammarParser grammar;

		public void SetFixture(GrammarParser data)
		{
			this.grammar = data;
		}

		[Fact]
		public void CanDeclareGrammar()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G:end");
			Assert.Equal("G", result.Name);
			Assert.Equal(MemberAttributes.Public, result.Attributes);
			Assert.True(result.IsClass);
			Assert.True(result.IsPartial);
			Assert.Empty(result.Members);
			Assert.Empty(result.Rules);
		}

		[Fact]
		public void CanDeclareGrammarAndClassSideBySide()
		{
			var result = (CodeCompileUnit)this.grammar.Parse("namespace N: grammar G:end class C:end end");
			Assert.Equal(1, result.Namespaces.Count);
			Assert.Equal(2, result.Namespaces[0].Types.Count);
			Assert.Equal("G", result.Namespaces[0].Types[0].Name);
			Assert.Equal("C", result.Namespaces[0].Types[1].Name);
			Assert.IsType<GrammarDeclaration>(result.Namespaces[0].Types[0]);
			Assert.IsType<TypeDeclaration>(result.Namespaces[0].Types[1]);
		}

		[Fact]
		public void GrammarCanDeclareConstructor()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: constructor:end end");
			Assert.Equal("G", result.Name);
			Assert.Equal(1, result.Members.Count);
			Assert.IsType<Constructor>(result.Members[0]);
		}

		[Fact]
		public void GrammarCanDeclareConstructorWithEmptyParameters()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: constructor():end end");
			Assert.Equal("G", result.Name);
			Assert.Equal(1, result.Members.Count);
			Assert.IsType<Constructor>(result.Members[0]);
		}

		[Fact]
		public void GrammarCanDeclareFieldAndRule()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: int x = 0; X = 0; end");
			Assert.Equal("G", result.Name);
			Assert.Equal(1, result.Members.Count);
			Assert.Equal(1, result.Rules.Count());
			Assert.IsType<Field>(result.Members[0]);
		}

		[Fact]
		public void GrammarCanDeclareMethod()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: void parse(A a): end end");
			Assert.Equal("G", result.Name);
			Assert.Equal(1, result.Members.Count);
			Assert.IsType<Method>(result.Members[0]);
		}

		[Fact]
		public void GrammarCannotDeclareVariablesWithKeywordNamedMatch()
		{
			var context = new PatternContext();
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: void M(object match): end end", context);
			Assert.Equal(2, context.Errors.Length);
		}
		
		[Fact]
		public void GrammarCanDeclarePublicMethod()
		{
			var result = (GrammarDeclaration)this.grammar.Parse(@"
				grammar G: 
					public void parse(A a): 
					end 
				end");

			Assert.Equal("G", result.Name);
			Assert.Equal(1, result.Members.Count);
			Assert.IsType<Method>(result.Members[0]);
		}

		[Fact]
		public void GrammarCanDeclarePublicStaticMethodWithGenericReturnType()
		{
			var result = (GrammarDeclaration)this.grammar.Parse(@"
				grammar G: 
					X = X | Y Z;
					public static A<B> parse(C c):
						return null;
					end 
				end");

			Assert.Equal("G", result.Name);
			Assert.Equal(1, result.Rules.Count());
			Assert.Equal(1, result.Members.Count);
			Assert.IsType<Method>(result.Members[0]);
		}

		[Fact]
		public void CanDeclareSuperTypeForGrammar()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G < S:end");
			Assert.Equal("G", result.Name);
		}

		[Fact]
		public void GrammarCanDeclareValuePattern()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: X = 7; end");
			Assert.Equal("X", result.Rules.ElementAt(0).Name);
			Assert.Equal(7, ((ValuePattern)result.Rules.ElementAt(0).Pattern).Value);
		}

		[Fact]
		public void GrammarCanDeclareReferencePattern()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: X = X; end");
			Assert.Equal("X", result.Rules.ElementAt(0).Name);
			Assert.Equal("X", ((ReferencePattern)result.Rules.ElementAt(0).Pattern).Name);
		}

		[Fact]
		public void GrammarCanDeclareOrPattern()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: X = X | Y; end");
			Assert.Equal("X", result.Rules.ElementAt(0).Name);
			var or = (OrPattern)result.Rules.ElementAt(0).Pattern;
			Assert.Equal(2, or.Patterns.Count());
			Assert.Equal("X", ((ReferencePattern)or.Patterns.ElementAt(0)).Name);
			Assert.Equal("Y", ((ReferencePattern)or.Patterns.ElementAt(1)).Name);
		}

		[Fact]
		public void SiblingOrPatternsAreFlattened()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: X = W | X | Y | Z; end");
			Assert.Equal("X", result.Rules.ElementAt(0).Name);
			var or = (OrPattern)result.Rules.ElementAt(0).Pattern;
			Assert.Equal(4, or.Patterns.Count());
			Assert.Equal("W", ((ReferencePattern)or.Patterns.ElementAt(0)).Name);
			Assert.Equal("X", ((ReferencePattern)or.Patterns.ElementAt(1)).Name);
			Assert.Equal("Y", ((ReferencePattern)or.Patterns.ElementAt(2)).Name);
			Assert.Equal("Z", ((ReferencePattern)or.Patterns.ElementAt(3)).Name);
		}

		[Fact]
		public void CanDeclareProjectionPattern()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: X = Y -> Z; end");
			Assert.Equal("X", result.Rules.ElementAt(0).Name);
			var projection = (ProjectionPattern)result.Rules.ElementAt(0).Pattern;
			Assert.Equal("Y", ((ReferencePattern)projection.Pattern).Name);
			Assert.Equal("Z", ((ReferenceExpression)projection.Projection.Cast<ExpressionStatement>().Single().Expression).PropertyName);
		}

		[Fact]
		public void OrPatternHasHigherPrecedenceThanProjectionPattern()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = A -> B | X -> Y; end");
			Assert.Equal("R", result.Rules.ElementAt(0).Name);
			var or = (OrPattern)result.Rules.ElementAt(0).Pattern;
			var left = or.Patterns.ElementAt(0);
			var right = or.Patterns.ElementAt(1);
			Assert.Equal(2, or.Patterns.Count());
			Assert.IsType<ProjectionPattern>(left);
			Assert.IsType<ProjectionPattern>(right);
		}

		[Fact]
		public void GrammarCanDeclareAndPattern()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = X Y; end");
			Assert.Equal("R", result.Rules.ElementAt(0).Name);
			var and = (AndPattern)result.Rules.ElementAt(0).Pattern;
			Assert.Equal(2, and.Patterns.Count());
			Assert.Equal("X", ((ReferencePattern)and.Patterns.ElementAt(0)).Name);
			Assert.Equal("Y", ((ReferencePattern)and.Patterns.ElementAt(1)).Name);
		}

		[Fact]
		public void SiblingAndPatternsAreFlattened()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = W X Y Z; end");
			Assert.Equal("R", result.Rules.ElementAt(0).Name);
			var and = (AndPattern)result.Rules.ElementAt(0).Pattern;
			Assert.Equal(4, and.Patterns.Count());
			Assert.Equal("W", ((ReferencePattern)and.Patterns.ElementAt(0)).Name);
			Assert.Equal("X", ((ReferencePattern)and.Patterns.ElementAt(1)).Name);
			Assert.Equal("Y", ((ReferencePattern)and.Patterns.ElementAt(2)).Name);
			Assert.Equal("Z", ((ReferencePattern)and.Patterns.ElementAt(3)).Name);
		}

		[Fact]
		public void GrammarCanDeclareZeroOrOnePattern()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = X?; end");
			Assert.Equal("R", result.Rules.ElementAt(0).Name);
			var p = (ZeroOrOnePattern)result.Rules.ElementAt(0).Pattern;
			Assert.IsType<ReferencePattern>(p.Pattern);
		}

		[Fact]
		public void GrammarCanDeclareZeroOrMorePattern()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = X*; end");
			Assert.Equal("R", result.Rules.ElementAt(0).Name);
			var p = (ZeroOrMorePattern)result.Rules.ElementAt(0).Pattern;
			Assert.IsType<ReferencePattern>(p.Pattern);
		}

		[Fact]
		public void GrammarCanDeclareOneOrMorePattern()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = X+; end");
			Assert.Equal("R", result.Rules.ElementAt(0).Name);
			var p = (OneOrMorePattern)result.Rules.ElementAt(0).Pattern;
			Assert.IsType<ReferencePattern>(p.Pattern);
		}

		[Fact]
		public void GrammarCanDeclareExactQuantityPattern()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = X#7; end");
			Assert.Equal("R", result.Rules.ElementAt(0).Name);
			var p = (ExactQuantityPattern)result.Rules.ElementAt(0).Pattern;
			Assert.IsType<ReferencePattern>(p.Pattern);
			Assert.Equal(7, p.Quantity);
		}

		[Fact]
		public void GrammarCanDeclareNotPattern()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = !X; end");
			Assert.Equal("R", result.Rules.ElementAt(0).Name);
			var p = (NotPattern)result.Rules.ElementAt(0).Pattern;
			Assert.IsType<ReferencePattern>(p.Pattern);
		}

		[Fact]
		public void NotPatternHasHigherPrecedenceThanOtherQuantifiers()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = !X+; end");
			Assert.Equal("R", result.Rules.ElementAt(0).Name);
			var p = (NotPattern)result.Rules.ElementAt(0).Pattern;
			Assert.IsType<OneOrMorePattern>(p.Pattern);

			result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = !X*; end");
			Assert.Equal("R", result.Rules.ElementAt(0).Name);
			p = (NotPattern)result.Rules.ElementAt(0).Pattern;
			Assert.IsType<ZeroOrMorePattern>(p.Pattern);

			result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = !X?; end");
			Assert.Equal("R", result.Rules.ElementAt(0).Name);
			p = (NotPattern)result.Rules.ElementAt(0).Pattern;
			Assert.IsType<ZeroOrOnePattern>(p.Pattern);

			result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = !X#11; end");
			Assert.Equal("R", result.Rules.ElementAt(0).Name);
			p = (NotPattern)result.Rules.ElementAt(0).Pattern;
			Assert.IsType<ExactQuantityPattern>(p.Pattern);
		}

		[Fact]
		public void CanGroupPatterns()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = (X Y); end");
			Assert.Equal("R", result.Rules.ElementAt(0).Name);
			Assert.IsType<AndPattern>(result.Rules.ElementAt(0).Pattern);

			var context = new PatternContext();
			result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = (X); end", context);
			Assert.Empty(result.Rules);
			Assert.Equal(1, context.Errors.Length);
		}

		[Fact]
		public void ParameterPatternCanBeGroupedAndQuantified()
		{
			//token:(WhitespaceInterleave | LineCommentInterleave | InlineCommentInterleave | any)*
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = p:(X Y)*; end");
			Assert.Equal("R", result.Rules.ElementAt(0).Name);
			var variable = (VariablePattern)result.Rules.ElementAt(0).Pattern;
			var quantifier = (ZeroOrMorePattern)variable.Pattern;
			var and = (AndPattern)quantifier.Pattern;
			var X = (ReferencePattern)and.Patterns.ElementAt(0);
			var Y = (ReferencePattern)and.Patterns.ElementAt(1);
			Assert.Equal("p", variable.Name);
			Assert.Equal("X", X.Name);
			Assert.Equal("Y", Y.Name);
			Assert.Equal(2, and.Patterns.Count());
		}

		[Fact]
		public void CanDeclareArrayPattern()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = [x]; end");
			Assert.Equal("R", result.Rules.ElementAt(0).Name);
			var array = (ArrayPattern)result.Rules.ElementAt(0).Pattern;
			Assert.Equal("x", ((ReferencePattern)array.Pattern).Name);
		}

		[Fact]
		public void CanDeclareArrayWithMultiplePatterns()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = [x y]; end");
			Assert.Equal("R", result.Rules.ElementAt(0).Name);
			var array = (ArrayPattern)result.Rules.ElementAt(0).Pattern;
			var and = (AndPattern)array.Pattern;
			Assert.Equal("x", ((ReferencePattern)and.Patterns.ElementAt(0)).Name);
			Assert.Equal("y", ((ReferencePattern)and.Patterns.ElementAt(1)).Name);
		}

		[Fact]
		public void CanDeclareDefaulPattern()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = default; end");
			Assert.Equal("R", result.Rules.ElementAt(0).Name);
			Assert.IsType<DefaultPattern>(result.Rules.ElementAt(0).Pattern);
		}

		[Fact]
		public void CanDeclareAnyPattern()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = any; end");
			Assert.Equal("R", result.Rules.ElementAt(0).Name);
			Assert.IsType<AnyPattern>(result.Rules.ElementAt(0).Pattern);
		}

		[Fact]
		public void CanDeclareFailPattern()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = fail; end");
			Assert.Equal("R", result.Rules.ElementAt(0).Name);
			Assert.IsType<FailPattern>(result.Rules.ElementAt(0).Pattern);
		}

		[Fact]
		public void CanDeclareNullPattern()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = null; end");
			Assert.Equal("R", result.Rules.ElementAt(0).Name);
			Assert.IsType<NullPattern>(result.Rules.ElementAt(0).Pattern);
		}

		[Fact]
		public void CanDeclareBooleanPatterns()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = true; end");
			Assert.True((bool)((ValuePattern)result.Rules.ElementAt(0).Pattern).Value);

			result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = false; end");
			Assert.False((bool)((ValuePattern)result.Rules.ElementAt(0).Pattern).Value);
		}

		[Fact]
		public void CanDeclareVariablePattern()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = p:X; end");
			var p = (VariablePattern)result.Rules.ElementAt(0).Pattern;
			Assert.Equal("p", p.Name);
			Assert.IsType<ReferencePattern>(p.Pattern);
		}

		[Fact]
		public void VariablePatternHasHigherPrecedenceThanQuantifierPatterns()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = p:X*; end");
			var p = (VariablePattern)result.Rules.ElementAt(0).Pattern;
			Assert.Equal("p", p.Name);
			Assert.IsType<ZeroOrMorePattern>(p.Pattern);

			result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = p:X?; end");
			p = (VariablePattern)result.Rules.ElementAt(0).Pattern;
			Assert.IsType<ZeroOrOnePattern>(p.Pattern);

			result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = p:X+; end");
			p = (VariablePattern)result.Rules.ElementAt(0).Pattern;
			Assert.IsType<OneOrMorePattern>(p.Pattern);
		}

		[Fact]
		public void VariablePatternHasLowerPrecedenceThanAndOrPatterns()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = a:X b:Y; end");
			var p = (AndPattern)result.Rules.ElementAt(0).Pattern;
			Assert.IsType<VariablePattern>(p.Patterns.ElementAt(0));
			Assert.IsType<VariablePattern>(p.Patterns.ElementAt(1));
		}

		[Fact]
		public void CanDeclareCallPattern()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = X(); end");
			var R = result.Rules.ElementAt(0);
			var call = (CallPattern)R.Pattern;
			Assert.Equal("X", call.Name);
			Assert.Empty(call.Parameters);
		}

		[Fact]
		public void CanDeclareParameterizedRule()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: A(x) = x; B(x,y) = x y; end");
			var A = result.Rules.ElementAt(0);
			var B = result.Rules.ElementAt(1);
			Assert.Equal(1, A.Parameters.Count());
			Assert.Equal(2, B.Parameters.Count());
		}

		[Fact]
		public void CanDeclareCallPatternWithParameters()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = X(Y, Z); end");
			var R = result.Rules.ElementAt(0);
			var call = (CallPattern)R.Pattern;
			Assert.Equal("X", call.Name);
			Assert.Equal(2, call.Parameters.Count());
		}

		[Fact]
		public void CanDeclareCallPatternWithComplexPattern()
		{
			// Must group complex patterns used as parameters, or put them in another rule and reference it.
			// Otherwise, it would be ambiguous with a Group(And) pattern.
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = W((X|Y), Z); end");
			var R = result.Rules.ElementAt(0);
			var call = (CallPattern)R.Pattern;
			Assert.Equal("W", call.Name);
			Assert.Equal(2, call.Parameters.Count());
			Assert.IsType<OrPattern>(call.Parameters.ElementAt(0));
		}

		[Fact]
		public void CallPatternIsNotMistakenForAGroup()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = X (Y() Z); end");
			var R = result.Rules.ElementAt(0);
			var and = (AndPattern)R.Pattern;
			var X = (ReferencePattern)and.Patterns.ElementAt(0);
			var Y = (CallPattern)and.Patterns.ElementAt(1);
			var Z = (ReferencePattern)and.Patterns.ElementAt(2);
			Assert.Equal("X", X.Name);
			Assert.Equal("Y", Y.Name);
			Assert.Equal("Z", Z.Name);
		}

		[Fact]
		public void CanAssignVariableToGroupPattern()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = v:(x | y); end");
			var R = result.Rules.ElementAt(0);
			var variablePattern = (VariablePattern)R.Pattern;
			var orPattern = (OrPattern)variablePattern.Pattern;
			Assert.Equal("x", ((ReferencePattern)orPattern.Patterns.ElementAt(0)).Name);
			Assert.Equal("y", ((ReferencePattern)orPattern.Patterns.ElementAt(1)).Name);
		}

		[Fact]
		public void RootLevelProjectionWithStatementGroupShouldNotHaveStatementEndToken()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = X -> { return X; } end");
			var R = result.Rules.ElementAt(0);
			var projection = (ProjectionPattern)R.Pattern;
			Assert.Equal(1, projection.Projection.Count());
			Assert.Equal("X", ((ReferenceExpression)((CodeMethodReturnStatement)projection.Projection.ElementAt(0)).Expression).PropertyName);
		}

		[Fact]
		public void RootLevelProjectionWithStatementGroupMissingEndShouldProduceError()
		{
			var context = new PatternContext();
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = X -> { return X; end", context);

			Assert.Equal(1, result.Rules.Count());
			Assert.NotEmpty(context.Errors); // actual result is somewhat undefined
		}

		[Fact]
		public void OrWithTrailingProjectionDoesNotNeedStatementEnd()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = Y | X -> { return X; } end");
			var R = result.Rules.ElementAt(0);
			var projection = (OrPattern)R.Pattern;
			Assert.Equal(2, projection.Patterns.Count());
		}

		[Fact]
		public void OrWithMultipleProjectionsDoesNotNeedStatementEnd()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = Y -> { return Y; } | X -> { return X; } end");
			var R = result.Rules.ElementAt(0);
			var projection = (OrPattern)R.Pattern;
			Assert.Equal(2, projection.Patterns.Count());
		}

		[Fact]
		public void OrWithExpressionProjectionAndBlockProjectionDoesNotNeedStatementEnd()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = Y -> Y | X -> { return X; } end");
			var R = result.Rules.ElementAt(0);
			var projection = (OrPattern)R.Pattern;
			Assert.Equal(2, projection.Patterns.Count());
		}

		[Fact]
		public void RootLevelProjectionWithSingleExpressionMustHaveStatementEndToken()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = X -> X; end");
			var R = result.Rules.ElementAt(0);
			var projection = (ProjectionPattern)R.Pattern;
			Assert.Equal(1, projection.Projection.Count());
			Assert.Equal("X", ((ReferenceExpression)((ExpressionStatement)projection.Projection.ElementAt(0)).Expression).PropertyName);
		}

		[Fact]
		public void ParameterizedRuleCanEndWithBlockProjection()
		{
			var context = new PatternContext();
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R(x) = Y -> { return true; } end", context);
			Assert.Equal(1, result.Rules.Count());
			Assert.Empty(context.Errors);

			var R = result.Rules.ElementAt(0);
			Assert.Equal("R", R.Name);
			Assert.Equal(1, R.Parameters.Count());
			Assert.Equal("x", R.Parameters.ElementAt(0));

			var body = (ProjectionPattern)R.Pattern;
			var ret = (ReturnStatement)body.Projection.ElementAt(0);
			var exp = (PrimitiveExpression)ret.Expression;
			Assert.True((bool)exp.Value);
		}

		[Fact]
		public void ParameterizedRuleCanEndWithSimpleProjection()
		{
			var context = new PatternContext();
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R(x) = Y -> true; end", context);
			Assert.Equal(1, result.Rules.Count());
			Assert.Empty(context.Errors);

			var R = result.Rules.ElementAt(0);
			Assert.Equal("R", R.Name);
			Assert.Equal(1, R.Parameters.Count());
			Assert.Equal("x", R.Parameters.ElementAt(0));

			var body = (ProjectionPattern)R.Pattern;
			var stmt = (ExpressionStatement)body.Projection.ElementAt(0);
			var exp = (PrimitiveExpression)stmt.Expression;
			Assert.True((bool)exp.Value);
		}

		[Fact]
		public void ParameterizedRuleCanHaveOrWithTwoSimpleProjections()
		{
			var context = new PatternContext();
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R(x) = Y -> true | Z -> false; end", context);
			Assert.Equal(1, result.Rules.Count());
			Assert.Empty(context.Errors);

			var R = result.Rules.ElementAt(0);
			Assert.Equal("R", R.Name);
			Assert.Equal(1, R.Parameters.Count());
			Assert.Equal("x", R.Parameters.ElementAt(0));

			var or = (OrPattern)R.Pattern;
			Assert.Equal(2, or.Patterns.Count());

			var left = (ProjectionPattern)or.Patterns.ElementAt(0);
			var stmt = (ExpressionStatement)left.Projection.ElementAt(0);
			var exp = (PrimitiveExpression)stmt.Expression;
			Assert.True((bool)exp.Value);

			var right = (ProjectionPattern)or.Patterns.ElementAt(1);
			stmt = (ExpressionStatement)right.Projection.ElementAt(0);
			exp = (PrimitiveExpression)stmt.Expression;
			Assert.False((bool)exp.Value);
		}

		[Fact]
		public void RootLevelProjectionWithSingleExpressionMissingStatementEndTokenProducesError()
		{
			var context = new PatternContext();
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = X -> X end", context);
			Assert.Equal(1, result.Rules.Count());
			Assert.Equal(1, context.Errors.Length);
		}

		[Fact]
		public void CanDeclareRangePattern()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = 'a'..'z'; end");
			var R = result.Rules.ElementAt(0);
			var range = (RangePattern)R.Pattern;
			Assert.Equal('a', range.Left);
			Assert.Equal('z', range.Right);
		}

		[Fact]
		public void CanOverrideRule()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: override R = X; end");
			var R = result.Rules.ElementAt(0);
			Assert.True(R.Override);

			result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = X; end");
			R = result.Rules.ElementAt(0);
			Assert.False(R.Override);
		}

		[Fact]
		public void CanDeclareAttributesOnRule()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: [A][B]R = X; end");
			var R = result.Rules.ElementAt(0);
			Assert.Equal(2, R.Attributes.Count());
			var A = R.Attributes.ElementAt(0);
			var B = R.Attributes.ElementAt(1);
			Assert.Equal("A", A.Name);
			Assert.Equal("B", B.Name);
		}


		[Fact]
		public void CanMatchSimpleObject()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = Foo { }; end");
			var R = result.Rules.ElementAt(0);
			var p = (ObjectPattern)R.Pattern;
			Assert.Equal("Foo", p.Type.BaseType);
			Assert.Empty(p.Properties);
		}

		[Fact]
		public void CanMatchObjectInOr()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = Foo { } | Bar { }; end");
			var R = result.Rules.ElementAt(0);
			var p = (OrPattern)R.Pattern;
			var l = (ObjectPattern)p.Patterns.ElementAt(0);
			var r = (ObjectPattern)p.Patterns.ElementAt(1);
			Assert.Equal("Foo", l.Type.BaseType);
			Assert.Equal("Bar", r.Type.BaseType);
		}

		[Fact]
		public void CanDifferentiateBetweenObjectPatternAndProjection()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = Foo { } -> { return x; } end");
			var R = result.Rules.ElementAt(0);
			var p = (ProjectionPattern)R.Pattern;
			var o = (ObjectPattern)p.Pattern;
			Assert.Equal("Foo", o.Type.BaseType);
		}

		[Fact]
		public void CanMatchObjectWithProperty()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = Foo { Bar }; end");
			var R = result.Rules.ElementAt(0);
			var p = (ObjectPattern)R.Pattern;
			Assert.Equal("Foo", p.Type.BaseType);
			Assert.Equal(1, p.Properties.Count());
			var prop = p.Properties.ElementAt(0);
			Assert.Equal("Bar", prop.Name);
			Assert.Null(prop.Alias);
			Assert.Null(prop.Pattern);
		}

		[Fact]
		public void CanMatchObjectWithMultipleProperties()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = Foo { Bar, Baz }; end");
			var R = result.Rules.ElementAt(0);
			var p = (ObjectPattern)R.Pattern;
			Assert.Equal("Foo", p.Type.BaseType);
			Assert.Equal(2, p.Properties.Count());
			var prop1 = p.Properties.ElementAt(0);
			var prop2 = p.Properties.ElementAt(1);
			Assert.Equal("Bar", prop1.Name);
			Assert.Null(prop1.Alias);
			Assert.Null(prop1.Pattern);
			Assert.Equal("Baz", prop2.Name);
			Assert.Null(prop2.Alias);
			Assert.Null(prop2.Pattern);
		}

		[Fact]
		public void CanOrMultipleObjectMatches()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = A { B } | C { D }; end");
			var R = result.Rules.ElementAt(0);
			var P = (OrPattern)R.Pattern;
			Assert.Equal(2, P.Patterns.Count());

			var A = (ObjectPattern)P.Patterns.ElementAt(0);
			Assert.Equal("A", A.Type.BaseType);
			Assert.Equal(1, A.Properties.Count());
			var B = A.Properties.ElementAt(0);
			Assert.Equal("B", B.Name);
			Assert.Null(B.Alias);
			Assert.Null(B.Pattern);

			var C = (ObjectPattern)P.Patterns.ElementAt(1);
			Assert.Equal("C", C.Type.BaseType);
			Assert.Equal(1, C.Properties.Count());
			var D = C.Properties.ElementAt(0);
			Assert.Equal("D", D.Name);
			Assert.Null(D.Alias);
			Assert.Null(D.Pattern);
		}

		[Fact]
		public void CanMatchObjectWithPropertyAndAlias()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = Foo { b:Bar }; end");
			var R = result.Rules.ElementAt(0);
			var p = (ObjectPattern)R.Pattern;
			Assert.Equal("Foo", p.Type.BaseType);
			Assert.Equal(1, p.Properties.Count());
			var prop = p.Properties.ElementAt(0);
			Assert.Equal("Bar", prop.Name);
			Assert.Equal("b", prop.Alias);
			Assert.Null(prop.Pattern);
		}

		[Fact]
		public void CanMatchObjectWithPropertyAndPattern()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = Foo { Bar = any }; end");
			var R = result.Rules.ElementAt(0);
			var p = (ObjectPattern)R.Pattern;
			Assert.Equal("Foo", p.Type.BaseType);
			Assert.Equal(1, p.Properties.Count());
			var prop = p.Properties.ElementAt(0);
			Assert.Equal("Bar", prop.Name);
			Assert.Null(prop.Alias);
			Assert.IsType<AnyPattern>(prop.Pattern);
		}

		[Fact]
		public void CanMatchObjectWithMultiplePropertiesAndPattern()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = Foo { Bar = any any, Baz = Baz { } }; end");
			var R = result.Rules.ElementAt(0);
			var p = (ObjectPattern)R.Pattern;
			Assert.Equal("Foo", p.Type.BaseType);
			Assert.Equal(2, p.Properties.Count());
			var bar = p.Properties.ElementAt(0);
			Assert.Equal("Bar", bar.Name);
			Assert.Null(bar.Alias);
			Assert.IsType<AndPattern>(bar.Pattern);
			var baz = p.Properties.ElementAt(1);
			Assert.Equal("Baz", baz.Name);
			Assert.Null(baz.Alias);
			Assert.IsType<ObjectPattern>(baz.Pattern);
		}

		[Fact]
		public void CanMatchErrorUntilRule()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = error until X; end");
			var R = result.Rules.ElementAt(0);
			var p = (ErrorUntilPattern)R.Pattern;
			var r = (ReferencePattern)p.Pattern;
			Assert.Equal("X", r.Name);
		}

		[Fact]
		public void CanMatchErrorUnlessRule()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = error unless X; end");
			var R = result.Rules.ElementAt(0);
			var p = (ErrorUnlessPattern)R.Pattern;
			var r = (ReferencePattern)p.Pattern;
			Assert.Equal("X", r.Name);
		}

		[Fact]
		public void CanMatchErrorUnlessAndUntilRule()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = error until X error unless Y; end");
			var R = result.Rules.ElementAt(0);
			var and = (AndPattern)R.Pattern;
			Assert.Equal(2, and.Patterns.Count());
			var left = (ErrorUntilPattern)and.Patterns.ElementAt(0);
			var right = (ErrorUnlessPattern)and.Patterns.ElementAt(1);

			var until = (ReferencePattern)left.Pattern;
			var unless = (ReferencePattern)right.Pattern;
			Assert.Equal("X", until.Name);
			Assert.Equal("Y", unless.Name);
		}

		[Fact]
		public void PatternsHaveLowerPriorityThanAnd()
		{
			var result = (GrammarDeclaration)this.grammar.Parse("grammar G: R = error until X Y; end");
			var R = result.Rules.ElementAt(0);
			var and = (AndPattern)R.Pattern;
			Assert.Equal(2, and.Patterns.Count());
			var left = (ErrorUntilPattern)and.Patterns.ElementAt(0);
			var right = (ReferencePattern)and.Patterns.ElementAt(1);

			var until = (ReferencePattern)left.Pattern;
			Assert.Equal("X", until.Name);
			Assert.Equal("Y", right.Name);
		}
	}
}
