﻿using System;
using System.CodeDom;
using System.CodeDom.Compiler;
using System.Linq;
using MetaSharp.Transformation.Grammar;
using MetaSharp.Transformation.Grammar.Ast;
using MetaSharp.Transformation.Lang.Ast;
using Xunit;

namespace MetaSharp.Transformation.Tests.Grammar
{
	public class GrammarTransformerTests : IUseFixture<GrammarTransformer>
	{
		private GrammarTransformer transformer;
		private Unit unit;
		private CodeNamespace @namespace;
		private GrammarDeclaration grammar;

		public void SetFixture(GrammarTransformer data)
		{
			this.transformer = data;
		}

		public GrammarTransformerTests()
		{
			this.unit = new Unit();
			this.@namespace = new Namespace("N");
			this.grammar = new GrammarDeclaration("G");
			@namespace.Types.Add(grammar);
			unit.Namespaces.Add(@namespace);
		}

		[Fact]
		public void GrammarUnitTransformsIntoCodeCompileUnit()
		{
			var unit = new Unit();
			var result = transformer.Transform(unit);
			Assert.NotSame(unit, result);
		}

		[Fact]
		public void NullUnitThrows()
		{
			Assert.Throws<ArgumentNullException>(() => transformer.Transform(null));
		}

		[Fact]
		public void SingleGrammarDeclarationSucceeds()
		{
			var result = this.transformer.Transform(this.unit);
			var test = result.Namespaces[0];
			Assert.Equal(1, test.Types.Count);
			Assert.Equal("G", test.Types[0].Name);
		}

		[Fact]
		public void MultipleGrammarDeclarationSucceeds()
		{
			@namespace.Types.Add(new GrammarDeclaration("G2"));
			var result = this.transformer.Transform(unit);
			var test = result.Namespaces[0];
			Assert.Equal(2, test.Types.Count);
			Assert.Equal("G", test.Types[0].Name);
			Assert.Equal("G2", test.Types[1].Name);
		}

		[Fact]
		public void GrammarWithSuperTypeSucceeds()
		{
			this.grammar.BaseTypes.Add(new CodeTypeReference("S"));
			var result = this.transformer.Transform(unit);
			var test = result.Namespaces[0];
			Assert.Equal(1, test.Types.Count);
			Assert.Equal("G", test.Types[0].Name);
			Assert.Equal("S", test.Types[0].BaseTypes[0].BaseType);
		}

		[Fact]
		public void SingleNamespaceDeclarationSucceeds()
		{
			var unit = new Unit();
			unit.Namespaces.Add(new Namespace("Test"));
			var result = this.transformer.Transform(unit);
			Assert.Equal(1, result.Namespaces.Count);
			Assert.Equal("Test", result.Namespaces[0].Name);
		}

		[Fact]
		public void MultipleNamespaceDeclarationSucceeds()
		{
			this.unit.Namespaces.Add(new Namespace("Test"));
			var result = this.transformer.Transform(unit);
			Assert.Equal(2, result.Namespaces.Count);
			Assert.Equal("N", result.Namespaces[0].Name);
			Assert.Equal("Test", result.Namespaces[1].Name);
		}

		[Fact]
		public void CanDeclareSimpleRule()
		{
			this.grammar.Rules = new[]{
				new RuleDeclaration {
					Name = "R",
					Pattern = new AnyPattern(),
				}
			};
			var result = this.transformer.Transform(unit);
			var generatedClass = (CodeTypeDeclaration)result.Namespaces[0].Types[0];
			var field = generatedClass.Members.OfType<CodeMemberField>().Single();
			Assert.False(generatedClass.Members.OfType<CodeConstructor>().Any());
			var method = generatedClass.Members.OfType<CodeMemberMethod>().Single();

			Assert.Equal("_R", field.Name);
			Assert.Equal("MetaSharp.Transformation.PatternMatch", field.Type.BaseType);
			Assert.Equal("R", method.Name);
			Assert.Equal("MetaSharp.Transformation.INode", method.ReturnType.BaseType);
			Assert.Equal(2, method.Parameters.Count);
			Assert.Equal("context", method.Parameters[0].Name);
			Assert.Equal("MetaSharp.Transformation.PatternContext", method.Parameters[0].Type.BaseType);
			Assert.Equal("stream", method.Parameters[1].Name);
			Assert.Equal("MetaSharp.Transformation.IStream", method.Parameters[1].Type.BaseType);
			Assert.Equal(2, method.Statements.Count);
			Assert.IsType<ConditionStatement>(method.Statements[0]);
			Assert.IsType<ReturnStatement>(method.Statements[1]);
		}

		[Fact]
		public void CanDeclareMultipleRules()
		{
			this.grammar.Rules = new[]{
				new RuleDeclaration {
					Name = "R1",
					Pattern = new AnyPattern(),
				},
				new RuleDeclaration {
					Name = "R2",
					Pattern = new AnyPattern()
				},
			};

			var result = this.transformer.Transform(unit);
			var generatedClass = (CodeTypeDeclaration)result.Namespaces[0].Types[0];
			Assert.Equal(2, generatedClass.Members.OfType<CodeMemberField>().Count());
			Assert.False(generatedClass.Members.OfType<CodeConstructor>().Any());
			Assert.Equal(2, generatedClass.Members.OfType<CodeMemberMethod>().Count());
		}

		[Fact]
		public void CanTransformProjectionWithSingleExpression()
		{
			this.grammar.Rules = new[] {
				new RuleDeclaration {
					Name = "R",
					Pattern = new ProjectionPattern {
						Pattern = new VariablePattern { Name = "V", Pattern = new ReferencePattern() },
						Projection = new Nodes<CodeStatement> { new ExpressionStatement(new PrimitiveExpression(null)) }
					}
				}
			};

			var result = this.transformer.Transform(unit);
			var generatedClass = (CodeTypeDeclaration)result.Namespaces[0].Types[0];
			Assert.Equal(1, generatedClass.Members.OfType<CodeMemberField>().Count());
			Assert.Equal(2, generatedClass.Members.OfType<CodeMemberMethod>().Count());

			var projectionMethod = generatedClass.Members.OfType<CodeMemberMethod>().Where(m => m.Name.Contains("projection")).Single();
			// Generates 3 states:
			// 4 variable declarations and a return.
			Assert.Equal(5, projectionMethod.Statements.Count);
			Assert.Equal(1, projectionMethod.Parameters.Count);
			Assert.Equal("__scope__", projectionMethod.Parameters[0].Name);
			Assert.Equal("MetaSharp.Transformation.Scope", projectionMethod.Parameters[0].Type.BaseType);
			Assert.IsType<VariableDeclarationStatement>(projectionMethod.Statements[0]);
			Assert.IsType<VariableDeclarationStatement>(projectionMethod.Statements[1]);
			Assert.IsType<VariableDeclarationStatement>(projectionMethod.Statements[2]);
			Assert.IsType<VariableDeclarationStatement>(projectionMethod.Statements[3]);
			var ret = (ReturnStatement)projectionMethod.Statements[4];
			Assert.Null(((PrimitiveExpression)((MethodInvokeExpression)ret.Expression).Parameters[0]).Value);
		}

		[Fact]
		public void ObjectMatchGeneratesCorrectProjectionIndexers()
		{
			this.grammar.Rules = new[] {
				new RuleDeclaration {
					Name = "R",
					Pattern = new ProjectionPattern {
						Pattern = new AndPattern {
							Patterns = new PatternDeclaration[] {
								new ReferencePattern { Name = "A" },
								new VariablePattern { 
									Name = "B", 
									Pattern = new ObjectPattern {
										Properties = new [] {
											new PropertyPattern {
												Alias = "C",
												Name = "P",
												Pattern = new AnyPattern()
											}
										}
									},
								},
								new ReferencePattern { Name = "D" }
							}
						},
						Projection = new Nodes<CodeStatement> { new ExpressionStatement(new PrimitiveExpression(null)) }
					}
				}
			};

			var result = this.transformer.Transform(unit);
			var generatedClass = (CodeTypeDeclaration)result.Namespaces[0].Types[0];
			var projectionMethod = generatedClass.Members.OfType<CodeMemberMethod>().Where(m => m.Name.Contains("projection")).Single();
			Assert.Equal(6, projectionMethod.Statements.Count);
			Assert.IsType<VariableDeclarationStatement>(projectionMethod.Statements[0]);
			var B = (MethodInvokeExpression)((CodeVariableDeclarationStatement)projectionMethod.Statements[0]).InitExpression;
			var C = (MethodInvokeExpression)((CodeVariableDeclarationStatement)projectionMethod.Statements[1]).InitExpression;
			var match = (MethodInvokeExpression)((CodeVariableDeclarationStatement)projectionMethod.Statements[2]).InitExpression;
			var context = (CodeVariableDeclarationStatement)projectionMethod.Statements[3];
			var position = (CodeVariableDeclarationStatement)projectionMethod.Statements[4];

			Assert.Equal("context", context.Name);
			Assert.Equal("Unwrap", match.Method.MethodName);
			Assert.Equal("Unwrap", B.Method.MethodName);
			Assert.Equal("Unwrap", C.Method.MethodName);
			Assert.Equal("MetaSharp.Transformation.PatternContext", context.Type.BaseType);
			Assert.Equal("MetaSharp.Transformation.Node", ((TypeReferenceExpression)match.Method.TargetObject).Type.BaseType);
			Assert.Equal("MetaSharp.Transformation.Node", ((TypeReferenceExpression)B.Method.TargetObject).Type.BaseType);
			Assert.Equal("MetaSharp.Transformation.Node", ((TypeReferenceExpression)C.Method.TargetObject).Type.BaseType);
			Assert.Equal(1, match.Parameters.Count);
			Assert.Equal(1, B.Parameters.Count);
			Assert.Equal(1, C.Parameters.Count);

			Assert.Equal("B", ((PrimitiveExpression)((ArrayIndexerExpression)B.Parameters[0]).Indices[0]).Value);
			Assert.Equal("Variables", ((ReferenceExpression)((ArrayIndexerExpression)B.Parameters[0]).TargetObject).PropertyName);
			Assert.Equal("__scope__", ((ReferenceExpression)((ReferenceExpression)((ArrayIndexerExpression)B.Parameters[0]).TargetObject).TargetObject).PropertyName);

			Assert.Equal("C", ((PrimitiveExpression)((ArrayIndexerExpression)C.Parameters[0]).Indices[0]).Value);
			Assert.Equal("Variables", ((ReferenceExpression)((ArrayIndexerExpression)C.Parameters[0]).TargetObject).PropertyName);
			Assert.Equal("__scope__", ((ReferenceExpression)((ReferenceExpression)((ArrayIndexerExpression)C.Parameters[0]).TargetObject).TargetObject).PropertyName);

			Assert.Equal("Match", ((ReferenceExpression)match.Parameters[0]).PropertyName);
			Assert.Equal("__scope__", ((ReferenceExpression)((ReferenceExpression)match.Parameters[0]).TargetObject).PropertyName);
			Assert.Null(((ReferenceExpression)((ReferenceExpression)match.Parameters[0]).TargetObject).TargetObject);

			Assert.Equal("Context", ((ReferenceExpression)context.InitExpression).PropertyName);
			Assert.Equal("__scope__", ((ReferenceExpression)((ReferenceExpression)context.InitExpression).TargetObject).PropertyName);

			Assert.Equal("Position", ((ReferenceExpression)position.InitExpression).PropertyName);
			Assert.Equal("__scope__", ((ReferenceExpression)((ReferenceExpression)position.InitExpression).TargetObject).PropertyName);

			Assert.IsType<ReturnStatement>(projectionMethod.Statements[5]);
		}

		[Fact]
		public void AliasedPropertyMatchGeneratesVariablePattern()
		{
			this.grammar.Rules = new[] {
				new RuleDeclaration {
					Name = "R",
					Pattern = new ObjectPattern {
						Properties = new [] {
							new PropertyPattern {
								Alias = "C",
								Name = "P",
								Pattern = new AnyPattern()
							}
						}
					},
				}
			};

			var result = this.transformer.Transform(unit);
			var generatedClass = (CodeTypeDeclaration)result.Namespaces[0].Types[0];
			var ruleMethod = generatedClass.Members.OfType<CodeMemberMethod>().Single();
			var condition = (ConditionStatement)ruleMethod.Statements[0];
			var ruleAssign = (CodeAssignStatement)condition.TrueStatements[0];
			//this._R = MetaSharp.Transformation.Pattern.Rule(
			//		MetaSharp.Transformation.Pattern.Object(
			//			MetaSharp.Transformation.Pattern.Type(typeof(object)), 
			//			MetaSharp.Transformation.Pattern.Variable("C", 
			//				MetaSharp.Transformation.Pattern.Property("P", MetaSharp.Transformation.Pattern.Any()))), 
			//				typeof(G).GetMethod("R"));
			var ruleExpression = (MethodInvokeExpression)ruleAssign.Right;
			var objectExpression = (MethodInvokeExpression)ruleExpression.Parameters[0];
			var typeExpression = (MethodInvokeExpression)objectExpression.Parameters[0];
			var propertyExpression = (MethodInvokeExpression)objectExpression.Parameters[1];
			var propertyName = (PrimitiveExpression)propertyExpression.Parameters[0];
			var variableExpression = (MethodInvokeExpression)propertyExpression.Parameters[1];
			var variableName = (PrimitiveExpression)variableExpression.Parameters[0];

			Assert.Equal("Rule", ruleExpression.Method.MethodName);
			Assert.Equal("Object", objectExpression.Method.MethodName);
			Assert.Equal("Type", typeExpression.Method.MethodName);
			Assert.Equal("Property", propertyExpression.Method.MethodName);
			Assert.Equal("Variable", variableExpression.Method.MethodName);
			Assert.Equal("C", variableName.Value);
			Assert.Equal("P", propertyName.Value);
		}

		[Fact]
		public void ProjectionWithoutVariablePatternsDoesNotGenerateAnyUnnecessaryStatements()
		{
			this.grammar.Rules = new[] {
				new RuleDeclaration {
					Name = "R",
					Pattern = new ProjectionPattern {
						Pattern = new ReferencePattern(),
						Projection = new Nodes<CodeStatement> { new ExpressionStatement(new PrimitiveExpression(null)) }
					}
				}
			};

			var result = this.transformer.Transform(unit);
			var generatedClass = (CodeTypeDeclaration)result.Namespaces[0].Types[0];
			var projectionMethod = generatedClass.Members.OfType<CodeMemberMethod>().Where(m => m.Name.Contains("projection")).Single();
			Assert.Equal(4, projectionMethod.Statements.Count); // three for the match, position and context variables and one for the actual statement.
		}

		[Fact]
		public void NestedProjectionDoesNotGenerateAnyUnnecessaryStatements()
		{
			this.grammar.Rules = new[] {
				new RuleDeclaration {
					Name = "R",
					Pattern = new ProjectionPattern {
						Pattern = new ProjectionPattern {
							Pattern = new VariablePattern { Pattern = new ReferencePattern() },
							Projection = new Nodes<CodeStatement> { new ExpressionStatement(new PrimitiveExpression(null)) }
						},
						Projection = new Nodes<CodeStatement> { new ExpressionStatement(new PrimitiveExpression(null)) }
					}
				}
			};

			var result = this.transformer.Transform(unit);
			var generatedClass = (CodeTypeDeclaration)result.Namespaces[0].Types[0];
			var projectionMethod0 = generatedClass.Members.OfType<CodeMemberMethod>().Where(m => m.Name.Contains("__projection0__")).Single(); // projections are build depth first
			Assert.Equal(5, projectionMethod0.Statements.Count);

			var projectionMethod1 = generatedClass.Members.OfType<CodeMemberMethod>().Where(m => m.Name.Contains("__projection1__")).Single();
			Assert.Equal(4, projectionMethod1.Statements.Count);
		}

		[Fact]
		public void MainAddsRuleInParameterlessConstructor()
		{
			this.grammar.Rules = new[] {
				new RuleDeclaration {
					Name = "Main",
					Pattern = new AnyPattern()
				}
			};

			var result = this.transformer.Transform(unit);
			var generatedClass = (CodeTypeDeclaration)result.Namespaces[0].Types[0];
			var constructor = generatedClass.Members.OfType<CodeConstructor>().Single();
			Assert.Equal(1, constructor.Statements.Count);
			var invoke = (MethodInvokeExpression)((ExpressionStatement)constructor.Statements[0]).Expression;
			Assert.Equal("Add", invoke.Method.MethodName);
			Assert.Equal(1, invoke.Parameters.Count);
		}

		[Fact]
		public void CanDeclareRuleWithFailPattern()
		{
			this.grammar.Rules = new[]{
				new RuleDeclaration {
					Name = "R",
					Pattern = new FailPattern(),
				}
			};

			var result = this.transformer.Transform(unit);
			var generatedClass = (CodeTypeDeclaration)result.Namespaces[0].Types[0];
			var method = generatedClass.Members.OfType<CodeMemberMethod>().Single();

			Assert.Equal("Fail", ((MethodInvokeExpression)((MethodInvokeExpression)((AssignStatement)((ConditionStatement)method.Statements[0]).TrueStatements[0]).Right).Parameters[0]).Method.MethodName);
		}

		[Fact]
		public void CanDeclareRuleAttributes()
		{
			this.grammar.Rules = new[]{
				new RuleDeclaration {
					Attributes = new [] { new AttributeDeclaration("A") },
					Name = "R",
					Pattern = new FailPattern(),
				}
			};

			var result = this.transformer.Transform(unit);
			var generatedClass = (CodeTypeDeclaration)result.Namespaces[0].Types[0];
			var method = generatedClass.Members.OfType<CodeMemberMethod>().Single();

			Assert.Equal(1, method.CustomAttributes.Count);
			Assert.Equal("A", method.CustomAttributes[0].Name);
		}

		[Fact]
		public void CanDeclareRuleWithNullPattern()
		{
			this.grammar.Rules = new[]{
				new RuleDeclaration {
					Name = "R",
					Pattern = new NullPattern(),
				}
			};

			var result = this.transformer.Transform(unit);
			var generatedClass = (CodeTypeDeclaration)result.Namespaces[0].Types[0];
			var method = generatedClass.Members.OfType<CodeMemberMethod>().Single();

			Assert.Equal("Null", ((MethodInvokeExpression)((MethodInvokeExpression)((AssignStatement)((ConditionStatement)method.Statements[0]).TrueStatements[0]).Right).Parameters[0]).Method.MethodName);
		}

		[Fact]
		public void IsTransformedGrammarClassHasGeneratedCodeAttribute()
		{
			var result = this.transformer.Transform(this.unit);
			var test = result.Namespaces[0];
			Assert.Equal(1, test.Types.Count);
			Assert.True(test.Types[0].CustomAttributes.OfType<AttributeDeclaration>().Any(a => a.Name == typeof(GeneratedCodeAttribute).FullName));
		}

		[Fact]
		public void IsTransformedClassWithinAGrammarHasGeneratedCodeAttribute()
		{
			var unit = new Unit();
			var @namespace = new Namespace("N");
			var type = new TypeDeclaration("T");
			@namespace.Types.Add(type);
			unit.Namespaces.Add(@namespace);

			var result = this.transformer.Transform(unit);
			Assert.True(result.Namespaces[0].Types[0].CustomAttributes.OfType<AttributeDeclaration>().Any(a => a.Name == typeof(GeneratedCodeAttribute).FullName));
		}

		[Fact]
		public void CanDeclareRuleWithErrorUntilPattern()
		{
			this.grammar.Rules = new[]{
				new RuleDeclaration {
					Name = "R",
					Pattern = new ErrorUntilPattern { Pattern = new NullPattern() },
				}
			};

			var result = this.transformer.Transform(unit);
			var generatedClass = (CodeTypeDeclaration)result.Namespaces[0].Types[0];
			var method = generatedClass.Members.OfType<CodeMemberMethod>().Single();

			var errorUntil = ((MethodInvokeExpression)((MethodInvokeExpression)((AssignStatement)((ConditionStatement)method.Statements[0]).TrueStatements[0]).Right).Parameters[0]);
			Assert.Equal("ErrorUntil", errorUntil.Method.MethodName);
			Assert.Equal("Null", ((MethodInvokeExpression)errorUntil.Parameters[0]).Method.MethodName);
		}

		[Fact]
		public void CanDeclareRuleWithErrorUnlessPattern()
		{
			this.grammar.Rules = new[]{
				new RuleDeclaration {
					Name = "R",
					Pattern = new ErrorUnlessPattern { Pattern = new NullPattern() },
				}
			};

			var result = this.transformer.Transform(unit);
			var generatedClass = (CodeTypeDeclaration)result.Namespaces[0].Types[0];
			var method = generatedClass.Members.OfType<CodeMemberMethod>().Single();

			var errorUntil = ((MethodInvokeExpression)((MethodInvokeExpression)((AssignStatement)((ConditionStatement)method.Statements[0]).TrueStatements[0]).Right).Parameters[0]);
			Assert.Equal("ErrorUnless", errorUntil.Method.MethodName);
			Assert.Equal("Null", ((MethodInvokeExpression)errorUntil.Parameters[0]).Method.MethodName);
		}
	}
}
