﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Xunit;
using Xunit.Extensions;
using MetaSharp.Transformation.Grammar;
using MetaSharp.Transformation.Lang.Ast;

namespace MetaSharp.Transformation.Tests.Grammar
{
	public class ParserTests : IUseFixture<ParserTests.TestParser>
	{
		private TestParser parser;
		public void SetFixture(TestParser data)
		{
			this.parser = data;
		}

		[Fact]
		public void CanParseIdentifierSuccessfully()
		{
			var text = "i:ABC";
			var result = (Nodes)this.parser.Parse(text);
			Assert.Equal("i", Node.Unwrap(result[0]));
			Assert.Equal(":", Node.Unwrap(result[1]));
			Assert.Equal("ABC", Node.Unwrap(result[2]));
			Assert.Equal(3, result.Count);
		}

		[Fact]
		public void IdentifierDoesNotParseInvalidText()
		{
			var text = "i:123";
			var result = this.parser.Parse(text);
			Assert.IsType<Fail>(result);
		}

		[Fact]
		public void CanParseNumberSuccessfully()
		{
			var text = "n:123";
			var result = (Nodes)this.parser.Parse(text);
			Assert.Equal("n", Node.Unwrap(result[0]));
			Assert.Equal(":", Node.Unwrap(result[1]));
			Assert.Equal("123", Node.Unwrap(result[2]));
			Assert.Equal(3, result.Count);
		}

		[Fact]
		public void NumberDoesNotParseInvalidText()
		{
			var text = "n:ABC";
			var result = this.parser.Parse(text);
			Assert.IsType<Fail>(result);
		}

		[Fact]
		public void CanParseLetterSuccessfully()
		{
			var text = "l:A";
			var result = (Nodes)this.parser.Parse(text);
			Assert.Equal("l", Node.Unwrap(result[0]));
			Assert.Equal(":", Node.Unwrap(result[1]));
			Assert.Equal("A", Node.Unwrap(result[2]));
			Assert.Equal(3, result.Count);
		}

		[Fact]
		public void LetterDoesNotParseInvalidText()
		{
			var text = "l:9";
			var result = this.parser.Parse(text);
			Assert.IsType<Fail>(result);
		}

		[Fact]
		public void CanParseDigitSuccessfully()
		{
			var text = "d:7";
			var result = (Nodes)this.parser.Parse(text);
			Assert.Equal("d", Node.Unwrap(result[0]));
			Assert.Equal(":", Node.Unwrap(result[1]));
			Assert.Equal("7", Node.Unwrap(result[2]));
			Assert.Equal(3, result.Count);
		}

		[Fact]
		public void DigitDoesNotParseInvalidText()
		{
			var text = "d:A";
			var result = this.parser.Parse(text);
			Assert.IsType<Fail>(result);
		}

		[Fact]
		public void CanParseHexSuccessfully()
		{
			var text = "h:A";
			var result = (Nodes)this.parser.Parse(text);
			Assert.Equal("h", Node.Unwrap(result[0]));
			Assert.Equal(":", Node.Unwrap(result[1]));
			Assert.Equal("A", Node.Unwrap(result[2]));
			Assert.Equal(3, result.Count);
		}

		[Fact]
		public void HexDoesNotParseInvalidText()
		{
			var text = "h:Z";
			var result = this.parser.Parse(text);
			Assert.IsType<Fail>(result);
		}

		[Theory]
		[InlineData("string")]
		[InlineData("")]
		[InlineData("\u1234")]
		[InlineData("\U0000FFFF")]
		[InlineData("\xF")]
		[InlineData("\x21")]
		[InlineData("\x021")]
		[InlineData("\x0021")]
		[InlineData("\xFZ")]
		[InlineData("\0")]
		[InlineData("\\0")]
		[InlineData("\\\\0")]
		[InlineData(@"\0")]
		[InlineData(@"\\0")]
		[InlineData(@"\\\\0")]
		[InlineData("\n")]
		[InlineData("\\n")]
		[InlineData("\\\\n")]
		[InlineData(@"\n")]
		[InlineData(@"\\n")]
		[InlineData(@"\\\\n")]
		[InlineData("\r")]
		[InlineData("\\r")]
		[InlineData("\\\\r")]
		[InlineData(@"\r")]
		[InlineData(@"\\r")]
		[InlineData(@"\\\\r")]
		[InlineData("\t")]
		[InlineData("\\t")]
		[InlineData("\\\\t")]
		[InlineData(@"\t")]
		[InlineData(@"\\t")]
		[InlineData(@"\\\\t")]
		//[InlineData(@"\a")] // CodeDom has bugs! Does not correct escape \(a|b|f|v). Weak sauce.
		public void CanParseStringLiteralSuccessfully(string input)
		{
			var quoted = "s:\"" + input + "\"";
			var result = (Nodes)this.parser.Parse(quoted);
			var s = Node.Unwrap(result[2]);
			var e = new PrimitiveExpression(s);
			var actual = e.Print().Trim('"');
			var expected = EscapeString(input);
			Assert.Equal(expected, actual);
		}

		private static string EscapeString(string str)
		{
			str = str.Replace("\0", @"\0");
			str = str.Replace("\n", @"\n");
			str = str.Replace("\r", @"\r");
			str = str.Replace("\t", @"\t");
			return str;
		}

		[Fact]
		public void StringLiteralDoesNotParseInvalidText()
		{
			var text = "s:'a'";
			var result = this.parser.Parse(text);
			Assert.IsType<Fail>(result);
		}

		[Theory]
		[InlineData(@"a", 'a')]
		[InlineData(@"\'", '\'')]
		[InlineData(@"\\", '\\')]
		[InlineData(@"\0", '\0')]
		[InlineData(@"\n", '\n')]
		[InlineData(@"\r", '\r')]
		[InlineData(@"\t", '\t')]
		[InlineData(@"\u0021", '\u0021')]
		[InlineData(@"\U00000021", '\U00000021')]
		[InlineData(@"\xF", '\xF')]
		[InlineData(@"\xF0", '\xF0')]
		[InlineData(@"\x0F0", '\x0F0')]
		[InlineData(@"\x00F0", '\x00F0')]
		public void CanParseCharacterLiteralSuccessfully(string escaped, char expected)
		{
			var text = "c:'" + escaped + "'";
			var result = (Nodes)this.parser.Parse(text);
			Assert.Equal(expected, Node.Unwrap(result[2]));
		}

		[Fact]
		public void CharacterLiteralDoesNotParseInvalidText()
		{
			var text = "c:\"abc\"";
			var result = this.parser.Parse(text);
			Assert.IsType<Fail>(result);
		}

		[Fact]
		public void CanParseListSuccessfully()
		{
			var text = "li:a.b.c";
			var result = (Nodes)this.parser.Parse(text);
			Assert.Equal("li", Node.Unwrap(result[0]));
			Assert.Equal(":", Node.Unwrap(result[1]));
			var list = (Nodes)result[2];
			Assert.Equal("a", Node.Unwrap(list[0]));
			Assert.Equal("b", Node.Unwrap(list[1]));
			Assert.Equal("c", Node.Unwrap(list[2]));
			Assert.Equal(3, result.Count);
		}

		[Fact]
		public void ListDoesNotParseInvalidText()
		{
			var text = "h:a,b,c";
			var result = this.parser.Parse(text);
			Assert.IsType<Fail>(result);
		}

		public class TestParser : Parser
		{
			public TestParser()
			{
				// Main 
				//	= "i" ":" Identifier
				//	| "n" ":" Number
				//	| "l" ":" Letter
				//	| "d" ":" Digit
				//	| "h" ":" Hex
				//	| "s" ":" StringLiteral
				//	| "c" ":" CharacterLiteral
				//	| "li" ":" List(Letter, ".")
				this.Add(Pattern.Or(
					Pattern.And(Pattern.Value("i"), Pattern.Value(":"), this.Identifier),
					Pattern.And(Pattern.Value("n"), Pattern.Value(":"), this.Number),
					Pattern.And(Pattern.Value("l"), Pattern.Value(":"), this.Letter),
					Pattern.And(Pattern.Value("d"), Pattern.Value(":"), this.Digit),
					Pattern.And(Pattern.Value("h"), Pattern.Value(":"), this.Hex),
					Pattern.And(Pattern.Value("s"), Pattern.Value(":"), this.StringLiteral),
					Pattern.And(Pattern.Value("c"), Pattern.Value(":"), this.CharacterLiteral),
					Pattern.And(Pattern.Value("li"), Pattern.Value(":"), Pattern.Call(this.List, this.Letter, Pattern.Value(".")))));
			}
		}
	}
}
