﻿using System;
using System.CodeDom;
using System.Linq;
using MetaSharp.Transformation.Grammar;
using MetaSharp.Transformation.Grammar.Ast;
using Xunit;
using MetaSharp.Transformation.Lang.Ast;
using g = MetaSharp.Transformation.Grammar.Ast;

namespace MetaSharp.Transformation.Tests.Grammar
{
	public class PatternTests : IUseFixture<GrammarTransformer>
	{
		private GrammarTransformer transformer;
		private Unit unit;
		private RuleDeclaration rule;

		public void SetFixture(GrammarTransformer data)
		{
			this.transformer = data;
		}

		public PatternTests()
		{
			this.unit = new Unit();
			var N = new Namespace("N");
			var G = new GrammarDeclaration
			{
				Name = "G",
				Rules = new[] { (this.rule = new RuleDeclaration { Name = "R" }) }
			};

			N.Imports.Add(new Import("System"));
			N.Imports.Add(new Import("System.Collections.Generic"));
			N.Types.Add(G);
			this.unit.Namespaces.Add(N);
		}

		[Fact]
		public void AndPatternCreatesCorrectInvoke()
		{
			this.rule.Pattern = new AndPattern()
			{
				Patterns = new[] { 
					new AnyPattern(), 
					new AnyPattern() 
				}
			};

			var result = this.transformer.Transform(this.unit);
			var invoke = GetPatternInvokeExpression(result);
			Assert.Equal("And", invoke.Method.MethodName);
			Assert.Equal(2, invoke.Parameters.Count);
		}

		[Fact]
		public void AnyPatternCreatesCorrectInvoke()
		{
			this.rule.Pattern = new AnyPattern();
			var result = this.transformer.Transform(this.unit);
			var invoke = GetPatternInvokeExpression(result);
			Assert.Equal("Any", invoke.Method.MethodName);
			Assert.Empty(invoke.Parameters);
		}

		[Fact]
		public void ArrayPatternCreatesCorrectInvoke()
		{
			this.rule.Pattern = new ArrayPattern { Pattern = new AnyPattern() };
			var result = this.transformer.Transform(this.unit);
			var invoke = GetPatternInvokeExpression(result);
			Assert.Equal("Array", invoke.Method.MethodName);
			Assert.Equal(1, invoke.Parameters.Count);
		}

		[Fact]
		public void DefaultPatternCreatesCorrectInvoke()
		{
			this.rule.Pattern = new DefaultPattern();
			var result = this.transformer.Transform(this.unit);
			var invoke = GetPatternInvokeExpression(result);
			Assert.Equal("Default", invoke.Method.MethodName);
			Assert.Empty(invoke.Parameters);
		}

		[Fact]
		public void ExactQuantityPatternCreatesCorrectInvoke()
		{
			this.rule.Pattern = new ExactQuantityPattern
			{
				Quantity = 7,
				Pattern = new AnyPattern()
			};

			var result = this.transformer.Transform(this.unit);
			var invoke = GetPatternInvokeExpression(result);
			Assert.Equal("ExactQuantity", invoke.Method.MethodName);
			Assert.Equal(2, invoke.Parameters.Count);
			Assert.NotNull(invoke.Parameters[0]);
			Assert.Equal(7, ((CodePrimitiveExpression)invoke.Parameters[1]).Value);
		}

		[Fact]
		public void NotPatternCreatesCorrectInvoke()
		{
			this.rule.Pattern = new NotPattern { Pattern = new AnyPattern() };
			var result = this.transformer.Transform(this.unit);
			var invoke = GetPatternInvokeExpression(result);
			Assert.Equal("Not", invoke.Method.MethodName);
			Assert.Equal(1, invoke.Parameters.Count);
			Assert.NotNull(invoke.Parameters[0]);
		}

		[Fact]
		public void ObjectPatternCreatesCorrectInvoke()
		{
			this.rule.Pattern = new ObjectPattern { Type = new TypeReference("T"), Properties = new[] { new PropertyPattern { Name = "A", Pattern = new AnyPattern() } } };
			var result = this.transformer.Transform(this.unit);
			var invoke = GetPatternInvokeExpression(result);
			Assert.Equal("Object", invoke.Method.MethodName);
			Assert.Equal(2, invoke.Parameters.Count);
		}

		[Fact]
		public void OneOrMoreCreatesCorrectInvoke()
		{
			this.rule.Pattern = new OneOrMorePattern { Pattern = new AnyPattern() };
			var result = this.transformer.Transform(this.unit);
			var invoke = GetPatternInvokeExpression(result);
			Assert.Equal("OneOrMore", invoke.Method.MethodName);
			Assert.Equal(1, invoke.Parameters.Count);
			Assert.NotNull(invoke.Parameters[0]);
		}

		[Fact]
		public void OrPatternCreatesCorrectInvoke()
		{
			this.rule.Pattern = new OrPattern { Patterns = new[] { new AnyPattern(), new AnyPattern() } };
			var result = this.transformer.Transform(this.unit);
			var invoke = GetPatternInvokeExpression(result);
			Assert.Equal("Or", invoke.Method.MethodName);
			Assert.Equal(2, invoke.Parameters.Count);
			Assert.NotNull(invoke.Parameters[0]);
			Assert.NotNull(invoke.Parameters[1]);
		}

		[Fact]
		public void ProjectionPatternCreatesInvokeAndMethod()
		{
			this.rule.Pattern = new ProjectionPattern { Pattern = new AnyPattern(), Projection = new Nodes<CodeStatement> { new ExpressionStatement(new ReferenceExpression("x")) } };
			var result = this.transformer.Transform(this.unit);
			var expression = result
				.Namespaces[0]
				.Types[0]
				.Members
				.OfType<CodeMemberMethod>()
				.Where(m => m.Name == "R")
				.Single()
				.Statements
				.OfType<CodeConditionStatement>()
				.Single()
				.TrueStatements
				.Cast<CodeAssignStatement>()
				.Single()
				.Right;

			var projectionMethod = result
				.Namespaces[0]
				.Types[0]
				.Members
				.OfType<CodeMemberMethod>()
				.Where(m => !(m is CodeConstructor) && m.Attributes == MemberAttributes.Private)
				.Single();

			var invoke = (CodeMethodInvokeExpression)((CodeMethodInvokeExpression)expression).Parameters[0];
			var projectionDelegate = (CodeDelegateCreateExpression)invoke.Parameters[1];
			Assert.Equal(projectionMethod.Name, projectionDelegate.MethodName);
			Assert.Equal(2, invoke.Parameters.Count);
			Assert.NotNull(invoke.Parameters[0]);
			Assert.NotNull(invoke.Parameters[1]);
		}

		[Fact]
		public void OtherPatternsCorrectlyPromoteProjectionMethods()
		{
			Action<CodeCompileUnit> verify = u =>
			{
				Assert.Equal(1, u
					.Namespaces[0]
					.Types[0]
					.Members
					.OfType<CodeMemberMethod>()
					.Where(m => !(m is CodeConstructor) && m.Attributes == MemberAttributes.Private)
					.Count());
			};

			var projection = new Nodes<CodeStatement> { new ExpressionStatement(new ReferenceExpression("x")) };
			this.rule.Pattern = new AndPattern { Patterns = new[] { new ProjectionPattern { Pattern = new AnyPattern(), Projection = projection } } };
			var result = this.transformer.Transform(this.unit);
			verify(result);

			this.rule.Pattern = new OrPattern { Patterns = new[] { new ProjectionPattern { Pattern = new AnyPattern(), Projection = projection } } };
			result = this.transformer.Transform(this.unit);
			verify(result);

			this.rule.Pattern = new ArrayPattern { Pattern = new ProjectionPattern { Pattern = new AnyPattern(), Projection = projection } };
			result = this.transformer.Transform(this.unit);
			verify(result);

			this.rule.Pattern = new ExactQuantityPattern { Pattern = new ProjectionPattern { Pattern = new AnyPattern(), Projection = projection } };
			result = this.transformer.Transform(this.unit);
			verify(result);

			this.rule.Pattern = new NotPattern { Pattern = new ProjectionPattern { Pattern = new AnyPattern(), Projection = projection } };
			result = this.transformer.Transform(this.unit);
			verify(result);

			this.rule.Pattern = new OneOrMorePattern { Pattern = new ProjectionPattern { Pattern = new AnyPattern(), Projection = projection } };
			result = this.transformer.Transform(this.unit);
			verify(result);

			this.rule.Pattern = new CallPattern { Name = "A", Parameters = new[] { new ProjectionPattern { Pattern = new AnyPattern(), Projection = projection } } };
			result = this.transformer.Transform(this.unit);
			verify(result);

			this.rule.Pattern = new ObjectPattern { Type = new TypeReference("A"), Properties = new[] { new PropertyPattern { Name = "P", Pattern = new ProjectionPattern { Pattern = new AnyPattern(), Projection = projection } } } };
			result = this.transformer.Transform(this.unit);
			verify(result);

			this.rule.Pattern = new ZeroOrMorePattern { Pattern = new ProjectionPattern { Pattern = new AnyPattern(), Projection = projection } };
			result = this.transformer.Transform(this.unit);
			verify(result);

			this.rule.Pattern = new ZeroOrOnePattern { Pattern = new ProjectionPattern { Pattern = new AnyPattern(), Projection = projection } };
			result = this.transformer.Transform(this.unit);
			verify(result);
		}

		[Fact]
		public void RangePatternCreatesCorrectExpression()
		{
			this.rule.Pattern = new RangePattern { Left = 0, Right = 10 };
			var result = this.transformer.Transform(this.unit);
			var invoke = GetPatternInvokeExpression(result);
			Assert.Equal("Range", invoke.Method.MethodName);
			Assert.Equal(2, invoke.Parameters.Count);
			Assert.Equal(0, ((CodePrimitiveExpression)invoke.Parameters[0]).Value);
			Assert.Equal(10, ((CodePrimitiveExpression)invoke.Parameters[1]).Value);
		}

		[Fact]
		public void ReferencePatternWithoutParametersCreatesCorrectExpression()
		{
			this.rule.Pattern = new ReferencePattern { Name = "X" };
			var result = this.transformer.Transform(this.unit);
			var invoke = (ReferenceExpression)GetPatternProjection(result);
			Assert.Equal("X", invoke.PropertyName);
		}

		[Fact]
		public void CallPatternWithParametersCreatesCorrectExpression()
		{
			this.rule.Pattern = new CallPattern { Name = "X", Parameters = new[] { new AnyPattern() } };
			var result = this.transformer.Transform(this.unit);
			var invoke = GetPatternInvokeExpression(result);
			Assert.Equal("Call", invoke.Method.MethodName);

			Assert.Equal(2, invoke.Parameters.Count);
			var reference = (CodeDelegateCreateExpression)invoke.Parameters[0];
			Assert.Equal("X", reference.MethodName);
			Assert.IsType(typeof(ThisExpression), reference.TargetObject);
			Assert.Equal("MetaSharp.Transformation.ParameterizedPatternMatch", reference.DelegateType.BaseType);

			var parameter = (CodeMethodInvokeExpression)invoke.Parameters[1];
			Assert.NotNull(parameter);
		}

		[Fact]
		public void ValuePatternCreatesCorrectExpression()
		{
			this.rule.Pattern = new ValuePattern { Value = 7 };
			var result = this.transformer.Transform(this.unit);
			var exp = (CodeMethodInvokeExpression)GetPatternProjection(result);
			Assert.Equal("Value", exp.Method.MethodName);
			Assert.Equal("MetaSharp.Transformation.Pattern", ((CodeTypeReferenceExpression)exp.Method.TargetObject).Type.BaseType);
			Assert.Equal(7, ((CodePrimitiveExpression)exp.Parameters[0]).Value);
			Assert.Equal(1, exp.Parameters.Count);
		}

		[Fact]
		public void ValuePatternWithNullCreatesCorrectExpression()
		{
			this.rule.Pattern = new ValuePattern { Value = null };
			var result = this.transformer.Transform(this.unit);
			var exp = (CodeMethodInvokeExpression)GetPatternProjection(result);
			Assert.Equal("Value", exp.Method.MethodName);
			Assert.Equal("MetaSharp.Transformation.Pattern", ((CodeTypeReferenceExpression)exp.Method.TargetObject).Type.BaseType);
			Assert.Null(((CodePrimitiveExpression)exp.Parameters[0]).Value);
			Assert.Equal(1, exp.Parameters.Count);
		}

		[Fact]
		public void ZeroOrMorePatternCreatsCorrectExpression()
		{
			this.rule.Pattern = new ZeroOrMorePattern { Pattern = new AnyPattern() };
			var result = this.transformer.Transform(this.unit);
			var exp = GetPatternInvokeExpression(result);
			Assert.Equal("ZeroOrMore", exp.Method.MethodName);
			Assert.Equal(1, exp.Parameters.Count);
			Assert.NotNull(exp.Parameters[0]);
		}

		[Fact]
		public void ZeroOrOnePatternCreatsCorrectExpression()
		{
			this.rule.Pattern = new ZeroOrOnePattern { Pattern = new AnyPattern() };
			var result = this.transformer.Transform(this.unit);
			var exp = GetPatternInvokeExpression(result);
			Assert.Equal("ZeroOrOne", exp.Method.MethodName);
			Assert.Equal(1, exp.Parameters.Count);
			Assert.NotNull(exp.Parameters[0]);
		}

		private static CodeMethodInvokeExpression GetPatternInvokeExpression(CodeCompileUnit unit)
		{
			var patternInvoke = (CodeMethodInvokeExpression)GetPatternProjection(unit);
			Assert.Equal("MetaSharp.Transformation.Pattern", ((CodeTypeReferenceExpression)patternInvoke.Method.TargetObject).Type.BaseType);
			return patternInvoke;
		}

		private static CodeExpression GetPatternProjection(CodeCompileUnit unit)
		{
			//namespace N
			//{
			//    using System;
			//    using System.Collections.Generic;

			//    public partial class G : MetaSharp.Transformation.Transformer
			//    {
			//        private MetaSharp.Transformation.PatternMatch _R;
			//        protected virtual object R(MetaSharp.Transformation.PatternContext context, MetaSharp.Transformation.IStream stream)
			//        {
			//            if ((this._R == null))
			//            {
			//                this._R = MetaSharp.Transformation.Pattern.Rule(MetaSharp.Transformation.Pattern.And(MetaSharp.Transformation.Pattern.Any(), MetaSharp.Transformation.Pattern.Any()));
			//            }
			//            return this._R(context, stream);
			//        }
			//    }
			//}

			var expression = unit
				.Namespaces[0]
				.Types[0]
				.Members
				.OfType<CodeMemberMethod>()
				.Single()
				.Statements
				.OfType<CodeConditionStatement>()
				.Single()
				.TrueStatements
				.OfType<CodeAssignStatement>()
				.Single()
				.Right;

			return ((CodeMethodInvokeExpression)expression).Parameters.Cast<CodeExpression>().First();
		}
	}
}
