﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Xunit;
using MetaSharp.Transformation.Grammar;
using Xunit.Extensions;

namespace MetaSharp.Transformation.Tests.Grammar
{
	public class TokenizerTests : IUseFixture<Tokenizer>
	{
		private Tokenizer tokenizer;
		public void SetFixture(Tokenizer data)
		{
			this.tokenizer = data;
		}

		[Fact]
		public void SingleCharacterIsAToken()
		{
			var input = "j";
			var result = (Nodes)this.tokenizer.Parse(input);
			Assert.Equal("j", Node.Unwrap(result[0]));
			Assert.Equal(1, result.Count);
		}

		[Fact]
		public void TwoNonIdentifiersAreSeparateTokens()
		{
			var input = "..";
			var result = (Nodes)this.tokenizer.Parse(input);
			Assert.Equal(".", Node.Unwrap(result[0]));
			Assert.Equal(".", Node.Unwrap(result[1]));
			Assert.Equal(2, result.Count);
		}

		[Theory]
		[InlineData(int.MinValue)]
		[InlineData(int.MaxValue)]
		[InlineData(-99)]
		[InlineData(-1)]
		[InlineData(0)]
		[InlineData(1)]
		[InlineData(99)]
		[InlineData(0xFFFFFF)]
		public void IntIsTokenized(int x)
		{
			var input = x.ToString();
			var result = (Nodes)this.tokenizer.Parse(input);
			Assert.Equal(input, Node.Unwrap(result[0]));
			Assert.Equal(1, result.Count);

			input = string.Format("0x{0:X}", x);
			result = (Nodes)this.tokenizer.Parse(input);
			Assert.Equal(input, Node.Unwrap(result[0]));
			Assert.Equal(1, result.Count);
		}

		[Theory]
		[InlineData("_a")]
		[InlineData("a_")]
		[InlineData("a1")]
		[InlineData("abc123")]
		[InlineData("_abcdefghijklmnopqrstuvwxyz_ABCDEFGHIJKLMNOPQRSTUVWXYZ_0123456789_")]
		public void IdentifierIsTokenized(string i)
		{
			var result = (Nodes)this.tokenizer.Parse(i);
			Assert.Equal(i, Node.Unwrap(result[0]));
			Assert.Equal(1, result.Count);
		}

		[Fact]
		public void NumberBeforeIdentifierIsTwoTokens()
		{
			var input = "123abc";
			var result = (Nodes)this.tokenizer.Parse(input);
			Assert.Equal("123", Node.Unwrap(result[0]));
			Assert.Equal("abc", Node.Unwrap(result[1]));
			Assert.Equal(2, result.Count);
		}

		[Theory]
		[InlineData(@"'\u09AF'")]
		[InlineData(@"'\U0189ABEF'")]
		[InlineData(@"'\x0'")]
		[InlineData(@"'\x01'")]
		[InlineData(@"'\x012'")]
		[InlineData(@"'\x0123'")]
		[InlineData(@"'\''")]
		[InlineData(@"'\\'")]
		[InlineData(@"'\0'")]
		[InlineData(@"'\n'")]
		[InlineData(@"'\r'")]
		[InlineData(@"'\t'")]
		[InlineData(@"'a'")]
		[InlineData(@"'a'")]
		[InlineData(@"'A'")]
		[InlineData(@"'<'")]
		[InlineData(@"'€'")]
		[InlineData(@"'‘'")]
		[InlineData(@"'_'")]
		public void CharacterLiteralIsTokenized(string i)
		{
			var result = (Nodes)this.tokenizer.Parse(i);
			Assert.Equal(i, Node.Unwrap(result[0]));
			Assert.Equal(1, result.Count);
		}

		[Theory]
		[InlineData(@"""""")]
		[InlineData(@"""\""a\""""")]
		[InlineData(@"""\""\\\0\n\r\t""")]
		[InlineData(@"""\u0123\U0123ABCD\x0\x01\x012\x0123""")]
		[InlineData(@"""abcdefghijklmnopqrstuvwxyz0123456789!@#$%^&*()_+-=[]\\;',./<>?:{}|`~€£¥""")]
		public void StringLiteralIsTokenized(string i)
		{
			var result = (Nodes)this.tokenizer.Parse(i);
			Assert.Equal(i, Node.Unwrap(result[0]));
			Assert.Equal(1, result.Count);
		}

		[Theory]
		[InlineData("a z")]
		[InlineData("a\0z")]
		[InlineData("a\az")]
		[InlineData("a\bz")]
		[InlineData("a\fz")]
		[InlineData("a\nz")]
		[InlineData("a\rz")]
		[InlineData("a\tz")]
		[InlineData("a\vz")]
		public void WhitespaceIsTokenized(string i)
		{
			var result = (Nodes)this.tokenizer.Parse(i);
			Assert.Equal("a", Node.Unwrap(result[0]));
			Assert.Equal(i[1].ToString(), Node.Unwrap(result[1]));
			Assert.Equal("z", Node.Unwrap(result[2]));
			Assert.Equal(3, result.Count);
		}

		[Fact]
		public void SingleLineAllLineCommentsAreTokenized()
		{
			var text = "//all are tokenized";
			var result = (Nodes)this.tokenizer.Parse(text);
			Assert.Equal(text, Node.Unwrap(result[0]));
			Assert.Equal(1, result.Count);
		}

		[Fact]
		public void SingleLineAllLineCommentsEndingWithSlashNAreTokenized()
		{
			var text = "//all are tokenized\n";
			var result = (Nodes)this.tokenizer.Parse(text);
			Assert.Equal(text, Node.Unwrap(result[0]));
			Assert.Equal(1, result.Count);
		}

		[Fact]
		public void TwoLinesAreSeparateTokens()
		{
			var text = "//all are tokenized";
			var result = (Nodes)this.tokenizer.Parse(text + "\n" + text);
			Assert.Equal(text + "\n", Node.Unwrap(result[0]));
			Assert.Equal(text, Node.Unwrap(result[1]));
			Assert.Equal(2, result.Count);
		}

		[Fact]
		public void FirstLineCommentedSecondLineNotIsTokenizedCorrectly()
		{
			var text = "//comment\nabc 123";
			var result = (Nodes)this.tokenizer.Parse(text);
			Assert.Equal("//comment\n", Node.Unwrap(result[0]));
			Assert.Equal("abc", Node.Unwrap(result[1]));
			Assert.Equal(" ", Node.Unwrap(result[2]));
			Assert.Equal("123", Node.Unwrap(result[3]));
			Assert.Equal(4, result.Count);
		}

		[Fact]
		public void LineEndingWithLineCommentIsTokenizedCorrectly()
		{
			var text = "abc 123 //comment";
			var result = (Nodes)this.tokenizer.Parse(text);
			Assert.Equal("abc", Node.Unwrap(result[0]));
			Assert.Equal(" ", Node.Unwrap(result[1]));
			Assert.Equal("123", Node.Unwrap(result[2]));
			Assert.Equal(" ", Node.Unwrap(result[3]));
			Assert.Equal("//comment", Node.Unwrap(result[4]));
			Assert.Equal(5, result.Count);
		}

		[Fact]
		public void SecondLineContainsLineCommentIsTokenizedCorrectly()
		{
			var text = "abc 123\n//comment";
			var result = (Nodes)this.tokenizer.Parse(text);
			Assert.Equal("abc", Node.Unwrap(result[0]));
			Assert.Equal(" ", Node.Unwrap(result[1]));
			Assert.Equal("123", Node.Unwrap(result[2]));
			Assert.Equal("\n", Node.Unwrap(result[3]));
			Assert.Equal("//comment", Node.Unwrap(result[4]));
			Assert.Equal(5, result.Count);
		}

		[Fact]
		public void SecondLineContainsLineCommentWithSlashNIsTokenizedCorrectly()
		{
			var text = "abc 123\n//comment\n";
			var result = (Nodes)this.tokenizer.Parse(text);
			Assert.Equal("abc", Node.Unwrap(result[0]));
			Assert.Equal(" ", Node.Unwrap(result[1]));
			Assert.Equal("123", Node.Unwrap(result[2]));
			Assert.Equal("\n", Node.Unwrap(result[3]));
			Assert.Equal("//comment\n", Node.Unwrap(result[4]));
			Assert.Equal(5, result.Count);
		}

		//InlineCommentToken = '/' '*' ((!'*' CharToken) | ('*' !'/'))* '*' '/';
		[Theory]
		[InlineData(@"/**/")]
		[InlineData(@"/***/")]
		[InlineData(@"/*/*/")]
		[InlineData(@"/*/**/")]
		[InlineData(@"/* */")]
		[InlineData(@"/** /*/")]
		[InlineData(@"/* * /*/")]
		[InlineData(@"/* * / */")]
		[InlineData(@"/*abc123*/")]
		public void ValidInlineCommentsTokenizeCorrectly(string inline)
		{
			var result = (Nodes)this.tokenizer.Parse(inline);
			Assert.Equal(inline, Node.Unwrap(result[0]));
			Assert.Equal(1, result.Count);
		}

		[Theory]
		[InlineData(@"/ **/")]
		[InlineData(@"/** /")]
		[InlineData(@"**/")]
		[InlineData(@"/**")]
		[InlineData(@"/*/")]
		[InlineData(@"/ /")]
		[InlineData(@"**")]
		[InlineData(@"*")]
		public void InvalidInlineCommentsTokenizeCorrectly(string inline)
		{
			var result = (Nodes)this.tokenizer.Parse(inline);
			for (int x = 0; x < inline.Length; x++)
				Assert.Equal(inline[x].ToString(), Node.Unwrap(result[x]));

			Assert.Equal(inline.Length, result.Count);
		}

		[Fact]
		public void InlineCommentBetweenIdentifiersTokenizesCorrectly()
		{
			var text = "abc/*123*/xyz";
			var result = (Nodes)this.tokenizer.Parse(text);
			Assert.Equal("abc", Node.Unwrap(result[0]));
			Assert.Equal("/*123*/", Node.Unwrap(result[1]));
			Assert.Equal("xyz", Node.Unwrap(result[2]));
			Assert.Equal(3, result.Count);
		}

		[Fact]
		public void InlineCommentBeforeIdentifierTokenizesCorrectly()
		{
			var text = "/*abc*/xyz";
			var result = (Nodes)this.tokenizer.Parse(text);
			Assert.Equal("/*abc*/", Node.Unwrap(result[0]));
			Assert.Equal("xyz", Node.Unwrap(result[1]));
			Assert.Equal(2, Node.Unwrap(result.Count));
		}

		[Fact]
		public void InlineCommentAfterIdentifierTokenizesCorrectly()
		{
			var text = "abc/*xyz*/";
			var result = (Nodes)this.tokenizer.Parse(text);
			Assert.Equal("abc", Node.Unwrap(result[0]));
			Assert.Equal("/*xyz*/", Node.Unwrap(result[1]));
			Assert.Equal(2, result.Count);
		}

		[Fact]
		public void MultiLineInlineCommentTokenizesCorrectly()
		{
			var text = "/*abc\nxyz*/";
			var result = (Nodes)this.tokenizer.Parse(text);
			Assert.Equal(text, Node.Unwrap(result[0]));
			Assert.Equal(1, result.Count);
		}

		[Fact]
		public void CustomTokensCanBeOverridden()
		{
			var text = "->";
			var result = (Nodes)this.tokenizer.Parse(text);
			Assert.Equal("-", Node.Unwrap(result[0]));
			Assert.Equal(">", Node.Unwrap(result[1]));
			Assert.Equal(2, result.Count);

			var customTokenizer = new CustomTokenizer();
			result = (Nodes)customTokenizer.Parse(text);
			Assert.Equal("->", Node.Unwrap(result[0]));
			Assert.Equal(1, result.Count);
		}

		private class CustomTokenizer : Tokenizer
		{
			private PatternMatch customTokens;
			public CustomTokenizer()
			{
				this.customTokens = Pattern.Rule(
					Pattern.And(
						Pattern.Value('-'),
						Pattern.Value('>')));
			}

			public override INode CustomTokens(PatternContext context, IStream stream)
			{
				return this.customTokens(context, stream);
			}
		}
	}
}
