﻿using System;
using System.CodeDom;
using System.CodeDom.Compiler;
using System.IO;
using System.Linq;
using System.Reflection;
using MetaSharp.Transformation.Grammar;
using Microsoft.Cci;
using Microsoft.CSharp;
using Xunit;
using MetaSharp.Transformation.Lang.Ast;
using MetaSharp.Transformation.Grammar.Ast;
using System.Text;

namespace MetaSharp.Transformation.Tests.Integration
{
	public class IntegrationTests : IUseFixture<GrammarTransformer>, IUseFixture<GrammarParser>
	{
		private GrammarParser parser;
		private GrammarTransformer transformer;
		public void SetFixture(GrammarTransformer data)
		{
			this.transformer = data;
		}

		public void SetFixture(GrammarParser data)
		{
			this.parser = data;
		}

		[Fact]
		public void CanCompileTokenizer()
		{
			CompileAndRunGrammar(
				this.parser,
				this.transformer,
				"Integration\\Tokenizer.g",
				"MetaSharp.Transformation.Tokenizer",
				(instance, parseMethod) =>
				{
					var result = (Nodes)parseMethod.Invoke(instance, new object[] { "a b c" });
					Assert.Equal("a", Node.Unwrap(result[0]));
					Assert.Equal(" ", Node.Unwrap(result[1]));
					Assert.Equal("b", Node.Unwrap(result[2]));
					Assert.Equal(" ", Node.Unwrap(result[3]));
					Assert.Equal("c", Node.Unwrap(result[4]));
					Assert.Equal(5, result.Count);
				});
		}

		[Fact]
		public void CanCompileInterleaver()
		{
			CompileAndRunGrammar(
				this.parser,
				this.transformer,
				"Integration\\Interleaver.g",
				"MetaSharp.Transformation.Interleaver",
				(instance, parseMethod) =>
				{
					var result = (Nodes)parseMethod.Invoke(instance, new object[] { "a b c" });
					Assert.Equal("a", Node.Unwrap(result[0]));
					Assert.Equal("b", Node.Unwrap(result[1]));
					Assert.Equal("c", Node.Unwrap(result[2]));
					Assert.Equal(3, result.Count);
				});
		}

		[Fact]
		public void CanCompileParser()
		{
			CompileAndRunGrammar(
				this.parser,
				this.transformer,
				"Integration\\Parser.g",
				"MetaSharp.Transformation.Parser",
				(instance, parseMethod) =>
				{
					var result = (Nodes)parseMethod.Invoke(instance, new object[] { "a b c" });
					Assert.Equal("a", Node.Unwrap(result[0]));
					Assert.Equal("b", Node.Unwrap(result[1]));
					Assert.Equal("c", Node.Unwrap(result[2]));
					Assert.Equal(3, result.Count);
				});
		}

		[Fact]
		public void CanCompileLang()
		{
			CompileAndRunGrammar(
				this.parser,
				this.transformer,
				"Integration\\LangParser.g",
				"MetaSharp.Transformation.Lang.LangParser",
				(instance, parseMethod) =>
				{
					var result = (CodeTypeDeclaration)parseMethod.Invoke(instance, new object[] { "class Foo: end" });
					Assert.Equal("Foo", result.Name);
				});
		}

		[Fact]
		public void CanCompileGrammar()
		{
			CompileAndRunGrammar(
				this.parser,
				this.transformer,
				"Integration\\GrammarParser.g",
				"MetaSharp.Transformation.Grammar.GrammarParser",
				(instance, parseMethod) =>
				{
					var result = (GrammarDeclaration)parseMethod.Invoke(instance, new object[] { "grammar Foo: end" });
					Assert.Equal("Foo", result.Name);
				});
		}

		[Fact]
		public void CanCompileExample()
		{
			CompileAndRunGrammar(
				this.parser,
				this.transformer,
				"Integration\\Example.g",
				"Example.HelloX",
				(instance, parseMethod) =>
				{
					var stdout = Console.Out;
					var builder = new StringBuilder();
					using (var writer = new StringWriter(builder))
					{
						Console.SetOut(writer);
						try
						{
							var result = parseMethod.Invoke(instance, new object[] { "World" });
							Assert.True((bool)Node.Unwrap(result));
						}
						finally
						{
							Console.SetOut(stdout);
						}
					}

					Assert.Equal("Hello World!\r\n", builder.ToString());
				});
		}

		private static void CompileAndRunGrammar(
			GrammarParser parser,
			GrammarTransformer transformer,
			string grammarPath,
			string transformerType,
			Action<object, MethodInfo> runAssembly)
		{
			var generatedCodeFile = Path.Combine(Path.GetTempPath(), "IntegrationTest.cs");
			var generatedAssembly = Path.Combine(Path.GetTempPath(), "IntegrationTest.dll");
			var generatedPdb = Path.ChangeExtension(generatedAssembly, "pdb");
			try
			{
				using (var provider = new CSharpCodeProvider())
				{
					var code = File.ReadAllText(grammarPath);
					var context = new PatternContext();
					var ast = parser.Parse(code, context);

					foreach (var error in context.Errors)
					{
						var start = (int)((IIndexed)error.Span.Start).Index;
						var end = (int)((IIndexed)error.Span.End).Index;
						Console.WriteLine("{0} at {1}: \"{2}\".", error.Message, start, code.Substring(start, end - start));
					}

					Assert.IsNotType<Fail>(ast);
					var unit = transformer.Transform((Unit)ast);
					var generatedCode = unit.Print(provider);
					File.WriteAllText(generatedCodeFile, generatedCode);

					var parameters = new CompilerParameters(
						new string[] { 
							typeof(string).Assembly.Location, 
							typeof(Pattern).Assembly.Location, 
							typeof(Enumerable).Assembly.Location,
							typeof(CodeDomProvider).Assembly.Location,
						},
						generatedAssembly,
						true);

					//var results = provider.CompileAssemblyFromFile(parameters, generatedCodeFile);
					var results = provider.CompileAssemblyFromDom(parameters, unit);
					if (results.Errors.HasErrors || results.Errors.HasWarnings)
					{
						foreach (CompilerError error in results.Errors)
							Console.WriteLine(error.ToString());

						System.Diagnostics.Debug.WriteLine(generatedCode);
						Assert.False(results.Errors.HasErrors);
					}

					// Use cci to load generated assembly as run and collect...
					using (var host = new PeReader.DefaultHost())
					{
						using (var pdbStream = File.OpenRead(generatedPdb))
						{
							var pdbreader = new PdbReader(pdbStream, host);
							var name = Path.GetFileNameWithoutExtension(generatedAssembly);
							var loaded = host.LoadAssembly(new AssemblyIdentity(host.NameTable.GetNameFor(name), null, new Version(), null, generatedAssembly));

							var dynamicLoader = new Microsoft.Cci.ReflectionEmitter.DynamicLoader(pdbreader, pdbreader);
							var dynamicAssembly = dynamicLoader.Load(loaded);
							var type = dynamicAssembly.GetType(transformerType);
							var instance = (Transformer)Activator.CreateInstance(type);
							runAssembly(instance, type.GetMethod("Parse", new[] { typeof(string) }));
						}
					}
				}
			}
			finally
			{
				GC.Collect();
				GC.WaitForPendingFinalizers();

				File.Delete(generatedCodeFile);
				File.Delete(generatedAssembly);
				if (File.Exists(generatedPdb))
					File.Delete(generatedPdb);
			}
		}
	}
}
