﻿using System.CodeDom;
using System.CodeDom.Compiler;
using System.Linq;
using System.Reflection;
using MetaSharp.Transformation.Lang;
using Xunit;
using MetaSharp.Transformation.Lang.Ast;

namespace MetaSharp.Transformation.Tests.Lang
{
	public class ClassTests : IUseFixture<LangParser>
	{
		private LangParser lang;
		public void SetFixture(LangParser data)
		{
			this.lang = data;
		}

		[Fact]
		public void CanDeclareClassInsideOfNamespace()
		{
			var result = (CodeCompileUnit)this.lang.Parse("namespace X: class Y: end end");
			Assert.Equal(1, result.Namespaces[0].Types.Count);
			var t = result.Namespaces[0].Types[0];
			Assert.Equal("Y", t.Name);
			Assert.Equal(MemberAttributes.Public, t.Attributes);
			Assert.Equal(TypeAttributes.Class, t.TypeAttributes);
			Assert.Empty(t.TypeParameters);
			Assert.Empty(t.Members);
			Assert.False(t.IsPartial);
			Assert.False(t.IsStruct);
			Assert.False(t.IsInterface);
			Assert.False(t.IsEnum);
			Assert.True(t.IsClass);
		}

		[Fact]
		public void CanEncapsulateAClass()
		{
			var result = (CodeTypeDeclaration)this.lang.Parse("class Y: end");
			Assert.True(result.IsClass);
			Assert.Equal(MemberAttributes.Public, result.Attributes);

			result = (CodeTypeDeclaration)this.lang.Parse("public class Y: end");
			Assert.True(result.IsClass);
			Assert.Equal(MemberAttributes.Public, result.Attributes);

			result = (CodeTypeDeclaration)this.lang.Parse("private class Y: end");
			Assert.True(result.IsClass);
			Assert.Equal(MemberAttributes.Private, result.Attributes);

			result = (CodeTypeDeclaration)this.lang.Parse("assembly class Y: end");
			Assert.True(result.IsClass);
			Assert.Equal(MemberAttributes.Assembly, result.Attributes);

			result = (CodeTypeDeclaration)this.lang.Parse("family class Y: end");
			Assert.True(result.IsClass);
			Assert.Equal(MemberAttributes.Family, result.Attributes);

			result = (CodeTypeDeclaration)this.lang.Parse("family and assembly class Y: end");
			Assert.True(result.IsClass);
			Assert.Equal(MemberAttributes.FamilyAndAssembly, result.Attributes);

			result = (CodeTypeDeclaration)this.lang.Parse("family or assembly class Y: end");
			Assert.True(result.IsClass);
			Assert.Equal(MemberAttributes.FamilyOrAssembly, result.Attributes);
		}

		[Fact]
		public void CanAddCustomAttributesToClass()
		{
			var result = (CodeTypeDeclaration)this.lang.Parse("[X]class Y: end");
			Assert.Equal("Y", result.Name);
			Assert.Equal(2, result.CustomAttributes.Count);
			Assert.Equal("X", result.CustomAttributes[0].Name);
		}

		[Fact]
		public void CanMakeAbstractFinalOrStaticClass()
		{
			var result = (CodeTypeDeclaration)this.lang.Parse("class Y: end");
			Assert.True(result.IsClass);
			Assert.Equal(MemberAttributes.Public, result.Attributes);

			result = (CodeTypeDeclaration)this.lang.Parse("abstract class Y: end");
			Assert.True(result.IsClass);
			Assert.Equal(MemberAttributes.Public | MemberAttributes.Abstract, result.Attributes);

			result = (CodeTypeDeclaration)this.lang.Parse("final class Y: end");
			Assert.True(result.IsClass);
			Assert.Equal(MemberAttributes.Public | MemberAttributes.Final, result.Attributes);

			result = (CodeTypeDeclaration)this.lang.Parse("static class Y: end");
			Assert.True(result.IsClass);
			Assert.Equal(MemberAttributes.Public | MemberAttributes.Static, result.Attributes);

			result = (CodeTypeDeclaration)this.lang.Parse("family abstract class Y: end");
			Assert.True(result.IsClass);
			Assert.Equal(MemberAttributes.Family | MemberAttributes.Abstract, result.Attributes);

			result = (CodeTypeDeclaration)this.lang.Parse("family final class Y: end");
			Assert.True(result.IsClass);
			Assert.Equal(MemberAttributes.Family | MemberAttributes.Final, result.Attributes);

			result = (CodeTypeDeclaration)this.lang.Parse("family static class Y: end");
			Assert.True(result.IsClass);
			Assert.Equal(MemberAttributes.Family | MemberAttributes.Static, result.Attributes);

		}

		[Fact]
		public void ClassCanInheritFromSuperClass()
		{
			var result = (CodeTypeDeclaration)this.lang.Parse("class X as Y: end");
			Assert.Equal("X", result.Name);
			Assert.Equal(1, result.BaseTypes.Count);
			Assert.Equal("Y", result.BaseTypes[0].BaseType);
			Assert.Empty(result.BaseTypes[0].TypeArguments);
		}

		[Fact]
		public void ClassCanInheritFromGenericSuperClass()
		{
			var result = (CodeTypeDeclaration)this.lang.Parse("class X as Y<Z>: end");
			Assert.Equal("X", result.Name);
			Assert.Equal(1, result.BaseTypes.Count);
			Assert.Equal("Y`1", result.BaseTypes[0].BaseType);
			Assert.Equal(1, result.BaseTypes[0].TypeArguments.Count);
			Assert.Equal("Z", result.BaseTypes[0].TypeArguments[0].BaseType);
		}

		[Fact]
		public void CanDeclareGenericClass()
		{
			var result = (CodeTypeDeclaration)this.lang.Parse("class X<T>: end");
			Assert.Equal("X", result.Name);
			Assert.Equal(1, result.TypeParameters.Count);
			Assert.Equal("T", result.TypeParameters[0].Name);
		}

		[Fact]
		public void CanDeclareGenericClassWithMultipleTypeParameters()
		{
			var result = (CodeTypeDeclaration)this.lang.Parse("class X<Y, Z>: end");
			Assert.Equal("X", result.Name);
			Assert.Equal(2, result.TypeParameters.Count);
			Assert.Equal("Y", result.TypeParameters[0].Name);
			Assert.Equal("Z", result.TypeParameters[1].Name);
		}

		[Fact]
		public void CanDeclareGenericClassWithConstraint()
		{
			var result = (CodeTypeDeclaration)this.lang.Parse("class X<T as Y>: end");
			Assert.Equal("X", result.Name);
			Assert.Equal(1, result.TypeParameters.Count);
			Assert.Equal("T", result.TypeParameters[0].Name);
			Assert.Equal(1, result.TypeParameters[0].Constraints.Count);
			Assert.Equal("Y", result.TypeParameters[0].Constraints[0].BaseType);
		}

		[Fact]
		public void CanConstrainGenericParameterWithConstructorConstraint()
		{
			var result = (CodeTypeDeclaration)this.lang.Parse("class X<T as new>: end");
			Assert.Equal("X", result.Name);
			Assert.Equal(1, result.TypeParameters.Count);
			Assert.Equal("T", result.TypeParameters[0].Name);
			Assert.Empty(result.TypeParameters[0].Constraints);
			Assert.True(result.TypeParameters[0].HasConstructorConstraint);
		}

		[Fact]
		public void CanDeclareGenericClassWithMultipleConstraintsIncludingConstructorConstraint()
		{
			var result = (CodeTypeDeclaration)this.lang.Parse("class X<T as new and Y>: end");
			Assert.Equal("X", result.Name);
			Assert.Equal(1, result.TypeParameters.Count);
			Assert.Equal("T", result.TypeParameters[0].Name);
			Assert.Equal(1, result.TypeParameters[0].Constraints.Count);
			Assert.True(result.TypeParameters[0].HasConstructorConstraint);
			Assert.Equal("Y", result.TypeParameters[0].Constraints[0].BaseType);
		}

		[Fact]
		public void CanDeclareGenericClassWithMultipleConstraints()
		{
			var result = (CodeTypeDeclaration)this.lang.Parse("class X<T as Y and Z>: end");
			Assert.Equal("X", result.Name);
			Assert.Equal(1, result.TypeParameters.Count);
			Assert.Equal("T", result.TypeParameters[0].Name);
			Assert.Equal(2, result.TypeParameters[0].Constraints.Count);
			Assert.Equal("Y", result.TypeParameters[0].Constraints[0].BaseType);
			Assert.Equal("Z", result.TypeParameters[0].Constraints[1].BaseType);
		}

		[Fact]
		public void CanDeclareGenericClassWithMultipleConstraintsAndMultipleParameters()
		{
			var result = (CodeTypeDeclaration)this.lang.Parse("class X<T1 as Y and Z, T2 as A and B>: end");
			Assert.Equal("X", result.Name);
			Assert.Equal(2, result.TypeParameters.Count);
			Assert.Equal("T1", result.TypeParameters[0].Name);
			Assert.Equal(2, result.TypeParameters[0].Constraints.Count);
			Assert.Equal("Y", result.TypeParameters[0].Constraints[0].BaseType);
			Assert.Equal("Z", result.TypeParameters[0].Constraints[1].BaseType);
			Assert.Equal("T2", result.TypeParameters[1].Name);
			Assert.Equal(2, result.TypeParameters[1].Constraints.Count);
			Assert.Equal("A", result.TypeParameters[1].Constraints[0].BaseType);
			Assert.Equal("B", result.TypeParameters[1].Constraints[1].BaseType);
		}

		[Fact]
		public void ClassCanHaveMethod()
		{
			var result = (CodeTypeDeclaration)this.lang.Parse("class X: void M: end end");
			var method = (CodeMemberMethod)result.Members[0];
			Assert.Empty(method.Parameters);
			Assert.Empty(method.TypeParameters);
			Assert.Empty(method.Statements);
			Assert.Empty(method.CustomAttributes);
			Assert.Equal("System.Void", method.ReturnType.BaseType);
		}

		[Fact]
		public void IsGeneratedCodeAttributeAddedToGeneratedClass()
		{
			var result = (CodeTypeDeclaration)this.lang.Parse("class Y: end");
			Assert.True(result.CustomAttributes.OfType<AttributeDeclaration>().Any(a => a.Name == typeof(GeneratedCodeAttribute).FullName));
		}
	}
}
