﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Xunit;
using MetaSharp.Transformation.Grammar;
using System.CodeDom;
using System.Reflection;
using MetaSharp.Transformation.Lang.Ast;
using MetaSharp.Transformation.Lang;

namespace MetaSharp.Transformation.Tests.Lang
{
	public class ConstructorTests : IUseFixture<LangParser>
	{
		private LangParser lang;
		public void SetFixture(LangParser data)
		{
			this.lang = data;
		}

		[Fact]
		public void ClassCanHaveConstructor()
		{
			var result = (CodeTypeDeclaration)this.lang.Parse("class X: constructor: end end");
			Assert.Equal("X", result.Name);
			Assert.Equal(1, result.Members.Count);
			Assert.IsAssignableFrom(typeof(CodeConstructor), result.Members[0]);
		}

		[Fact]
		public void CanDeclareStaticConstructor()
		{
			var result = (CodeConstructor)this.lang.Parse("static constructor: end");
			Assert.NotNull(result);
			Assert.Equal(MemberAttributes.Public | MemberAttributes.Static, result.Attributes);
			Assert.Empty(result.BaseConstructorArgs);
			Assert.Empty(result.ChainedConstructorArgs);
			Assert.Empty(result.CustomAttributes);
			Assert.Empty(result.ImplementationTypes);
			Assert.Empty(result.Parameters);
			Assert.Empty(result.Statements);
			Assert.Empty(result.TypeParameters);
			Assert.Equal(".ctor", result.Name);
		}

		[Fact]
		public void CanDeclareEncapsulatedConstructor()
		{
			var result = (CodeConstructor)this.lang.Parse("constructor: end");
			Assert.Equal(MemberAttributes.Public, result.Attributes);

			result = (CodeConstructor)this.lang.Parse("public constructor: end");
			Assert.Equal(MemberAttributes.Public, result.Attributes);

			result = (CodeConstructor)this.lang.Parse("private constructor: end");
			Assert.Equal(MemberAttributes.Private, result.Attributes);

			result = (CodeConstructor)this.lang.Parse("assembly constructor: end");
			Assert.Equal(MemberAttributes.Assembly, result.Attributes);

			result = (CodeConstructor)this.lang.Parse("family constructor: end");
			Assert.Equal(MemberAttributes.Family, result.Attributes);

			result = (CodeConstructor)this.lang.Parse("family or assembly constructor: end");
			Assert.Equal(MemberAttributes.FamilyOrAssembly, result.Attributes);

			result = (CodeConstructor)this.lang.Parse("family and assembly constructor: end");
			Assert.Equal(MemberAttributes.FamilyAndAssembly, result.Attributes);
		}

		[Fact]
		public void CanDeclareConstructorWithNoParameters()
		{
			var result = (CodeConstructor)this.lang.Parse("constructor: end");
			Assert.Empty(result.Parameters);
		}

		[Fact]
		public void CanDeclareConstructorWithEmptyParameters()
		{
			var result = (CodeConstructor)this.lang.Parse("constructor(): end");
			Assert.Empty(result.Parameters);
		}

		[Fact]
		public void CanDeclareConstructorWithSingleParameter()
		{
			var result = (CodeConstructor)this.lang.Parse("constructor(int x): end");
			Assert.Equal(1, result.Parameters.Count);
			Assert.Equal("System.Int32", result.Parameters[0].Type.BaseType);
			Assert.Equal("x", result.Parameters[0].Name);
		}

		[Fact]
		public void CanDeclareConstructorWithMultipleParameters()
		{
			var result = (CodeConstructor)this.lang.Parse("constructor(int x, double y, string z): end");
			Assert.Equal(3, result.Parameters.Count);
			Assert.Equal("System.Int32", result.Parameters[0].Type.BaseType);
			Assert.Equal("System.Double", result.Parameters[1].Type.BaseType);
			Assert.Equal("System.String", result.Parameters[2].Type.BaseType);
			Assert.Equal("x", result.Parameters[0].Name);
			Assert.Equal("y", result.Parameters[1].Name);
			Assert.Equal("z", result.Parameters[2].Name);
		}

		[Fact]
		public void CanDeclareConstructorWithParameterAndSuper()
		{
			var result = (CodeConstructor)this.lang.Parse("constructor(int x) super(): end");
			Assert.Equal(1, result.Parameters.Count);
			Assert.Equal("System.Int32", result.Parameters[0].Type.BaseType);
			Assert.Equal("x", result.Parameters[0].Name);
			Assert.Empty(result.BaseConstructorArgs);
		}

		[Fact]
		public void CanDeclareConstructorWithEmptySuperArgs()
		{
			var result = (CodeConstructor)this.lang.Parse("constructor super(): end");
			Assert.Empty(result.BaseConstructorArgs);
		}

		[Fact]
		public void CanDeclareConstructorWithSeperatedEmptySuperArgs()
		{
			var result = (CodeConstructor)this.lang.Parse("constructor super (): end");
			Assert.Empty(result.BaseConstructorArgs);
		}

		[Fact]
		public void CanDeclareConstructorWithSuperArgs()
		{
			var result = (CodeConstructor)this.lang.Parse("constructor(int x) super(x): end");
			Assert.Equal(1, result.BaseConstructorArgs.Count);
			Assert.Equal("x", ((ReferenceExpression)result.BaseConstructorArgs[0]).PropertyName);
		}

		[Fact]
		public void CanDeclareConstructorWithSeperatedSuperArgs()
		{
			var result = (CodeConstructor)this.lang.Parse("constructor(int x) super (x): end");
			Assert.Equal(1, result.BaseConstructorArgs.Count);
			Assert.Equal("x", ((ReferenceExpression)result.BaseConstructorArgs[0]).PropertyName);
		}

		[Fact]
		public void CanDeclareConstructorWithEmptyChainArgs()
		{
			var result = (CodeConstructor)this.lang.Parse("constructor this(): end");
			Assert.Empty(result.ChainedConstructorArgs);
		}

		[Fact]
		public void CanDeclareConstructorWithSeperatedEmptyChainArgs()
		{
			var result = (CodeConstructor)this.lang.Parse("constructor this (): end");
			Assert.Empty(result.ChainedConstructorArgs);
		}

		[Fact]
		public void CanDeclareConstructorWithChainArgs()
		{
			var result = (CodeConstructor)this.lang.Parse("constructor(int x) this(x): end");
			Assert.Equal(1, result.ChainedConstructorArgs.Count);
			Assert.Equal("x", ((ReferenceExpression)result.ChainedConstructorArgs[0]).PropertyName);
		}
		[Fact]
		public void CanDeclareConstructorWithSeperatedChainArgs()
		{
			var result = (CodeConstructor)this.lang.Parse("constructor(int x) this (x): end");
			Assert.Equal(1, result.ChainedConstructorArgs.Count);
			Assert.Equal("x", ((ReferenceExpression)result.ChainedConstructorArgs[0]).PropertyName);
		}
	}
}
