﻿using System.CodeDom;
using MetaSharp.Transformation.Grammar;
using Xunit;
using Xunit.Extensions;
using MetaSharp.Transformation.Lang.Ast;
using MetaSharp.Transformation.Lang;

namespace MetaSharp.Transformation.Tests.Lang
{
	public class ExpressionTests : IUseFixture<ExpressionParser>
	{
		private ExpressionParser lang;
		public void SetFixture(ExpressionParser data)
		{
			this.lang = data;
		}

		[Fact]
		public void CanCreateReferenceExpression()
		{
			var result = (ReferenceExpression)this.lang.Parse("x");
			Assert.Equal("x", result.PropertyName);
		}

		[Fact]
		public void CanCreateLogicalOrExpression()
		{
			var result = (CodeBinaryOperatorExpression)this.lang.Parse("x or y");
			Assert.Equal(CodeBinaryOperatorType.BooleanOr, result.Operator);
			Assert.Equal("x", ((ReferenceExpression)result.Left).PropertyName);
			Assert.Equal("y", ((ReferenceExpression)result.Right).PropertyName);
		}

		[Fact]
		public void CanCreateLogicalAndExpression()
		{
			var result = (CodeBinaryOperatorExpression)this.lang.Parse("x and y");
			Assert.Equal(CodeBinaryOperatorType.BooleanAnd, result.Operator);
			Assert.Equal("x", ((ReferenceExpression)result.Left).PropertyName);
			Assert.Equal("y", ((ReferenceExpression)result.Right).PropertyName);
		}

		[Fact]
		public void CanCreateIsAndAsExpression()
		{
			//constraint is string and constraint as string == "new"
			var result = (CodeBinaryOperatorExpression)this.lang.Parse("x is y and x as y");
			Assert.Equal(CodeBinaryOperatorType.BooleanAnd, result.Operator);
		}

		[Fact]
		public void CanCreateEqualityAsExpression()
		{
			//constraint is string and constraint as string == "new"
			var result = (CodeBinaryOperatorExpression)this.lang.Parse("x as y == \"z\"");
			Assert.Equal(CodeBinaryOperatorType.IdentityEquality, result.Operator);
		}

		[Fact]
		public void CanCreateIsAndEqualityAsExpression()
		{
			var result = (CodeBinaryOperatorExpression)this.lang.Parse("x is y and (x as y) == z");
			Assert.Equal(CodeBinaryOperatorType.BooleanAnd, result.Operator);
		}

		[Fact]
		public void LogicalOrAndAreOrderDependent()
		{
			dynamic result = this.lang.Parse("x or y and z");
			Assert.Equal(CodeBinaryOperatorType.BooleanAnd, result.Operator);
			Assert.Equal(CodeBinaryOperatorType.BooleanOr, result.Left.Operator);
			Assert.Equal("x", result.Left.Left.PropertyName);
			Assert.Equal("y", result.Left.Right.PropertyName);
			Assert.Equal("z", result.Right.PropertyName);

			result = this.lang.Parse("x or y or z");
			Assert.Equal(CodeBinaryOperatorType.BooleanOr, result.Operator);
			Assert.Equal(CodeBinaryOperatorType.BooleanOr, result.Left.Operator);
			Assert.Equal("x", result.Left.Left.PropertyName);
			Assert.Equal("y", result.Left.Right.PropertyName);
			Assert.Equal("z", result.Right.PropertyName);

			result = this.lang.Parse("x and y and z");
			Assert.Equal(CodeBinaryOperatorType.BooleanAnd, result.Operator);
			Assert.Equal(CodeBinaryOperatorType.BooleanAnd, result.Left.Operator);
			Assert.Equal("x", result.Left.Left.PropertyName);
			Assert.Equal("y", result.Left.Right.PropertyName);
			Assert.Equal("z", result.Right.PropertyName);
		}

		[Fact]
		public void CanGroupExpressions()
		{
			dynamic result = this.lang.Parse("(a or b) and (x or y)");
			Assert.Equal(CodeBinaryOperatorType.BooleanAnd, result.Operator);
			Assert.Equal(CodeBinaryOperatorType.BooleanOr, result.Left.Operator);
			Assert.Equal(CodeBinaryOperatorType.BooleanOr, result.Right.Operator);
			Assert.Equal("a", result.Left.Left.PropertyName);
			Assert.Equal("b", result.Left.Right.PropertyName);
			Assert.Equal("x", result.Right.Left.PropertyName);
			Assert.Equal("y", result.Right.Right.PropertyName);

			result = this.lang.Parse("a or (b and x) or y");
			Assert.Equal(CodeBinaryOperatorType.BooleanOr, result.Operator);
			Assert.Equal(CodeBinaryOperatorType.BooleanOr, result.Left.Operator);
			Assert.Equal(CodeBinaryOperatorType.BooleanAnd, result.Left.Right.Operator);
			Assert.Equal("a", result.Left.Left.PropertyName);
			Assert.Equal("b", result.Left.Right.Left.PropertyName);
			Assert.Equal("x", result.Left.Right.Right.PropertyName);
			Assert.Equal("y", result.Right.PropertyName);
		}

		[Fact]
		public void CanCreateNotExpression()
		{
			var result = (CodeBinaryOperatorExpression)this.lang.Parse("!x");
			Assert.Equal(CodeBinaryOperatorType.ValueEquality, result.Operator);
			Assert.Equal("x", ((ReferenceExpression)result.Left).PropertyName);
			Assert.False((bool)((CodePrimitiveExpression)result.Right).Value);
		}

		[Fact]
		public void CanCreateGroupedNotExpression()
		{
			var result = (CodeBinaryOperatorExpression)this.lang.Parse("!(x and y)");
			Assert.Equal(CodeBinaryOperatorType.ValueEquality, result.Operator);
			var exp = (CodeBinaryOperatorExpression)result.Left;
			Assert.Equal("x", ((ReferenceExpression)exp.Left).PropertyName);
			Assert.Equal("y", ((ReferenceExpression)exp.Right).PropertyName);
			Assert.False((bool)((CodePrimitiveExpression)result.Right).Value);
		}

		[Fact]
		public void CanCreateGroupReferenceExpression()
		{
			//(reference as ReferenceExpression).PropertyName
			var result = (ReferenceExpression)this.lang.Parse("(x).y");
			var exp = (ReferenceExpression)result.TargetObject;
			Assert.Equal("x", exp.PropertyName);
			Assert.Equal("y", result.PropertyName);
		}

		[Fact]
		public void NotExpressionHasHigherPriorityThanLogicalExpressions()
		{
			var result = (CodeBinaryOperatorExpression)this.lang.Parse("!x and !y");
			Assert.Equal(CodeBinaryOperatorType.BooleanAnd, result.Operator);
			var left = (CodeBinaryOperatorExpression)result.Left;
			var right = (CodeBinaryOperatorExpression)result.Right;
			Assert.Equal(CodeBinaryOperatorType.ValueEquality, left.Operator);
			Assert.Equal(CodeBinaryOperatorType.ValueEquality, right.Operator);
			Assert.False((bool)((CodePrimitiveExpression)left.Right).Value);
			Assert.False((bool)((CodePrimitiveExpression)right.Right).Value);
			Assert.Equal("x", ((ReferenceExpression)left.Left).PropertyName);
			Assert.Equal("y", ((ReferenceExpression)right.Left).PropertyName);
		}

		[Fact]
		public void CanCreateEqualityExpression()
		{
			var result = (CodeBinaryOperatorExpression)this.lang.Parse("x == y");
			Assert.Equal(CodeBinaryOperatorType.IdentityEquality, result.Operator);
			Assert.Equal("x", ((ReferenceExpression)result.Left).PropertyName);
			Assert.Equal("y", ((ReferenceExpression)result.Right).PropertyName);
		}

		[Fact]
		public void CanCreateInequalityExpression()
		{
			var result = (CodeBinaryOperatorExpression)this.lang.Parse("x != y");
			Assert.Equal(CodeBinaryOperatorType.IdentityInequality, result.Operator);
			Assert.Equal("x", ((ReferenceExpression)result.Left).PropertyName);
			Assert.Equal("y", ((ReferenceExpression)result.Right).PropertyName);
		}

		[Fact]
		public void CanCreateGreaterThanOrEqualExpression()
		{
			var result = (CodeBinaryOperatorExpression)this.lang.Parse("x >= y");
			Assert.Equal(CodeBinaryOperatorType.GreaterThanOrEqual, result.Operator);
			Assert.Equal("x", ((ReferenceExpression)result.Left).PropertyName);
			Assert.Equal("y", ((ReferenceExpression)result.Right).PropertyName);
		}

		[Fact]
		public void CanCreateGreaterThanExpression()
		{
			var result = (CodeBinaryOperatorExpression)this.lang.Parse("x > y");
			Assert.Equal(CodeBinaryOperatorType.GreaterThan, result.Operator);
			Assert.Equal("x", ((ReferenceExpression)result.Left).PropertyName);
			Assert.Equal("y", ((ReferenceExpression)result.Right).PropertyName);
		}

		[Fact]
		public void CanCreateLessThanExpression()
		{
			var result = (CodeBinaryOperatorExpression)this.lang.Parse("x < y");
			Assert.Equal(CodeBinaryOperatorType.LessThan, result.Operator);
			Assert.Equal("x", ((ReferenceExpression)result.Left).PropertyName);
			Assert.Equal("y", ((ReferenceExpression)result.Right).PropertyName);
		}

		[Fact]
		public void CanCreateLessThanOrEqualExpression()
		{
			var result = (CodeBinaryOperatorExpression)this.lang.Parse("x <= y");
			Assert.Equal(CodeBinaryOperatorType.LessThanOrEqual, result.Operator);
			Assert.Equal("x", ((ReferenceExpression)result.Left).PropertyName);
			Assert.Equal("y", ((ReferenceExpression)result.Right).PropertyName);
		}

		[Fact]
		public void EqualityExpressionsMustBeQualified()
		{
			dynamic result = this.lang.Parse("x == y == z");
			Assert.IsType<Fail>(result);

			result = this.lang.Parse("x == (y == z)");
			Assert.Equal(CodeBinaryOperatorType.IdentityEquality, result.Operator);
			Assert.Equal(CodeBinaryOperatorType.IdentityEquality, result.Right.Operator);
			Assert.Equal("x", result.Left.PropertyName);
			Assert.Equal("y", result.Right.Left.PropertyName);
			Assert.Equal("z", result.Right.Right.PropertyName);
		}

		[Fact]
		public void LogicalExpressionsHaveHigherPrecedenceThanEqualityExpressions()
		{
			dynamic result = this.lang.Parse("a == b and x == y");
			Assert.Equal(CodeBinaryOperatorType.BooleanAnd, result.Operator);
			Assert.Equal(CodeBinaryOperatorType.IdentityEquality, result.Left.Operator);
			Assert.Equal(CodeBinaryOperatorType.IdentityEquality, result.Right.Operator);
			Assert.Equal("a", result.Left.Left.PropertyName);
			Assert.Equal("b", result.Left.Right.PropertyName);
			Assert.Equal("x", result.Right.Left.PropertyName);
			Assert.Equal("y", result.Right.Right.PropertyName);
		}

		[Fact]
		public void CanCreateBitwiseOrExpression()
		{
			dynamic result = this.lang.Parse("x | y");
			Assert.Equal(CodeBinaryOperatorType.BitwiseOr, result.Operator);
			Assert.Equal("x", result.Left.PropertyName);
			Assert.Equal("y", result.Right.PropertyName);
		}

		[Fact]
		public void CanCreateBitwiseAndExpression()
		{
			dynamic result = this.lang.Parse("x & y");
			Assert.Equal(CodeBinaryOperatorType.BitwiseAnd, result.Operator);
			Assert.Equal("x", result.Left.PropertyName);
			Assert.Equal("y", result.Right.PropertyName);
		}

		[Fact]
		public void BitwiseExpressionsAreOrderDependent()
		{
			dynamic result = this.lang.Parse("x & y | z");
			Assert.Equal(CodeBinaryOperatorType.BitwiseOr, result.Operator);
			Assert.Equal(CodeBinaryOperatorType.BitwiseAnd, result.Left.Operator);
			Assert.Equal("x", result.Left.Left.PropertyName);
			Assert.Equal("y", result.Left.Right.PropertyName);
			Assert.Equal("z", result.Right.PropertyName);
		}

		[Fact]
		public void EqualityExpressionsHaveHigherPriorityThanBitwiseExpressions()
		{
			dynamic result = this.lang.Parse("a | b == x & y");
			Assert.Equal(CodeBinaryOperatorType.IdentityEquality, result.Operator);
			Assert.Equal(CodeBinaryOperatorType.BitwiseOr, result.Left.Operator);
			Assert.Equal(CodeBinaryOperatorType.BitwiseAnd, result.Right.Operator);
			Assert.Equal("a", result.Left.Left.PropertyName);
			Assert.Equal("b", result.Left.Right.PropertyName);
			Assert.Equal("x", result.Right.Left.PropertyName);
			Assert.Equal("y", result.Right.Right.PropertyName);
		}

		[Fact]
		public void CanCreateAddExpression()
		{
			var result = (CodeBinaryOperatorExpression)this.lang.Parse("x + y");
			Assert.Equal(CodeBinaryOperatorType.Add, result.Operator);
			Assert.Equal("x", ((ReferenceExpression)result.Left).PropertyName);
			Assert.Equal("y", ((ReferenceExpression)result.Right).PropertyName);
		}

		[Fact]
		public void CanCreateSubtractExpression()
		{
			var result = (CodeBinaryOperatorExpression)this.lang.Parse("x - y");
			Assert.Equal(CodeBinaryOperatorType.Subtract, result.Operator);
			Assert.Equal("x", ((ReferenceExpression)result.Left).PropertyName);
			Assert.Equal("y", ((ReferenceExpression)result.Right).PropertyName);
		}

		[Fact]
		public void AddSubtractExpressionsAreOrderDependent()
		{
			var result = (CodeBinaryOperatorExpression)this.lang.Parse("x + y - z");
			Assert.Equal(CodeBinaryOperatorType.Subtract, result.Operator);

			var left = (CodeBinaryOperatorExpression)result.Left;
			Assert.Equal(CodeBinaryOperatorType.Add, left.Operator);
			Assert.Equal("x", ((ReferenceExpression)left.Left).PropertyName);
			Assert.Equal("y", ((ReferenceExpression)left.Right).PropertyName);
			Assert.Equal("z", ((ReferenceExpression)result.Right).PropertyName);
		}

		[Fact]
		public void BitwiseExpressionsHaveHigherPriorityThanAddSubtractExpressions()
		{
			var result = (CodeBinaryOperatorExpression)this.lang.Parse("a + b | x - y");
			Assert.Equal(CodeBinaryOperatorType.BitwiseOr, result.Operator);

			var left = (CodeBinaryOperatorExpression)result.Left;
			var right = (CodeBinaryOperatorExpression)result.Right;
			Assert.Equal(CodeBinaryOperatorType.Add, left.Operator);
			Assert.Equal(CodeBinaryOperatorType.Subtract, right.Operator);
			Assert.Equal("a", ((ReferenceExpression)left.Left).PropertyName);
			Assert.Equal("b", ((ReferenceExpression)left.Right).PropertyName);
			Assert.Equal("x", ((ReferenceExpression)right.Left).PropertyName);
			Assert.Equal("y", ((ReferenceExpression)right.Right).PropertyName);
		}

		[Fact]
		public void CanCreateMultiplyExpression()
		{
			var result = (CodeBinaryOperatorExpression)this.lang.Parse("x * y");
			Assert.Equal(CodeBinaryOperatorType.Multiply, result.Operator);
			Assert.Equal("x", ((ReferenceExpression)result.Left).PropertyName);
			Assert.Equal("y", ((ReferenceExpression)result.Right).PropertyName);
		}

		[Fact]
		public void CanCreateDivideExpression()
		{
			var result = (CodeBinaryOperatorExpression)this.lang.Parse("x / y");
			Assert.Equal(CodeBinaryOperatorType.Divide, result.Operator);
			Assert.Equal("x", ((ReferenceExpression)result.Left).PropertyName);
			Assert.Equal("y", ((ReferenceExpression)result.Right).PropertyName);
		}

		[Fact]
		public void CanCreateModulusExpression()
		{
			var result = (CodeBinaryOperatorExpression)this.lang.Parse("x % y");
			Assert.Equal(CodeBinaryOperatorType.Modulus, result.Operator);
			Assert.Equal("x", ((ReferenceExpression)result.Left).PropertyName);
			Assert.Equal("y", ((ReferenceExpression)result.Right).PropertyName);
		}

		[Fact]
		public void AddMultiplyDivideModulusExpressionsAreOrderDependent()
		{
			dynamic result = this.lang.Parse("w * x / y % z");
			Assert.Equal(CodeBinaryOperatorType.Modulus, result.Operator);
			Assert.Equal(CodeBinaryOperatorType.Divide, result.Left.Operator);
			Assert.Equal(CodeBinaryOperatorType.Multiply, result.Left.Left.Operator);
			Assert.Equal("w", result.Left.Left.Left.PropertyName);
			Assert.Equal("x", result.Left.Left.Right.PropertyName);
			Assert.Equal("y", result.Left.Right.PropertyName);
			Assert.Equal("z", result.Right.PropertyName);
		}

		[Fact]
		public void AddSubtractExpressionsHaveHigherPriorityThanMultiplyDivideModulusExpressions()
		{
			dynamic result = this.lang.Parse("a * b + x / y");
			Assert.Equal(CodeBinaryOperatorType.Add, result.Operator);
			Assert.Equal(CodeBinaryOperatorType.Multiply, result.Left.Operator);
			Assert.Equal(CodeBinaryOperatorType.Divide, result.Right.Operator);
			Assert.Equal("a", result.Left.Left.PropertyName);
			Assert.Equal("b", result.Left.Right.PropertyName);
			Assert.Equal("x", result.Right.Left.PropertyName);
			Assert.Equal("y", result.Right.Right.PropertyName);
		}

		[Fact]
		public void CanCreateInvokeExpression()
		{
			var result = (CodeDelegateInvokeExpression)this.lang.Parse("x()");
			Assert.Equal("x", ((ReferenceExpression)result.TargetObject).PropertyName);
			Assert.Empty(result.Parameters);
		}

		[Fact]
		public void CanCreateInvokeExpressionWithGenericArguments()
		{
			var result = (CodeMethodInvokeExpression)this.lang.Parse("x<Y,Z>()");
			Assert.Equal("x", result.Method.MethodName);
			Assert.Empty(result.Parameters);
			Assert.Equal(2, result.Method.TypeArguments.Count);
			Assert.Equal("Y", result.Method.TypeArguments[0].BaseType);
			Assert.Equal("Z", result.Method.TypeArguments[1].BaseType);
		}

		[Fact]
		public void CanCreateInvokeExpressionWithGenericArgumentsAndTarget()
		{
			var result = (CodeMethodInvokeExpression)this.lang.Parse("w.x<Y,Z>()");
			Assert.Equal("x", result.Method.MethodName);
			Assert.Empty(result.Parameters);
			Assert.Equal(2, result.Method.TypeArguments.Count);
			Assert.Equal("Y", result.Method.TypeArguments[0].BaseType);
			Assert.Equal("Z", result.Method.TypeArguments[1].BaseType);
			var w = (ReferenceExpression)result.Method.TargetObject;
			Assert.Equal("w", w.PropertyName);
			Assert.Null(w.TargetObject);
		}

		[Fact]
		public void CanCreateReferenceExpressionWithTarget()
		{
			var result = (ReferenceExpression)this.lang.Parse("x.y");
			Assert.Equal("y", result.PropertyName);
			var x = (ReferenceExpression)result.TargetObject;
			Assert.Equal("x", x.PropertyName);
			Assert.Null(x.TargetObject);
		}

		[Fact]
		public void CanCreateNestedReferenceExpressionsWithTarget()
		{
			var z = (ReferenceExpression)this.lang.Parse("x.y.z");
			Assert.Equal("z", z.PropertyName);
			var y = (ReferenceExpression)z.TargetObject;
			Assert.Equal("y", y.PropertyName);
			var x = (ReferenceExpression)y.TargetObject;
			Assert.Equal("x", x.PropertyName);
			Assert.Null(x.TargetObject);
		}

		[Fact]
		public void CanCreateReferenceExpressionsWithInvokeTarget()
		{
			var z = (ReferenceExpression)this.lang.Parse("x().y.z");
			Assert.Equal("z", z.PropertyName);
			var y = (ReferenceExpression)z.TargetObject;
			Assert.Equal("y", y.PropertyName);
			var invoke = (DelegateInvokeExpression)y.TargetObject;
			var x = (ReferenceExpression)invoke.TargetObject;
			Assert.Empty(invoke.Parameters);
			Assert.Equal("x", x.PropertyName);
			Assert.Null(x.TargetObject);
		}

		[Fact]
		public void CanCreateReferenceExpressionsWithInvokeTargetAndReferenceTarget()
		{
			var z = (ReferenceExpression)this.lang.Parse("x.y().z");
			Assert.Equal("z", z.PropertyName);
			var invoke = (DelegateInvokeExpression)z.TargetObject;
			Assert.Empty(invoke.Parameters);
			var y = (ReferenceExpression)invoke.TargetObject;
			Assert.Equal("y", y.PropertyName);
			var x = (ReferenceExpression)y.TargetObject;
			Assert.Equal("x", x.PropertyName);
			Assert.Null(x.TargetObject);
		}

		[Fact]
		public void CanCreateInvokeWithNestedReferenceTargets()
		{
			var invoke = (DelegateInvokeExpression)this.lang.Parse("x.y.z()");
			Assert.Empty(invoke.Parameters);
			var z = (ReferenceExpression)invoke.TargetObject;
			Assert.Equal("z", z.PropertyName);
			var y = (ReferenceExpression)z.TargetObject;
			Assert.Equal("y", y.PropertyName);
			var x = (ReferenceExpression)y.TargetObject;
			Assert.Equal("x", x.PropertyName);
			Assert.Null(x.TargetObject);
		}


		[Fact]
		public void CanCreateInvokeExpressionWithParameters()
		{
			var result = (CodeDelegateInvokeExpression)this.lang.Parse("x(y, z)");
			Assert.Equal("x", ((ReferenceExpression)result.TargetObject).PropertyName);
			Assert.Equal(2, result.Parameters.Count);
			Assert.Equal("y", ((ReferenceExpression)result.Parameters[0]).PropertyName);
			Assert.Equal("z", ((ReferenceExpression)result.Parameters[1]).PropertyName);
		}

		[Fact]
		public void CanCreateInvokeExpressionWithParametersAndTarget()
		{
			var invoke = (DelegateInvokeExpression)this.lang.Parse("w.x(y, z)");
			Assert.Equal(2, invoke.Parameters.Count);
			Assert.Equal("y", ((ReferenceExpression)invoke.Parameters[0]).PropertyName);
			Assert.Equal("z", ((ReferenceExpression)invoke.Parameters[1]).PropertyName);
			var x = (ReferenceExpression)invoke.TargetObject;
			Assert.Equal("x", x.PropertyName);
			var w = (ReferenceExpression)x.TargetObject;
			Assert.Equal("w", w.PropertyName);
			Assert.Null(w.TargetObject);
		}

		[Fact]
		public void CanInvokeWithInOutRefDirections()
		{
			var result = (CodeDelegateInvokeExpression)this.lang.Parse("x(in a, out b, ref c)");
			Assert.Equal("x", ((ReferenceExpression)result.TargetObject).PropertyName);
			Assert.Equal(3, result.Parameters.Count);
			var exp1 = (CodeDirectionExpression)result.Parameters[0];
			var exp2 = (CodeDirectionExpression)result.Parameters[1];
			var exp3 = (CodeDirectionExpression)result.Parameters[2];
			Assert.Equal(FieldDirection.In, exp1.Direction);
			Assert.Equal(FieldDirection.Out, exp2.Direction);
			Assert.Equal(FieldDirection.Ref, exp3.Direction);
			Assert.Equal("a", ((ReferenceExpression)exp1.Expression).PropertyName);
			Assert.Equal("b", ((ReferenceExpression)exp2.Expression).PropertyName);
			Assert.Equal("c", ((ReferenceExpression)exp3.Expression).PropertyName);
		}

		[Fact]
		public void CanInvokeAnInvoke()
		{
			var result = (CodeDelegateInvokeExpression)this.lang.Parse("x()()");
			var target1 = (CodeDelegateInvokeExpression)result.TargetObject;
			var target2 = (ReferenceExpression)target1.TargetObject;
			Assert.Equal("x", target2.PropertyName);
		}

		[Fact]
		public void CanInvokeStaticMethodOnPrimitiveType()
		{
			var result = (CodeDelegateInvokeExpression)this.lang.Parse("char.Parse(str)");
			var parse = (CodePropertyReferenceExpression)result.TargetObject;
			var @char = (ReferenceExpression)parse.TargetObject;
			var p = (ReferenceExpression)result.Parameters[0];
			Assert.Equal("System.Char", @char.PropertyName);
			Assert.Equal("Parse", parse.PropertyName);
			Assert.Equal("str", p.PropertyName);
			Assert.Equal(1, result.Parameters.Count);
		}

		[Fact]
		public void CanCreateArrayIndexerExpression()
		{
			var result = (CodeArrayIndexerExpression)this.lang.Parse("x[y, z]");
			Assert.Equal("x", ((ReferenceExpression)result.TargetObject).PropertyName);
			Assert.Equal(2, result.Indices.Count);
			Assert.Equal("y", ((ReferenceExpression)result.Indices[0]).PropertyName);
			Assert.Equal("z", ((ReferenceExpression)result.Indices[1]).PropertyName);
		}

		[Fact]
		public void CanIndexAnIndexer()
		{
			var result = (CodeArrayIndexerExpression)this.lang.Parse("x[y][z]");
			var t1 = (CodeArrayIndexerExpression)result.TargetObject;
			var t2 = (ReferenceExpression)t1.TargetObject;
			Assert.Equal("x", t2.PropertyName);
			Assert.Equal(1, result.Indices.Count);
			Assert.Equal(1, t1.Indices.Count);
			Assert.Equal("z", ((ReferenceExpression)result.Indices[0]).PropertyName);
			Assert.Equal("y", ((ReferenceExpression)t1.Indices[0]).PropertyName);
		}

		[Fact]
		public void CanCreateSuperReferenceExpression()
		{
			var result = (CodeBaseReferenceExpression)this.lang.Parse("super");
			Assert.NotNull(result);
		}

		[Fact]
		public void CanCreateThisReferenceExpression()
		{
			var result = (CodeThisReferenceExpression)this.lang.Parse("this");
			Assert.NotNull(result);
		}

		[Fact]
		public void CanCreatePrimitiveBooleanExpressions()
		{
			var result = (CodePrimitiveExpression)this.lang.Parse("true");
			Assert.True((bool)result.Value);

			result = (CodePrimitiveExpression)this.lang.Parse("false");
			Assert.False((bool)result.Value);
		}

		[Fact]
		public void CanCreatePrimitiveNullExpression()
		{
			var result = (CodePrimitiveExpression)this.lang.Parse("null");
			Assert.Null(result.Value);
		}

		[Theory]
		[InlineData(int.MinValue)]
		[InlineData(-1)]
		[InlineData(0)]
		[InlineData(1)]
		[InlineData(7)]
		[InlineData(11)]
		[InlineData(1024)]
		[InlineData(int.MaxValue)]
		public void CanCreatePrimitiveIntExpressions(object expected)
		{
			var result = (CodePrimitiveExpression)this.lang.Parse(expected.ToString());
			Assert.Equal(expected, result.Value);
		}

		[Theory]
		[InlineData("0x00", 0x00)]
		[InlineData("0x01", 0x01)]
		[InlineData("0xFF", 0xFF)]
		[InlineData("0x0000", 0x0000)]
		[InlineData("0x0101", 0x0101)]
		[InlineData("0xFFFF", 0xFFFF)]
		[InlineData("0x000000", 0x000000)]
		[InlineData("0x010101", 0x010101)]
		[InlineData("0xFFFFFF", 0xFFFFFF)]
		public void CanCreatePrimitiveHexExpressions(string hex, int expected)
		{
			var result = (CodePrimitiveExpression)this.lang.Parse(hex);
			Assert.Equal(expected, result.Value);
		}

		[Theory]
		[InlineData("0.0", 0.0)]
		[InlineData("-0.0", 0.0)]
		[InlineData("0.1", 0.1)]
		[InlineData("1.0", 1.0)]
		[InlineData("123456789.123456789", 123456789.123456789)]
		public void CanCreatePrimitiveDoubleExpressions(string hex, double expected)
		{
			var result = (CodePrimitiveExpression)this.lang.Parse(hex);
			Assert.Equal(expected, result.Value);
		}

		[Fact]
		public void CanCreateStringLiteralExpression()
		{
			var result = (CodePrimitiveExpression)this.lang.Parse("\"string\"");
			Assert.Equal("string", (string)result.Value);
		}

		[Fact]
		public void StringLiteralDoesNotTrimInnerQuotes()
		{
			var expected = @"""\""string\""""";
			var result = (CodePrimitiveExpression)this.lang.Parse(expected);
			Assert.Equal(expected, result.Print());
		}

		[Fact]
		public void StringLiteralExpressionUnescapesString()
		{
			var expected = @"""\\\0\n\r\t""";
			var result = (CodePrimitiveExpression)this.lang.Parse(expected);
			var actual = result.Print();
			Assert.Equal(expected, actual);
		}

		[Theory]
		[InlineData(@"'c'", 'c')]
		[InlineData(@"'\''", '\'')]
		[InlineData(@"'\\'", '\\')]
		[InlineData(@"'\0'", '\0')]
		[InlineData(@"'\n'", '\n')]
		[InlineData(@"'\r'", '\r')]
		[InlineData(@"'\t'", '\t')]
		public void CanCreateCharacterLiteral(string code, char expected)
		{
			var result = (CodePrimitiveExpression)this.lang.Parse(code);
			Assert.Equal(expected, (char)result.Value);
		}

		[Fact]
		public void CanCreatePropertySetValueExpression()
		{
			var result = (CodePropertySetValueReferenceExpression)this.lang.Parse("value");
			Assert.NotNull(result);
		}

		[Fact]
		public void CanCreateCastExpression()
		{
			var result = (CastExpression)this.lang.Parse("(x as y)");
			Assert.Equal("y", result.TargetType.BaseType);
			Assert.Equal("x", ((ReferenceExpression)result.Expression).PropertyName);
		}

		[Fact]
		public void CastHasHigherPriorityThanBooleanLogicalExpressions()
		{
			var result = (CodeBinaryOperatorExpression)this.lang.Parse("x as y and z");
			Assert.IsType<CastExpression>(result.Left);
		}

		[Fact]
		public void CastHasHigherPriorityThanEqualityExpressions()
		{
			var result = (CodeBinaryOperatorExpression)this.lang.Parse("x > z as y");
			Assert.IsType<CastExpression>(result.Right);
		}

		[Fact]
		public void CastExpressionIsHigherPriority()
		{
			var result = (CodeBinaryOperatorExpression)this.lang.Parse("y + x as z");
			Assert.IsType<CastExpression>(result.Right);

			result = (CodeBinaryOperatorExpression)this.lang.Parse("y as x + z");
			Assert.IsType<CastExpression>(result.Left);
		}

		[Fact]
		public void CanCreateTypeOfExpression()
		{
			var result = (CodeTypeOfExpression)this.lang.Parse("typeof(x)");
			Assert.Equal("x", result.Type.BaseType);
		}

		[Fact]
		public void CanCreateIsExpression()
		{
			var result = (CodeBinaryOperatorExpression)this.lang.Parse("x is X");
			var left = (CodeBinaryOperatorExpression)result.Left;
			var right = (CodeMethodInvokeExpression)result.Right;

			Assert.Equal(CodeBinaryOperatorType.BooleanAnd, result.Operator);
			Assert.Equal(CodeBinaryOperatorType.IdentityInequality, left.Operator);
			Assert.Null(((CodePrimitiveExpression)left.Right).Value);
			Assert.Equal("x", ((ReferenceExpression)left.Left).PropertyName);

			Assert.Equal(1, right.Parameters.Count);
			var getTypeOfX = (CodeMethodInvokeExpression)right.Parameters[0];
			Assert.Equal("GetType", getTypeOfX.Method.MethodName);
			Assert.Equal("x", ((ReferenceExpression)getTypeOfX.Method.TargetObject).PropertyName);
			Assert.Equal("X", ((CodeTypeOfExpression)right.Method.TargetObject).Type.BaseType);
			Assert.Equal("IsAssignableFrom", right.Method.MethodName);
		}

		[Fact]
		public void IsExpressionIsHigherPriorityThanAnd()
		{
			var result = (CodeBinaryOperatorExpression)this.lang.Parse("x is X and x is Y");
			Assert.Equal(CodeBinaryOperatorType.BooleanAnd, result.Operator);
			Assert.Equal(CodeBinaryOperatorType.BooleanAnd, ((CodeBinaryOperatorExpression)result.Left).Operator);
			Assert.Equal(CodeBinaryOperatorType.BooleanAnd, ((CodeBinaryOperatorExpression)result.Right).Operator);
		}

		[Fact]
		public void IsExpressionIsLowerPriorityThanAdd()
		{
			var result = (CodeBinaryOperatorExpression)this.lang.Parse("x + y is z");
			Assert.Equal(CodeBinaryOperatorType.BooleanAnd, result.Operator);
			Assert.Equal(CodeBinaryOperatorType.IdentityInequality, ((CodeBinaryOperatorExpression)result.Left).Operator);
			Assert.Equal(CodeBinaryOperatorType.Add, ((CodeBinaryOperatorExpression)((CodeBinaryOperatorExpression)result.Left).Left).Operator);
		}

		[Fact]
		public void CanCreateObjectCreateExpression()
		{
			var result = (CodeObjectCreateExpression)this.lang.Parse("new Foo()");
			Assert.Equal("Foo", result.CreateType.BaseType);
			Assert.Empty(result.Parameters);
		}

		[Fact]
		public void CanCreateObjectCreateExpressionWithParameters()
		{
			var result = (CodeObjectCreateExpression)this.lang.Parse("new Foo(x, y, z)");
			Assert.Equal("Foo", result.CreateType.BaseType);
			Assert.Equal(3, result.Parameters.Count);
			Assert.Equal("x", ((ReferenceExpression)result.Parameters[0]).PropertyName);
			Assert.Equal("y", ((ReferenceExpression)result.Parameters[1]).PropertyName);
			Assert.Equal("z", ((ReferenceExpression)result.Parameters[2]).PropertyName);
		}

		[Fact]
		public void CanCreateDefaultExpression()
		{
			var result = (CodeDefaultValueExpression)this.lang.Parse("default(x)");
			Assert.Equal("x", result.Type.BaseType);
		}

		[Fact]
		public void CanCreateArrayCreateExpression()
		{
			var result = (CodeArrayCreateExpression)this.lang.Parse("new Foo[]");
			Assert.Equal("Foo", result.CreateType.BaseType);
			Assert.Empty(result.Initializers);
			Assert.Null(result.SizeExpression);
		}

		[Fact]
		public void CanCreateArrayCreateExpressionWithSize()
		{
			var result = (CodeArrayCreateExpression)this.lang.Parse("new Foo[255]");
			Assert.Equal("Foo", result.CreateType.BaseType);
			Assert.Empty(result.Initializers);
			Assert.Equal(255, ((CodePrimitiveExpression)result.SizeExpression).Value);
		}
	}
}
