﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Xunit;
using MetaSharp.Transformation.Grammar;
using System.CodeDom;
using System.Reflection;
using MetaSharp.Transformation.Lang.Ast;
using MetaSharp.Transformation.Lang;

namespace MetaSharp.Transformation.Tests.Lang
{
	public class MethodTests : IUseFixture<LangParser>
	{
		private LangParser lang;
		public void SetFixture(LangParser data)
		{
			this.lang = data;
		}

		[Fact]
		public void CanHaveMethodWithEmptyParameters()
		{
			var result = (CodeMemberMethod)this.lang.Parse("void M(): end");
			Assert.Empty(result.Parameters);
		}

		[Fact]
		public void CanHaveMethodWithSeperatedEmptyParameters()
		{
			var result = (CodeMemberMethod)this.lang.Parse("void M (): end");
			Assert.Empty(result.Parameters);
		}

		[Fact]
		public void CanHaveMethodWithParameters()
		{
			var result = (CodeMemberMethod)this.lang.Parse("void M(X x, Y y): end");
			Assert.Equal(2, result.Parameters.Count);
			Assert.Equal("X", result.Parameters[0].Type.BaseType);
			Assert.Equal("Y", result.Parameters[1].Type.BaseType);
			Assert.Equal("x", result.Parameters[0].Name);
			Assert.Equal("y", result.Parameters[1].Name);
		}

		[Fact]
		public void ClassCanHaveMethodWithTypeParameters()
		{
			var method = (CodeMemberMethod)this.lang.Parse("void M<X,Y>: end");
			Assert.Equal(2, method.TypeParameters.Count);
			Assert.Equal("X", method.TypeParameters[0].Name);
			Assert.Equal("Y", method.TypeParameters[1].Name);
		}

		[Fact]
		public void ClassCanHaveMethodWithTypeParametersAndEmptyParameters()
		{
			var method = (CodeMemberMethod)this.lang.Parse("void M<X,Y>(): end");
			Assert.Equal(2, method.TypeParameters.Count);
			Assert.Equal("X", method.TypeParameters[0].Name);
			Assert.Equal("Y", method.TypeParameters[1].Name);
		}

		[Fact]
		public void ClassCanHaveMethodWithStatements()
		{
			var method = (CodeMemberMethod)this.lang.Parse("void M: return; end");
			Assert.Equal(1, method.Statements.Count);
			Assert.IsType<ReturnStatement>(method.Statements[0]);
		}

		[Fact]
		public void ClassCanHaveMethodWithEncapsulation()
		{
			var method = (CodeMemberMethod)this.lang.Parse("public void M: end");
			Assert.Equal(MemberAttributes.Public, method.Attributes);

			method = (CodeMemberMethod)this.lang.Parse("private void M: return; end");
			Assert.Equal(MemberAttributes.Private, method.Attributes);

			method = (CodeMemberMethod)this.lang.Parse("family void M: return; end");
			Assert.Equal(MemberAttributes.Family, method.Attributes);

			method = (CodeMemberMethod)this.lang.Parse("assembly void M: return; end");
			Assert.Equal(MemberAttributes.Assembly, method.Attributes);

			method = (CodeMemberMethod)this.lang.Parse("family and assembly void M: return; end");
			Assert.Equal(MemberAttributes.FamilyAndAssembly, method.Attributes);

			method = (CodeMemberMethod)this.lang.Parse("family or assembly void M: return; end");
			Assert.Equal(MemberAttributes.FamilyOrAssembly, method.Attributes);
		}

		[Fact]
		public void MethodCanReturnArray()
		{
			var method = (CodeMemberMethod)this.lang.Parse("public object[] M: end");
			Assert.Equal(1, method.ReturnType.ArrayRank);
			Assert.Equal("System.Object", method.ReturnType.BaseType);
		}

		[Fact]
		public void MethodCanReturnMultiDimensionalArray()
		{
			var method = (CodeMemberMethod)this.lang.Parse("public object[,] M: end");
			Assert.Equal(2, method.ReturnType.ArrayRank);
			Assert.Equal("System.Object", method.ReturnType.BaseType);
		}

		[Fact]
		public void CanReturnGenericType()
		{
			var context = new PatternContext();
			var method = (CodeMemberMethod)this.lang.Parse(@"
				public static IEnumerable<INode> Flatten(object match):
					return null;
				end",
				context);

			Assert.Empty(context.Errors);
			Assert.Equal("IEnumerable`1", method.ReturnType.BaseType);
			Assert.Equal(1, method.ReturnType.TypeArguments.Count);
			Assert.Equal("INode", method.ReturnType.TypeArguments[0].BaseType);
		}
	}
}
