﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Xunit;
using MetaSharp.Transformation.Grammar;
using System.CodeDom;
using MetaSharp.Transformation.Lang;
using MetaSharp.Transformation.Lang.Ast;

namespace MetaSharp.Transformation.Tests.Lang
{
	public class PropertyDeclarationTests : IUseFixture<LangParser>
	{
		private LangParser lang;
		public void SetFixture(LangParser data)
		{
			this.lang = data;
		}

		[Fact]
		public void ClassCanDeclareProperty()
		{
			var result = (CodeTypeDeclaration)this.lang.Parse("class X: P P: get; set; end end");
			Assert.Equal(2, result.Members.Count);
			Assert.IsType<Field>(result.Members[0]);
			Assert.IsType<Property>(result.Members[1]);
		}

		[Fact]
		public void CanDeclarePropertyWithAutoGetSet()
		{
			var result = (Nodes)this.lang.Parse("P P: get; set; end");
			Assert.Equal(2, result.Count);
			var field = (CodeMemberField)result[0];
			var prop = (CodeMemberProperty)result[1];
			Assert.Equal("P", field.Type.BaseType);
			Assert.Equal("__P", field.Name);
			Assert.Equal("P", prop.Type.BaseType);
			Assert.Equal("P", prop.Name);
			Assert.True(prop.HasSet);
			Assert.True(prop.HasGet);

			Assert.Equal(1, prop.GetStatements.Count);
			var ret = (ReturnStatement)prop.GetStatements[0];
			Assert.Equal("__P", ((ReferenceExpression)ret.Expression).PropertyName);
			Assert.Equal(1, prop.SetStatements.Count);
			var assign = (AssignStatement)prop.SetStatements[0];
			Assert.Equal("__P", ((ReferenceExpression)assign.Left).PropertyName);
			Assert.IsType<PropertySetValueReferenceExpression>(assign.Right);
		}

		[Fact]
		public void CanDeclarePropertyWithGetSet()
		{
			var prop = (CodeMemberProperty)this.lang.Parse("P P: get: return __P; end set: __P = value; end end");
			Assert.Equal("P", prop.Type.BaseType);
			Assert.Equal("P", prop.Name);
			Assert.True(prop.HasSet);
			Assert.True(prop.HasGet);

			Assert.Equal(1, prop.GetStatements.Count);
			var ret = (ReturnStatement)prop.GetStatements[0];
			Assert.Equal("__P", ((ReferenceExpression)ret.Expression).PropertyName);
			Assert.Equal(1, prop.SetStatements.Count);
			var assign = (AssignStatement)prop.SetStatements[0];
			Assert.Equal("__P", ((ReferenceExpression)assign.Left).PropertyName);
			Assert.IsType<PropertySetValueReferenceExpression>(assign.Right);
		}

		[Fact]
		public void CanDeclarePropertyWithGet()
		{
			var prop = (CodeMemberProperty)this.lang.Parse("P P: get: return __P; end end");
			Assert.Equal("P", prop.Type.BaseType);
			Assert.Equal("P", prop.Name);
			Assert.False(prop.HasSet);
			Assert.True(prop.HasGet);

			Assert.Equal(1, prop.GetStatements.Count);
			var ret = (ReturnStatement)prop.GetStatements[0];
			Assert.Equal("__P", ((ReferenceExpression)ret.Expression).PropertyName);
			Assert.Empty(prop.SetStatements);
		}

		[Fact]
		public void CanDeclarePropertyWithSet()
		{
			var prop = (CodeMemberProperty)this.lang.Parse("P P: set: __P = value; end end");
			Assert.Equal("P", prop.Type.BaseType);
			Assert.Equal("P", prop.Name);
			Assert.True(prop.HasSet);
			Assert.False(prop.HasGet);

			Assert.Empty(prop.GetStatements);
			Assert.Equal(1, prop.SetStatements.Count);
			var assign = (AssignStatement)prop.SetStatements[0];
			Assert.Equal("__P", ((ReferenceExpression)assign.Left).PropertyName);
			Assert.IsType<PropertySetValueReferenceExpression>(assign.Right);
		}

		[Fact]
		public void CanDeclarePropertyWithSetGet()
		{
			var prop = (CodeMemberProperty)this.lang.Parse("P P: set: __P = value; end get: return __P; end end");
			Assert.Equal("P", prop.Type.BaseType);
			Assert.Equal("P", prop.Name);
			Assert.True(prop.HasSet);
			Assert.True(prop.HasGet);

			Assert.Equal(1, prop.GetStatements.Count);
			var ret = (ReturnStatement)prop.GetStatements[0];
			Assert.Equal("__P", ((ReferenceExpression)ret.Expression).PropertyName);
			Assert.Equal(1, prop.SetStatements.Count);
			var assign = (AssignStatement)prop.SetStatements[0];
			Assert.Equal("__P", ((ReferenceExpression)assign.Left).PropertyName);
			Assert.IsType<PropertySetValueReferenceExpression>(assign.Right);
		}

		[Fact]
		public void CanDeclarePropertyWithAccess()
		{
			var result = (Nodes)this.lang.Parse("P P: get; set; end");
			var prop = (Property)result[1];
			Assert.Equal(MemberAttributes.Public, prop.Attributes);

			result = (Nodes)this.lang.Parse("public  P P: get; set; end");
			prop = (Property)result[1];
			Assert.Equal(MemberAttributes.Public, prop.Attributes);

			result = (Nodes)this.lang.Parse("private P P: get; set; end");
			prop = (Property)result[1];
			Assert.Equal(MemberAttributes.Private, prop.Attributes);

			result = (Nodes)this.lang.Parse("assembly P P: get; set; end");
			prop = (Property)result[1];
			Assert.Equal(MemberAttributes.Assembly, prop.Attributes);

			result = (Nodes)this.lang.Parse("family P P: get; set; end");
			prop = (Property)result[1];
			Assert.Equal(MemberAttributes.Family, prop.Attributes);

			result = (Nodes)this.lang.Parse("family or assembly P P: get; set; end");
			prop = (Property)result[1];
			Assert.Equal(MemberAttributes.FamilyOrAssembly, prop.Attributes);

			result = (Nodes)this.lang.Parse("family and assembly P P: get; set; end");
			prop = (Property)result[1];
			Assert.Equal(MemberAttributes.FamilyAndAssembly, prop.Attributes);
		}

		[Fact]
		public void CanDeclarePropertyWithOptions()
		{
			var result = (Nodes)this.lang.Parse("static P P: get; set; end");
			var prop = (Property)result[1];
			Assert.True((MemberAttributes.Static & prop.Attributes) == MemberAttributes.Static);

			result = (Nodes)this.lang.Parse("new P P: get; set; end");
			prop = (Property)result[1];
			Assert.True((MemberAttributes.New & prop.Attributes) == MemberAttributes.New);

			result = (Nodes)this.lang.Parse("final P P: get; set; end");
			prop = (Property)result[1];
			Assert.True((MemberAttributes.Final & prop.Attributes) == MemberAttributes.Final);

			result = (Nodes)this.lang.Parse("override P P: get; set; end");
			prop = (Property)result[1];
			Assert.True((MemberAttributes.Override & prop.Attributes) == MemberAttributes.Override);
		}

		[Fact]
		public void CanDeclarePropertyWithAccessAndOptions()
		{
			var result = (Nodes)this.lang.Parse("family override P P: get; set; end");
			var prop = (Property)result[1];
			Assert.True((MemberAttributes.Family & prop.Attributes) == MemberAttributes.Family);
			Assert.True((MemberAttributes.Override & prop.Attributes) == MemberAttributes.Override);
		}
	}
}
