﻿using System.CodeDom;
using MetaSharp.Transformation.Grammar;
using Xunit;
using Xunit.Extensions;
using MetaSharp.Transformation.Lang.Ast;
using MetaSharp.Transformation.Lang;

namespace MetaSharp.Transformation.Tests.Lang
{
	public class StatementTests : IUseFixture<LangParser>
	{
		private LangParser lang;
		public void SetFixture(LangParser data)
		{
			this.lang = data;
		}

		[Fact]
		public void CanCreateMethodReturnStatement()
		{
			var result = (CodeMethodReturnStatement)this.lang.Parse("return;");
			Assert.Null(result.Expression);
		}

		[Fact]
		public void CanCreateMethodReturnStatementWithExpression()
		{
			var result = (CodeMethodReturnStatement)this.lang.Parse("return x;");
			Assert.Equal("x", ((ReferenceExpression)result.Expression).PropertyName);
		}

		[Fact]
		public void CanCreateMinimalIterationStatement()
		{
			var result = (CodeIterationStatement)this.lang.Parse("for(;;):end");
			Assert.Null(result.InitStatement);
			Assert.Null(result.TestExpression);
			Assert.Null(result.IncrementStatement);
			Assert.Empty(result.Statements);
		}

		[Fact]
		public void CanCreateIterationStatementWithVariableDeclarationStatement()
		{
			var result = (CodeIterationStatement)this.lang.Parse("for(int i;;):end");
			Assert.Null(result.IncrementStatement);
			Assert.Null(result.TestExpression);
			Assert.Empty(result.Statements);
			var decl = (CodeVariableDeclarationStatement)result.InitStatement;
			Assert.Equal("i", decl.Name);
		}

		[Fact]
		public void CanCreateIterationStatementWithTestExpression()
		{
			var result = (CodeIterationStatement)this.lang.Parse("for(;x;):end");
			Assert.Null(result.InitStatement);
			Assert.Null(result.IncrementStatement);
			Assert.Empty(result.Statements);
			Assert.Equal("x", ((ReferenceExpression)result.TestExpression).PropertyName);
		}

		[Fact]
		public void CanCreateIterationStatementWithIncrementStatement()
		{
			dynamic result = (CodeIterationStatement)this.lang.Parse("for(;;i += 1):end");
			Assert.Null(result.InitStatement);
			Assert.Null(result.TestExpression);
			Assert.Empty(result.Statements);
			Assert.Equal("i", result.IncrementStatement.Left.PropertyName);
			Assert.Equal(1, result.IncrementStatement.Right.Right.Value);
		}

		[Fact]
		public void CanCreateForInStatement()
		{
			var result = (CodeIterationStatement)this.lang.Parse("for(int x in items):end");
			var init = (CodeVariableDeclarationStatement)result.InitStatement;
			var test = (CodeMethodInvokeExpression)result.TestExpression;
			Assert.NotNull(result.IncrementStatement);
			Assert.Equal(1, result.Statements.Count);

			Assert.Equal("_x_enumerator", init.Name);
			Assert.Equal("System.Collections.IEnumerator", init.Type.BaseType);
			var getEnumerator = (CodeMethodInvokeExpression)init.InitExpression;
			Assert.Equal("GetEnumerator", getEnumerator.Method.MethodName);
			Assert.Empty(getEnumerator.Parameters);
			var itemsCast = (CodeCastExpression)getEnumerator.Method.TargetObject;
			Assert.Equal("items", ((ReferenceExpression)itemsCast.Expression).PropertyName);
			Assert.Equal("System.Collections.IEnumerable", itemsCast.TargetType.BaseType);

			Assert.Empty(test.Parameters);
			Assert.Equal("MoveNext", test.Method.MethodName);
			var moveNextOn = (ReferenceExpression)test.Method.TargetObject;
			Assert.Equal("_x_enumerator", moveNextOn.PropertyName);

			var assignX = (CodeVariableDeclarationStatement)result.Statements[0];
			Assert.Equal("x", assignX.Name);
			var castCurrent = (CodeCastExpression)assignX.InitExpression;
			Assert.Equal("System.Int32", castCurrent.TargetType.BaseType);
			var current = (CodePropertyReferenceExpression)castCurrent.Expression;
			Assert.Equal("Current", current.PropertyName);
			Assert.Equal("_x_enumerator", ((ReferenceExpression)current.TargetObject).PropertyName);
		}

		[Fact]
		public void ForInStatementsCanHaveStatements()
		{
			var result = (IterationStatement)this.lang.Parse("for(int x in items):y += x;end");
			Assert.Equal(2, result.Statements.Count);
			var yPlusEqualX = result.Statements[1];
			Assert.IsType(typeof(AssignStatement), yPlusEqualX);
		}

		[Fact]
		public void CanCreateWhileStatement()
		{
			var result = (CodeIterationStatement)this.lang.Parse("while(x):end");
			Assert.Null(result.InitStatement);
			Assert.IsType<ExpressionStatement>(result.IncrementStatement);
			Assert.Equal("x", ((ReferenceExpression)result.TestExpression).PropertyName);
			Assert.Empty(result.Statements);
		}

		[Fact]
		public void WhileStatementCanHaveStatements()
		{
			var result = (CodeIterationStatement)this.lang.Parse("while(x):x=y;end");
			Assert.Equal(1, result.Statements.Count);
			Assert.IsType(typeof(AssignStatement), result.Statements[0]);
		}

		[Fact]
		public void CanCreateAssignmentStatement()
		{
			var result = (AssignStatement)this.lang.Parse("x = y;");
			Assert.Equal("x", ((ReferenceExpression)result.Left).PropertyName);
			Assert.Equal("y", ((ReferenceExpression)result.Right).PropertyName);
		}

		[Fact]
		public void CanCombineAssingmentWithExpressions()
		{
			dynamic result = (AssignStatement)this.lang.Parse("x = y + z;");
			Assert.Equal("x", result.Left.PropertyName);
			Assert.Equal("y", result.Right.Left.PropertyName);
			Assert.Equal("z", result.Right.Right.PropertyName);
		}

		[Fact]
		public void CanDoPlusEqualStatement()
		{
			dynamic result = this.lang.Parse("x += y;");
			Assert.Equal("x", result.Left.PropertyName);
			Assert.Equal("x", result.Right.Left.PropertyName);
			Assert.Equal(CodeBinaryOperatorType.Add, result.Right.Operator);
			Assert.Equal("y", result.Right.Right.PropertyName);
		}

		[Fact]
		public void CanDoMinusEqualStatement()
		{
			dynamic result = this.lang.Parse("x -= y;");
			Assert.Equal("x", result.Left.PropertyName);
			Assert.Equal("x", result.Right.Left.PropertyName);
			Assert.Equal(CodeBinaryOperatorType.Subtract, result.Right.Operator);
			Assert.Equal("y", result.Right.Right.PropertyName);
		}

		[Fact]
		public void CanDeclaraVariables()
		{
			var result = (CodeIterationStatement)this.lang.Parse("for(;;): int x; end");
			var variable = (CodeVariableDeclarationStatement)result.Statements[0];
			Assert.Equal("x", variable.Name);
			Assert.Equal("System.Int32", variable.Type.BaseType);
			Assert.Null(variable.InitExpression);
		}

		[Fact]
		public void CanDeclaraVariableWithInitExpression()
		{
			var result = (CodeIterationStatement)this.lang.Parse("for(;;): int x = 7; end");
			var variable = (CodeVariableDeclarationStatement)result.Statements[0];
			Assert.Equal("x", variable.Name);
			Assert.Equal("System.Int32", variable.Type.BaseType);
			Assert.Equal(7, ((CodePrimitiveExpression)variable.InitExpression).Value);
		}

		[Theory]
		[InlineData("char")]
		//[InlineData("end")]
		[InlineData("true")]
		[InlineData("null")]
		[InlineData("if")]
		[InlineData("typeof")]
		[InlineData("constructor")]
		[InlineData("class")]
		[InlineData("or")]
		[InlineData("new")]
		[InlineData("static")]
		[InlineData("public")]
		public void CanNotDeclareVariableWithInvalidName(string invalidName)
		{
			var context = new PatternContext();
			var result = this.lang.Parse("void Foo: object " + invalidName + "; end", context);
			Assert.IsType<Method>(result);
			Assert.NotEmpty(context.Errors);
		}

		[Fact]
		public void CanNotDeclareVariableNamedEnd()
		{
			var context = new PatternContext();
			var result = (Method)this.lang.Parse("void Foo: object end; end", context);

			Assert.Empty(result.Statements);
			Assert.NotEmpty(context.Errors);
		}

		[Fact]
		public void CallCanBeValidExpressionStatement()
		{
			var context = new PatternContext();
			var result = this.lang.Parse("void Foo: x(); end", context);
			Assert.IsNotType<Fail>(result);
			Assert.Empty(context.Errors);
		}

		[Fact]
		public void NewObjectCanBeValidExpressionStatement()
		{
			var context = new PatternContext();
			var result = this.lang.Parse("void Foo: new x(); end", context);
			Assert.IsNotType<Fail>(result);
			Assert.Empty(context.Errors);
		}

		[Fact]
		public void ReferenceExpressionIsNotValidExpressionStatement()
		{
			var context = new PatternContext();
			var result = this.lang.Parse("void Foo: x; end", context);
			Assert.IsNotType<Fail>(result);
			Assert.NotEmpty(context.Errors);
		}

		[Fact]
		public void IndexerExpressionIsNotValidExpressionStatement()
		{
			var context = new PatternContext();
			var result = this.lang.Parse("void Foo: x[0]; end", context);
			Assert.IsNotType<Fail>(result);
			Assert.NotEmpty(context.Errors);
		}

		[Fact]
		public void CanDeclareIfStatement()
		{
			var result = (CodeConditionStatement)this.lang.Parse("if(x):end");
			Assert.Equal("x", ((ReferenceExpression)result.Condition).PropertyName);
			Assert.Empty(result.TrueStatements);
			Assert.Empty(result.FalseStatements);
		}

		[Fact]
		public void CanDeclareIfStatementWithStatements()
		{
			var result = (CodeConditionStatement)this.lang.Parse("if(x):x=y;end");
			Assert.Equal(1, result.TrueStatements.Count);
			Assert.IsType(typeof(AssignStatement), result.TrueStatements[0]);
		}

		[Fact]
		public void CanDeclareIfElseStatement()
		{
			var result = (CodeConditionStatement)this.lang.Parse("if(x):else:end");
			Assert.Equal("x", ((ReferenceExpression)result.Condition).PropertyName);
			Assert.Empty(result.TrueStatements);
			Assert.Empty(result.FalseStatements);
		}

		[Fact]
		public void CanDeclareIfElseIfStatement()
		{
			var result = (CodeConditionStatement)this.lang.Parse("if(x):else if(y):end");
			Assert.Equal("x", ((ReferenceExpression)result.Condition).PropertyName);
			Assert.Empty(result.TrueStatements);
			Assert.Equal(1, result.FalseStatements.Count);
			Assert.Equal("y", ((ReferenceExpression)((CodeConditionStatement)result.FalseStatements[0]).Condition).PropertyName);
		}

		[Fact]
		public void CanDeclareTryFinally()
		{
			var result = (CodeTryCatchFinallyStatement)this.lang.Parse("try:x=y;finally:x=y;end");
			Assert.Equal(1, result.TryStatements.Count);
			Assert.Empty(result.CatchClauses);
			Assert.Equal(1, result.FinallyStatements.Count);
		}

		[Fact]
		public void CanDeclareTryCatch()
		{
			var result = (CodeTryCatchFinallyStatement)this.lang.Parse("try:x=y;catch:x=y;end");
			Assert.Equal(1, result.TryStatements.Count);
			Assert.Equal(1, result.CatchClauses.Count);
			Assert.Equal(1, result.CatchClauses[0].Statements.Count);
			Assert.Empty(result.CatchClauses[0].LocalName);
			Assert.Equal("System.Exception", result.CatchClauses[0].CatchExceptionType.BaseType);
		}

		[Fact]
		public void CanDeclareTryCatchWithParameter()
		{
			var result = (CodeTryCatchFinallyStatement)this.lang.Parse("try:x=y;catch(Exception ex):x=y;end");
			Assert.Equal(1, result.TryStatements.Count);
			Assert.Equal(1, result.CatchClauses.Count);
			Assert.Equal(1, result.CatchClauses[0].Statements.Count);
			Assert.Equal("ex", result.CatchClauses[0].LocalName);
			Assert.Equal("Exception", result.CatchClauses[0].CatchExceptionType.BaseType);
		}

		[Fact]
		public void CanDeclareTryCatchWithMultipleParameters()
		{
			var result = (CodeTryCatchFinallyStatement)this.lang.Parse("try:x=y;catch(A a, B b):x=y;end");
			Assert.Equal(1, result.TryStatements.Count);
			Assert.Equal(2, result.CatchClauses.Count);
			Assert.Equal(1, result.CatchClauses[0].Statements.Count);
			Assert.Equal("a", result.CatchClauses[0].LocalName);
			Assert.Equal("A", result.CatchClauses[0].CatchExceptionType.BaseType);
			Assert.Equal(1, result.CatchClauses[1].Statements.Count);
			Assert.Equal("b", result.CatchClauses[1].LocalName);
			Assert.Equal("B", result.CatchClauses[1].CatchExceptionType.BaseType);
		}

		[Fact]
		public void CanDeclareTryWithMultipleCatches()
		{
			var result = (CodeTryCatchFinallyStatement)this.lang.Parse("try:x=y;catch(A a):x=y;catch(B b):x=y;end");
			Assert.Equal(1, result.TryStatements.Count);
			Assert.Equal(2, result.CatchClauses.Count);
			Assert.Equal(1, result.CatchClauses[0].Statements.Count);
			Assert.Equal("a", result.CatchClauses[0].LocalName);
			Assert.Equal("A", result.CatchClauses[0].CatchExceptionType.BaseType);
			Assert.Equal(1, result.CatchClauses[1].Statements.Count);
			Assert.Equal("b", result.CatchClauses[1].LocalName);
			Assert.Equal("B", result.CatchClauses[1].CatchExceptionType.BaseType);
		}

		[Fact]
		public void CanDeclareTryWithCatchAndFinally()
		{
			var result = (CodeTryCatchFinallyStatement)this.lang.Parse("try:x=y;catch:x=y;finally:x=y;end");
			Assert.Equal(1, result.TryStatements.Count);
			Assert.Equal(1, result.CatchClauses.Count);
			Assert.Equal(1, result.FinallyStatements.Count);
		}

		[Fact]
		public void CanDeclareTryCatchWithoutParameterName()
		{
			var result = (CodeTryCatchFinallyStatement)this.lang.Parse("try:x=y;catch(X):end");
			Assert.Equal(1, result.TryStatements.Count);
			Assert.Equal(1, result.CatchClauses.Count);
			Assert.Empty(result.CatchClauses[0].LocalName);
			Assert.Equal("X", result.CatchClauses[0].CatchExceptionType.BaseType);
		}

		[Fact]
		public void CanDeclareLabel()
		{
			var result = (CodeLabeledStatement)this.lang.Parse("label foo:x=y;");
			Assert.Equal("foo", result.Label);
			Assert.IsType(typeof(AssignStatement), result.Statement);
		}

		[Fact]
		public void CanThrow()
		{
			var result = (CodeThrowExceptionStatement)this.lang.Parse("throw;");
			Assert.Null(result.ToThrow);
		}

		[Fact]
		public void CanThrowExpression()
		{
			var result = (CodeThrowExceptionStatement)this.lang.Parse("throw x;");
			Assert.NotNull(result.ToThrow);
		}

		[Fact]
		public void CanThrowComplexExpression()
		{
			var result = (CodeThrowExceptionStatement)this.lang.Parse("throw x().y.z;");
			Assert.NotNull(result.ToThrow);
		}

		[Fact]
		public void CanDeclareGoTo()
		{
			var result = (CodeGotoStatement)this.lang.Parse("goto x;");
			Assert.Equal("x", result.Label);
		}

		[Fact]
		public void CanDeclareArray()
		{
			var result = (Field)this.lang.Parse("object x = new object[0];");
			var arrayCreate = (ArrayCreateExpression)result.InitExpression;
			Assert.Equal("System.Object", arrayCreate.CreateType.BaseType);
			Assert.Equal(0, ((PrimitiveExpression)arrayCreate.SizeExpression).Value);
			Assert.Empty(arrayCreate.Initializers);
		}

		[Fact]
		public void FieldMissingStatementEndIsInvalid()
		{
			var context = new PatternContext();
			var result = this.lang.Parse("int x = 0", context);
			var errors = context.Errors;
			Assert.IsType<Field>(result);
			Assert.Equal(1, errors.Length);
		}

		[Fact]
		public void VariableDeclarationMissingStatementEndIsInvalid()
		{
			var context = new PatternContext();
			var result = (ConditionStatement)this.lang.Parse("if(true): int x = 0 end", context);
			var errors = context.Errors;
			Assert.IsType<VariableDeclarationStatement>(result.TrueStatements[0]);
			Assert.Equal(1, errors.Length);
		}

		[Fact]
		public void FirstOfTwoStatementsMissingStatementEndResultsInError()
		{
			var context = new PatternContext();
			var result = (ConditionStatement)this.lang.Parse("if ( true ) : int x = 0 return x; end", context);
			var errors = context.Errors;
			Assert.Equal(1, errors.Length);
			Assert.Equal(2, result.TrueStatements.Count);

			Assert.Equal(23, ((IIndexed)errors[0].Span.Start).Index);
			Assert.Equal(24, ((IIndexed)errors[0].Span.End).Index);
		}

		[Fact]
		public void IfContainingInvalidCodeResultsInError()
		{
			var context = new PatternContext();
			var result = (ConditionStatement)this.lang.Parse("if(true): #$%^ end", context);
			var errors = context.Errors;
			Assert.Empty(result.TrueStatements);
			Assert.Equal(1, errors.Length);
			Assert.Equal(10, ((IIndexed)errors[0].Span.Start).Index);
			Assert.Equal(13, ((IIndexed)errors[0].Span.End).Index);
		}

		[Fact]
		public void ElseContainingInvalidCodeResultsInError()
		{
			var context = new PatternContext();
			var result = (ConditionStatement)this.lang.Parse("if(true): else: #$%^ end", context);
			var errors = context.Errors;
			Assert.Empty(result.TrueStatements);
			Assert.Empty(result.FalseStatements);
			Assert.Equal(1, errors.Length);
			Assert.Equal(16, ((IIndexed)errors[0].Span.Start).Index);
			Assert.Equal(19, ((IIndexed)errors[0].Span.End).Index);
		}

		[Fact]
		public void ElseIfContainingInvalidCodeResultsInError()
		{
			var context = new PatternContext();
			var result = (ConditionStatement)this.lang.Parse("if(true): else if(false): #$%^ end", context);
			var errors = context.Errors;
			Assert.Empty(result.TrueStatements);
			Assert.Equal(1, result.FalseStatements.Count);
			var elif = (ConditionStatement)result.FalseStatements[0];
			Assert.Empty(elif.TrueStatements);
			Assert.Empty(elif.FalseStatements);

			Assert.Equal(1, errors.Length);
			Assert.Equal(26, ((IIndexed)errors[0].Span.Start).Index);
			Assert.Equal(29, ((IIndexed)errors[0].Span.End).Index);
		}
	}
}
