﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Xunit;
using MetaSharp.Transformation.Grammar;
using System.CodeDom;
using System.Reflection;
using MetaSharp.Transformation.Lang;

namespace MetaSharp.Transformation.Tests.Lang
{
	public class UnitAndNamespaceTests : IUseFixture<LangParser>
	{
		private LangParser lang;
		public void SetFixture(LangParser data)
		{
			this.lang = data;
		}

		[Fact]
		public void CanParseSingleAssemblyLevelAttributeWithNoParameters()
		{
			var result = (CodeCompileUnit)this.lang.Parse("[assembly:X]");
			Assert.Empty(result.Namespaces);
			Assert.Equal(1, result.AssemblyCustomAttributes.Count);
			Assert.Equal("X", result.AssemblyCustomAttributes[0].Name);
			Assert.Empty(result.AssemblyCustomAttributes[0].Arguments);
		}

		[Fact]
		public void CanParseSingleAssemblyLevelAttributeWithEmptyParameters()
		{
			var result = (CodeCompileUnit)this.lang.Parse("[assembly:X()]");
			Assert.Empty(result.Namespaces);
			Assert.Equal(1, result.AssemblyCustomAttributes.Count);
			Assert.Equal("X", result.AssemblyCustomAttributes[0].Name);
			Assert.Empty(result.AssemblyCustomAttributes[0].Arguments);
		}

		[Fact]
		public void CanParseMultipleAssemblyLevelAttributes()
		{
			var result = (CodeCompileUnit)this.lang.Parse("[assembly:X][assembly:Y][assembly:Z]");
			Assert.Empty(result.Namespaces);
			Assert.Equal(3, result.AssemblyCustomAttributes.Count);
			Assert.Equal("X", result.AssemblyCustomAttributes[0].Name);
			Assert.Equal("Y", result.AssemblyCustomAttributes[1].Name);
			Assert.Equal("Z", result.AssemblyCustomAttributes[2].Name);
		}

		[Fact]
		public void CanParseTypeReferenceWithGenericParameter()
		{
			var result = (CodeCompileUnit)this.lang.Parse("[assembly:X<Y>]");
			Assert.Empty(result.Namespaces);
			Assert.Equal(1, result.AssemblyCustomAttributes.Count);
			Assert.Equal("X`1", result.AssemblyCustomAttributes[0].AttributeType.BaseType);
			Assert.Equal(1, result.AssemblyCustomAttributes[0].AttributeType.TypeArguments.Count);
			Assert.Equal("Y", result.AssemblyCustomAttributes[0].AttributeType.TypeArguments[0].BaseType);
		}

		[Fact]
		public void CanParseTypeReferenceWithMultipleGenericParameters()
		{
			var result = (CodeCompileUnit)this.lang.Parse("[assembly:X<Y,Z>]");
			Assert.Empty(result.Namespaces);
			Assert.Equal(1, result.AssemblyCustomAttributes.Count);
			Assert.Equal("X`2", result.AssemblyCustomAttributes[0].AttributeType.BaseType);
			Assert.Equal(2, result.AssemblyCustomAttributes[0].AttributeType.TypeArguments.Count);
			Assert.Equal("Y", result.AssemblyCustomAttributes[0].AttributeType.TypeArguments[0].BaseType);
			Assert.Equal("Z", result.AssemblyCustomAttributes[0].AttributeType.TypeArguments[1].BaseType);
		}

		[Fact]
		public void CanParseTypeReferenceWithNestedGenericParameters()
		{
			var result = (CodeCompileUnit)this.lang.Parse("[assembly:X<Y<Z>>]");
			Assert.Empty(result.Namespaces);
			Assert.Equal(1, result.AssemblyCustomAttributes.Count);
			Assert.Equal("X`1", result.AssemblyCustomAttributes[0].AttributeType.BaseType);
			Assert.Equal(1, result.AssemblyCustomAttributes[0].AttributeType.TypeArguments.Count);
			Assert.Equal("Y`1", result.AssemblyCustomAttributes[0].AttributeType.TypeArguments[0].BaseType);
			Assert.Equal("Z", result.AssemblyCustomAttributes[0].AttributeType.TypeArguments[0].TypeArguments[0].BaseType);
		}

		[Fact]
		public void CanDeclareEmptyNamespace()
		{
			var result = (CodeCompileUnit)this.lang.Parse("namespace X: end");
			Assert.Equal(1, result.Namespaces.Count);
			Assert.Equal("X", result.Namespaces[0].Name);
			Assert.Empty(result.Namespaces[0].Imports);
			Assert.Empty(result.Namespaces[0].Types);
		}

		[Fact]
		public void CanDeclareMultipleNamespaces()
		{
			var result = (CodeCompileUnit)this.lang.Parse("namespace X: end namespace Y: end");
			Assert.Equal(2, result.Namespaces.Count);
			Assert.Equal("X", result.Namespaces[0].Name);
			Assert.Equal("Y", result.Namespaces[1].Name);
		}

		[Fact]
		public void CanDeclareAssemblyAttributeAndNamesapce()
		{
			var result = (CodeCompileUnit)this.lang.Parse("[assembly:X] namespace Y: end");
			Assert.Equal(1, result.AssemblyCustomAttributes.Count);
			Assert.Equal("X", result.AssemblyCustomAttributes[0].Name);
			Assert.Equal(1, result.Namespaces.Count);
			Assert.Equal("Y", result.Namespaces[0].Name);
			Assert.Empty(result.Namespaces[0].Imports);
			Assert.Empty(result.Namespaces[0].Types);
		}

		[Fact]
		public void CanDeclareNamespaceWithImport()
		{
			var result = (CodeCompileUnit)this.lang.Parse("namespace X: import Y; end");
			Assert.Equal(1, result.Namespaces.Count);
			Assert.Equal("X", result.Namespaces[0].Name);
			Assert.Empty(result.Namespaces[0].Types);
			Assert.Equal(1, result.Namespaces[0].Imports.Count);
			Assert.Equal("Y", result.Namespaces[0].Imports[0].Namespace);
		}

		[Fact]
		public void CanDeclareNamespaceWithMultipleImports()
		{
			var result = (CodeCompileUnit)this.lang.Parse("namespace X: import Y; import Z; end");
			Assert.Equal(1, result.Namespaces.Count);
			Assert.Equal("X", result.Namespaces[0].Name);
			Assert.Empty(result.Namespaces[0].Types);
			Assert.Equal(2, result.Namespaces[0].Imports.Count);
			Assert.Equal("Y", result.Namespaces[0].Imports[0].Namespace);
			Assert.Equal("Z", result.Namespaces[0].Imports[1].Namespace);
		}

		[Fact]
		public void CanDeclareImportWithComplexNamespace()
		{
			var result = (CodeCompileUnit)this.lang.Parse("namespace X: import A.B.C; end");
			Assert.Equal(1, result.Namespaces[0].Imports.Count);
			Assert.Equal("A.B.C", result.Namespaces[0].Imports[0].Namespace);
		}
	}
}
