﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Xunit;

namespace MetaSharp.Transformation.Tests.Patterns
{
	public class ErrorTests
	{
		private PatternContext context = new PatternContext();

		[Fact]
		public void SingleValueMatchedProducesNoErrors()
		{
			var R = Pattern.ErrorUnless(Pattern.Value('a'));
			var result = Pattern.Parse(context, "a", R);
			Assert.Equal('a', Node.Unwrap(result));
			Assert.Empty(context.Errors);
		}

		[Fact]
		public void SingleValueUnmatchedProducesError()
		{
			var R = Pattern.ErrorUnless(Pattern.Value('a'));
			var result = Pattern.Parse(context, "z", R);
			var errors = context.Errors;
			Assert.Null(Node.Unwrap(result));
			Assert.Equal(2, errors.Length);
			Assert.Equal(-1, ((IIndexed)errors[0].Span.Start).Index);
			Assert.Equal(0, ((IIndexed)errors[0].Span.End).Index);
			Assert.Equal(0, ((IIndexed)errors[1].Span.Start).Index);
			Assert.Equal(0, ((IIndexed)errors[1].Span.End).Index);
		}

		[Fact]
		public void SomeValuesMatchedProcudesError()
		{
			var R = Pattern.ErrorUnless(Pattern.And(Pattern.Value('a'), Pattern.Value('a')));
			var result = Pattern.Parse(context, "az", R);
			var errors = context.Errors;
			Assert.Null(Node.Unwrap(result));
			Assert.Equal(2, errors.Length);
			Assert.Equal(-1, ((IIndexed)errors[0].Span.Start).Index);
			Assert.Equal(0, ((IIndexed)errors[0].Span.End).Index);
			Assert.Equal(0, ((IIndexed)errors[1].Span.Start).Index);
			Assert.Equal(1, ((IIndexed)errors[1].Span.End).Index);
		}

		[Fact]
		public void SomeValuesMatchedProcudesSpecificError()
		{
			var R1 = Pattern.ErrorUnless(Pattern.Value('a'));
			var R2 = Pattern.And(R1, R1);
			var result = (Nodes)Pattern.Parse(context, "az", R2);
			var errors = context.Errors;
			Assert.Equal('a', Node.Unwrap(result[0]));
			Assert.Null(Node.Unwrap(result[1]));
			Assert.Equal(2, errors.Length);
			Assert.Equal(0, ((IIndexed)errors[0].Span.Start).Index);
			Assert.Equal(1, ((IIndexed)errors[0].Span.End).Index);
			Assert.Equal(1, ((IIndexed)errors[1].Span.Start).Index);
			Assert.Equal(1, ((IIndexed)errors[1].Span.End).Index);
		}

		[Fact]
		public void ErrorUnlessConsumesNoInput()
		{
			var R1 = Pattern.ErrorUnless(Pattern.Value('a'));
			var R2 = Pattern.Value('z');
			var R3 = Pattern.And(R1, R1, R2);
			var result = (Nodes)Pattern.Parse(context, "az", R3);
			var errors = context.Errors;

			Assert.Equal(3, result.Count);
			Assert.Equal('a', Node.Unwrap(result[0]));
			Assert.Null(Node.Unwrap(result[1]));
			Assert.Equal('z', Node.Unwrap(result[2]));
			Assert.Equal(1, errors.Length);
			Assert.Equal(0, ((IIndexed)errors[0].Span.Start).Index);
			Assert.Equal(1, ((IIndexed)errors[0].Span.End).Index);
		}

		[Fact]
		public void ErrorUnlessProcudeErrorAcrossMultipleSteps()
		{
			var R1 = Pattern.Rule(
				Pattern.OneOrMore(
					Pattern.Rule(
						Pattern.Projection(
							Pattern.Any(),
							c => Node.Wrap(Tokenizer.Stringify(c.Match))))));

			var R2 = Pattern.And(
				Pattern.ErrorUnless(Pattern.Value("a")),
				Pattern.ErrorUnless(Pattern.Value("z")));

			var result = (Nodes)Pattern.Parse(context, "aa", R1, R2);
			var errors = context.Errors;

			Assert.Equal("a", Node.Unwrap(result[0]));
			Assert.Null(Node.Unwrap(result[1]));

			Assert.Equal(2, errors.Length);
			Assert.Equal(1, ((IIndexed)errors[0].Span.Start).Index);
			Assert.Equal(2, ((IIndexed)errors[0].Span.End).Index);
			Assert.Equal(1, ((IIndexed)errors[1].Span.Start).Index);
			Assert.Equal(2, ((IIndexed)errors[1].Span.End).Index);
		}

		[Fact]
		public void UnexpectedValuesConsumesAllUntilEnd()
		{
			var R = Pattern.ErrorUntil(Pattern.Value('a'));
			var result = Pattern.Parse(context, "zzz", R);
			var errors = context.Errors;

			Assert.Null(Node.Unwrap(result));
			Assert.Equal(1, errors.Length);
			Assert.Equal(0, ((IIndexed)errors[0].Span.Start).Index);
			Assert.Equal(2, ((IIndexed)errors[0].Span.End).Index);
		}

		[Fact]
		public void UnexpectedValuesConsumedUntilExpectedValue()
		{
			var R = Pattern.ErrorUntil(Pattern.Value('a'));
			var result = Pattern.Parse(context, "za", R);
			var errors = context.Errors;

			Assert.Equal('a', Node.Unwrap(result));
			Assert.Equal(1, errors.Length);
			Assert.Equal(0, ((IIndexed)errors[0].Span.Start).Index);
			Assert.Equal(0, ((IIndexed)errors[0].Span.End).Index);
		}

		[Fact]
		public void UnexpectedValuesSurroundedByExpectedValuesConsumed()
		{
			var R = Pattern.And(
				Pattern.Value('a'),
				Pattern.ErrorUntil(Pattern.Value('a')));

			var result = (Nodes)Pattern.Parse(context, "azzza", R);
			var errors = context.Errors;
			Assert.Equal(2, result.Count);
			Assert.Equal('a', Node.Unwrap(result[0]));
			Assert.Equal('a', Node.Unwrap(result[1]));

			Assert.Equal(1, errors.Length);
			Assert.Equal(1, ((IIndexed)errors[0].Span.Start).Index);
			Assert.Equal(3, ((IIndexed)errors[0].Span.End).Index);
		}

		[Fact]
		public void UnexpectedValuesUntilEndProducesErrorAndFail()
		{
			var R = Pattern.And(
				Pattern.Value('a'),
				Pattern.ErrorUntil(Pattern.Value('a')));

			var result = (Nodes)Pattern.Parse(context, "azzz", R);
			var errors = context.Errors;
			Assert.Equal('a', Node.Unwrap(result[0]));
			Assert.Null(Node.Unwrap(result[1]));

			Assert.Equal(1, errors.Length);
			Assert.Equal(1, ((IIndexed)errors[0].Span.Start).Index);
			Assert.Equal(3, ((IIndexed)errors[0].Span.End).Index);
		}

		[Fact]
		public void ErrorUntilProcudeErrorAcrossMultipleSteps()
		{
			var R1 = Pattern.Rule(
				Pattern.OneOrMore(
					Pattern.Rule(
						Pattern.Projection(
							Pattern.Any(),
							c => Node.Wrap(Tokenizer.Stringify(c.Match))))));

			var R2 = Pattern.And(
				Pattern.Value("a"),
				Pattern.ErrorUntil(Pattern.Value("a")));

			var result = (Nodes)Pattern.Parse(context, "azzza", R1, R2);
			var errors = context.Errors;

			Assert.Equal(1, errors.Length);
			Assert.Equal(1, ((IIndexed)errors[0].Span.Start).Index);
			Assert.Equal(3, ((IIndexed)errors[0].Span.End).Index);
		}

		[Fact]
		public void TransformRuleHasImplicitErrorUntil()
		{
			var context = new PatternContext();
			var R = Pattern.Value('a');
			var result = Pattern.Transform(context, "zzzz", R);
			Assert.Null(Node.Unwrap(result));
			Assert.Equal(1, context.Errors.Length);
			Assert.Equal(0, ((IIndexed)context.Errors[0].Span.Start).Index);
			Assert.Equal(3, ((IIndexed)context.Errors[0].Span.End).Index);
		}

	}
}
