﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Xunit;

namespace MetaSharp.Transformation.Tests.Patterns
{
	public class FailTests
	{
		private PatternContext context = new PatternContext();
		private static PatternMatch fail = Pattern.Rule(Pattern.Fail());

		[Fact]
		public void FailInAndTracesSuccessfully()
		{
			var result = Pattern.Parse(context, "abc", Pattern.And(
				Pattern.Value('a'),
				Pattern.Value('b'),
				fail));

			Assert.IsType<Null>(result);
			Assert.Equal(1, context.Errors.Length);
			Assert.Equal(0, ((IIndexed)context.Errors[0].Span.Start).Index);
			Assert.Equal(2, ((IIndexed)context.Errors[0].Span.End).Index);
		}

		[Fact]
		public void FailInAnyTracesSuccessfully()
		{
			var result = Pattern.Parse(context, "", Pattern.Rule(Pattern.Any()));
			Assert.IsType<Null>(result);
		}

		[Fact]
		public void FailWhenArrayDoesntReadToEndTracesSuccessfully()
		{
			var result = Pattern.Transform(context, new[] { new[] { 7, 11 } }, Pattern.Rule(Pattern.Array(Pattern.Any())));
			var e = context.Errors[0];
			Assert.Equal(0, ((IIndexed)e.Span.Start).Index);
			Assert.Equal(0, ((IIndexed)e.Span.End).Index);
		}

		[Fact]
		public void FailWhenArrayDoesntMatchTracesSuccessfully()
		{
			var result = Pattern.Transform(context, new[] { new[] { 7 } }, Pattern.Rule(Pattern.Array(fail)));

			Assert.IsType<Null>(result);
			Assert.Equal(1, context.Errors.Length);
			Assert.Equal(0, ((IIndexed)context.Errors[0].Span.Start).Index);
			Assert.Equal(0, ((IIndexed)context.Errors[0].Span.End).Index);
		}

		[Fact]
		public void FailWhenObjectIsntArrayTracesSuccessfully()
		{
			var result = Pattern.Transform(context, new[] { 7 }, Pattern.Rule(Pattern.Array(Pattern.Any())));

			Assert.IsType<Null>(result);
			Assert.Equal(1, context.Errors.Length);
			Assert.Equal(0, ((IIndexed)context.Errors[0].Span.Start).Index);
			Assert.Equal(0, ((IIndexed)context.Errors[0].Span.End).Index);
		}

		[Fact]
		public void FailArrayOnEmptyStreamTracesSuccessfully()
		{
			var result = Pattern.Parse(context, "", Pattern.Rule(Pattern.Array(Pattern.Any())));
			Assert.IsType<Null>(result);
			Assert.Empty(context.Errors);
		}

		[Fact]
		public void FailInOrTracesSuccessfully()
		{
			var result = Pattern.Parse(context, "ab", Pattern.Rule(
				Pattern.Or(
					Pattern.And(Pattern.Value('a'), fail),
					Pattern.And(Pattern.Value('a'), Pattern.Value('b'), fail))));

			Assert.IsType<Null>(result);
			Assert.Equal(1, context.Errors.Length);
			Assert.Equal(0, ((IIndexed)context.Errors[0].Span.Start).Index);
			Assert.Equal(1, ((IIndexed)context.Errors[0].Span.End).Index);
		}

		[Fact]
		public void FailToReadToEndSingleRule()
		{
			var result = (Nodes)Pattern.Parse(context, "abc123", Pattern.Rule(
				Pattern.OneOrMore(
					Pattern.Range('a', 'z'))));

			Assert.Equal(3, result.Count);
			Assert.Equal('a', Node.Unwrap(result[0]));
			Assert.Equal('b', Node.Unwrap(result[1]));
			Assert.Equal('c', Node.Unwrap(result[2]));
			Assert.Equal(1, context.Errors.Length);
			Assert.Equal(3, ((IIndexed)context.Errors[0].Span.Start).Index);
			Assert.Equal(5, ((IIndexed)context.Errors[0].Span.End).Index);
		}

		[Fact]
		public void FailToReadToEndRuleAtEachPosition()
		{
			var result = (Nodes)Pattern.Parse(context, "abc123", Pattern.Rule(
				Pattern.OneOrMore(
					Pattern.Rule(
						Pattern.Range('a', 'z')))));

			Assert.Equal(3, result.Count);
			Assert.Equal('a', Node.Unwrap(result[0]));
			Assert.Equal('b', Node.Unwrap(result[1]));
			Assert.Equal('c', Node.Unwrap(result[2]));
			Assert.Equal(1, context.Errors.Length);
			Assert.Equal(3, ((IIndexed)context.Errors[0].Span.Start).Index);
			Assert.Equal(5, ((IIndexed)context.Errors[0].Span.End).Index);
		}

		[Fact]
		public void FailToReadToEndRuleAtEachPositionWithMultiplePaths()
		{
			var result = Pattern.Parse(context, "abc123", Pattern.Rule(
				Pattern.And(
					Pattern.Rule(
						Pattern.OneOrMore(
								Pattern.Range('a', 'z'))),
					Pattern.Rule(
						Pattern.ExactQuantity(
							Pattern.Any(),
							4)))));

			Assert.IsType<Null>(result);
			Assert.Equal(1, context.Errors.Length);
			Assert.Equal(0, ((IIndexed)context.Errors[0].Span.Start).Index);
			Assert.Equal(5, ((IIndexed)context.Errors[0].Span.End).Index);

		}

		[Fact]
		public void FailDueToIncompleteReadWithNoRulesReturnsStreamPosition()
		{
			var result = (Nodes)Pattern.Parse(context, "abc123",
				Pattern.And(
					Pattern.OneOrMore(
						Pattern.Range('a', 'z'))));

			var items = (Nodes)result[0];
			Assert.Equal(1, result.Count);
			Assert.Equal('a', Node.Unwrap(items[0]));
			Assert.Equal('b', Node.Unwrap(items[1]));
			Assert.Equal('c', Node.Unwrap(items[2]));
			Assert.Equal(1, context.Errors.Length);
			Assert.Equal(3, ((IIndexed)context.Errors[0].Span.Start).Index);
			Assert.Equal(5, ((IIndexed)context.Errors[0].Span.End).Index);
		}

		[Fact]
		public void ErrorWithMultipleStepsReturnsOriginalPositionCorrectly()
		{
			var R1 = Pattern.OneOrMore(Pattern.Rule(Pattern.Value('a')));
			var R2 = Pattern.And(
				Pattern.Value('a'),
				Pattern.ErrorUnless(Pattern.Value('b')),
				Pattern.Value('a'));

			var result = (Nodes)Pattern.Parse(context, "aa", R1, R2);
			var trace = context.Memos.Trace(result);
			var errors = context.Errors;

			Assert.Equal(3, result.Count);
			Assert.Equal('a', Node.Unwrap(result[0]));
			Assert.Null(Node.Unwrap(result[1]));
			Assert.Equal('a', Node.Unwrap(result[2]));

			Assert.Equal(0, ((IIndexed)trace.Start).Index);
			Assert.Equal(2, ((IIndexed)trace.End).Index);

			Assert.Equal(1, errors.Length);
			Assert.Equal(1, ((IIndexed)errors[0].Span.Start).Index);
			Assert.Equal(2, ((IIndexed)errors[0].Span.End).Index);
		}
	}
}
