﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Xunit;
using NSubstitute;
using System.Reflection;

namespace MetaSharp.Transformation.Tests.Patterns
{
	public class MetadataTests
	{
		private PatternContext context = new PatternContext();

		[Fact]
		public void CanGetMetadata()
		{
			var result = Pattern.Parse(context, "a", R1);
			var metadata = context.Memos.GetMetadata(result);

			Assert.Equal(1, metadata.Length);
			var atts = metadata[0].Metadata;
			Assert.Equal(1, atts.Length);
			Assert.IsType<TestMetadataAttribute>(atts[0]);
		}

		[Fact]
		public void GetsMetadataOnceForRuleMultipleTimes()
		{
			var result = Pattern.Parse(context, "a", R1, R1);
			var metadata = context.Memos.GetMetadata(result);

			Assert.Equal(2, metadata.Length);
			var atts0 = metadata[0].Metadata;
			var atts1 = metadata[1].Metadata;
			Assert.Equal(1, atts0.Length);
			Assert.IsType<TestMetadataAttribute>(atts0[0]);
			Assert.Equal(1, atts1.Length);
			Assert.IsType<TestMetadataAttribute>(atts1[0]);
		}

		[Fact]
		public void CanGetMetadataAcrossMultipleSteps()
		{
			var result = Pattern.Parse(context, "a", R1, R4);
			var metadata = context.Memos.GetMetadata(result);

			Assert.Equal(2, metadata.Length);
			var atts0 = metadata[0].Metadata;
			var atts1 = metadata[1].Metadata;
			Assert.Equal(1, atts0.Length);
			Assert.Equal(1, atts1.Length);
			Assert.IsType<TestMetadata4Attribute>(atts0[0]);
			Assert.IsType<TestMetadataAttribute>(atts1[0]);
		}

		[Fact]
		public void GetsOnlyCorrectMetadata()
		{
			var result = (Nodes)Pattern.Parse(context, "abc", Pattern.And(R1, R2, R3));

			var metadata = context.Memos.GetMetadata(result[0]);
			Assert.Equal(1, metadata.Length);
			var atts = metadata[0].Metadata;
			Assert.Equal(1, atts.Length);
			Assert.IsType<TestMetadataAttribute>(atts[0]);

			metadata = context.Memos.GetMetadata(result[1]);
			Assert.Equal(1, metadata.Length);
			atts = metadata[0].Metadata;
			Assert.Equal(1, atts.Length);
			Assert.IsType<TestMetadata2Attribute>(atts[0]);

			metadata = context.Memos.GetMetadata(result[2]);
			Assert.Equal(1, metadata.Length);
			atts = metadata[0].Metadata;
			Assert.Equal(1, atts.Length);
			Assert.IsType<TestMetadata3Attribute>(atts[0]);
		}

		[Fact]
		public void GetsAllMetadataWhenProjected()
		{
			var result = Pattern.Parse(context, "abc",
				Pattern.Rule(
					Pattern.Projection(
						Pattern.And(
							R1,
							Pattern.Variable("r", R2),
							R3),
						c => c.Variables["r"])));

			var metadata = context.Memos.GetMetadata(result);

			Assert.Equal(3, metadata.Length);
			Assert.Equal(1, metadata[0].Metadata.Length);
			Assert.Equal(1, metadata[1].Metadata.Length);
			Assert.Equal(1, metadata[2].Metadata.Length);
			Assert.IsType<TestMetadataAttribute>(metadata[0].Metadata[0]);
			Assert.IsType<TestMetadata2Attribute>(metadata[1].Metadata[0]);
			Assert.IsType<TestMetadata3Attribute>(metadata[2].Metadata[0]);
		}

		[Fact]
		public void MethodsWithNoMetadataAreNotInResults()
		{
			var result = Pattern.Parse(context, "a", Pattern.Rule(R1));
			var metadata = context.Memos.GetMetadata(result);
			Assert.Equal(1, metadata.Length);
			Assert.Equal(1, metadata[0].Metadata.Length);
			Assert.IsType<TestMetadataAttribute>(metadata[0].Metadata[0]);
		}

		[Fact]
		public void GetsRightMetadataWhenOrd()
		{
			var result = Pattern.Parse(context, "b", Pattern.Or(R1, R2, R3));
			var metadata = context.Memos.GetMetadata(result);

			Assert.Equal(1, metadata.Length);
			Assert.Equal(1, metadata[0].Metadata.Length);
			Assert.IsType<TestMetadata2Attribute>(metadata[0].Metadata[0]);
		}

		[Fact]
		public void GetsRightMetadataDuringDirectLeftRecursion()
		{
			var result = Pattern.Parse(context, "aaac", Pattern.Rule(
				Pattern.Or(
					Pattern.And(RecursionRule, R2),
					Pattern.And(RecursionRule, R3))));

			var metadata = context.Memos.GetMetadata(result);

			Assert.Equal(5, metadata.Length);
			Assert.Equal(1, metadata[0].Metadata.Length);
			Assert.Equal(1, metadata[1].Metadata.Length);
			Assert.Equal(1, metadata[2].Metadata.Length);
			Assert.Equal(1, metadata[3].Metadata.Length);
			Assert.Equal(1, metadata[4].Metadata.Length);
			Assert.IsType<RecursionMetadataAttribute>(metadata[0].Metadata[0]);
			Assert.IsType<TestMetadataAttribute>(metadata[1].Metadata[0]);
			Assert.IsType<TestMetadataAttribute>(metadata[2].Metadata[0]);
			Assert.IsType<TestMetadataAttribute>(metadata[3].Metadata[0]);
			Assert.IsType<TestMetadata3Attribute>(metadata[4].Metadata[0]);
		}

		private static PatternMatch _r1, _r2, _r3, _r4, recursionRule;

		[TestMetadata]
		private static INode R1(PatternContext context, IStream stream)
		{
			if (_r1 == null)
				_r1 = Pattern.Rule(Pattern.Value('a'), typeof(MetadataTests).GetMethod("R1", BindingFlags.Static | BindingFlags.NonPublic));
			return _r1(context, stream);
		}

		[TestMetadata2]
		private static INode R2(PatternContext context, IStream stream)
		{
			if (_r2 == null)
				_r2 = Pattern.Rule(Pattern.Value('b'), typeof(MetadataTests).GetMethod("R2", BindingFlags.Static | BindingFlags.NonPublic));
			return _r2(context, stream);
		}

		[TestMetadata3]
		private static INode R3(PatternContext context, IStream stream)
		{
			if (_r3 == null)
				_r3 = Pattern.Rule(Pattern.Value('c'), typeof(MetadataTests).GetMethod("R3", BindingFlags.Static | BindingFlags.NonPublic));
			return _r3(context, stream);
		}

		[TestMetadata4]
		private static INode R4(PatternContext context, IStream stream)
		{
			if (_r4 == null)
				_r4 = Pattern.Rule(Pattern.Any(), typeof(MetadataTests).GetMethod("R4", BindingFlags.Static | BindingFlags.NonPublic));
			return _r4(context, stream);
		}

		[RecursionMetadata]
		private static INode RecursionRule(PatternContext context, IStream stream)
		{
			if (recursionRule == null)
				recursionRule = Pattern.Rule(
					Pattern.Or(
						Pattern.And(
							RecursionRule,
							R1),
						R1),
					typeof(MetadataTests).GetMethod("RecursionRule", BindingFlags.Static | BindingFlags.NonPublic));
			return recursionRule(context, stream);
		}

	}

	public class TestMetadataAttribute : MetadataAttribute
	{
	}
	public class TestMetadata2Attribute : MetadataAttribute
	{
	}
	public class TestMetadata3Attribute : MetadataAttribute
	{
	}
	public class TestMetadata4Attribute : MetadataAttribute
	{
	}
	public class RecursionMetadataAttribute : MetadataAttribute
	{
	}
}