﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Xunit;
using MetaSharp.Transformation.Properties;
using System.Globalization;
using NSubstitute;

namespace MetaSharp.Transformation.Tests.Patterns
{
	public class ObjectTests
	{
		[Fact]
		public void MatchesObjectSuccessfully()
		{
			var expected = Substitute.For<INode>();
			var actual = Pattern.Transform(expected, Pattern.Object(Pattern.Type(typeof(object))));
			Assert.Same(expected, actual);
		}

		[Fact]
		public void MatchesWrappedTypeSuccessfully()
		{
			var expected = 7;
			var actual = Pattern.Transform(expected, Pattern.Object(Pattern.Type(typeof(int))));
			Assert.Equal(expected, Node.Unwrap(actual));
		}

		[Fact]
		public void MatchesGenericTypeSuccessfully()
		{
			Lazy<int> expected = new Lazy<int>();
			var actual = Pattern.Transform(expected, Pattern.Object(Pattern.Type(typeof(Lazy<int>))));
			Assert.Equal(expected, Node.Unwrap(actual));
		}

		[Fact]
		public void MatchesNullableType()
		{
			int? expected = 7;
			var actual = Pattern.Transform(expected, Pattern.Object(Pattern.Type(typeof(int?))));
			Assert.Equal(expected, Node.Unwrap(actual));
		}

		[Fact]
		public void MatchesInheritanceHierarchySuccessfully()
		{
			var expected = new ArgumentException();
			var actual = Pattern.Transform(expected, Pattern.Object(Pattern.Type(typeof(Exception))));
			Assert.Same(expected, Node.Unwrap(actual));
		}

		[Fact]
		public void MatchesInterfaceInheritanceHierarchySuccessfully()
		{
			var expected = 42;
			var actual = Pattern.Transform(expected, Pattern.Object(Pattern.Type(typeof(IComparable))));
			Assert.Equal(expected, Node.Unwrap(actual));
		}

		[Fact]
		public void StructMatchesBoxedTypeSuccessfully()
		{
			var expected = 42;
			var actual = Pattern.Transform(expected, Pattern.Object(Pattern.Type(typeof(object))));
			Assert.Equal(expected, Node.Unwrap(actual));
		}

		[Fact]
		public void ValueTypeMatchesSuccessfully()
		{
			var expected = 42;
			var actual = Pattern.Transform(expected, Pattern.Object(Pattern.Type(typeof(int))));
			Assert.Equal(expected, Node.Unwrap(actual));
		}

		[Fact]
		public void ObjectPatternFailsOnEmptyStream()
		{
			var result = Pattern.Transform(new object[] { }, Pattern.Object(Pattern.Type(typeof(object))));
			Assert.IsType<Null>(result);
		}

		[Fact]
		public void ObjectPatternFailsOnNull()
		{
			var result = Pattern.Transform(new object[] { null }, Pattern.Object(Pattern.Type(typeof(object))));
			Assert.IsType<Fail>(result);
		}

		[Fact]
		public void IncorrectTypeFailsToMatch()
		{
			var result = Pattern.Transform(new object(), Pattern.Object(Pattern.Type(typeof(char))));
			Assert.IsType<Fail>(result);
		}

		[Fact]
		public void ObjectPatternWithFailedPropertyMatchDoesNotConsumeInput()
		{
			var result = Pattern.Transform(new[] { new { x = 7 } },
				Pattern.ZeroOrOne( // object fails but this will return Null, but Transform will still fail since there is still input.
					Pattern.Object(
						Pattern.Type(typeof(char)),
						Pattern.Property("y"))));

			Assert.IsType<Fail>(result);
		}
	}
}
