﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Xunit;

namespace MetaSharp.Transformation.Tests.Patterns
{
	public class PropertyTests
	{
		[Fact]
		public void MatchesAnObjectAndASingleProperty()
		{
			var obj = Node.Wrap(new { A = 7 });
			var result = Pattern.Transform(obj,
				Pattern.Object(
					Pattern.Type(typeof(object)),
					Pattern.Property("A")));

			Assert.Equal(obj, result);
		}

		[Fact]
		public void RequiresMatchToNotBeNullIfSpecified()
		{
			Assert.Throws<ArgumentNullException>(() => Pattern.Transform(new { A = 7 }, Pattern.Object(Pattern.Type(typeof(object)), Pattern.Property("A", null))));
		}

		[Fact]
		public void MatchesAnObjectPropertyAndValue()
		{
			var obj = Node.Wrap(new { A = 7 });
			var result = (Nodes)Pattern.Transform(obj,
				Pattern.Projection(
					Pattern.Object(
						Pattern.Type(typeof(object)),
						Pattern.Property("A", Pattern.Variable("a", Pattern.Value(7)))),
					ctx => new Nodes(ctx.Match, ctx.Variables["a"])));

			Assert.Equal(obj, result[0]);
			Assert.Equal(7, Node.Unwrap(result[1]));
		}

		[Fact]
		public void MatchesAnObjectTwoPropertiesAndTheirValues()
		{
			var obj = Node.Wrap(new { A = 7, B = 11 });
			var result = (Nodes)Pattern.Transform(obj,
				Pattern.Projection(
					Pattern.Object(
						Pattern.Type(typeof(object)),
						Pattern.Property("A", Pattern.Variable("l", Pattern.Value(7))),
						Pattern.Property("B", Pattern.Variable("r", Pattern.Value(11)))),
					ctx => new Nodes(ctx.Match, ctx.Variables["l"], ctx.Variables["r"])));

			Assert.Equal(obj, result[0]);
			Assert.Equal(7, Node.Unwrap(result[1]));
			Assert.Equal(11, Node.Unwrap(result[2]));
		}

		[Fact]
		public void MatchesAnObjectAndOneOfTwoProperties()
		{
			var obj = Node.Wrap(new { A = 7, B = 11 });
			var result = (Nodes)Pattern.Transform(obj,
				Pattern.Projection(
					Pattern.Object(
						Pattern.Type(typeof(object)),
						Pattern.Property("B", Pattern.Variable("b", Pattern.Any()))),
					ctx => new Nodes(ctx.Match, ctx.Variables["b"])));

			Assert.Equal(obj, result[0]);
			Assert.Equal(11, Node.Unwrap(result[1]));
		}

		[Fact]
		public void MatchesNestedObjectAndProperty()
		{
			var b = Node.Wrap(new { B = 7 });
			var a = Node.Wrap(new { A = b });
			var result = Pattern.Transform(a,
				Pattern.Object(
					Pattern.Type(typeof(object)),
					Pattern.Property("A",
						Pattern.Object(
							Pattern.Type(typeof(object)),
							Pattern.Property("B")))));

			Assert.Equal(a, result);
		}

		[Fact]
		public void MatchesNestedArray()
		{
			var obj = Node.Wrap(new { A = new[] { 7, 11 } });
			var result = (Nodes)Pattern.Transform(obj,
				Pattern.Projection(
					Pattern.Object(
						Pattern.Type(typeof(object)),
						Pattern.Property("A",
							Pattern.Array(Pattern.And(
								Pattern.Variable("l", Pattern.Value(7)),
								Pattern.Variable("r", Pattern.Value(11)))))),
					ctx => new Nodes(ctx.Match, ctx.Variables["l"], ctx.Variables["r"])));

			Assert.Equal(obj, result[0]);
			Assert.Equal(7, Node.Unwrap(result[1]));
			Assert.Equal(11, Node.Unwrap(result[2]));
		}

		[Fact]
		public void MatchesMultipleObjects()
		{
			var obj = new object[] { new { A = 7 }, new { B = 11 }, 23 };
			var result = (Nodes)Pattern.Transform(obj, Pattern.And(
				Pattern.Projection(
					Pattern.Object(Pattern.Type(typeof(object)), Pattern.Property("A", Pattern.Variable("a", Pattern.Any()))),
					ctx => ctx.Variables["a"]),
				Pattern.Projection(
					Pattern.Object(Pattern.Type(typeof(object)), Pattern.Property("B", Pattern.Variable("b", Pattern.Any()))),
					ctx => ctx.Variables["b"]),
				Pattern.Value(23)));

			Assert.Equal(7, Node.Unwrap(result[0]));
			Assert.Equal(11, Node.Unwrap(result[1]));
			Assert.Equal(23, Node.Unwrap(result[2]));
		}

		[Fact]
		public void NestedObjectMatchIsMemoized()
		{
			// The inner most property match is done as a rule, which will cause memoization at that position. Subsequent calls
			// to that inner property will be not be called but instead memoized.
			var count = 0;
			var r = Pattern.Rule(Pattern.Projection(Pattern.Value(7), s => { count++; return s.Match; }));
			var n = Pattern.Projection(
				Pattern.Object(
				Pattern.Type(typeof(object)),
				Pattern.Property("A",
					Pattern.Object(
						Pattern.Type(typeof(object)),
						Pattern.Property("B", Pattern.Variable("v", r))))),
				ctx => ctx.Variables["v"]);

			var obj = new object[] { new { A = new { B = 7 } }, 11 };
			var result = (Nodes)Pattern.Transform(obj, Pattern.Or(Pattern.And(n, Pattern.Value(23)), Pattern.And(n, Pattern.Value(11))));
			Assert.Equal(1, count);
			Assert.Equal(7, Node.Unwrap(result[0]));
			Assert.Equal(11, Node.Unwrap(result[1]));
		}
	}
}
