﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Xunit;

namespace MetaSharp.Transformation.Tests.Patterns
{
	public class RuleTests
	{
		[Fact]
		public void ParsesSingleRuleSuccessfully()
		{
			var rule = Pattern.Rule(Pattern.Value('a'));
			var result = Pattern.Parse("a", rule);
			Assert.Equal('a', Node.Unwrap(result));
		}

		[Fact]
		public void RecursiveArrayToTreeRule()
		{
			// Tree = Tree any | any;
			var tree = default(PatternMatch);
			tree = Pattern.Rule(
				 Pattern.Or(
					Pattern.And(
						(c, s) => tree(c, s),
						Pattern.Any()),
					 Pattern.Any()));

			var result = (Nodes)Pattern.Transform(new[] { 1, 2, 3 }, tree);
			Assert.Equal(3, Node.Unwrap(result[1]));
			result = (Nodes)result[0];
			Assert.Equal(2, Node.Unwrap(result[1]));
			Assert.Equal(1, Node.Unwrap(result[0]));
		}

		[Fact]
		public void ParsesRuleExactlyOnce()
		{
			int projectionCount = 0;
			var aAtIndex0 = Pattern.Rule(Pattern.Projection(Pattern.Value('a'), s => { projectionCount++; return s.Match; }));

			var result = (Nodes)Pattern.Parse(
				"ac",
				Pattern.Or(
					Pattern.And(aAtIndex0, Pattern.Value('b')),
					Pattern.And(aAtIndex0, Pattern.Value('c'))));

			Assert.Equal(1, projectionCount);
			Assert.Equal('a', Node.Unwrap(result[0]));
			Assert.Equal('c', Node.Unwrap(result[1]));
		}

		[Fact]
		public void MatchInsideArrayIsCalledExactlyOnce()
		{
			int projectionCount = 0;
			var two = Pattern.Rule(Pattern.Projection(Pattern.Value(2), s => { projectionCount++; return s.Match; }));

			var result = (Nodes)Pattern.Transform(
				new object[] { 1, new[] { 2 }, 3 },
				Pattern.Or(
					Pattern.And(Pattern.Value(1), Pattern.Array(two), Pattern.Value(0)),
					Pattern.And(Pattern.Value(1), Pattern.Array(two), Pattern.Value(3))));

			Assert.Equal(1, projectionCount);
			Assert.Equal(1, Node.Unwrap(result[0]));
			Assert.Equal(2, Node.Unwrap(result[1]));
			Assert.Equal(3, Node.Unwrap(result[2]));
		}

		[Fact]
		public void DirectLeftRecursionIsSuccessful()
		{
			var result = (Nodes)Pattern.Parse("1+2+3", DlrTest.exp);
			Assert.Equal('1', Node.Unwrap(((Nodes)result[0])[0]));
			Assert.Equal('+', Node.Unwrap(((Nodes)result[0])[1]));
			Assert.Equal('2', Node.Unwrap(((Nodes)result[0])[2]));
			Assert.Equal('+', Node.Unwrap(result[1]));
			Assert.Equal('3', Node.Unwrap(result[2]));
		}

		class DlrTest
		{
			//num = 0..9;
			//exp = exp '+' num | num;
			private static readonly PatternMatch num = Pattern.Rule(Pattern.Range('0', '9'));
			public static readonly PatternMatch exp = Pattern.Rule(
				Pattern.Or(
					Pattern.And((c, s) => exp(c, s), Pattern.Value('+'), num),
					num));
		}

		[Fact]
		public void IndirectLeftRecursionFails()
		{
			Assert.Throws<TransformationException>(() => Pattern.Parse("111", IlrTest.exp));
		}

		class IlrTest
		{
			//num = exp;
			//exp = num;
			private static readonly PatternMatch num = Pattern.Rule((c, s) => exp(c, s));
			public static readonly PatternMatch exp = Pattern.Rule((c, s) => num(c, s));
		}

		[Fact]
		public void CanOverrideRuleReference()
		{
			var super = new OverrideTestSuper();
			var result = Pattern.Transform(11, super.Test);
			Assert.Equal(11, Node.Unwrap(result));

			var sub = new OverrideTestSub();
			result = Pattern.Transform(11, sub.Test);
			Assert.IsType<Fail>(result);

			super = sub;
			result = Pattern.Transform(11, super.Test);
			Assert.IsType<Fail>(result);
		}

		class OverrideTestSuper
		{
			private PatternMatch test;

			public OverrideTestSuper()
			{
				this.test = Pattern.Rule(Pattern.Any());
			}

			public virtual PatternMatch Test
			{
				get { return this.test; }
			}
		}

		class OverrideTestSub : OverrideTestSuper
		{
			private PatternMatch test;
			public OverrideTestSub()
			{
				this.test = Pattern.Rule(Pattern.Value(7));
			}

			public override PatternMatch Test
			{
				get { return this.test; }
			}
		}
	}
}
