﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Xunit;
using NSubstitute;

namespace MetaSharp.Transformation.Tests.Patterns
{
	public class TraceTests
	{
		private PatternContext context = new PatternContext();

		[Fact]
		public void CanTraceSingleItem()
		{
			var rule = Pattern.Rule(Pattern.Value('a'));
			var result = Pattern.Parse(context, "a", rule);
			var trace = context.Memos.Trace(result);
			Assert.Equal('a', Node.Unwrap(result));
			Assert.Equal(0, ((IIndexed)trace.Start).Index);
			Assert.Equal(1, ((IIndexed)trace.End).Index);
		}

		[Fact]
		public void CanTraceLines()
		{
			var rule = Pattern.Rule(Pattern.OneOrMore(Pattern.Rule(Pattern.Any())));
			var result = (Nodes)Pattern.Parse(context, "a\nb\nc", rule);
			var trace = context.Memos.Trace(result);

			Assert.Equal(0, ((CharacterPosition)trace.Start).Line);
			Assert.Equal(2, ((CharacterPosition)trace.End).Line);

			trace = context.Memos.Trace(result[0]);
			Assert.Equal(0, ((CharacterPosition)trace.Start).Line);
			Assert.Equal(0, ((CharacterPosition)trace.End).Line);

			trace = context.Memos.Trace(result[2]);
			Assert.Equal(1, ((CharacterPosition)trace.Start).Line);
			Assert.Equal(1, ((CharacterPosition)trace.End).Line);

			trace = context.Memos.Trace(result[4]);
			Assert.Equal(2, ((CharacterPosition)trace.Start).Line);
			Assert.Equal(2, ((CharacterPosition)trace.End).Line);
		}

		[Fact]
		public void CanTraceColumnsAcrossLines()
		{
			var rule = Pattern.Rule(Pattern.OneOrMore(Pattern.Rule(Pattern.Any())));
			var result = (Nodes)Pattern.Parse(context, "a\nbc\nxyz", rule);

			var trace = context.Memos.Trace(result[0]);
			Assert.Equal(0, ((CharacterPosition)trace.Start).Line);
			Assert.Equal(0, ((CharacterPosition)trace.End).Line);
			Assert.Equal(0, ((CharacterPosition)trace.Start).Column);
			Assert.Equal(1, ((CharacterPosition)trace.End).Column);

			trace = context.Memos.Trace(result[2]);
			Assert.Equal(1, ((CharacterPosition)trace.Start).Line);
			Assert.Equal(1, ((CharacterPosition)trace.End).Line);
			Assert.Equal(0, ((CharacterPosition)trace.Start).Column);
			Assert.Equal(1, ((CharacterPosition)trace.End).Column);

			trace = context.Memos.Trace(result[3]);
			Assert.Equal(1, ((CharacterPosition)trace.Start).Line);
			Assert.Equal(1, ((CharacterPosition)trace.End).Line);
			Assert.Equal(1, ((CharacterPosition)trace.Start).Column);
			Assert.Equal(2, ((CharacterPosition)trace.End).Column);

			trace = context.Memos.Trace(result[5]);
			Assert.Equal(2, ((CharacterPosition)trace.Start).Line);
			Assert.Equal(2, ((CharacterPosition)trace.End).Line);
			Assert.Equal(0, ((CharacterPosition)trace.Start).Column);
			Assert.Equal(1, ((CharacterPosition)trace.End).Column);

			trace = context.Memos.Trace(result[6]);
			Assert.Equal(2, ((CharacterPosition)trace.Start).Line);
			Assert.Equal(2, ((CharacterPosition)trace.End).Line);
			Assert.Equal(1, ((CharacterPosition)trace.Start).Column);
			Assert.Equal(2, ((CharacterPosition)trace.End).Column);

			trace = context.Memos.Trace(result[7]);
			Assert.Equal(2, ((CharacterPosition)trace.Start).Line);
			Assert.Equal(2, ((CharacterPosition)trace.End).Line);
			Assert.Equal(2, ((CharacterPosition)trace.Start).Column);
			Assert.Equal(3, ((CharacterPosition)trace.End).Column);
		}


		[Fact]
		public void CanTraceProjection()
		{
			var expected = Substitute.For<INode>();
			var rule = Pattern.Rule(Pattern.Projection(Pattern.Value('a'), s => expected));
			var result = Pattern.Parse(context, "a", rule);
			var trace = context.Memos.Trace(result);
			Assert.Same(expected, result);
			Assert.Equal(0, ((IIndexed)trace.Start).Index);
			Assert.Equal(1, ((IIndexed)trace.End).Index);
		}

		[Fact]
		public void TwoItemsTraceSuccessfully()
		{
			var rule = Pattern.Rule(Pattern.And(Pattern.Any(), Pattern.Any()));
			var result = (Nodes)Pattern.Parse(context, "ab", rule);
			var trace = context.Memos.Trace(result);
			Assert.Equal('a', Node.Unwrap(result[0]));
			Assert.Equal('b', Node.Unwrap(result[1]));
			Assert.Equal(0, ((IIndexed)trace.Start).Index);
			Assert.Equal(2, ((IIndexed)trace.End).Index);
		}

		[Fact]
		public void NestedRulesTraceSuccessfully()
		{
			var a = Pattern.Rule(Pattern.Value('a'));
			var b = Pattern.Rule(Pattern.Value('b'));
			var rule = Pattern.Rule(Pattern.And(a, b));
			var result = (Nodes)Pattern.Parse(context, "ab", rule);
			Assert.Equal('a', Node.Unwrap(result[0]));
			Assert.Equal('b', Node.Unwrap(result[1]));

			var traceResult = context.Memos.Trace(result);
			Assert.Equal(0, ((CharacterPosition)traceResult.Start).Index);
			Assert.Equal(2, ((CharacterPosition)traceResult.End).Index);

			var traceA = context.Memos.Trace(result[0]);
			Assert.Equal(0, ((CharacterPosition)traceA.Start).Index);
			Assert.Equal(1, ((CharacterPosition)traceA.End).Index);

			var traceB = context.Memos.Trace(result[1]);
			Assert.Equal(1, ((CharacterPosition)traceB.Start).Index);
			Assert.Equal(2, ((CharacterPosition)traceB.End).Index);
		}

		[Fact]
		public void TracesSingleAcrossMultipleSteps()
		{
			var rule1 = Pattern.Rule(Pattern.Any());
			var rule2 = Pattern.Rule(Pattern.Any());

			var result = Pattern.Parse(context, "a", rule1, rule2);
			var trace = context.Memos.Trace(result);
			Assert.Equal('a', Node.Unwrap(result));
			Assert.Equal(0, ((IIndexed)trace.Start).Index);
			Assert.Equal(1, ((IIndexed)trace.End).Index);
		}

		[Fact]
		public void TracesMultipleAcrossMultipleSteps()
		{
			var a = Pattern.Rule(Pattern.Any());
			var b = Pattern.Rule(Pattern.Any());
			var c = Pattern.Rule(Pattern.Any());
			var d = Pattern.Rule(Pattern.Any());
			var rule1 = Pattern.Rule(Pattern.And(a, b));
			var rule2 = Pattern.Rule(Pattern.And(c, d));

			var result = (Nodes)Pattern.Parse(context, "ab", rule1, rule2);
			var traceResult = context.Memos.Trace(result);
			Assert.Equal(0, ((IIndexed)traceResult.Start).Index);
			Assert.Equal(2, ((IIndexed)traceResult.End).Index);

			var traceA = context.Memos.Trace(result[0]);
			Assert.Equal(0, ((IIndexed)traceA.Start).Index);
			Assert.Equal(1, ((IIndexed)traceA.End).Index);

			var traceB = context.Memos.Trace(result[1]);
			Assert.Equal(1, ((IIndexed)traceB.Start).Index);
			Assert.Equal(2, ((IIndexed)traceB.End).Index);
		}

		[Fact]
		public void CanTraceProjectedArrays()
		{
			var expected = Substitute.For<INode>();
			var r1 = Pattern.Rule(
				Pattern.Projection(
					Pattern.And(
						Pattern.Value('a'),
						Pattern.OneOrMore(Pattern.Value('b')),
						Pattern.Value('c')),
					s => expected));

			var result = Pattern.Parse(context, "abbbbbc", r1);
			var trace = context.Memos.Trace(result);
			Assert.Same(expected, result);
			Assert.Equal(0, ((IIndexed)trace.Start).Index);
			Assert.Equal(7, ((IIndexed)trace.End).Index);
		}

		[Fact]
		public void CanTraceAcrossComplexObjects()
		{
			var obj1 = Substitute.For<INode>();
			var obj2 = Node.Wrap(new { Test = obj1 });

			var r2 = Pattern.Rule(Pattern.Any());
			var r1 = Pattern.Rule(
				Pattern.Projection(
					Pattern.Object(
						Pattern.Type(typeof(object)),
						Pattern.Property("Test", Pattern.Variable("t", r2))),
					ctx => new Nodes(ctx.Match, ctx.Variables["t"])));

			var result = (Nodes)Pattern.Transform(context, new NodeStream(obj2), r1);
			var trace = context.Memos.Trace(result);
			Assert.Same(obj2, result[0]);
			Assert.Same(obj1, result[1]);
			Assert.Equal(0, ((IIndexed)trace.Start).Index);
			Assert.Equal(1, ((IIndexed)trace.End).Index);

			var traceObj1 = context.Memos.Trace(result[1]);
			Assert.Equal("Test", ((PropertyPosition)traceObj1.Start).Property.Name);
			Assert.Same(obj1, ((ValuePosition)traceObj1.End).Current);
		}

		[Fact]
		public void CanTraceAcross3Transforms()
		{
			var obj1 = Substitute.For<INode>();
			var obj2 = Node.Wrap(new { Test = obj1 });

			var identifier = Pattern.Rule(
				Pattern.Projection(
					Pattern.OneOrMore(Pattern.Range('a', 'z')),
					s => Node.Wrap(new string(s.Match.Cast<Node>().Select(ch => (char)ch.Value).ToArray()))));

			var space = Pattern.Rule(
				Pattern.Projection(
					Pattern.Value(' '),
					s => Node.Wrap(" ")));

			var r1 = Pattern.Rule(
					Pattern.OneOrMore(
						Pattern.Or(
							identifier,
							space)));

			var r2 = Pattern.Rule(
				Pattern.Projection(
					Pattern.OneOrMore(
						Pattern.Or(
							Pattern.Projection(Pattern.Value(" "), s => null),
							Pattern.Rule(Pattern.Any()))),
					s => ((Nodes)s.Match).Where(o => Node.Unwrap(o) != null).ToNodes()));

			int count = 0;
			var foo1 = Substitute.For<INode>();
			var bar = Substitute.For<INode>();
			var foo2 = Substitute.For<INode>();
			var keywords = Pattern.Rule(
					Pattern.Or(
					Pattern.Projection(Pattern.Value("foo"), s => count++ == 0 ? foo1 : foo2),
						Pattern.Projection(Pattern.Value("bar"), s => bar)));

			var r3 = Pattern.Rule(
				Pattern.OneOrMore(
					keywords));

			var result = (Nodes)Pattern.Parse(context, "foo bar foo", r1, r2, r3);
			Assert.Equal(2, count);
			Assert.Same(foo1, result[0]);
			Assert.Same(bar, result[1]);
			Assert.Same(foo2, result[2]);

			var trace = context.Memos.Trace(result);
			Assert.Equal(0, ((IIndexed)trace.Start).Index);
			Assert.Equal(11, ((IIndexed)trace.End).Index);

			var trace0 = context.Memos.Trace(result[0]);
			Assert.Equal(0, ((IIndexed)trace0.Start).Index);
			Assert.Equal(3, ((IIndexed)trace0.End).Index);

			var trace1 = context.Memos.Trace(result[1]);
			Assert.Equal(4, ((IIndexed)trace1.Start).Index);
			Assert.Equal(7, ((IIndexed)trace1.End).Index);

			var trace2 = context.Memos.Trace(result[2]);
			Assert.Equal(8, ((IIndexed)trace2.Start).Index);
			Assert.Equal(11, ((IIndexed)trace2.End).Index);
		}
	}
}
