﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Xunit;
using NSubstitute;

namespace MetaSharp.Transformation.Tests.Patterns
{
	public class ValueTests
	{
		private PatternContext context = new PatternContext();

		[Fact]
		public void CanReadChar()
		{
			var expected = 'a';
			var actual = Pattern.Parse("a", Pattern.Value(expected));
			Assert.Equal(expected, Node.Unwrap(actual));
		}

		[Fact]
		public void CanReadInt16()
		{
			short expected = 16;
			var actual = Pattern.Transform(expected, Pattern.Value(expected));
			Assert.Equal(expected, Node.Unwrap(actual));
		}
		[Fact]
		public void CanReadInt32()
		{
			int expected = 32;
			var actual = Pattern.Transform(expected, Pattern.Value(expected));
			Assert.Equal(expected, Node.Unwrap(actual));
		}

		[Fact]
		public void CanReadInt64()
		{
			var expected = 64;
			var actual = Pattern.Transform(expected, Pattern.Value(expected));
			Assert.Equal(expected, Node.Unwrap(actual));
		}

		[Fact]
		public void CanReadFloat()
		{
			float expected = 7.11F;
			var actual = Pattern.Transform(expected, Pattern.Value(expected));
			Assert.Equal(expected, Node.Unwrap(actual));
		}

		[Fact]
		public void CanReadDouble()
		{
			var expected = 7.11;
			var actual = Pattern.Transform(expected, Pattern.Value(expected));
			Assert.Equal(expected, Node.Unwrap(actual));
		}

		[Fact]
		public void CanReadDecimal()
		{
			decimal expected = 7.11M;
			var actual = Pattern.Transform(expected, Pattern.Value(expected));
			Assert.Equal(expected, Node.Unwrap(actual));
		}

		[Fact]
		public void CanReadBool()
		{
			bool expected = true;
			var actual = Pattern.Transform(expected, Pattern.Value(expected));
			Assert.Equal(expected, Node.Unwrap(actual));
		}

		[Fact]
		public void CanReadByte()
		{
			byte expected = 8;
			var actual = Pattern.Transform(expected, Pattern.Value(expected));
			Assert.Equal(expected, Node.Unwrap(actual));
		}

		enum TestEnum1 { Success, Fail }
		[Fact]
		public void CanReadEnum()
		{
			var expected = TestEnum1.Success;
			var actual = Pattern.Transform(expected, Pattern.Value(expected));
			Assert.Equal(expected, Node.Unwrap(actual));
		}

		[Fact]
		public void CanReadGuid()
		{
			var expected = Guid.NewGuid();
			var actual = Pattern.Transform(expected, Pattern.Value(expected));
			Assert.Equal(expected, Node.Unwrap(actual));
		}

		[Fact]
		public void CanReadDateTime()
		{
			var expected = DateTime.Now;
			var actual = Pattern.Transform(expected, Pattern.Value(expected));
			Assert.Equal(expected, Node.Unwrap(actual));
		}

		[Fact]
		public void CanReadObject()
		{
			object expected = new object();
			var actual = Pattern.Transform(expected, Pattern.Value(expected));
			Assert.Equal(expected, Node.Unwrap(actual));
		}

		class TestClass1 : INode
		{
			public int Value { get; set; }
			public override bool Equals(object obj)
			{
				return (obj is TestClass1) && (this.Value == ((TestClass1)obj).Value);
			}

			public override int GetHashCode() { return this.Value; }
		}

		[Fact]
		public void CanReadCustom()
		{
			TestClass1 expected = new TestClass1 { Value = 7 };
			var actual = Pattern.Transform(expected, Pattern.Value(expected));
			Assert.Equal(expected, actual);
		}

		[Fact]
		public void CanReadEmptyArray()
		{
			var expected = new object[] { };
			var actual = Pattern.Transform(new[] { expected }, Pattern.Value(expected));
			Assert.Equal(expected, Node.Unwrap(actual));
		}

		[Fact]
		public void CanReadNonEmptyArray()
		{
			var expected = new object[] { 7, 11.13, "foo" };
			var actual = Pattern.Transform(new[] { expected }, Pattern.Value(expected));
			Assert.Equal(expected, Node.Unwrap(actual));
		}

		[Fact]
		public void CanReadBoxedValue()
		{
			object expected = 7.11;
			var actual = Pattern.Transform(expected, Pattern.Value(7.11));
			Assert.Equal(expected, Node.Unwrap(actual));
		}

		[Fact]
		public void CanReadNull()
		{
			object expected = null;
			var actual = Pattern.Transform(expected, Pattern.Value(expected));
			Assert.Equal(expected, Node.Unwrap(actual));
		}

		[Fact]
		public void CanReadWrappedNull()
		{
			var expected = Node.Wrap(null);
			var actual = Pattern.Transform(expected, Pattern.Value(null));
			Assert.Same(expected, actual);
		}


		[Fact]
		public void CanReadNodeSuccessfully()
		{
			INode expected = Substitute.For<INode>();
			var actual = Pattern.Transform(expected, Pattern.Value(expected));
			Assert.Equal(expected, actual);
		}

		[Fact]
		public void ValueIsNullFails()
		{
			object expected = new object();
			var actual = Pattern.Transform(expected, Pattern.Value(null));
			Assert.IsType<Fail>(actual);
		}

		[Fact]
		public void InputIsNullFails()
		{
			object expected = null;
			var actual = Pattern.Transform(context, expected, Pattern.Value(7.11));
			Assert.IsType<Null>(actual);
			Assert.Equal(1, context.Errors.Length);
		}

		[Fact]
		public void ValuesDontMatchFails()
		{
			var expected = 7;
			var actual = Pattern.Transform(expected, Pattern.Value(11));
			Assert.IsType<Fail>(actual);
		}

		[Fact]
		public void ValueTypesDontMatchFails()
		{
			var expected = 7;
			var actual = Pattern.Transform(expected, Pattern.Value(new object()));
			Assert.IsType<Fail>(actual);
		}
	}
}
