﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Xunit;

namespace MetaSharp.Transformation.Tests.Patterns
{
	public class VariablePatterns
	{
		[Fact]
		public void CanDeclareVariablePatternAndProject()
		{
			var expected = 7;
			var actual = Pattern.Transform(
				expected,
				Pattern.Projection(
					Pattern.Variable("x", Pattern.Value(7)),
					s => s.Variables["x"]));

			Assert.Equal(expected, Node.Unwrap(actual));
		}

		[Fact]
		public void CanNotAccessNonDeclaredVariablesInProjection()
		{
			Assert.Throws<KeyNotFoundException>(() =>
				Pattern.Transform(
				new[] { 7, 11 },
				Pattern.Projection(
					Pattern.And(
						Pattern.Variable("x", Pattern.Value(7)),
						Pattern.Value(11)),
					s => s.Variables["y"])));
		}

		[Fact]
		public void CanAccessVariablesDeclaredInRule()
		{
			var expected = new object[] { 7, 11 };

			var rule = Pattern.Rule(Pattern.Projection(Pattern.Variable("x", Pattern.Any()), s => s.Variables["x"]));
			var actual = (Nodes)Pattern.Transform(
				expected,
				Pattern.Projection(
					Pattern.And(
						Pattern.Variable("x", rule),
						Pattern.Variable("y", rule)),
					s => new Nodes(s.Variables["x"], s.Variables["y"])));

			Assert.Equal(expected[0], Node.Unwrap(actual[0]));
			Assert.Equal(expected[1], Node.Unwrap(actual[1]));
		}

		[Fact]
		public void CanAccessVariablesDeclaredInHigherProjection()
		{
			var expected = 7;
			var actual = Pattern.Transform(
					expected,
					Pattern.Projection(
						Pattern.And(
							Pattern.Variable("x", Pattern.Any()),
							Pattern.Projection(
								Pattern.Switch("x", Pattern.Any()),
								s => s.Match)),
						s => s.Variables["x"]));

			Assert.Equal(expected, Node.Unwrap(actual));
		}

		[Fact]
		public void CanOverwriteVariablesDeclaredInHigherProjection()
		{
			var expected = new[] { 7, 11 };
			var actual = (Nodes)Pattern.Transform(
					expected,
					Pattern.Projection(
						Pattern.And(
							Pattern.Variable("x", Pattern.Any()),
							Pattern.Projection(
								Pattern.And(
									Pattern.Variable("x", Pattern.Any()),
									Pattern.Switch("x", Pattern.Any())),
								s => s.Variables["x"])),
						s => s.Match));

			Assert.Equal(expected[0], Node.Unwrap(actual[0]));
			Assert.Equal(expected[1], Node.Unwrap(actual[1]));
		}

		[Fact]
		public void CannotAccessVariablesDeclaredInLowerProjection()
		{
			Assert.Throws<KeyNotFoundException>(() =>
				Pattern.Transform(
					new[] { 7, 11 },
					Pattern.Projection(
						Pattern.And(
							Pattern.Any(),
							Pattern.Projection(
								Pattern.Switch("x", Pattern.Any()),
								s => null)),
						s => s.Variables["x"])));
		}

		[Fact]
		public void CanSwitchToVariable()
		{
			var expected = 7;

			// rule = int { }
			// main = x:any rule(x) -> x
			var rule = Pattern.Rule(Pattern.Object(Pattern.Type(typeof(int))));
			var actual = Pattern.Transform(
				expected,
				Pattern.Projection(
					Pattern.And(
						Pattern.Variable("x", Pattern.Any()),
						Pattern.Switch("x", rule)),
					s => s.Variables["x"]));

			Assert.Equal(expected, Node.Unwrap(actual));
		}

		[Fact]
		public void CannotSwitchToWrongVariable()
		{
			// rule = int { }
			// main = x:any rule(y)
			var rule = Pattern.Rule(Pattern.Object(Pattern.Type(typeof(int))));
			Assert.Throws<KeyNotFoundException>(() =>
				Pattern.Transform(
					7,
					Pattern.And(
						Pattern.Variable("x", Pattern.Any()),
						Pattern.Switch("y", rule))));
		}

		[Fact]
		public void CanAssignObjectAndPropertyToDifferentVariables()
		{
			var rule = Pattern.Projection(
				Pattern.Object(
					Pattern.Variable("o", Pattern.Type(typeof(object))),
					Pattern.Property("x", Pattern.Variable("x", Pattern.Any()))),
				s => new Nodes(s.Variables["o"], s.Variables["x"]));


			var expected = new { x = 7 };
			var actual = (Nodes)Pattern.Transform(
				expected,
				rule);

			Assert.Same(expected, Node.Unwrap(actual[0]));
			Assert.Equal(7, Node.Unwrap(actual[1]));
		}

		[Fact]
		public void DirectLeftRecursionWithVerificationSwitchIsSuccessful()
		{
			var result = (expr)Pattern.Parse("x.y.z", DlrTest2.A);
			Assert.Equal("z", result.Name);
			Assert.Equal("y", result.Target.Name);
			Assert.Equal("x", result.Target.Target.Name);
			Assert.Null(result.Target.Target.Target);
		}

		[Fact]
		public void DirectLeftRecursionWithVerificationSwitchCorrectlyFails()
		{
			var result = Pattern.Parse("x.y.x", DlrTest2.A);
			Assert.IsType<Fail>(result);
		}

		[Fact]
		public void CallParameterizedPattern()
		{
			var and = Pattern.Closure(p => Pattern.And(p[0], p[1]));
			var result = (Nodes)Pattern.Parse("ab", Pattern.Call(and, Pattern.Value('a'), Pattern.Value('b')));
			Assert.Equal('a', Node.Unwrap(result[0]));
			Assert.Equal('b', Node.Unwrap(result[1]));
		}

		[Fact]
		public void DirectLeftRecursionOnCallablePatternIsSuccessful()
		{
			var result = (expr)Pattern.Parse("x.y.z", DlrTest2.A);
			Assert.Equal("z", result.Name);
			Assert.Equal("y", result.Target.Name);
			Assert.Equal("x", result.Target.Target.Name);
			Assert.Null(result.Target.Target.Target);
		}

		class expr : INode
		{
			public string Name { get; set; }
			public expr Target { get; set; }
		}

		class DlrTest2
		{
			// A = b:B 'z'(b) -> b; //switch variable b
			// B = b:B '.' c:C -> { c.Target = b; return c; }
			//   | C
			// C = c:'a'..'z' -> new expr { Name = c.ToString() };
			public static PatternMatch _C;
			public static PatternMatch _B;
			public static PatternMatch _A;

			public static INode A(PatternContext context, IStream stream)
			{
				if (_A == null)
					_A = Pattern.Rule(
						Pattern.Projection(
							Pattern.And(
								Pattern.Variable("b", B),
								Pattern.Switch(
									"b",
									Pattern.Object(
										Pattern.Type(typeof(expr)),
										Pattern.Property(
											"Name",
											Pattern.Value("z"))))),
							s => s.Variables["b"]));

				return _A(context, stream);
			}

			public static INode B(PatternContext context, IStream stream)
			{
				if (_B == null)
					_B = Pattern.Rule(
						Pattern.Or(
							Pattern.Projection(
								Pattern.And(
									Pattern.Variable("b", B),
									Pattern.Value('.'),
									Pattern.Variable("c", C)),
								s =>
								{
									var l = (expr)s.Variables["b"];
									var r = (expr)s.Variables["c"];
									r.Target = l;
									return r;
								}),
							C));

				return _B(context, stream);
			}

			public static INode C(PatternContext context, IStream stream)
			{
				if (_C == null)
					_C = Pattern.Rule(
						Pattern.Projection(
							Pattern.Variable("n", Pattern.Range('a', 'z')),
							s => new expr { Name = Node.Unwrap(s.Variables["n"]).ToString() }));

				return _C(context, stream);
			}
		}

	}
}
