﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Collections;
using System.CodeDom;
using System.CodeDom.Compiler;
using System.IO;
using Microsoft.CSharp;
using MetaSharp.Transformation.Lang.Ast;
using System.Reflection;

namespace MetaSharp.Transformation
{
	public static class Extensions
	{
		public static IEnumerable<T> Cast<T>(this object value)
		{
			if (Node.Unwrap(value) == null)
				return Enumerable.Empty<T>();

			var enumerableOfTValue = value as IEnumerable<T>;
			if (enumerableOfTValue != null)
				return enumerableOfTValue;

			var enumerableValue = value as IEnumerable;
			if (enumerableValue != null)
				return Enumerable.Cast<T>(enumerableValue).Where(i => i != null);

			return Enumerable.Cast<T>(new[] { value });
		}

		public static IEnumerable<T> Unwrap<T>(this IEnumerable<INode> items)
		{
			return items.Select(i => (T)Node.Unwrap(i));
		}

		public static INode ToNodes(this IEnumerable<INode> items)
		{
			return new Nodes(items);
		}

		public static IEnumerable<INode> Ancestors(this Position position)
		{
			var pos = position;
			while (pos != null)
			{
				yield return pos.Current;
				pos = pos.Parent;
			}
		}
	}

	public static class CodeDomExtensions
	{
		public static string Print(this CodeObject codeObject, CodeDomProvider provider = null, CodeGeneratorOptions options = null)
		{
			if (codeObject == null)
				throw new ArgumentNullException("codeObject");
			if (provider == null)
				provider = new CSharpCodeProvider();
			if (options == null)
				options = new CodeGeneratorOptions
				{
					BlankLinesBetweenMembers = false,
					BracingStyle = "C",
					ElseOnClosing = false,
					IndentString = "\t",
					VerbatimOrder = true,
				};

			var builder = new StringBuilder();
			using (var writer = new StringWriter(builder))
			{
				CodeCompileUnit codeUnit = null;
				CodeNamespace codeNamespace = null;
				CodeTypeDeclaration codeTypeDeclaration = null;
				CodeTypeMember codeTypeMember = null;
				CodeStatement codeStatement = null;
				CodeExpression codeExpression = null;
				if ((codeUnit = codeObject as CodeCompileUnit) != null)
				{
					provider.GenerateCodeFromCompileUnit(codeUnit, writer, options);
				}
				else if ((codeNamespace = codeObject as CodeNamespace) != null)
				{
					provider.GenerateCodeFromNamespace(codeNamespace, writer, options);
				}
				else if ((codeTypeDeclaration = codeObject as CodeTypeDeclaration) != null)
				{
					provider.GenerateCodeFromType(codeTypeDeclaration, writer, options);
				}
				else if ((codeTypeMember = codeObject as CodeTypeMember) != null)
				{
					provider.GenerateCodeFromMember(codeTypeMember, writer, options);
				}
				else if ((codeStatement = codeObject as CodeStatement) != null)
				{
					provider.GenerateCodeFromStatement(codeStatement, writer, options);
				}
				else if ((codeExpression = codeObject as CodeExpression) != null)
				{
					provider.GenerateCodeFromExpression(codeExpression, writer, options);
				}
				else
				{
					throw new NotSupportedException();
				}
			}

			return builder.ToString();
		}

		public static TypeReference ToReference(this Type type)
		{
			return new TypeReference(type);
		}

		public static ReferenceExpression ToReference(this CodeMemberField field)
		{
			return new ReferenceExpression(new ThisExpression(), field.Name);
		}

		public static TypeReferenceExpression ToExpression(this TypeReference reference)
		{
			return new TypeReferenceExpression(reference);
		}
	}
}
