﻿namespace MetaSharp.Transformation.Grammar:
	import System;
	import System.Collections.Generic;
	import System.Linq;
	import System.Text;
	import MetaSharp.Transformation.Grammar.Ast;
	import System.CodeDom;
	import MetaSharp.Transformation.Lang;
	import MetaSharp.Transformation.Lang.Ast;

	grammar GrammarParser < LangParser:

		override ValidBlocks
			= RuleDeclaration
			| super;

		override TypeDeclaration
			= GrammarDeclaration
			| super;

		GrammarDeclaration = 
			Grammar
			name:ValidIdentifier
			superType:(GrammarImplements t:TypeReference -> t)?
			BlockBegin
				members:GrammarMemberDeclaration*
			error until BlockEnd -> {
				GrammarDeclaration g = new GrammarDeclaration(name as string);
				g.Members.AddRange(members.Cast<object>().OfType<CodeTypeMember>().ToArray());
				g.Rules = members.Cast<object>().OfType<RuleDeclaration>().ToArray();
				if (superType != null): g.BaseTypes.Add(superType as CodeTypeReference); end
				return g;
			}

		GrammarMemberDeclaration
			= RuleDeclaration
			| ClassMemberDeclaration;

		RuleDeclaration = 
			att:AttributeDeclaration*
			overridden:(Override -> true | default -> false)
			name:ValidIdentifier
			parameters:(GroupBegin p:List(ValidIdentifier, ParameterSeparator) GroupEnd -> p)?
			Assign
			pattern:(p:PatternDeclaration (EndsWithMultiStatementProjection(p) | error unless StatementEnd) -> p)
			-> {
				RuleDeclaration rule = new RuleDeclaration();
				rule.Attributes = new Nodes<AttributeDeclaration>(att.Cast<AttributeDeclaration>());
				rule.Name = name as string;
				rule.Pattern = pattern as PatternDeclaration;
				rule.Override = overridden as bool;
				rule.Parameters = new Nodes<string>(parameters.Cast<INode>().Unwrap<string>());
				return rule;
			}

		EndsWithMultiStatementProjection
			= OrPattern { Patterns = [p:ToTree EndsWithMultiStatementProjection(p)] }
			| [any ProjectionPattern { WithExpression = false }]
			| ProjectionPattern { WithExpression = false };

		PatternDeclaration = OrPattern;
		OrPattern 
			=	left:OrPattern "|" right:ProjectionPattern -> {
					Nodes<PatternDeclaration> patterns = new Nodes<PatternDeclaration>();
					for(PatternDeclaration decl in left
						.Cast<PatternDeclaration>()
						.Concat(right.Cast<PatternDeclaration>())):
						if(decl is OrPattern):
							patterns.AddRange((decl as OrPattern).Patterns);
						else:
							patterns.Add(decl);
						end
					end
					OrPattern orPattern = new OrPattern();
					orPattern.Patterns = patterns;
					return orPattern;
				}
			|	ProjectionPattern;

		ProjectionPattern
			= ProjectionPatternWithStatements
			| ProjectionPatternWithExpression
			| AndPattern;

		ProjectionPatternWithExpression =
			pattern:AndPattern
			Projection
			exp:BinaryAddSubtractExpression -> {
				ProjectionPattern proj = new ProjectionPattern();
				proj.WithExpression = true;
				proj.Pattern = pattern as PatternDeclaration;
				proj.Projection = new Nodes<CodeStatement>(new ExpressionStatement(exp as CodeExpression));
				return proj;
			}

		ProjectionPatternWithStatements =
			pattern:AndPattern
			Projection
			ProjectionBegin
				statements:Statement*
			error unless ProjectionEnd -> {
				ProjectionPattern proj = new ProjectionPattern();
				proj.Pattern = pattern as PatternDeclaration;
				proj.Projection = new Nodes<CodeStatement>(statements.Cast<CodeStatement>());
				return proj;
			}

		AndPattern
			=	left:AndPattern right:ErrorPattern -> {
					Nodes<PatternDeclaration> patterns = new Nodes<PatternDeclaration>();
					for(PatternDeclaration decl in left
						.Cast<PatternDeclaration>()
						.Concat(right.Cast<PatternDeclaration>())):
						if(decl is AndPattern):
							patterns.AddRange((decl as AndPattern).Patterns);
						else:
							patterns.Add(decl);
						end
					end
					AndPattern andPattern = new AndPattern();
					andPattern.Patterns = patterns;
					return andPattern;
				}
			| ErrorPattern;

		ErrorPattern
			= Error Until p:NotPattern -> {
					ErrorUntilPattern e = new ErrorUntilPattern();
					e.Pattern = p as PatternDeclaration;
					return e;
				}
			| Error Unless p:NotPattern -> {
					ErrorUnlessPattern e = new ErrorUnlessPattern();
					e.Pattern = p as PatternDeclaration;
					return e;
				}
			| NotPattern;

		NotPattern
			= "!" p:QuantifierPattern -> {
					NotPattern not = new NotPattern();
					not.Pattern = p as PatternDeclaration;
					return not;
				}
			| QuantifierPattern;

		QuantifierPattern
			= ZeroOrOnePattern
			| ZeroOrMorePattern
			| OneOrMorePattern
			| ExactQuantityPattern
			| RangePattern;

		ZeroOrOnePattern = p:RangePattern "?" -> {
				ZeroOrOnePattern zoop = new ZeroOrOnePattern();
				zoop.Pattern = p as PatternDeclaration;
				return zoop;
			}

		ZeroOrMorePattern = p:RangePattern "*" -> {
				ZeroOrMorePattern zomp = new ZeroOrMorePattern();
				zomp.Pattern = p as PatternDeclaration;
				return zomp;
			}

		OneOrMorePattern = p:RangePattern "+" -> {
				OneOrMorePattern oomp = new OneOrMorePattern();
				oomp.Pattern = p as PatternDeclaration;
				return oomp;
			}

		ExactQuantityPattern = p:RangePattern "#" i:NumberPrimitive -> {
				ExactQuantityPattern eqp = new ExactQuantityPattern();
				eqp.Pattern = p as PatternDeclaration;
				eqp.Quantity = (i as PrimitiveExpression).Value as int;
				return eqp;
			}

		RangePattern
			= l:StringLiteral Range r:StringLiteral -> {
					RangePattern rp = new RangePattern();
					rp.Left = l as string;
					rp.Right = r as string;
					return rp;
				}
			| l:CharacterLiteral Range r:CharacterLiteral -> {
					RangePattern rp = new RangePattern();
					rp.Left = l as char;
					rp.Right = r as char;
					return rp;
				}
			| ValuePattern;

			
		ValuePattern
			= GroupPattern
			| PrimitiveExpression -> {
					ValuePattern vp = new ValuePattern();
					vp.Value = (match as PrimitiveExpression).Value;
					return vp;
				}
			| ArrayPattern
			| ObjectPattern
			| DefaultPattern
			| AnyPattern
			| FailPattern
			| NullPattern
			| SuperPattern
			| VariablePattern
			| CallPattern
			| ReferencePattern;

		GroupPattern = 
			GroupBegin 
			p:PatternDeclaration
			ValidateGroup(p)
			GroupEnd ->
				p;

		ValidateGroup = ProjectionPattern { } | OrPattern { } | AndPattern { };

		ArrayPattern =
			IndexBegin
			pattern:PatternDeclaration?
			IndexEnd -> {
				ArrayPattern a = new ArrayPattern();
				a.Pattern = pattern as PatternDeclaration;
				return a;
			}

		DefaultPattern = Default -> new DefaultPattern();
		AnyPattern = Any -> new AnyPattern();
		NullPattern = Null -> new NullPattern();
		SuperPattern = Super -> new SuperPattern();
		FailPattern = Fail -> new FailPattern();
		VariablePattern = i:ValidIdentifier VariableAssign p:QuantifierPattern -> {
				VariablePattern pattern = new VariablePattern();
				pattern.Name = i as string;
				pattern.Pattern = p as PatternDeclaration;
				return pattern;
			}

		ObjectPattern =
			t:ObjectTypeReference
			ObjectBegin
			p:List(ObjectPropertyPattern, ParameterSeparator)
			ObjectEnd -> {
				ObjectPattern op = new ObjectPattern();
				op.Type = t as TypeReference;
				op.Properties = p.Cast<PropertyPattern>().ToArray();
				return op;
			}

		ObjectPropertyPattern =
			alias:(i:ValidIdentifier VariableAssign -> i)?
			name:Identifier
			p:(Assign p:PatternDeclaration -> p)? -> {
				PropertyPattern pp = new PropertyPattern();
				pp.Alias = alias as string;
				pp.Name = name as string;
				pp.Pattern = p as PatternDeclaration;
				return pp;
			}

		ReferencePattern = i:ValidIdentifier -> {
				ReferencePattern p = new ReferencePattern();
				p.Name = i as string;
				return p;
			}

		CallPattern =
			i:ValidIdentifier
			GroupBegin
			p:List((QuantifierPattern | GroupPattern), ParameterSeparator)
			GroupEnd -> {
				CallPattern call = new CallPattern();
				call.Name = i as string;
				call.Parameters = p.Cast<PatternDeclaration>().ToArray();
				return call;
			}

		override ReferenceExpression
			= Match -> new MatchExpression()
			| super;

		override Keywords 
			= Grammar
			| Default
			| Any
			| Match
			| Fail
			| Error
			| Unless
			| Until
			| super;

		override CustomTokens
			= '-' '>'
			| '.' '.'
			| super;

		VariableAssign = ":";
		ProjectionBegin = "{";
		ProjectionEnd = "}";
		ObjectBegin = "{";
		ObjectEnd = "}";
		GrammarImplements = "<";
		Projection = "->";
		Range = "..";

		[Keyword]Grammar = "grammar";
		[Keyword]Default = "default";
		[Keyword]Any = "any";
		[Keyword]Match = "match";
		[Keyword]Fail = "fail";
		[Keyword]Error = "error";
		[Keyword]Unless = "unless";
		[Keyword]Until = "until";
	end
end