﻿namespace MetaSharp.Transformation.Lang:
	import System;
	import System.CodeDom;
	import System.CodeDom.Compiler;
	import System.Collections;
	import System.Collections.Generic;
	import System.Globalization;
	import System.Linq;
	import System.Reflection;
	import MetaSharp.Transformation.Lang.Ast;

	grammar LangParser < Parser:
		Main = ValidBlocks;

		ValidBlocks 
			= Unit
			| ImportDeclaration
			| TypeDeclaration 
			| ClassMemberDeclaration 
			| Statement
			| Expression;

		Unit =
			attributes:UnitAttributeDeclaration*
			namespaces:NamespaceDeclaration+ ->  {
				Unit unit = new Unit();
				unit.AssemblyCustomAttributes.AddRange(attributes.Cast<CodeAttributeDeclaration>().ToArray());
				unit.Namespaces.AddRange(namespaces.Cast<CodeNamespace>().ToArray());
				return unit;
			}
			| UnitAttributeDeclaration+ -> {
				Unit unit = new Unit();
				unit.AssemblyCustomAttributes.AddRange(match.Cast<CodeAttributeDeclaration>().ToArray());
				return unit;
			}

		UnitAttributeDeclaration = 
			AttributeBegin
			Assembly
			BlockBegin
			t:ObjectTypeReference
			a:(GroupBegin e:List(Expression, ParameterSeparator) GroupEnd -> e)?
			AttributeEnd -> 
				new AttributeDeclaration(
					t as TypeReference, 
					a.Cast<AttributeArgument>().ToArray());

		NamespaceDeclaration = 
			Namespace
			name:TypeIdentifier
			BlockBegin
			imports:ImportDeclaration*
			types:TypeDeclaration*
			error until BlockEnd -> {
				Namespace ns = new Namespace(name as string);
				ns.Imports.AddRange(imports.Cast<CodeNamespaceImport>().ToArray());
				ns.Types.AddRange(types.Cast<CodeTypeDeclaration>().ToArray());
				return ns;
			}

		ImportDeclaration = 
			Import
			i:TypeIdentifier
			StatementEnd ->
				new Import(i as string);

		// Types
		TypeDeclaration 
			= ClassDeclaration 
			| EnumDeclaration 
			| InterfaceDeclaration 
			| DelegateDeclaration 
			| StructDeclaration;

		ClassDeclaration = 
			att:AttributeDeclaration*
			access:AccessModifier?
			options:TypeOptions?
			isPartial:IsPartial
			Class
			name:Identifier
			typeParameters:TypeParameterDeclarations?
			superTypes:(Implements t:List(TypeReference, ParameterSeparator) -> t | default -> new Nodes())
			BlockBegin
			members:ClassMemberDeclaration*
			error until BlockEnd -> {
				TypeDeclaration type = new TypeDeclaration(name as string);
				type.TypeAttributes = TypeAttributes.Class;
				type.IsPartial = isPartial as bool;
				if(access != null): type.Attributes = access as MemberAttributes;
				else: type.Attributes = MemberAttributes.Public; end
				if(options != null): type.Attributes = type.Attributes | options as MemberAttributes; end

				type.CustomAttributes.AddRange(att.Cast<CodeAttributeDeclaration>().ToArray());

				AttributeDeclaration codeAttributeDeclaration = new AttributeDeclaration
				(
					new TypeReference(typeof(GeneratedCodeAttribute)),
					new AttributeArgument(new CodePrimitiveExpression("MetaSharp")),
					new AttributeArgument(new CodePrimitiveExpression(System.Reflection.Assembly.GetExecutingAssembly().GetName().Version.ToString()))
				);
	
				bool isFound = false;

				for(CodeAttributeDeclaration attributeDeclaration in type.CustomAttributes.OfType<AttributeDeclaration>()):
					if (attributeDeclaration.Name==typeof(GeneratedCodeAttribute).FullName):
						isFound = true;
						goto Done;
					end
				end
				label Done:

				if (!isFound):
					type.CustomAttributes.Add(codeAttributeDeclaration);
				end

				type.BaseTypes.AddRange(superTypes.Cast<CodeTypeReference>().ToArray());
				type.TypeParameters.AddRange(typeParameters.Cast<CodeTypeParameter>().ToArray());
				for(object item in members):
					for(CodeTypeMember member in item.Cast<CodeTypeMember>()):
						type.Members.Add(member);
					end
				end
				return type;
			}

		AccessModifier
			= Public -> MemberAttributes.Public
			| Private -> MemberAttributes.Private
			| Family Or Assembly -> MemberAttributes.FamilyOrAssembly
			| Family And Assembly -> MemberAttributes.FamilyAndAssembly
			| Assembly -> MemberAttributes.Assembly
			| Family -> MemberAttributes.Family;

		TypeOptions
			= Abstract -> MemberAttributes.Abstract
			| Static -> MemberAttributes.Static
			| Final -> MemberAttributes.Final;

		IsPartial
			= Partial -> true
			| default -> false;

		AttributeDeclaration = 
			AttributeBegin
			t:TypeReference 
			a:(GroupBegin p:List(Expression, ParameterSeparator)? GroupEnd -> p)?
			AttributeEnd -> 
				new AttributeDeclaration(t as TypeReference, a.Cast<AttributeArgument>().ToArray());

		ClassMemberDeclaration 
			= TypeDeclaration
			| TypeConstructorDeclaration
			| ConstructorDeclaration
			| FieldDeclaration
			| PropertyDeclaration
			| AbstractPropertyDeclaration
			| MethodDeclaration
			| AbstractMethodDeclaration;

		EnumDeclaration = fail;
		InterfaceDeclaration = fail;
		DelegateDeclaration = fail;
		StructDeclaration = fail;

		TypeConstructorDeclaration =
			att:AttributeDeclaration*						//[example]
			Static Constructor BlockBegin					//static constructor:
				s:Statement*								//	foo = bar;
			error until BlockEnd -> {						//end
				CodeConstructor c = new Constructor();
				c.Attributes = MemberAttributes.Public | MemberAttributes.Static;
				c.CustomAttributes.AddRange(att.Cast<CodeAttributeDeclaration>().ToArray());
				c.Statements.AddRange(s.Cast<CodeStatement>().ToArray());
				return c;
			}
			
		ConstructorDeclaration =
			att:AttributeDeclaration*
			access:AccessModifier?
			Constructor 
			parameters:(GroupBegin p:List(ParameterDeclarationExpression, ParameterSeparator) GroupEnd -> p)?
			superArgs:ConstructorSuperArgs?
			chainArgs:ConstructorChainArgs?
			BlockBegin
				statements:Statement*
			error until BlockEnd -> {
				Constructor c = new Constructor();
				c.Attributes = MemberAttributes.Public;
				if(access != null): c.Attributes = access as MemberAttributes;
				else: c.Attributes = MemberAttributes.Public; end
				c.CustomAttributes.AddRange(att.Cast<CodeAttributeDeclaration>().ToArray());
				c.Parameters.AddRange(parameters.Cast<CodeParameterDeclarationExpression>().ToArray());
				c.BaseConstructorArgs.AddRange(superArgs.Cast<CodeExpression>().ToArray());
				c.ChainedConstructorArgs.AddRange(chainArgs.Cast<CodeExpression>().ToArray());
				c.Statements.AddRange(statements.Cast<CodeStatement>().ToArray());
				return c;
			}

		ParameterDeclarationExpression = 
			att:AttributeDeclaration*
			dir:Direction?
			type:TypeReference
			name:ValidIdentifier -> {
				ParameterDeclarationExpression decl = new ParameterDeclarationExpression(type as TypeReference, name as string);
				if(dir!=null): decl.Direction = dir as FieldDirection; end
				decl.CustomAttributes.AddRange(att.Cast<CodeAttributeDeclaration>().ToArray());
				return decl;
			}

		ConstructorSuperArgs =
			Super
			GroupBegin
			args:List(Expression, ParameterSeparator)
			GroupEnd ->
				args;

		ConstructorChainArgs = 
			This
			GroupBegin
			args:List(Expression, ParameterSeparator)
			GroupEnd ->
				args;

		FieldDeclaration=
			att:AttributeDeclaration*
			access:AccessModifier?
			options:FieldMemberOptions?
			type:TypeReference
			name:ValidIdentifier
			init:( (Assign e:Expression error unless StatementEnd -> e) | StatementEnd -> null ) -> {
				Field f = new Field(type as TypeReference, name as string);
				f.Attributes = MemberAttributes.Private;
				if(access != null): f.Attributes = access as MemberAttributes; end
				if(options != null): f.Attributes = f.Attributes | options as MemberAttributes; end
				f.CustomAttributes.AddRange(att.Cast<CodeAttributeDeclaration>().ToArray());
				f.InitExpression = init as CodeExpression;
				return f;
			}

		FieldMemberOptions 
			= Static -> MemberAttributes.Static
			| New -> MemberAttributes.New
			| Constant -> MemberAttributes.Const;

		MethodDeclaration =
			att:AttributeDeclaration*
			access:AccessModifier?
			options:MethodMemberOptions?
			type:TypeReference
			name:ValidIdentifier
			typeParameters:TypeParameterDeclarations?
			parameters:(GroupBegin p:List(ParameterDeclarationExpression, ParameterSeparator) GroupEnd -> p)?
			BlockBegin
				s:Statement*
			error until BlockEnd -> {
				Method method = new Method();
				method.Name = name as string;
				if(access!=null): method.Attributes = access as MemberAttributes;
				else: method.Attributes = MemberAttributes.Family; end
				if(options!=null): method.Attributes = method.Attributes | options as MemberAttributes; end
				method.CustomAttributes.AddRange(att.Cast<CodeAttributeDeclaration>().ToArray());
				method.ReturnType = type as CodeTypeReference;
				method.Parameters.AddRange(parameters.Cast<CodeParameterDeclarationExpression>().ToArray());
				method.TypeParameters.AddRange(typeParameters.Cast<CodeTypeParameter>().ToArray());
				method.Statements.AddRange(s.Cast<CodeStatement>().ToArray());
				return method;
			}
			
		MethodMemberOptions
			= Static -> MemberAttributes.Static
			| New -> MemberAttributes.New
			| Final -> MemberAttributes.Final
			| Override -> MemberAttributes.Override;

		PropertyDeclaration 
			= ExpandedPropertyDeclaration
			| AutoPropertyDeclaration;

		ExpandedPropertyDeclaration = 
			att:AttributeDeclaration*
			a:AccessModifier?
			o:PropertyMemberOptions?
			t:TypeReference
			n:Identifier
			BlockBegin
			gs:( 
			  g:PropertyGetDeclaration s:PropertySetDeclaration -> new Nodes(g as INode, s as INode)
			| s:PropertySetDeclaration g:PropertyGetDeclaration -> new Nodes(g as INode, s as INode)
			| g:PropertyGetDeclaration                          -> new Nodes(g as INode, new Nodes())
			| s:PropertySetDeclaration                          -> new Nodes(new Nodes(), s as INode))
			error until BlockEnd -> {
				Nodes getset = gs as Nodes;
				Property prop = new Property();
				if(a != null): prop.Attributes = a as MemberAttributes;
				else: prop.Attributes = MemberAttributes.Public; end
				if(o != null): prop.Attributes = prop.Attributes | o as MemberAttributes; end
				prop.CustomAttributes.AddRange(att.Cast<CodeAttributeDeclaration>().ToArray());
				prop.Type = t as TypeReference;
				prop.Name = n as string;
				prop.GetStatements.AddRange(getset[0].Cast<CodeStatement>().ToArray());
				prop.SetStatements.AddRange(getset[1].Cast<CodeStatement>().ToArray());
				return prop;
			}

		PropertyGetDeclaration =
			Get
			BlockBegin
				s:Statement*
			error until BlockEnd ->
				s;

		PropertySetDeclaration =
			Set
			BlockBegin
				s:Statement*
			error until BlockEnd ->
				s;

		AutoPropertyDeclaration =
			att:AttributeDeclaration*
			a:AccessModifier?
			o:PropertyMemberOptions?
			t:TypeReference
			n:Identifier
			BlockBegin
				(Get StatementEnd Set StatementEnd | Set StatementEnd Get StatementEnd)
			error until BlockEnd -> {
				Nodes members = new Nodes();
				string fieldName = "__" + n as string;
				members.Add(new Field(t as TypeReference, fieldName));
				Property prop = new Property();
				if(a != null): prop.Attributes = a as MemberAttributes;
				else: prop.Attributes = MemberAttributes.Public; end
				if(o != null): prop.Attributes = prop.Attributes | o as MemberAttributes; end
				prop.CustomAttributes.AddRange(att.Cast<CodeAttributeDeclaration>().ToArray());
				prop.Type = t as TypeReference;
				prop.Name = n as string;
				prop.GetStatements.Add(new ReturnStatement(new ReferenceExpression(fieldName)));
				prop.SetStatements.Add(new AssignStatement(new ReferenceExpression(fieldName), new PropertySetValueReferenceExpression()));
				members.Add(prop);
				return members;
			}
			
		PropertyMemberOptions
			= Static -> MemberAttributes.Static
			| New -> MemberAttributes.New
			| Final -> MemberAttributes.Final
			| Override -> MemberAttributes.Override;

		// TODO: implement
		AbstractMethodDeclaration = fail;
		AbstractPropertyDeclaration = fail;


		// Type rules
		TypeReference 
			= ArrayTypeReference 
			| ObjectTypeReference;

		ObjectTypeReference = 
			i:(TypeIdentifier | PrimitiveTypes)
			a:TypeArguments?
			n:(Nullable -> true | default -> false) -> {
				TypeReference t = new TypeReference(
					i as string, 
					a.Cast<TypeReference>().ToArray());

				if(n as bool):
					t = new TypeReference("System.Nullable", t);
				end
				return t;
			}

		ArrayTypeReference = 
			type:ObjectTypeReference
			IndexBegin
			ranks:ParameterSeparator*
			IndexEnd -> {
				int rank = ranks.Cast<object>().Count() + 1;
				CodeTypeReference typeRef = type as CodeTypeReference;
				typeRef.ArrayRank = rank;
				return typeRef;
			}

		TypeIdentifier = 
			ValidIdentifier
			(TypeDelimiter ValidIdentifier)* ->
				Parser.Stringify(match);

		TypeArguments = 
			TypeParameterBegin
			args:List(TypeReference, ParameterSeparator)
			TypeParameterEnd
			-> args;

		TypeParameterDeclarations = 
			TypeParameterBegin
			p:List(TypeParameter, ParameterSeparator)
			TypeParameterEnd
			-> p;

		TypeParameter = 
			name:ValidIdentifier
			constraints:(As c:List((New | TypeReference), And) -> c)? -> {
				TypeParameter p = new TypeParameter(name as string);
				if(constraints != null):
					for(object c in constraints):
						object constraint = Node.Unwrap(c);
						if(constraint is string and (constraint as string) == "new"):
							p.HasConstructorConstraint = true;
						else:
							p.Constraints.Add(constraint as TypeReference);
						end
					end
				end
				return p;
			}

		Statement
			= BlockStatements
			| LineStatements;

		BlockStatements
			= TryCatchFinallyStatement
			| IterationStatement
			| ForEachStatement
			| WhileStatement
			| ConditionStatement
			| LabeledStatement;

		LineStatements = s:
			( VariableDeclarationStatement
			| MethodReturnStatement
			| ThrowStatement
			| GotoStatement 
			| ExpressionStatement)
			error unless e:StatementEnd ->
			s;

		ExpressionStatement = exp:StatementExpression -> {
				INode statement = null;
				if(exp is CodeStatement):
					statement = exp as INode;
				else:
					statement = new ExpressionStatement(exp as CodeExpression);
				end
				return statement;
			}

		StatementExpression
			= AssignmentStatement
			| AddAssignExpression
			| SubtractAssignExpression
			| e:CallableExpression v:ValidCallableStatementExpression(e)? -> {
				if(v == null):
					context.LogError(
						position.Next(),
						ErrorLocation.Span,
						ErrorSeverity.Error,
						"Only assignment, call, increment, decrement, and new object expressions can be used as a statement");
					return new Fail();
				end
				return e;
			}

		ValidCallableStatementExpression
			= DelegateInvokeExpression { }
			| NewExpression { }
			| MethodInvokeExpression { };

		AssignmentStatement = 
			left:CallableExpression
			Assign
			right:Expression ->
				new AssignStatement(left as CodeExpression, right as CodeExpression);

		AddAssignExpression = 
			left:CallableExpression
			PlusEqual
			right:Expression ->
				new AssignStatement(
					left as CodeExpression,
					new BinaryOperatorExpression(
						left as CodeExpression,
						CodeBinaryOperatorType.Add,
						right as CodeExpression));

		SubtractAssignExpression = 
			left:CallableExpression
			MinusEqual
			right:Expression ->
				new AssignStatement(
					left as CodeExpression,
					new BinaryOperatorExpression(
						left as CodeExpression,
						CodeBinaryOperatorType.Subtract,
						right as CodeExpression));

		MethodReturnStatement = 
			Return
			exp:Expression? ->
				new ReturnStatement(exp as CodeExpression);

		IterationStatement = 
			For
			GroupBegin
			init:(v:VariableDeclarationStatement? StatementEnd -> v)
			test:Expression?
			StatementEnd
			increment:StatementExpression?
			GroupEnd
			BlockBegin
				statements:Statement*
			error until BlockEnd ->
				new IterationStatement(
					init as CodeVariableDeclarationStatement,
					test as CodeExpression,
					increment as CodeStatement,
					statements.Cast<CodeStatement>().ToArray());

		VariableDeclarationStatement
			= t:TypeReference i:ValidIdentifier exp:(Assign e:Expression -> e)? ->
				new VariableDeclarationStatement(
					t as CodeTypeReference,
					i as string,
					exp as CodeExpression);

		ForEachStatement =
			For
			GroupBegin
				type:TypeReference
				name:ValidIdentifier
				In
				exp:Expression
			GroupEnd
			BlockBegin
				statements:Statement*
			error until BlockEnd -> {
				// for(IEnumerator _ie = ((IEnumerable)exp).GetEnumerator();_ie.MoveNext();):
				//		type i = (_ie.Current as type);
				//		statements
				// end

				string enumeratorName = "_" + name + "_enumerator";
				VariableDeclarationStatement init = new VariableDeclarationStatement(
					typeof(IEnumerator),
					enumeratorName,
					new MethodInvokeExpression(
						new CastExpression(
							typeof(IEnumerable),
							exp as CodeExpression),
						"GetEnumerator"));

				MethodInvokeExpression condition = new MethodInvokeExpression(
					new ReferenceExpression(enumeratorName),
					"MoveNext");

				VariableDeclarationStatement assign = new VariableDeclarationStatement(
					type as CodeTypeReference,
					name as string,
					new CastExpression(
						type as CodeTypeReference,
						new ReferenceExpression(
							new ReferenceExpression(enumeratorName),
							"Current")));

				return new IterationStatement(
					init,
					condition,
					new ExpressionStatement(new NewExpression(typeof(object))), // in the gheeeetttoooo...
					assign.Cast<CodeStatement>().Concat(statements.Cast<CodeStatement>()).ToArray());
			}

		WhileStatement = 
			While
			GroupBegin
				condition:Expression
			GroupEnd
			BlockBegin
				statements:Statement*
			error until BlockEnd ->
				new IterationStatement(
					null,
					condition as CodeExpression,
					new ExpressionStatement(new NewExpression(typeof(object))),
					statements.Cast<CodeStatement>().ToArray());

		ConditionStatement =
			If
			GroupBegin
			condition:BooleanExpression
			GroupEnd
			BlockBegin
				statements:Statement*
			elseStatements:(ElseStatement | error until BlockEnd -> null) ->
				new ConditionStatement(
					condition as CodeExpression,
					statements.Cast<CodeStatement>().ToArray(),
					elseStatements.Cast<CodeStatement>().ToArray());

		ElseStatement
			= Else BlockBegin s:Statement* error until BlockEnd -> s
			| Else c:ConditionStatement -> c;

		TryCatchFinallyStatement =
			Try
			BlockBegin
			statements:Statement*
			catches:CatchWithParameterClause*
			catchesWithoutParameters:CatchWithoutParameterClause*
			finallyStatements:FinallyClause?
			error until BlockEnd -> {
				List<CodeCatchClause> catchClauses = new List<CodeCatchClause>();
				for(IEnumerable c in catches):
					catchClauses.AddRange(c.Cast<CodeCatchClause>());
				end
				for(CodeCatchClause clause in catchesWithoutParameters):
					catchClauses.Add(clause);
				end

				return new TryCatchFinallyStatement(
					statements.Cast<CodeStatement>().ToArray(),
					catchClauses.ToArray(),
					finallyStatements.Cast<CodeStatement>().ToArray());
			}

		CatchWithParameterClause =
			Catch
			GroupBegin
			parameters:List(CatchClauseParameter, ParameterSeparator)
			GroupEnd
			BlockBegin
			statements:Statement* -> {
				Nodes clauses = new Nodes();
				for (ParameterDeclarationExpression p in parameters):
					clauses.Add(new CatchClause(
						p.Name, 
						p.Type, 
						statements.Cast<CodeStatement>().ToArray()));
				end
				return clauses;
			}

		CatchClauseParameter = 
			t:TypeReference
			i:ValidIdentifier? ->
				new ParameterDeclarationExpression(
					t as TypeReference,
					i as string);

		CatchWithoutParameterClause = 
			Catch
			BlockBegin
			statements:Statement* ->
				new CatchClause(
					null,
					null,
					statements.Cast<CodeStatement>().ToArray());

		FinallyClause = 
			Finally
			BlockBegin
			s:Statement* -> s;

		ThrowStatement = 
			Throw
			exp:Expression? ->
				new ThrowExceptionStatement(exp as CodeExpression);

		LabeledStatement =
			Label
			l:ValidIdentifier
			BlockBegin
			statement:Statement ->
				new LabeledStatement(
					l as string,
					statement as CodeStatement);

		GotoStatement = 
			Goto
			i:ValidIdentifier ->
				new GotoStatement(i as string);

		// Expressions
		Expression = BooleanExpression;

		TypeOfExpression =
			TypeOf
			GroupBegin
			t:TypeReference
			GroupEnd ->
				new TypeOfExpression(t as CodeTypeReference);

		ObjectCreateExpression = 
			New
			t:TypeReference
			GroupBegin
			p:List(InvokeParameterExpression, ParameterSeparator)
			GroupEnd ->
				new NewExpression(
					t as CodeTypeReference,
					p.Cast<CodeExpression>().ToArray());

		DefaultValueExpression = 
			DefaultValue
			GroupBegin
			t:TypeReference
			GroupEnd ->
				new DefaultValueExpression(t as TypeReference);

		ArrayCreateExpression =
			New
			t:ObjectTypeReference
			IndexBegin
				size:Expression?
			IndexEnd -> {
				ArrayCreateExpression create = new ArrayCreateExpression(t as CodeTypeReference);
				create.SizeExpression = size as CodeExpression;
				return create;
			}

		BooleanExpression
			= left:BooleanExpression op:BooleanLogicalOperators right:IsExpression ->
				new BinaryOperatorExpression(
					left as CodeExpression,
					op as CodeBinaryOperatorType,
					right as CodeExpression)
			| IsExpression;

		IsExpression
			=	exp:BooleanEqualityExpression 
				Is 
				t:TypeReference ->
					new BinaryOperatorExpression(
						new BinaryOperatorExpression(
							exp as CodeExpression,
							CodeBinaryOperatorType.IdentityInequality,
							new PrimitiveExpression(null)),
						CodeBinaryOperatorType.BooleanAnd,
						new MethodInvokeExpression(
							new TypeOfExpression(t as CodeTypeReference),
							"IsAssignableFrom",
							new MethodInvokeExpression(
								exp as CodeExpression,
								"GetType")))
			| BooleanEqualityExpression;

		BooleanEqualityExpression
			=	left:BinaryBitwiseExpression
				op:BooleanEqualityOperators
				right:BinaryBitwiseExpression ->
					new BinaryOperatorExpression(
						left as CodeExpression,
						op as CodeBinaryOperatorType,
						right as CodeExpression)
			|	BinaryBitwiseExpression;

		BinaryBitwiseExpression
			=	left:BinaryBitwiseExpression
				op:BooleanBitwiseOperators
				right:BinaryAddSubtractExpression ->
					new BinaryOperatorExpression(
						left as CodeExpression,
						op as CodeBinaryOperatorType,
						right as CodeExpression)
			| BinaryAddSubtractExpression;

		BinaryAddSubtractExpression
			=	left:BinaryAddSubtractExpression
				op:BooleanAddSubtractOperators
				right:BinaryMultiplyDivideExpression ->
					new BinaryOperatorExpression(
						left as CodeExpression,
						op as CodeBinaryOperatorType,
						right as CodeExpression)
			| BinaryMultiplyDivideExpression;

		BinaryMultiplyDivideExpression
			=	left:BinaryMultiplyDivideExpression
				op:BooleanMultiplyDivideOperators
				right:CallableExpression ->
					new BinaryOperatorExpression(
						left as CodeExpression,
						op as CodeBinaryOperatorType,
						right as CodeExpression)
			|	CallableExpression;

		NotExpression 
			=	Not e:CallableExpression ->
					new BinaryOperatorExpression(
						e as CodeExpression,
						CodeBinaryOperatorType.ValueEquality,
						new PrimitiveExpression(false));

		CallableExpression
			=	NotExpression
			|	target:CallableExpression TypeDelimiter reference:ReferenceExpression ->
					new ReferenceExpression(
						target as CodeExpression,
						(reference as ReferenceExpression).PropertyName)
			|	t:CallableExpression p:InvokeExpression ->
					new DelegateInvokeExpression(
						t as CodeExpression,
						p.Cast<CodeExpression>().ToArray())
			|	t:CallableExpression a:TypeArguments e:InvokeExpression ->{
					ReferenceExpression target = t as ReferenceExpression;
					return new MethodInvokeExpression(
						new MethodReferenceExpression(
							target.TargetObject,
							target.PropertyName,
							a.Cast<CodeTypeReference>().ToArray()),
						e.Cast<CodeExpression>().ToArray());
				}
			|	t:CallableExpression i:ArrayIndexerExpression ->
					new ArrayIndexerExpression(
						t as CodeExpression,
						i.Cast<CodeExpression>().ToArray())
			|	target:CallableExpression As type:TypeReference ->
					new CastExpression(
						type as CodeTypeReference,
						target as CodeExpression)
			|	TypeOfExpression
			|	DefaultValueExpression
			|	NullExpression
			|	ObjectCreateExpression
			|	ArrayCreateExpression
			|	AccessReferenceExpression
			|	PrimitiveExpression
			|	PropertySetValueReferenceExpression
			|	SuperExpression
			|	ThisExpression
			|	ReferenceExpression
			|	GroupBegin e:Expression GroupEnd -> e;
			
		SuperExpression = Super -> new SuperExpression();
		ThisExpression = This -> new ThisExpression();
		ReferenceExpression = r:(PrimitiveTypes | ValidIdentifier) -> new ReferenceExpression(r as string);
		InvokeExpression = InvokeBegin p:List(InvokeParameterExpression, ParameterSeparator) InvokeEnd -> p;
		InvokeParameterExpression = 
			d:Direction?
			e:Expression -> {
				if(d!=null):
					return new DirectionExpression(d as FieldDirection, e as CodeExpression);
				else:
					return e as INode;
				end
			}

		ArrayIndexerExpression = IndexBegin p:List(Expression, ParameterSeparator) IndexEnd -> p;
		AccessReferenceExpression
			= SuperExpression
			| ThisExpression;
		NullExpression = Null -> new PrimitiveExpression(null);
		PrimitiveExpression
			= True -> new PrimitiveExpression(match)
			| False -> new PrimitiveExpression(match)
			| NumberPrimitive
			| StringLiteral -> new PrimitiveExpression(match)
			| CharacterLiteral -> new PrimitiveExpression(match);

		NumberPrimitive = Number -> {
			string val = match as string;
			int i;
			if (int.TryParse(val, NumberStyles.Integer, CultureInfo.InvariantCulture, out i)):
				return new PrimitiveExpression(i);
			end

			string hex = val.Replace("0x", "");
			if(int.TryParse(hex, NumberStyles.HexNumber, CultureInfo.InvariantCulture, out i)):
				return new PrimitiveExpression(i);
			end

			double d;
			if(double.TryParse(val, NumberStyles.AllowLeadingSign | NumberStyles.AllowDecimalPoint, CultureInfo.InvariantCulture, out d)):
				return new PrimitiveExpression(d);
			end

			//return Fail;
			throw new NotImplementedException();
		}

		PropertySetValueReferenceExpression = SetValue ->
			new PropertySetValueReferenceExpression();

		//MatchExpression = Match ->
		//	new MatchExpression();

		// Tokens and Keywords
		ValidIdentifier = !Keywords i:Identifier -> i;

		Keywords
			= PrimitiveTypes 
			| BooleanLogicalOperators 
			| BlockEnd
			| False | True | Null
			| Class | Enum | Struct | Import | Namespace
			| Constructor | Event | Implements
			| Null | If | Else | Is | As | For | While | Goto | Label | Return
			| Throw | Try | Finally | Catch
			| DefaultValue | Delegate | TypeOf | New
			| Super | This 
			| In | Out | Ref
			| Abstract | Partial | Static | Override | Virtual | Final | Constant | Transient
			| Public | Private | Assembly | Family;
		
		[Keyword]
		PrimitiveTypes
			= "bool" -> "System.Boolean"
			| "byte" -> "System.Byte"
			| "char" -> "System.Char"
			| "decimal" -> "System.Decimal"
			| "double" -> "System.Double"
			| "single" -> "System.Single"
			| "short" -> "System.Int16"
			| "int" -> "System.Int32"
			| "long" -> "System.Int64"
			| "uint" -> "System.UInt32"
			| "ulong" -> "System.UInt64"
			| "ushort" -> "System.UInt16"
			| "sbyte" -> "System.SByte"
			| "string" -> "System.String"
			| "unique" -> "System.Guid"
			| "void" -> "System.Void"
			| "object" -> "System.Object";

		BooleanAddSubtractOperators
			= Add -> CodeBinaryOperatorType.Add
			| Subtract -> CodeBinaryOperatorType.Subtract;

		BooleanMultiplyDivideOperators
			= Multiply -> CodeBinaryOperatorType.Multiply
			| Divide -> CodeBinaryOperatorType.Divide
			| Modulus -> CodeBinaryOperatorType.Modulus;

		BooleanBitwiseOperators
			= BitwiseOr -> CodeBinaryOperatorType.BitwiseOr
			| BitwiseAnd -> CodeBinaryOperatorType.BitwiseAnd;
		
		BooleanEqualityOperators
			= Inequality -> CodeBinaryOperatorType.IdentityInequality
			| Equality -> CodeBinaryOperatorType.IdentityEquality
			| GreaterThanOrEqual -> CodeBinaryOperatorType.GreaterThanOrEqual
			| LessThanOrEqual -> CodeBinaryOperatorType.LessThanOrEqual
			| GreaterThan -> CodeBinaryOperatorType.GreaterThan
			| LessThan -> CodeBinaryOperatorType.LessThan;

		BooleanLogicalOperators
			= And -> CodeBinaryOperatorType.BooleanAnd
			| Or -> CodeBinaryOperatorType.BooleanOr;

		[Keyword]False = "false" -> false;
		[Keyword]True = "true" -> true;

		[Keyword]Abstract = "abstract";
		[Keyword]Partial = "partial";
		[Keyword]Static = "static";
		[Keyword]Override = "override";
		[Keyword]Virtual = "virtual";
		[Keyword]Final = "final";
		[Keyword]Constant = "const";
		[Keyword]Transient = "transient";

		[Keyword]Public = "public";
		[Keyword]Private = "private";
		[Keyword]Assembly = "assembly";
		[Keyword]Family = "family";

		[Keyword]Class = "class";
		[Keyword]Enum = "enum";
		[Keyword]Interface = "interface";
		[Keyword]Struct = "struct";
		[Keyword]Import = "import";
		[Keyword]Namespace = "namespace";
		[Keyword]Constructor = "constructor";
		[Keyword]Event = "event";

		[Keyword]Null = "null";
		[Keyword]Implements = "as";
		[Keyword]Block = "block";
		[Keyword]If = "if";
		[Keyword]Else = "else";
		[Keyword]As = "as";
		[Keyword]For = "for";
		[Keyword]While = "while";
		[Keyword]Goto = "goto";
		[Keyword]Label = "label";
		[Keyword]Return = "return";
		[Keyword]Throw = "throw";
		[Keyword]Try = "try";
		[Keyword]Finally = "finally"; 
		[Keyword]Catch = "catch";
		[Keyword]DefaultValue = "default";
		[Keyword]Delegate = "delegate";
		[Keyword]TypeOf = "typeof";
		[Keyword]New = "new";
		[Keyword]SetValue = "value";
		[Keyword]Super = "super";
		[Keyword]This = "this" ;
		[Keyword]In = "in";
		[Keyword]Out = "out";
		[Keyword]Ref = "ref";
		[Keyword]Set = "set";
		[Keyword]Get = "get";
		//Match = "match";

		Direction
			= In -> FieldDirection.In
			| Ref -> FieldDirection.Ref
			| Out -> FieldDirection.Out;

		BlockBegin = ":";
		[Keyword]BlockEnd = "end";
		ParameterSeparator = ",";
		TypeDelimiter = ".";
		StatementEnd = ";";
		Nullable = "?";
		AttributeBegin = "[";
		AttributeEnd = "]";
		GroupBegin = "(";
		GroupEnd = ")";
		InvokeBegin = "(";
		InvokeEnd = ")";
		IndexBegin = "[";
		IndexEnd = "]";
		TypeParameterBegin = "<";
		TypeParameterEnd = ">";

		Add = "+";
		Subtract = "-";
		Multiply = "*";
		Divide = "/";
		Modulus = "%";
		Assign = "=";
		BitwiseOr = "|";
		BitwiseAnd = "&";
		PlusEqual = "+=";
		MinusEqual = "-=";
		Not = "!";

		Inequality = "!=";
		Equality = "==";
		LessThan = "<";
		LessThanOrEqual = "<=";
		GreaterThan = ">";
		GreaterThanOrEqual = ">=";
		[Keyword]And = "and";
		[Keyword]Or = "or";
		[Keyword]Is = "is";

		override CustomTokens
			= '+' '='
			| '-' '='
			| '=' '='
			| '!' '='
			| '<' '='
			| '>' '='
			| super;
	end
end
