﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Runtime.InteropServices;

namespace MetaSharp.Transformation
{
	public sealed class MemoTable
	{
		private class PrimaryIndex : Dictionary<MemoKey, Memo> { }
		private class References : Dictionary<MemoKey, MemoKey[]> { }
		private class ProjectionIndex : Dictionary<INode, MemoKey> { }

		private int currentStep;
		private List<MemoKey> memoStack = new List<MemoKey>();
		private List<PrimaryIndex> memos = new List<PrimaryIndex> { new PrimaryIndex() };
		private List<References> references = new List<References> { new References() };
		private List<ProjectionIndex> projections = new List<ProjectionIndex> { new ProjectionIndex() };

		public MemoKey Top
		{
			get
			{
				var count = this.memoStack.Count;
				if (count > 0)
					return this.memoStack[this.memoStack.Count - 1];

				return null;
			}
		}

		public int MemoCount
		{
			get { return this.memoStack.Count; }
		}

		public bool GetMemo(PatternMatch method, Position position, out MemoKey id, out Memo memo)
		{
			id = new MemoKey(method, position);
			return this.memos[this.currentStep].TryGetValue(id, out memo);
		}

		public void SetMemo(MemoKey id, Memo memo)
		{
			if (!this.memos[this.currentStep].ContainsKey(id))
				this.memos[this.currentStep].Add(id, memo);

			this.memoStack.Add(id);
		}

		public void Step()
		{
			currentStep++;
			this.memoStack.Clear();
			this.memos.Add(new PrimaryIndex());
			this.references.Add(new References());
			this.projections.Add(new ProjectionIndex());
		}

		public void UpdateMemo(MemoKey id, Memo memo, int memoIndex)
		{
			var p = memo.Projection;
			var memoCount = this.memoStack.Count - memoIndex;
			MemoKey[] refs = null;

			if (p is Fail)
			{
				this.memoStack.RemoveRange(memoIndex, memoCount);
				memoCount = 0;
			}

			refs = new MemoKey[memoCount];
			if (memoCount > 0)
				this.memoStack.CopyTo(memoIndex, refs, 0, memoCount);

			this.memos[this.currentStep][id] = memo;

			// update references
			if (!this.references[this.currentStep].ContainsKey(id))
				this.references[this.currentStep].Add(id, refs);
			else
				this.references[this.currentStep][id] = refs;

			// update projection indexes
			if (!this.projections[this.currentStep].ContainsKey(p))
				this.projections[this.currentStep].Add(p, id);
			else
				this.projections[this.currentStep][p] = id;
		}

		public void Reset(int memoIndex)
		{
			var memoCount = this.memoStack.Count - memoIndex;
			this.memoStack.RemoveRange(memoIndex, memoCount);
		}

		public void Load(MemoKey id)
		{
			var refs = this.references[this.currentStep][id];
			this.memoStack.AddRange(refs);
		}

		public SourceSpan Trace(INode node)
		{
			return Trace(node, memos.Count - 1);
		}

		private SourceSpan Trace(INode node, int step)
		{
			var memoStep = this.memos[step];
			SourceSpan span = null;
			MemoKey id;
			if (this.projections[step].TryGetValue(node, out id))
			{
				var m = memoStep[id];
				span = Expand(step - 1, m.Start, m.End);
			}
			else if (step > 0)
			{
				span = Trace(node, step - 1);
			}

			return span;
		}

		private SourceSpan Expand(int step, Position start, Position end)
		{
			SourceSpan span = null;
			if (step == -1)
			{
				if (start is IIndexed)
					start = start.Next();
				if (end is IIndexed)
					end = end.Next();

				span = new SourceSpan { Start = start, End = end };
			}
			else if (start is IIndexed)
			{
				var startIndex = ((IIndexed)start).Index;
				var endIndex = Math.Max(((IIndexed)end).Index, startIndex + 1);
				var pos = start;
				for (var x = startIndex; x < endIndex; x++)
				{
					pos = pos.Next();
					span += Expand(step, pos);
				}
			}
			else
			{
				span = Expand(step, start) + Expand(step, end);
			}

			return span;
		}

		private SourceSpan Expand(int step, Position pos)
		{
			var result = default(SourceSpan);
			if (!pos.IsEnd)
			{
				result = Trace(pos.Current, step);
				while (result == null && pos.Parent != null)
				{
					pos = pos.Parent;
					result = Trace(pos.Current, step);
				}
			}

			return result;
		}

		public IImmutableArray<RuleMetadata> GetMetadata(INode node)
		{
			var metadata = new ArrayList<RuleMetadata>();
			GetMetadata(node, this.currentStep, metadata);
			return metadata;
		}

		private void GetMetadata(INode node, int step, ArrayList<RuleMetadata> metadata)
		{
			MemoKey id;
			if (this.projections[step].TryGetValue(node, out id))
			{
				var refs = this.references[step][id];
				for (int x = 0; x < refs.Length; x++)
				{
					var memo = this.memos[step][refs[x]];
					if (memo.Method != null)
					{
						var atts = (MetadataAttribute[])memo.Method.GetCustomAttributes(typeof(MetadataAttribute), true);
						if (atts.Length > 0)
						{
							var span = Trace(memo.Projection, step);
							metadata.Add(new RuleMetadata(memo.Method, memo.Projection, span, atts));
						}

						if (step > 0)
						{
							var next = memo.Start.Next();
							if (!next.IsEnd)
								GetMetadata(next.Current, step - 1, metadata);
						}
					}
					else if (step > 0 && x == 0 && memo.Start.Current != null)
					{
						GetMetadata(memo.Start.Current, step - 1, metadata);
					}
				}
			}
			else if (step > 0)
				GetMetadata(node, step - 1, metadata);
		}
	}
}
