﻿namespace MetaSharp.Transformation
{
	using System;
	using System.Collections.Generic;
	using System.Diagnostics.CodeAnalysis;
	using System.Linq;
	using System.IO;
	using System.Collections;
	using System.Reflection;
	using System.Linq.Expressions;
	using System.Globalization;
	using MetaSharp.Transformation.Properties;

	public delegate INode Projection(Scope scope);
	public delegate INode PatternMatch(PatternContext context, IStream stream);
	public delegate INode ParameterizedPatternMatch(PatternContext context, IStream stream, params PatternMatch[] parameters);

	public class PropertyMatch
	{
		private string propertyName;
		private PatternMatch match;

		public PropertyMatch(string propertyName, PatternMatch match)
		{
			if (propertyName == null)
				throw new ArgumentNullException();

			if (match == null)
				throw new ArgumentNullException("match", "Match must not be null. Try passing Pattern.All() instead.");


			this.propertyName = propertyName;
			this.match = match;
		}

		public string PropertyName { get { return this.propertyName; } }
		public PatternMatch Match { get { return this.match; } }
	}

	public class Null : Node { public Null() : base(null) { } }
	public class Fail : INode { }
	public abstract class Pattern
	{
		#region Fail
		public static PatternMatch Fail()
		{
			return (c, s) => new Fail();
		}
		#endregion

		#region Null
		public static PatternMatch Null()
		{
			return (c, s) => new Null();
		}
		#endregion

		#region Transform
		public static INode Parse(string text, params PatternMatch[] patterns)
		{
			var context = new PatternContext();
			var result = Parse(context, text, patterns);
			if (context.Errors.Any())
				result = new Fail();
			return result;
		}

		public static INode Parse(PatternContext context, string text, params PatternMatch[] patterns)
		{
			return Transform(context, new StringStream(text), patterns);
		}

		public static INode Transform(object input, params PatternMatch[] patterns)
		{
			return Transform(input as IEnumerable ?? new[] { input }, patterns);
		}

		public static INode Transform(IEnumerable input, params PatternMatch[] patterns)
		{
			var context = new PatternContext();
			var result = Transform(context, new NodeStream(input), patterns);
			if (context.Errors.Any())
				result = new Fail();
			return result;
		}

		public static INode Transform(PatternContext context, object input, params PatternMatch[] patterns)
		{
			return Transform(context, new NodeStream(input), patterns);
		}

		public static INode Transform(PatternContext context, IEnumerable input, params PatternMatch[] patterns)
		{
			return Transform(context, new NodeStream(input), patterns);
		}

		public static INode Transform(PatternContext context, IStream stream, params PatternMatch[] patterns)
		{
			INode projection = null;
			var count = patterns.Length;
			for (int x = 0; x < count; x++)
			{
				var pattern = Wrap(patterns[x]);
				projection = pattern(context, stream);

				if (projection is Fail || stream.MoveNext())
					throw new InvalidOperationException("stream is not in a valid state.");

				if (context.Errors.Any(e => e.Severity == ErrorSeverity.Error))
					break;

				if (x < count - 1)
				{
					context.Memos.Step();
					stream = new NodeStream(projection);
				}
			}

			return projection;
		}

		private static PatternMatch Wrap(PatternMatch pattern)
		{
			// Main = p:P? error until !any -> p
			var wrapped = Pattern.Rule(
				Pattern.Projection(
					Pattern.And(
						Pattern.Variable("p", Pattern.ZeroOrOne(pattern)),
						Pattern.ErrorUntil(Pattern.Not(Pattern.Any()))),
					c => c.Variables["p"]));

			return wrapped;
		}
		#endregion

		#region Common Patterns

		#region And
		public static PatternMatch And(params PatternMatch[] patterns)
		{
			return (c, s) => And(c, s, patterns);
		}

		public static INode And(PatternContext context, IStream stream, IEnumerable<PatternMatch> patterns)
		{
			var pos = stream.Position;
			var projection = new Nodes();
			foreach (var pattern in patterns)
			{
				var match = pattern(context, stream);
				if (match is Fail)
				{
					stream.Position = pos;
					return match;
				}

				projection.Add(match);
			}

			return projection;
		}
		#endregion

		#region Or
		public static PatternMatch Or(params PatternMatch[] patterns)
		{
			return (c, s) => Or(c, s, patterns);
		}

		private static INode Or(PatternContext context, IStream stream, params PatternMatch[] patterns)
		{
			return Or(context, stream, (IEnumerable<PatternMatch>)patterns);
		}

		public static INode Or(PatternContext context, IStream stream, IEnumerable<PatternMatch> patterns)
		{
			var memos = context.Memos;
			var memoIndex = memos.MemoCount;
			var position = stream.Position;
			foreach (var pattern in patterns)
			{
				var match = pattern(context, stream);
				if (match is Fail)
				{
					memos.Reset(memoIndex);
					stream.Position = position;
					continue;
				}

				return match;
			}

			return new Fail();
		}
		#endregion

		#region Any
		private static readonly PatternMatch any = new PatternMatch(Any);

		public static PatternMatch Any()
		{
			return any;
		}

		private static INode Any(PatternContext context, IStream stream)
		{
			var position = stream.Position;
			if (stream.MoveNext())
				return Node.Wrap(stream.Current);

			stream.Position = position;
			return new Fail();
		}
		#endregion

		#region Default
		private static readonly PatternMatch @default = new PatternMatch(Default);

		public static PatternMatch Default()
		{
			return @default;
		}

		private static INode Default(PatternContext context, IStream stream)
		{
			return new Null();
		}
		#endregion

		#region Not
		public static PatternMatch Not(PatternMatch pattern)
		{
			return (c, s) => Not(c, s, pattern);
		}

		private static INode Not(PatternContext context, IStream stream, PatternMatch pattern)
		{
			var position = stream.Position;
			try
			{
				var match = pattern(context, stream);
				if (!(match is Fail))
					return new Fail();

				return new Null();
			}
			finally
			{
				// Not never consumes any input.
				stream.Position = position;
			}
		}
		#endregion

		#region Value
		public static PatternMatch Value(object value)
		{
			return (c, s) => Value(c, s, value);
		}

		private static INode Value(PatternContext context, IStream stream, object value)
		{
			var position = stream.Position;
			if (stream.MoveNext())
			{
				var current = stream.Current;
				var val = Node.Unwrap(current);
				if ((value != null && value.Equals(val)) ||
					(value == null && val == null))
					return Node.Wrap(current);
			}

			stream.Position = position;
			return new Fail();
		}
		#endregion

		#region Range
		public static PatternMatch Range<T>(T left, T right)
			where T : IComparable
		{
			return (c, s) => Range<T>(c, s, left, right);
		}

		private static INode Range<T>(PatternContext context, IStream stream, T left, T right)
			where T : IComparable
		{
			var position = stream.Position;
			if (stream.MoveNext())
			{
				var current = stream.Current;
				var value = Node.Unwrap(current);
				if (value is T &&
					left.CompareTo(value) <= 0 &&
					right.CompareTo(value) >= 0)
					return Node.Wrap(current);
			}

			stream.Position = position;
			return new Fail();
		}
		#endregion

		#region Array
		private static readonly PatternMatch emptyArray = Array(Not(Any()));

		public static PatternMatch Array()
		{
			return emptyArray;
		}

		public static PatternMatch Array(PatternMatch pattern)
		{
			return (c, s) => Array(c, s, pattern);
		}

		private static INode Array(PatternContext context, IStream stream, PatternMatch pattern)
		{
			var position = stream.Position;
			if (stream.MoveNext())
			{
				var arrayPosition = stream.Position;
				if (stream.MoveInto())
				{
					var match = pattern(context, stream);
					if (!(match is Fail))
					{
						if (!stream.MoveNext())
						{
							stream.Position = arrayPosition;
							return Node.Wrap(match);
						}
					}
				}
			}

			stream.Position = position;
			return new Fail();
		}
		#endregion

		#region OneOrMore
		public static PatternMatch OneOrMore(PatternMatch pattern)
		{
			return (c, s) => OneOrMore(c, s, pattern);
		}

		private static INode OneOrMore(PatternContext context, IStream stream, PatternMatch pattern)
		{
			var projection = new Nodes();
			var position = stream.Position;

			while (true)
			{
				var match = pattern(context, stream);
				if (match is Fail)
					break;

				projection.Add(match);
			}

			if (projection.Count > 0)
				return projection;

			stream.Position = position;
			return new Fail();
		}
		#endregion

		#region ZeroOrMore
		public static PatternMatch ZeroOrMore(PatternMatch pattern)
		{
			return (c, s) => ZeroOrMore(c, s, pattern);
		}

		private static INode ZeroOrMore(PatternContext context, IStream stream, PatternMatch pattern)
		{
			var projection = new Nodes();
			while (true)
			{
				var match = pattern(context, stream);
				if (match is Fail)
					break;

				projection.Add(match);
			}

			return projection;
		}
		#endregion

		#region ZeroOrOne
		public static PatternMatch ZeroOrOne(PatternMatch pattern)
		{
			return (c, s) => ZeroOrOne(c, s, pattern);
		}

		private static INode ZeroOrOne(PatternContext context, IStream stream, PatternMatch pattern)
		{
			var match = pattern(context, stream);
			if (!(match is Fail))
				return match;

			return new Null();
		}
		#endregion

		#region Exact Quantity
		public static PatternMatch ExactQuantity(PatternMatch pattern, int quantity)
		{
			return (c, s) => ExactQuantity(c, s, pattern, quantity);
		}

		private static INode ExactQuantity(PatternContext context, IStream stream, PatternMatch pattern, int quantity)
		{
			if (quantity < 0)
				throw new ArgumentOutOfRangeException("quantity", "Quantity must be a positive number or 0.");

			var projection = new Nodes();
			var position = stream.Position;
			while (projection.Count != quantity)
			{
				var match = pattern(context, stream);
				if (match is Fail)
					break;

				projection.Add(match);
			}

			if (projection.Count == quantity)
				return projection;

			stream.Position = position;
			return new Fail();
		}
		#endregion

		#region Projection
		public static PatternMatch Projection(PatternMatch pattern, Projection projectionMethod)
		{
			return (c, s) => Projection(c, s, pattern, projectionMethod);
		}

		private static INode Projection(PatternContext context, IStream stream, PatternMatch pattern, Projection projection)
		{
			INode match = null;
			try
			{
				stream.Scope.Push(stream.Position);
				match = pattern(context, stream);
				if (!(match is Fail))
				{
					var scope = stream.Scope.Current;
					scope.Context = context;
					scope.Match = match;
					match = projection(scope) ?? new Null();
				}
			}
			finally
			{
				stream.Scope.Pop();
			}

			return match;
		}
		#endregion

		#region Variable / Switch
		public static PatternMatch Switch(string name, PatternMatch pattern)
		{
			return (c, s) => Switch(c, s, name, pattern);
		}

		private static INode Switch(PatternContext context, IStream stream, string name, PatternMatch pattern)
		{
			var value = stream.Scope.Current.Variables[name];
			var position = stream.Position;
			INode match = null;
			if (stream.StepTo(name))
				match = pattern(context, stream);

			stream.Position = position;
			return match ?? new Fail();
		}

		public static PatternMatch Variable(string name)
		{
			return Variable(name, Pattern.Any());
		}

		public static PatternMatch Variable(string name, PatternMatch pattern)
		{
			return (c, s) => Variable(c, s, name, pattern);
		}

		private static INode Variable(PatternContext context, IStream stream, string name, PatternMatch pattern)
		{
			var match = pattern(context, stream);
			var scope = stream.Scope.Current;
			if (!scope.Variables.ContainsKey(name))
				scope.Variables.Add(name, match);
			else
				scope.Variables[name] = match;

			return match;
		}
		#endregion

		#region Call / Closure
		public static PatternMatch Call(ParameterizedPatternMatch pattern, params PatternMatch[] parameters)
		{
			return (c, s) => pattern(c, s, parameters);
		}

		public static ParameterizedPatternMatch Closure(Func<PatternMatch[], PatternMatch> pattern)
		{
			return (c, s, p) => Closure(c, s, pattern, p);
		}

		private static INode Closure(PatternContext context, IStream stream, Func<PatternMatch[], PatternMatch> pattern, params PatternMatch[] parameters)
		{
			return pattern(parameters)(context, stream);
		}
		#endregion

		#region Object
		public static PatternMatch Object(PatternMatch typeMatch, params PropertyMatch[] propertyMatches)
		{
			return (c, s) => Object(c, s, typeMatch, propertyMatches);
		}

		private static INode Object(PatternContext context, IStream stream, PatternMatch typeMatch, params PropertyMatch[] propertyMatches)
		{
			var position = stream.Position;
			var match = typeMatch(context, stream);
			if (!(match is Fail))
			{
				if (propertyMatches.Any())
				{
					foreach (var propertyMatch in propertyMatches)
					{
						var projection = Pattern.Property(context, stream, propertyMatch.PropertyName, propertyMatch.Match);
						if (projection is Fail)
							goto Fail;
					}
				}

				return match;
			}

		Fail:
			stream.Position = position;
			return new Fail();
		}

		public static PatternMatch Type(Type type)
		{
			return (c, s) => Type(c, s, type);
		}

		private static INode Type(PatternContext context, IStream stream, Type type)
		{
			var position = stream.Position;
			if (stream.MoveNext())
			{
				var current = stream.Current;
				var value = Node.Unwrap(current);
				if (value != null)
				{
					var valType = value.GetType();
					if (type.IsAssignableFrom(valType))
						return Node.Wrap(current);
				}
			}

			stream.Position = position;
			return new Fail();
		}

		public static PropertyMatch Property(string propertyName)
		{
			return Pattern.Property(propertyName, Pattern.any);
		}

		public static PropertyMatch Property(string propertyName, PatternMatch match)
		{
			return new PropertyMatch(propertyName, match);
		}

		private static INode Property(PatternContext context, IStream stream, string propertyName, PatternMatch match)
		{
			INode value = null;
			var position = stream.Position;
			if (stream.MoveInto(propertyName))
			{
				if (match != null)
					value = match(context, stream);
				else
					value = Node.Wrap(stream.Current);
			}
			else
			{
				value = new Fail();
			}

			stream.Position = position;
			return value ?? new Null();
		}
		#endregion

		#region Rule
		public static PatternMatch Rule(PatternMatch pattern, MethodInfo method = null)
		{
			return (c, s) => Rule(c, s, pattern, method);
		}

		private static INode Rule(PatternContext context, IStream stream, PatternMatch pattern, MethodInfo method)
		{
			var p = stream.Position;
			var m = default(Memo);
			MemoKey id;
			if (!context.Memos.GetMemo(pattern, p, out id, out m))
			{
				LR lr = new LR();
				m = new Memo { Projection = lr, Start = p, Method = method };
				var memoIndex = context.Memos.MemoCount;
				context.Memos.SetMemo(id, m);

				stream.Scope.Break(stream.Position);
				INode projection = m.Projection = pattern(context, stream);
				stream.Scope.Pop();

				m.End = stream.Position;
				context.Memos.UpdateMemo(id, m, memoIndex);

				if (lr.Detected && !(m.Projection is Fail))
				{
					projection = Pattern.Grow(context, stream, pattern, m, id, memoIndex);
					lr.Detected = false;
				}

				return projection;
			}
			else
			{
				if (m.Start != p)
					throw new AmbiguousMatchException(string.Format("Hashcode collision detected.\nstream position: '{0}'\n memo position: '{1}'", p, m.Start));

				stream.Position = m.End;
				var lr = m.Projection as LR;
				if (lr != null)
				{
					// Throw if recursion is indirect LR.
					if (context.Memos.Top != id)
						throw new TransformationException(string.Format(
							CultureInfo.CurrentCulture,
							"Indirect left recursion detected in rule {0}.",
							method));

					// Direct left-recursion detected.
					lr.Detected = true;
					return new Fail();
				}
				else
				{
					context.Memos.Load(id);
					return m.Projection;
				}
			}
		}

		private static INode Grow(PatternContext context, IStream stream, PatternMatch pattern, Memo memo, MemoKey id, int memoIndex)
		{
			INode projection = memo.Projection;
			var memos = context.Memos;
			while (true)
			{
				memos.Reset(memoIndex);
				stream.Position = memo.Start;
				projection = pattern(context, stream);
				if (projection is Fail || stream.Position <= memo.End)
				{
					memos.Reset(memoIndex);
					memos.Load(id);
					projection = memo.Projection;
					break;
				}

				memo.End = stream.Position;
				memo.Projection = projection;
				memos.UpdateMemo(id, memo, memoIndex);
			}

			stream.Position = memo.End;
			return projection;
		}

		private sealed class LR : INode
		{
			public bool Detected { get; set; }
		}
		#endregion

		#endregion

		#region Error Patterns

		public static PatternMatch ErrorUnless(PatternMatch pattern)
		{
			return (c, s) => ErrorUnless(c, s, pattern);
		}

		public static INode ErrorUnless(PatternContext context, IStream stream, PatternMatch pattern)
		{
			var match = pattern(context, stream);
			if (match is Fail)
			{
				match = new Null();
				context.LogError(
					stream.Position,
					ErrorLocation.After,
					ErrorSeverity.Error,
					"Expected {0}",
					pattern.Method.Name);
			}

			return match;
		}

		public static PatternMatch ErrorUntil(PatternMatch pattern)
		{
			return (c, s) => ErrorUntil(c, s, pattern);
		}

		public static INode ErrorUntil(PatternContext context, IStream stream, PatternMatch pattern)
		{
			var start = stream.Position;
			var end = start;
			try
			{
				while (!stream.Position.IsEnd)
				{
					var match = pattern(context, stream);
					if (!(match is Fail))
						return match;

					if (!stream.MoveNext())
						break;

					end = stream.Position;
				}

				return new Null();
			}
			finally
			{
				if (start != end)
				{
					context.LogError(
						new SourceSpan { Start = start.Next(), End = end },
						ErrorLocation.Span,
						ErrorSeverity.Error,
						"Values unexpected");
				}
			}
		}

		#endregion
	}
}
