﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Reflection;
using System.Diagnostics;
using System.Globalization;
using System.Collections;
using System.Runtime.InteropServices;

namespace MetaSharp.Transformation
{
	public class PatternContext
	{
		private MemoTable memos = new MemoTable();
		private Stack<Dictionary<object, SourceSpan>> steps = new Stack<Dictionary<object, SourceSpan>>();
		private Dictionary<Type, object> services = new Dictionary<Type, object>();
		private ArrayList<Error> errors = new ArrayList<Error>();

		public IEnumerable<Type> ServiceTypes()
		{
			return this.services.Keys;
		}

		public void Register<T>(T service)
		{
			this.Register(typeof(T), service);
		}

		public void Register(Type key, object service)
		{
			if (!this.services.ContainsKey(key))
				this.services.Add(key, service);
			else
				this.services[key] = service;
		}

		public T Locate<T>()
		{
			if (this.services.ContainsKey(typeof(T)))
				return (T)this.services[typeof(T)];

			return default(T);
		}

		public object Locate(Type key)
		{
			return this.services[key];
		}

		public MemoTable Memos
		{
			[DebuggerStepThrough]
			get { return this.memos; }
		}

		public IImmutableArray<Error> Errors
		{
			get { return this.errors; }
		}

		public void LogError(Position position, ErrorLocation location, ErrorSeverity severity, string format, params object[] args)
		{
			this.LogError(new SourceSpan { Start = position, End = position }, location, severity, format, args);
		}

		internal void LogError(
			SourceSpan span,
			ErrorLocation location,
			ErrorSeverity severity,
			string format,
			params object[] args)
		{

			var start = span.Start;
			if (start.Current != null)
			{
				var trace = memos.Trace(start.Current);
				if (trace != null)
					start = trace.Start;
			}

			var end = span.End;
			if (end.Current != null)
			{
				var trace = memos.Trace(end.Current);
				if (trace != null)
				{
					end = span.Start == span.End
						? trace.End
						: trace.Start;
				}
			}

			switch (location)
			{
				case ErrorLocation.Span:
					break;
				case ErrorLocation.After:
					start = end; //.IsEnd ? end : end.Next();
					end = start.IsEnd ? start : start.Next();
					break;
				default:
					throw new NotSupportedException("invalid location.");
			}

			this.errors.Add(new Error(
				new SourceSpan { Start = start, End = end },
				severity,
				string.Format(format, args)));
		}
	}
}
