﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;

namespace MetaSharp.Transformation
{
	public abstract class Position : IComparable<Position>
	{
		private int id;
		private string root;
		private long[] path;

		private Position parent;
		private Position into;
		private Position to;
		private Lazy<Dictionary<string, Position>> properties = new Lazy<Dictionary<string, Position>>(() => new Dictionary<string, Position>());

		protected Position(Position parent)
		{
			this.parent = parent;
		}

		public Position Parent
		{
			get { return this.parent; }
		}

		public virtual Position Next()
		{
			return null;
		}

		public virtual bool IsEnd { get { return false; } }

		public Position Into()
		{
			if (this.into == null)
			{
				var current = Node.Unwrap(this.Current);

				var text = current as string;
				var items = current as IEnumerable;
				if (text != null)
				{
					this.into = new CharacterPosition(this, text);
				}
				else if (items != null)
				{
					this.into = new ArrayPosition(this, items);
				}
			}

			return this.into;
		}

		public Position Into(string propertyName)
		{
			if (!this.properties.Value.ContainsKey(propertyName))
			{
				var current = this.Current;
				if (current != null)
				{
					var value = Node.Unwrap(current);
					var property = value.GetType().GetProperty(propertyName);
					if (property != null)
						this.properties.Value.Add(propertyName, new PropertyPosition(this, current, property));
				}
			}

			Position p = null;
			this.properties.Value.TryGetValue(propertyName, out p);
			return p;
		}

		public Position To(string name, object value)
		{
			if (this.to == null)
				this.to = new SwitchPosition(this, name, value);

			return this.to;
		}

		public abstract INode Current { get; }

		public static bool operator <(Position p1, Position p2)
		{
			return op_Spaceship(p1, p2) < 0;
		}

		public static bool operator <=(Position p1, Position p2)
		{
			return op_Spaceship(p1, p2) < 1;
		}

		public static bool operator >(Position p1, Position p2)
		{
			return op_Spaceship(p1, p2) > 0;
		}

		public static bool operator >=(Position p1, Position p2)
		{
			return op_Spaceship(p1, p2) > -1;
		}

		private static long[] EnsurePath(Position position)
		{
			if (position == null)
				return new long[0];

			if (position.path == null)
			{
				var path = EnsurePath(position.parent);
				var indexed = position as IIndexed;
				if (indexed != null)
				{
					var buffer = new long[path.Length + 1];
					Buffer.BlockCopy(path, 0, buffer, 0, path.Length * sizeof(long));
					buffer[path.Length] = indexed.Index;
					path = buffer;
				}

				position.path = path;
			}

			return position.path;
		}

		public override string ToString()
		{
			if (root == null)
			{
				var builder = new StringBuilder();
				if (this.parent != null)
					builder.Append(this.parent.ToString());
				BuildString(builder);
				root = builder.ToString();
			}

			return root;
		}

		protected abstract void BuildString(StringBuilder builder);

		public override int GetHashCode()
		{
			if (id == 0)
				id = ToString().GetHashCode();

			return id;
		}

		public int CompareTo(Position other)
		{
			return op_Spaceship(this, other);
		}

		// spaceship operator: <=>
		private static int op_Spaceship(Position p1, Position p2)
		{
			if (object.ReferenceEquals(p1, p2))
				return 0;

			var p1Path = EnsurePath(p1);
			var p2Path = EnsurePath(p2);

			var len = Math.Min(p1Path.Length, p2Path.Length);
			for (int x = 0; x < len; x++)
			{
				var c = p1Path[x].CompareTo(p2Path[x]);
				if (c != 0)
					return c;
			}

			return p1Path.Length.CompareTo(p2Path.Length);
		}
	}
}
