﻿namespace MetaSharp.Transformation:
	import System;
	import System.Collections.Generic;
	import System.Linq;
	import System.Text;
	import System.Collections;

	grammar Tokenizer < Transformer:

		Main = Token*;

		Token
			= token:
			(  CustomTokens
			|  WhitespaceToken
			|  LineCommentToken
			|  InlineCommentToken
			|  IdentifierToken
			|  StringLiteralToken
			|  CharacterLiteralToken
			|  NumberToken
			|  CharToken)
			-> {
				if (token is char or
					token is string or
					token is IEnumerable):
					token = Node.Wrap(Tokenizer.Stringify(token));
				end
				return token as INode;
			}
			
		CustomTokens = !default; // fail

		LineCommentToken = LineCommentBeginToken (!'\n' CharToken)* ('\n' | !any);
		LineCommentBeginToken = '/' '/';

		InlineCommentToken = '/' '*' ((!'*' CharToken) | ('*' !'/'))* '*' '/';

		WhitespaceToken = ' ' | '\0' | '\n' | '\r' | '\t';

		IdentifierToken = LetterToken+ (LetterToken | DigitToken)*;

		StringLiteralToken
			= str:
			( '"' 
			( '\\' '\\'
			| '\\' '"'
			| '\\' '0'
			| '\\' 'n'
			| '\\' 'r'
			| '\\' 't'
			| '\\' 'u' HexToken#4
			| '\\' 'U' HexToken#8
			| '\\' 'x' HexToken HexToken? HexToken? HexToken?
			| !'\\' !'"' CharToken)*
			'"' );

		CharacterLiteralToken
		    = '\'' '\\' 'u' HexToken#4 '\''
		    | '\'' '\\' 'U' HexToken#8 '\''
		    | '\'' '\\' 'x' HexToken HexToken? HexToken? HexToken? '\''
		    | '\'' '\\' ('\'' | '\\' | '0' | 'n' | 'r' | 't') '\''
		    | '\'' !'\\' CharToken '\'';

		NumberToken
		    =  '0' 'x' (HexToken#8 | HexToken#6 | HexToken#4 | HexToken#2 | HexToken)
		    |  '-'? DigitToken+ ('.' DigitToken+)?;

		LetterToken = 'a'..'z' | 'A'..'Z' | '_';

		DigitToken = '0'..'9';

		HexToken = '0'..'9' | 'a'..'f' | 'A'..'F';

		CharToken = '\u0000'..'\uffff';

		public INode Parse(string text):
			PatternContext context = new PatternContext();
			INode result = Parse(text, context);
			if(context.Errors.Length > 0):
				result = new Fail();
			end
			return result;
		end

		public INode Parse(string text, PatternContext context):
			if (text == null):
				throw new ArgumentNullException("text");
			end

			if (context == null):
				context = new PatternContext();
			end

			if (context.Locate<IIncrementService>() == null):
				context.Register<IIncrementService>(new IncrementService());
			end

			IStream stream = new StringStream(text);
			return super.Transform(context, stream);
		end
		
		public static string Stringify(object value):
			StringBuilder builder = new StringBuilder();
			Stringify(Node.Unwrap(value), builder);
			return builder.ToString();
		end

		private static void Stringify(object value, StringBuilder builder):
			if (value == null):
				return;
			end

			if (value is char):
				builder.Append(value as char);
				return;
			end

			if (value is string):
				builder.Append(value as string);
				return;
			end

			if (!(value is IEnumerable)):
				return;
			end

			for (object o in value.Cast<object>()):
				object val = Node.Unwrap(o);
				Stringify(val, builder);
			end
		end
	end
end